package gov.nist.nlpir.irf.broker;

import gov.nist.nlpir.irf.handle.Handle;
import gov.nist.nlpir.irf.broker.PersistenceBroker;
import gov.nist.nlpir.irf.util.Support;

/** 
 * This is an abstract class which represents what's common to all
 * flat file brokers for support of persistent objects - their
 * materialization and dematerialization.  <br><br>
 * 
 * @version $Revision: 1.2 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 *                             
 */

public abstract class FileBroker extends PersistenceBroker {
  

  /** 
   * Returns the passed file name with the current path (to where
   * persistent objects are kept) prepended. Handle case where
   * path is null or the null string
   * @param file the file name
   * @return the path + (separator + ) file
   */       
  public final String addPath(String file){
    String path = getCurrentLocationOfBrokeredStorage();
    if (path != null && path.length() > 0)
      if (!path.endsWith(Support.fileSeparator) && 
	  !file.startsWith(Support.fileSeparator))
	return path + Support.fileSeparator + file;
      else
	return path + file;
    else
      return file;
  }



  /** 
   * Materializes the object associated with the designated Oid.
   * Handle any part of the materialization that is common to
   * all objects using this type of broker.
   * @param handle The identifier unique of the object with the handle
   * @return Materialize the designated object
   */          
  public Object materializeWith(Handle handle) {
    /* Add code to construct the argument for file-based storage scheme */
    return currentRecordAsObject(handle);
  }
  


  /** 
   * Dematerializes the object associated with the returned handle.
   * Handle any part of the materialization that is common to
   * all objects using this type of broker.
   * @param object to be dematerialized
   * @return handle with which to recover object
   */          
  public Handle dematerialize(Object obj) {
    /* Add code to construct the argument for file-based storage scheme */
    return(currentObjectAsRecord(obj));
  }



  /** 
   * Dematerializes the updated object passed
   * Handle any part of the materialization that is common to
   * all objects using this type of broker.
   * @param oldHandle handle of the object before update
   * @param object to be dematerialized
   * @return handle with which to recover the updated object
   */          
  public Handle dematerialize(Handle oldHandle,Object obj) {
    /* Add code to construct the argument for file-based storage scheme */
    return(currentObjectAsRecord(oldHandle, obj));
  }
  


  /** 
   * Materializes the object associated with the designated Oid.
   * The concrete version will handle any part of the materialization
   * which is specific to the object.
   * @param arg The information needed to get the desired object
   * @return Materialize the designated object
   */          
  public abstract Object currentRecordAsObject(Object arg);


  
  /** 
   * Dematerializes the object.
   * The concrete version will handle any part of the materialization
   * which is specific to the object.
   * @param obj The object
   * @return handle The information needed to recover the object
   */          
  public abstract Handle currentObjectAsRecord(Object obj);



  /** 
   * Dematerializes the object.
   * The concrete version will handle any part of the materialization
   * which is specific to the object.
   * @param obj The object
   * @return handle The information needed to recover the object
   */          
  public Handle currentObjectAsRecord(Handle oldHandle, Object obj){
  return (Handle)null;}
  // For now an empty version of this method will be overridden only
  // by the Index broker. Later this should be abstract and all 
  // brokers will have to implement it.
}
