package gov.nist.nlpir.irf.broker;

import gov.nist.nlpir.irf.irfman.IrfGlobal;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.handle.Handle;

import java.lang.reflect.Method;

/** 
 * This is an abstract class which represents what's common to all
 * brokers (object relational, relational, flat file, etc.) for support
 * of persistent objects - their materialization and dematerialization.
 * <br><br>
 * 
 * @version $Revision: 1.1 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public abstract class PersistenceBroker {
  

  /** 
   * Returns the current high-level information about where brokered
   * (persistent) objects are kept.
   * @return the location information
   */       
  public final String getCurrentLocationOfBrokeredStorage(){
    return( IrfGlobal.gIrfManager.getInfoServer().
	    getLocationOfBrokeredStorage() );
  }



  /** 
   * Returns the passed file name with the current path (to where
   * persistent objects are kept) prepended.
   * @param file the file name
   * @return the path + (separator + ) file
   */       
  public String addPath(String file){
    String path = getCurrentLocationOfBrokeredStorage();
    if ( path != null)
      {
      if( !path.endsWith(Support.fileSeparator) && 
	  !file.startsWith(Support.fileSeparator) )
	return  path + Support.fileSeparator + file;
      else
	return  path + file;
      }
    else
      return file;
  }



  /** 
   * Returns the object associated with the passed handle
   * @param handle The info need to materialize a persistent object
   * @return The object associated with the handle passed as a parameter
   */          
  public Object objectWith(Handle handle) {
    Object obj = inCache(handle);  
    /*----------------------------------------------------------- 
      If the object was found in cache then return it; 
      Otherwise, materialize it 
      ------------------------------------------------------------*/
    if (obj != null)
      return obj;
    return materializeWith(handle);
  }
  


  /** 
   * Saves in persistent storage the object passed
   * @param obj The object to save
   * @return The handle which allows retrieval of the saved object
   */            
  public Handle handleOf(Object obj) {
    /*----------------------------------------------------------- 
      Dematerialize it 
      ------------------------------------------------------------*/
    if (obj != null)
      return dematerialize(obj);
    return null;
  }
  


  /** 
   * Saves in persistent storage the updated object passed
   * @param oldHandle The handle of the object already saved
   * @param obj The updated object to save
   * @return The handle which allows retrieval of the saved object
   */        
  public Handle handleOf(Handle oldHandle, Object obj) {
    /*----------------------------------------------------------- 
      Dematerialize it 
      ------------------------------------------------------------*/
    if (obj != null)
      return dematerialize(oldHandle, obj);
    return null;
  }
  


  /** 
   * Returns from cache the object associated with the designated handle
   * @param handle The info needed to materialize a persistent object
   * @return The requested object if found in the cache, otherwise
   * return null
   */          
  public Object inCache(Handle handle) {
    //Add code to get the object if it's there
    return null;
  }
  


  /** 
   * Materializes the object associated with the designated handle
   * @param handle The info needed to materialize a persistent object
   * @return the materialized object
   */          
  public abstract Object materializeWith(Handle handle);
  


  /** 
   * Dematerializes the object and associate it with the returned handle
   * @param obj object to dematerialize.
   * @return handleThe info needed to materialize a persistent object.
   */          
  public abstract Handle dematerialize(Object obj);



  /** 
   * Dematerializes the updated object and associate it with the returned handle
   * @param oldHandle Handle of the object before update
   * @param obj updated object to dematerialize.
   * @return handle The info needed to materialize the updated object.
   */          
  public abstract Handle dematerialize(Handle oldHandle, Object obj);
}
