package gov.nist.nlpir.irf.broker.braf;

import gov.nist.nlpir.irf.broker.FileBroker;
import gov.nist.nlpir.irf.handle.braf.BufferedRandomAccessFileHandle;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.irfman.IrfGlobal;

import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectInputStream;
import java.io.IOException;

/** 
 * This is an abstract class which represents what's common to all buffered
 * random access flat file brokers for support of persistent objects - their
 * materialization and dematerialization.  <br><br>
 * 
 * @version $Revision: 1.2 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public abstract class BufferedRandomAccessFileBroker extends FileBroker {
  
  /** A table of open files used to hold persistent objects.
   * Could be replaced by a cache mechanism with a finite number of files. */
  protected static Hashtable openFiles = new Hashtable(10);


  /** 
   * Byte size of the object being (de)serialized.
   * The readObject here  needs this information, but most overriding ones 
   * won't, so it doesn't appear in the signature. As it is protected, an 
   * overriding method may still use it if needed. */
  protected int sizeOfRecord;



  /** 
   * Reads the object associated with the designated argument from a file.
   * @param arg Information needed to input and materialize the object
   * @return the designated object matrialized
   */                 
  public Object readRecordFromFile(BufferedRandomAccessFileHandle handle) {
    Object s = null;
    
    try {
      /* Get the directory in which the file-based brokers will
	 store their objects and add to the file name (with the
	 file separator for the current platform if need be) so
	 that the file name is fully qualified. */
      String fileName = handle.getFile();
      // Get buffered random read and write access to the file.
      BufferedRandomAccessFile braf = 
	(BufferedRandomAccessFile)openFiles.get(fileName);
      if (braf == null) {
	braf = new BufferedRandomAccessFile(addPath(fileName),
					    BufferedRandomAccessFile.READ |
					    BufferedRandomAccessFile.WRITE);
	openFiles.put(fileName, braf);
      }
      synchronized (braf) {
	// Position the file so the next read will start at the
	// begining desired object.
	braf.seek(handle.getStartOfRecord());
	// Read the desired object into b.
	// Materialize the object in b which was read from the file.
	sizeOfRecord = handle.getLengthOfRecord();
	s = readObject(braf);
      }
    } catch (IOException ioe) {
      Support.errorMessage("BufferedRandomAccessFileBroker.readRecordFromFile: "
			   + ioe + "\nWas using  " + handle);
    }
    return s;
  }
  
  

  /** 
   * Writes the supplied object associated with the designated
   * handle to a file.
   * @param handle Information needed to dematerialize the object.
   * @param arg Object to be stored.
   * @return handle Information needed to materialize the object
   */                 
  public void writeRecordToFile(BufferedRandomAccessFileHandle handle,
				Object obj) {
    try {
      /* Get the directory in which the file-based brokers will store
	 their objects and add to the file name (with the file
	 separator for the current platform if need be) so that the
	 file name is fully qualified. */
      String fileName = handle.getFile();
      // Get buffered random read and write access to the file.
      BufferedRandomAccessFile braf = 
	(BufferedRandomAccessFile)openFiles.get(fileName);
      if (braf == null) {
	braf = new BufferedRandomAccessFile(addPath(fileName),
					    BufferedRandomAccessFile.READ |
					    BufferedRandomAccessFile.WRITE);
	openFiles.put(fileName, braf);
      }

      synchronized (braf) {      
	// Get the current length of the file. This will be used to
	// rember where the current record begins.
	long brafLength = braf.getLength();
	// Position the file so the next write will append the current
	// object to the file.
	braf.seek(brafLength);
	// Write the current object.
	// The toByteArray method is invoked to get a copy of the 
	// serialized object because the buffer of the
	// ByteArrayOutputStream is protected and I see no method to
	// gain access to it.
	writeObject(obj, braf);
	// Place the information about the offset into the file for the
	// start of the object and the object's length in the handle.
	handle.setPositionOfRecord(brafLength,
				   (int)(braf.getLength() - brafLength));
      }
    } catch(IOException ioe) {
      Support.errorMessage("BRAF_PB.writeRecordToFile: " + ioe +
			   " Object=" + obj);
    }
  }
  


  /**
   * Serializes an object to a file.
   * Every subclass may redefine this method to provide a tuned serialization
   * mechanism using this BufferedRandomAccessFile. This method is the default
   * mechanism, which uses the Serializable interface. It is very slow and 
   * disk space wasting, so it is OK to use it for large objects, thus in 
   * this case the serialization overhead is negligeable, but it should be 
   * tuned for small objects, like it is for DeIntegers for example (see 
   * the method pointed below). Nota: if the writing mechanism is tuned, the 
   * reading mechanism must be symmetric in the <code>readObject()</code> 
   * method.
   * @see gov.nist.nlpir.irf.broker.braf.DeIntegerBrafBroker#writeObject
   */
  public void writeObject(Object obj, BufferedRandomAccessFile braf)
    throws java.io.IOException {
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    ObjectOutputStream oos = new ObjectOutputStream(baos);
    oos.writeObject(obj);
    oos.flush();
    baos.flush();
    braf.write(baos.toByteArray());
    baos.close();
    oos.close();
  }
  


  /**
   * This method reads an object from the place where it is in the given file.
   * It uses the default deserialization mechanism, and thus may be overriden
   * by inheriting classes to use something more efficient. It must be 
   * symmetric to the the <code>writeObject()</code> method. For an example 
   * of a tuned mechanism:
   * @see gov.nist.nlpir.irf.broker.braf.DeIntegerBrafBroker#readObject
   */
  public Object readObject(BufferedRandomAccessFile braf)
    throws java.io.IOException {
    byte b[] = new byte[sizeOfRecord];
    try {
      int nbRead = braf.read(b);
      if (nbRead != sizeOfRecord)
	throw new IOException("BRAF_PB.readObject read " + nbRead +
			      " bytes instead of " + sizeOfRecord + ".");
      ByteArrayInputStream bais = new ByteArrayInputStream(b);
      ObjectInputStream ois = new ObjectInputStream(bais);
      Object read =  ois.readObject();
      ois.close();
      return read;
    } catch (java.lang.ClassNotFoundException cnfe) {
      Support.errorMessage("BRAF_PB: " + cnfe);
    }
    return null;
  }



  /** This static code registers the class with InfoServer so that it will
   * receive a shutdown signal (received by the following method) when the
   * application is about to quit. This way, everything that must be done
   * before leaving the application can be done in the shutdown method, as
   * the InfoServer will invoke it. 
   **/
  static {
    IrfGlobal.gIrfManager.getInfoServer().
      registerForShutdown(BufferedRandomAccessFileBroker.class);
  }



  /**
   * Needs to be called before the class disappears, otherwise the file will
   * be corrupted. As this class registered with the InfoServer thanks to the
   * <code>registerForShutdown()</code> method, this InfoServer will invoke
   * the current method when it receives itself a shutdown signal, so none
   * has to call it if the InfoServer mechanism is used.
   */
  public static void shutdown() {
    Enumeration files = openFiles.elements();
    while (files.hasMoreElements())
      try {
	((BufferedRandomAccessFile)files.nextElement()).close();
      } catch (IOException ioe) {
	// Will never happen: exceptions are not handled during finalization
      }
    Enumeration fileNames = openFiles.keys();
    while (fileNames.hasMoreElements())
      openFiles.remove(fileNames.nextElement());
  }
}
