package gov.nist.nlpir.irf.broker.braf;

import gov.nist.nlpir.irf.handle.Handle;
import gov.nist.nlpir.irf.broker.PersistenceBroker;
import gov.nist.nlpir.irf.broker.FileBroker;
import gov.nist.nlpir.irf.handle.braf.FeatureListBrafHandle;
import gov.nist.nlpir.irf.handle.braf.BufferedRandomAccessFileHandle;
import gov.nist.nlpir.irf.irfman.IrfGlobal;
import gov.nist.nlpir.irf.index.braf.LinkedListPool;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;

import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;

/**
 * This class models a file-based broker for to storing and retrieving
 *  Vectors from LinkedListPools.
 *
 * @version $Revision: 1.3 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public class FeatureListBrafBroker extends FileBroker {


  /** The link list pools served by this broker */
  private static Vector theLLPs;


  /** Static reference to the singleton instance of the broker */
  private static FeatureListBrafBroker broker = null;


  /** Constructur with no arguments */
  public FeatureListBrafBroker()
    throws IllegalArgumentException, IOException, SecurityException {
    theLLPs = new Vector();
  }



  /**
   * Returns the instance of the singleton class: FL file persistence
   * broker, creating the a broker instance if it doesn't already
   * exist.
   * @returns instance of IF file persistence broker.
   */
  public static PersistenceBroker instance() {
    if (broker == null)
      try {
	broker = new FeatureListBrafBroker();
      } catch (Exception e) {
	System.err.println("Error: " + e);
      }
    return broker;
  }



  /** 
   * Materializes the object associated with the designated Oid.
   * The concrete version will handle any part of the materialization
   * which is specific to the object.
   * @param handle Must be the FeatureListBrafHandle of the
   * FeatureList to store.
   * @return Materialize the designated object
   */          
  public Object currentRecordAsObject(Object handle) {
    FeatureListBrafHandle castHandle = (FeatureListBrafHandle)handle;
    try {
      return ((LinkedListPool)theLLPs.elementAt(castHandle.getFilePoolCode())).getList(castHandle);
    } catch (NullPointerException npe) {
      addLLP(castHandle.getFilePoolCode());
      return currentRecordAsObject(handle);
    } catch (ArrayIndexOutOfBoundsException npe) {
      addLLP(castHandle.getFilePoolCode());
      return currentRecordAsObject(handle);
    }
  }



  /** 
   * Dematerializes the object associated with the designated argument.
   * This concrete version will handle any part of the materialization
   * which is specific to the object.
   * @param obj object to be stored
   * @return handle of stored object representation in file.
   */
  public Handle currentObjectAsRecord(Object obj) {
    Support.errorMessage("currentObjectAsRecord should never be called on a FeatureListBrafBroker");
    return null;
  }



  /**
   * Adds an indexing feature to a feature list in the pool
   * @param toAdd the indexing feature to be added
   * @param the offset of the last feature added
   * @return the handle of the feature list added to
   */
  public void addElement(ProxyIndexingFeature toAdd,
			 FeatureListBrafHandle handle) {
    try {
      ((LinkedListPool)theLLPs.elementAt(handle.getFilePoolCode())).put(handle, toAdd);
    } catch (NullPointerException npe) {
      addLLP(handle.getFilePoolCode());
      addElement(toAdd, handle);
    } catch (ArrayIndexOutOfBoundsException npe) {
      addLLP(handle.getFilePoolCode());
      addElement(toAdd, handle);
    }
  }



  /**
   * Adds an indexing feature at the specified index to a feature list in 
   * the pool. <blink>DOESN'T WORK CURRENTLY</blink>.
   * @param toAdd the indexing feature to be added
   * @param position index at which to add
   * @param the offset of the last feature added
   * @return the handle of the feature list added to
   */
  public FeatureListBrafHandle insertElement(ProxyIndexingFeature toAdd,
						   int position, 
						   long previousOffset) {
    // To be written.
    return null;
  }



  /**
   * Adds a LinkedListPool to this broker so that indexes don't share
   * the same pool files, which seems more appropriate.
   * @param filePoolCode The index of the LinkedListPool to be added
   * in the theLLPs Vector
   */
  private void addLLP(int filePoolCode) {
    String fileName = (String)BufferedRandomAccessFileHandle.fileNames.elementAt(filePoolCode);
    if (theLLPs.size() < filePoolCode + 1)
      theLLPs.setSize(filePoolCode + 1);
    theLLPs.setElementAt(new LinkedListPool(fileName), filePoolCode);
  }



  /** This static code registers the class with InfoServer so that it will
      receive a shutdown signal (received by the following method) when the
      application is about to quit. This way, everything that must be done
      before leaving the application can be done in the shutdown method, as
      the InfoServer will invoke it. */
  static {
    IrfGlobal.gIrfManager.getInfoServer().
      registerForShutdown(FeatureListBrafBroker.class);
  }



  /**
   * Needs to be called before the class disappears, otherwise the file will
   * be corrupted. As this class registered with the InfoServer thanks to the
   * <code>registerForShutdown()</code> method, this InfoServer will invoke
   * the current method when it receives itself a shutdown signal, so none
   * has to call it if the InfoServer mechanism is used.
   */
  public static void shutdown() {
    Enumeration llps = theLLPs.elements();
    while (llps.hasMoreElements())
      try {
	((LinkedListPool)llps.nextElement()).shutdown();
      } catch (NullPointerException ioe) {
	// It's normal: there are "holes" in the Vector.
      }
  }
}
