package gov.nist.nlpir.irf.conversion;

import java.lang.reflect.Field;

/**
 * The class models a rule for use in converting tagged data into
 * Document objects
 *
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.
 * @version $Revision: 1.1 $
 */

public class ConversionRule {
  
  /** The tag whose presence in the raw data triggers the application of 
   * this rule */
  public String tag;

  /** A code indicating what conversion action is to be taken when this
   * rule is applied */
  public int action;

  /** Possible action code - ignore data associated with this tag */
  public static final int NONE = 0;

  /** Possible action code - convert data associated with this tag */
  public static final int CONVERT = 1;

  /** Possible action code - create a new document */
  public static final int START_DOCUMENT = 2;

  /** An instance of a field to be created by this rule */
  public Field targetField;

  /** An instance of the class for a field to be created by this rule */
  public Class targetFieldClass;



  /** 
   * Creates an empty conversion rule 
   */
  public ConversionRule() {
    tag = null;
    action = NONE;
    targetField = null;
    targetFieldClass = null;
  }



  /** 
   * Creates a conversion rule based on the specified parameters
   * @param tag the triggering the rule's application
   * @param action the code for the conversion action to be carried out
   * @param the name of the field to be created
   * @param the name field class to be created
   * @param the name document class to be created
   */
  public ConversionRule(String tag,
			int action,
			String targetFieldName,
			String targetFieldClassName,
			String targetDocClassName) {
    this.tag = tag;
    this.action = action;
    if (targetFieldClassName != null)
      {
      try 
	{
	targetFieldClass = Class.forName(targetFieldClassName);
	} 
      catch (ClassNotFoundException cnfe) 
	{
	System.err.println("ConversionRule(): " + cnfe);
	}
      }
    else
      targetFieldClass = null;

    Class targetDocClass = null;
    try {
      targetDocClass = Class.forName(targetDocClassName);
    } catch (ClassNotFoundException cnfe) {
      System.err.println("ConversionRule(): " + cnfe);
    }

    if (targetFieldName != null)
      {
      try 
	{
	targetField = targetDocClass.getField(targetFieldName);
	} 
      catch (NoSuchFieldException nsfe) 
	{
	System.err.println("ConversionRule, no such field: \"" 
			   + targetFieldName + "\" in class \""+ 
			   targetDocClassName + "\"" +nsfe);
	} 
      catch (SecurityException se) 
	{
	System.err.println("Security Exception: " + se);
	}
      }
    else
      targetField = null;
  }



  /** 
   * Returns the tag for this rule
   * @return the tag for this rule
   */
  public String getTag() {
    return tag;
  }  



  /**
   * Returns a string representation of the rule
   * @return a string representation of the rule
   */
  public String toString() {
    return "\nConversionRule: tag = \""+tag+"\";"+
           " action = "+action+";"+
           " targetfield = \""+targetField.getName()+"\";"+
           " targetFieldClass = \""+targetFieldClass.getName()+"\"";
  }
}
