package gov.nist.nlpir.irf.conversion;

import gov.nist.nlpir.irf.conversion.IrfConverter;
import gov.nist.nlpir.irf.conversion.Ascii2HtmlConverter;
import gov.nist.nlpir.irf.conversion.ConversionRules;
import gov.nist.nlpir.irf.conversion.ConversionRule;

import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.document.Document;

import gov.nist.nlpir.irf.de.ProxyDeString;
import gov.nist.nlpir.irf.de.ProxyDeHtml;
import gov.nist.nlpir.irf.de.ProxyDeInteger;
import gov.nist.nlpir.irf.de.ProxyDeDouble;
import gov.nist.nlpir.irf.de.ProxyDePersonName;
import gov.nist.nlpir.irf.de.DataElem;

import gov.nist.nlpir.irf.proxy.VirtualProxy;

import gov.nist.nlpir.irf.util.Support;

import java.util.Stack;
import java.util.Vector;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;


/**
 * Models generic conversion of SGML-tagged documents (as formatted by 
 * the SGML parser/checker nsgmls) in a file of proxy Documents as 
 * dictated by conversion rules.
 *
 * @version $Revision: 1.2 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 * */

public class Sgml2AppDocConverter extends IrfConverter {


  /** The rules to be used in conversion */
  private ConversionRules rules = null;

  /** The reader to be used to read the raw data */
  private BufferedReader file = null;


  /** Working constant */
  private static final int EOF = -1;   // end of file marker

  /** Working rule used in start state and when no rule is found for a 
   * tag. */
  ConversionRule defaultRule = new ConversionRule();

  /** Working Infield for start state*/
  InField defaultInField = new InField("Convert1's start state",defaultRule);



  /** 
   * Makes a new converter with no rules
   */
  public Sgml2AppDocConverter() {
    this.rules = null;
  }



  /** 
   * Makes a new converter given a set of conversion rules 
   * @param conversionRules collection-specific rules to guide conversion
   */
  public Sgml2AppDocConverter(ConversionRules rules) {
    this.rules = rules;
  }



  /** 
   * Sets the conversion rules
   * @param conversionRules collection-specific rules to guide conversion
   */
  public void setConversionRules(ConversionRules rules) {
    this.rules = rules;
  }
  

  /** 
   * Returns  the conversion rules
   * @return  the collection-specific rules that guide conversion
   */
  public ConversionRules getConversionRules() {
    return rules;
  }
  

  /**
   * Local class for objects on inFieldStack. Used to keep track of
   * input document fields and their associated conversion rules
   * and data values
   */
  class InField {           
    public String tag;
    public ConversionRule rule;
    public StringBuffer value;
      
    public InField(String tag, ConversionRule rule) {
      this.tag = tag;
      this.rule = rule;
      value = new StringBuffer();
    }	

    public void present() {
      System.out.println("InField: tag="+tag+
			 "\nInField: rule="+rule+
			 "\nInfield: value=\""+value+"\"");
    }
  }



  /**
   * Gives the converter the information it needs to access the raw
   * data for the documents to be converted. In the case of this 
   * converter this will be a fully qualified file name.
   *
   * @param loc fully qualified name of the file with the raw doc data
   */
  public void setRawDocLocation(String loc)
    throws FileNotFoundException {
    file = new BufferedReader(new FileReader(loc));
  }



  /** 
   * Advances the reader to just beyond the end of the Nth 
   * (numToIgnore-th) document, if possible.
   *
   * @param file a file opened using a BufferedReader
   * @param numToIgnore number of documents to skip
   * @return number of documents skipped
   */
  public int ignoreN(int numToIgnore) throws IOException {

    String line = "";
    String startTag = null;
    ConversionRule rule = null;
    String tag = null;
    char c = ' ';
    int docCount = 0;

    /*-----------------------------------------------------------------*
     * Find the first instance of the START_DOCUMENT  o p e n   tag
     *-----------------------------------------------------------------*/
    line = file.readLine();
    while (line != null && startTag == null) {
      c = line.charAt(0);
      //if no start doc yet & this is an open tag open
      if (startTag == null && c == '(') {
	tag = line.substring(1).trim();
	rule = rules.findRuleForTag(tag);
	if (rule != null && rule.action == ConversionRule.START_DOCUMENT)
	  startTag = tag;
      }    
      line = file.readLine(); // when tag found this loops read next line
      // for next loop
    } // end while
    
    /*-----------------------------------------------------------------*
     * Find Nth first instance of the START_DOCUMENT  c l o s e  tag
     *-----------------------------------------------------------------*/
    while (line != null) {
      if (line.charAt(0) == ')') { //if have start & this close
	tag = line.substring(1).trim();
	if (tag.compareTo(startTag) == 0) {
	  docCount++;
	  if (docCount == numToIgnore)
	    break; // don't read the next line !
	}
      }
      line = file.readLine();
    } // end while
    
    return docCount;
  }







  /** 
   * Converts one document from a file opened using a 
   * BufferedReader.
   *
   * @param file a file opened using a BufferedReader
   * @return a proxy IR document or null
   */
  public ProxyDocument convert1() throws IOException {
    
    // This stack keeps track of the actions within the scope of the
    // current field in the input document. 
    Stack inFieldStack = new Stack(); 
    
    ConversionRule rule = null;
    InField curInField = null;
    String tag = "";
    String line = "";
    int lineNum = 0;
    char c = ' ';

    ProxyDocument proxyIR_Doc = (ProxyDocument)null; // Return value

    // This vector keeps track of data for fields to be built within
    // the ProxyDocument to be returned.
    Vector outFields = new Vector();

    // Start off with a dummy InField entry and the default conversion rule 
    inFieldStack.push(defaultInField);

    boolean docComplete = false;
    line = file.readLine();
    lineNum++;

    /*-------------------------------------------------------------------*
     * A well-formed document will being with an open tag - skip
     * lines until one with an open tag is found or end of file is
     * reached. Useful for eating non-document tags at end file too.
     *-------------------------------------------------------------------*/
    while (line != null)
      if (line.charAt(0) == '(')
	break;
      else {
	line = file.readLine();
	lineNum++;
      }
    
    /*-------------------------------------------------------------------*
     * Look at input file a line at a time, determining from the first
     * character of each line  what sort of line it is and, if a tag, 
     * from the conversion rules how to treat the line.
     *-------------------------------------------------------------------*/
    while (line != null && docComplete == false) {
      c = line.charAt(0);
      /*---------------------------
       * Open tag ?
       *--------------------------*/
      if      ( c == '(' )  {
	/*----------------------------------------------------------------*
	 * Find the rule for the just encountered start tag and push 
	 * a reference to it or the default rule onto the InField
	 * Stack which keeps track of the current tag and rule.
	 *----------------------------------------------------------------*/
	tag = line.substring(1).trim();
	rule = rules.findRuleForTag(tag);
	
	if (rule != null)
	  curInField = new InField(tag, rule);
	else
	  curInField = new InField(tag, defaultRule);
	
	
	inFieldStack.push(curInField); 
	
	if (curInField.rule.action == ConversionRule.CONVERT) {
	  // Create a entry in the outFields vector
	  // iff one is not already there.
	  boolean found = false;
	  int outFieldsSize = outFields.size();
	  for (int i = 0; i < outFieldsSize; i++)
	    if ( ((InField)outFields.elementAt(i)).rule.targetField ==
		  curInField.rule.targetField ) 
	      {
	      found = true;
	      break;
	      }
	  if (!found) {
	    InField copyOfCurInField = 
	      new InField(curInField.tag, curInField.rule);
	    outFields.addElement(copyOfCurInField);
	    if (Support.DETAIL_TRACE <= Support.traceLevel){
	      System.out.print("Sgml2AppDocConverter:Adding outField ");
	      copyOfCurInField.present();
	    }
	  }
	}
	
      } // end ( c == '(' )

      /*---------------------------
       * Close tag ?
       *--------------------------*/
      else if ( c == ')' ) {
	tag = line.substring(1).trim();
	
	// Since the input documents are formatted by nsgmls
	// which supplies closing tags if omitted from the 
	// original document and since all start tags cause
	// an InField entry to be pushed on the inFieldStack,
	// the tag should match the one on top of the inFieldStack
	if (tag.compareTo(curInField.tag) != 0) {
	  Support.errorMessage("Sgml2AppDocConverter.convert1: Mismatch" +
			       " between expected closing tag \"" +
			       curInField.tag + "\"" +
			       " and tag encountered \"" + tag + "\"");
	  System.exit(-1);
	} 
	

	if ( curInField.rule.action == ConversionRule.CONVERT ) {
	  // If this field's data is to be saved, it should already 
	  // represented in the outFields vector. Find the entry and
	  // than append the value from the curInField to the value 
	  // in the field vector entry. 
	  int foundAtIndex = -1;
	  int outFieldsSize = outFields.size();
	  for (int i = 0; i < outFieldsSize; i++)
	    if ( ((InField)outFields.elementAt(i)).rule.targetField ==
		 curInField.rule.targetField ) 
	      {
	      foundAtIndex = i;
	      break;
	      }
	  if (foundAtIndex != -1)
	    ((InField)outFields.elementAt(foundAtIndex)).value.append(curInField.value);
	  else {
	    Support.errorMessage("Sgml2AppDocConverter.convert1: no entry"+
				 " in field vector for input document"+ 
				 " field with end tag \"" + tag);
	    System.exit(-1);
	  } // end if converting	  
	} else { // open tag's action isn't CONVERT, maybe it's START_DOCUMENT
	  
	  // Is this the end of a document?
	  rule = rules.findRuleForTag(tag);
	  if (rule != null && rule.action == ConversionRule.START_DOCUMENT) 
	    {     
	    docComplete = true; 
	    // Create the document to be returned
	    proxyIR_Doc = buildDoc(getProxyDocClass(),
				   getRealDocClass(), 
				   outFields);
	    continue;
	    } 
	}
	
	inFieldStack.pop(); // always clean up if doc not complete
	curInField = (InField)inFieldStack.peek();
	
      } // end else if ( c == ')' )
      
      
      /*---------------------------
       * Attribute ?
       *--------------------------*/
      else if ( c == 'A' ) {
	// Ignore this input line
      }
      
      /*---------------------------
       * Data
       *--------------------------*/
      else if ( c == '-' ) {
	if (curInField.rule.action == ConversionRule.CONVERT) {
	  int lineLen = line.length();
	  char lineArray[] = new char[lineLen];
	  String recEndMarker = "\\n";
	  int startCopySearchFromIndex = 1; // skip initial "-"
	  int copyToIndex = 0;
	  int foundAtIndex = 
	    line.indexOf(recEndMarker,startCopySearchFromIndex);
	  while( foundAtIndex != -1 && startCopySearchFromIndex < lineLen) {
	    
	    // Copy chars from where we left off up to the instance of "\n"
	    while(startCopySearchFromIndex < foundAtIndex) {
	      lineArray[copyToIndex] = line.charAt(startCopySearchFromIndex);
	      startCopySearchFromIndex++;
	      copyToIndex++;
	    }

	    // Insert a newline in the output array in place of "\n"
	    // and advance the indexes
	    lineArray[copyToIndex] = '\n';
	    copyToIndex++;
	    startCopySearchFromIndex = startCopySearchFromIndex + 2;
	    
	    // See if there is another instance of "\n"
	    foundAtIndex= line.indexOf(recEndMarker,startCopySearchFromIndex);
	  }   
	  
	  // Copy any remaining characters
	  while (startCopySearchFromIndex < lineLen) {
	    lineArray[copyToIndex] = line.charAt(startCopySearchFromIndex);
	    startCopySearchFromIndex++;
	    copyToIndex++;
	  }
	  
	  curInField.value.append(lineArray, 0, copyToIndex);
	  }
	}

      /*---------------------------
       * End of good parse or eof?
       *--------------------------*/
      else if ( c == 'C' || c == EOF ) {
	Support.errorMessage("Sgml2AppDocConverter.convert1: encountered"+
			     " end of file or parse before end of document"+
			     " on line " + lineNum);
	System.exit(-1);	
      }
      
      /*--------------------------------
       * Unexpected line type? - Ignore
       *--------------------------------*/
      else {
	//Support.errorMessage("Sgml2AppDocConverter.convert1: encountered"+
	//" unexpected ngsmls line type beginning with"+
	//" the character \"" + c + "\"");
	//System.exit(-1);
      }
      
      line = file.readLine();
      lineNum++;
      
    } // end while (line.compareTo("") != 0)
    
    return proxyIR_Doc;
  }



  /** 
   * Builds a ProxyDocument using the data on each field in the
   * outFields vector.
   * @param InField data for Start Document tag
   * @param vector of data on fields to be constructed 
   * @return a proxy IR document
   */
  public ProxyDocument buildDoc(Class proxyDocClass,
				   Class realDocClass,
				   Vector outFields) {

    ProxyDocument proxyDoc = null;
    Document realDoc = null;
    DataElem proxyDE = null;

    Object constructorArgs[] = new Object[1];
    Class argTypes[] = new Class[1];
    argTypes[0] = java.lang.String.class;

    String targetFieldClassName = null;

    /*--------------------------------------------------------------*
     * Get an instance of the application's real Document using
     * its passed Class class.
     *--------------------------------------------------------------*/
    try {
      realDoc = (Document)realDocClass.newInstance();
    } catch (IllegalAccessException e)  {
      System.err.println("Sgml2AppDocConverter can't create new"+
			 " real application document - illegal"+
			 " access exception. " +e);
    } catch (InstantiationException e) {
      System.err.println("Sgml2AppDocConverter can't create new"+
			 " real application document -"+
			 " instantiation exception. " +e);
    }  
    
    
    /*--------------------------------------------------------------*
     * Loop through the member objects to be created in the real document,
     * getting a constuctor for each, invoking the constructor to create
     * the member object, and setting the field to refer to the object.
     *--------------------------------------------------------------*/
    int outFieldsSize = outFields.size();
    for (int i = 0; i < outFieldsSize; i++) {
      targetFieldClassName = ((InField)outFields.elementAt(i)).
	rule.targetFieldClass.getName();
      
      if ((targetFieldClassName.compareTo("gov.nist.nlpir.irf.de.ProxyDeString")==0)
	  || (targetFieldClassName.compareTo("gov.nist.nlpir.irf.de.ProxyDeHtml")==0)
	  || (targetFieldClassName.compareTo("gov.nist.nlpir.irf.de.ProxyDePersonName")==0)
	  || (targetFieldClassName.compareTo("gov.nist.nlpir.irf.de.ProxyDeDouble")==0)
	  || (targetFieldClassName.compareTo("gov.nist.nlpir.irf.de.ProxyDeInteger")==0) ) {
        // For now we assume everything is represented in the raw 
        // documents as strings
	// argTypes[0] = java.lang.String.class; this line moved to a static
	// initialization at the beginning of the method.
        constructorArgs[0] = ((InField)outFields.elementAt(i)).value.toString();
      } else { // unexpected target class
        Support.errorMessage("Sgml2AppDocConverter.convert1: Request to"+
			     " create field of unexpected class \""+
			     ((InField)outFields.elementAt(i)).rule.
			     targetFieldClass.getName() + "\"");
        System.exit(-1);
      }
      /*--------------------------------------------------------------*
       * Get the constructor with the correct parameters and use
       * it to create a new object of the type to be assigned to
       * the field we are creating in the proxy application document
       *--------------------------------------------------------------*/
      try {
        proxyDE = (DataElem)((InField)outFields.elementAt(i)).rule.
	  targetFieldClass.getConstructor(argTypes).
	  newInstance(constructorArgs);
      } catch (IllegalAccessException e) {
        System.err.println("Sgml2AppDocConverter can't construct"+
			   " proxy data element due to"+
			   " illegal access exception. " +e);
      } catch (NoSuchMethodException e) {
        System.err.println("Sgml2AppDocConverter can't construct"+
			   " proxy data element due to"+
			   " no such method exception. " +e);
      } catch (InstantiationException e) {
        System.err.println("Sgml2AppDocConverter can't construct"+
			   " proxy data element due to"+
			   " instantiation exception. " +e);
      } catch (InvocationTargetException e) {
        System.err.println("Sgml2AppDocConverter can't construct"+
			   " proxy data element due to"+
			   " invocation target exception. " +e);
      }  
      /*--------------------------------------------------------------*
       * Set the field in the real document to the refer to the
       * proxy DE object just created.
       *--------------------------------------------------------------*/
      try {
        ((InField)outFields.elementAt(i)).rule.targetField.set(realDoc,proxyDE);
      } catch (IllegalAccessException e) {
        System.err.println("Sgml2AppDocConverter can't set field"+
			   " within proxy application document -"+
			   " illegal access exception. " +e);
      }  
      
    } // end for each field to be created in the ProxyDocument
    
    /*--------------------------------------------------------------*
     * Get the one argument (realObject) constructor for the proxy
     * document class and invoke it passing the realDoc to create
     * the desired proxy document which refers to the just built 
     * real document and its fields.
     *--------------------------------------------------------------*/

    argTypes[0] = realDocClass;
    constructorArgs[0] = realDoc;
    try {
      proxyDoc = 
        (ProxyDocument)proxyDocClass.
	getConstructor(argTypes).newInstance(constructorArgs);
    } catch (IllegalAccessException e) {
      System.err.println("Sgml2AppDocConverter can't construct"+
			 " proxy data element due to"+
			 " illegal access exception. " +e);
    } catch (NoSuchMethodException e) {
      System.err.println("Sgml2AppDocConverter can't construct"+
			 " proxy data element due to"+
			 " no such method exception. " +e);
    } catch (InstantiationException e) {
      System.err.println("Sgml2AppDocConverter can't construct"+
			 " proxy data element due to"+
			 " instantiation exception. " +e);
    } catch (InvocationTargetException e) {
      System.err.println("Sgml2AppDocConverter can't construct"+
			 " proxy data element due to"+
			 " invocation target exception. " +e);
    }  
    
    return proxyDoc;
  } // end buildDoc
} 
