package gov.nist.nlpir.irf.de;

import gov.nist.nlpir.irf.InfoObj;
import gov.nist.nlpir.irf.modality.IndexingModalities;
import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.query.QueryRep;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Vector; 

/**
 * This interface provides the skeleton all data element classes must comply
 * with<br><br>
 * IRF provides several base Data Elements but developpers can write their 
 * own ones to handle specific types of information, like speech or images. 
 * DataElems must provide function to be compared to other DataElems so that 
 * indexing and retrieval can be done.<br>
 * DataElems must also provide a hashCode() function based on the value they
 * represent so that they can be indexed.
 * 
 * @version $Revision: 1.1 $ 	
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public interface DataElem {
  
  /**
   * Each DataElem has to provide a function to present itself
   * so that the tree of DataElems can be browsed and checked.
   */
  abstract public void present();
  


  /**
   * Same as present() but the presentation can be shorter.
   */
  abstract public void presentBrief();
  


  /**
   * Matches score of this DataElem with another.
   * @param de the DataElem this is compared with.
   * @return the score of the matching process, given a convention.
   */
  abstract public float match(DataElem de);
  


  /**
   * Compares this DataElem to another.
   * @param de the DataElem this is compared with.
   * @return true if the DataElem are similar, <br> false otherwise
   */
  abstract public boolean exactMatch(DataElem de);
  


  /**
   * Returns the indexing features of this data element.
   * @param source source document of data element 
   * @param im the controlling indexing modalities
   * @param sourceField the document field from which the data element came
   * @return collection of indexing features for data element
   */
  abstract public Vector getIndexingFeatures(ProxyDocument source,
					     IndexingModalities im, 
					     DataElem sourceField);
  

  /**
   * Returns the string representation of the value of the DataElem
   * Note: may be different of getAsString(), when object embeds other ones
   * for instance.
   * @return the string representation of the value of the DataElem.<br>
   */
  abstract public String getValueAsString();
  


  /**
   * Returns the string representation of the DataElem
   * @return the string representation of the DataElem
   */
  abstract public String toString();
  


  /**
   * Compares the current object to the given object, which must be of the 
   * same type.
   * @param o the object this is compared to. Must be of the same type.
   * @return the difference between the two objects converted to an int. 
   * A convention must be chosen, including a value for incompatible types 
   * of arguments.
   */
  abstract public int compareTo(DataElem o);
  


  /** 
   *  Extends the default definition of hashCode() in Object.
   *  Definition of hashCode must be based on the value of the object.
   *  @return hash code of the DataElem, used for Index, and Hashtable
   *  @see java.util.Hashtable
   *  @see java.lang.Object
   */
  abstract public int hashCode();
  


  /** 
   *  Extends the default definition of equals(Object) in Object.
   *  Definition of equals must be based on the value of the object.
   *  @return true if objects are equals, "IRFly" speaking
   *  @see java.lang.Object#equals
   */
  abstract public boolean equals(Object other);



  /**
   * Adds the value of the DataElem to the string given in parameter.
   * @param baseString the string to which the HTML value will be appended.
   * Caution : the String is modified.
   * @param maxNrChar maximum number of characters this function will add 
   * to the string.
   * @param form if true, the text will be used to create a form, otherwise 
   * it won't be.
   * @return the HTML String resulting from the addition.
   */
  abstract public void addHTML(StringBuffer baseString, int maxNrChar, 
			       boolean form);
  


  /**
   * Appends a HTML representation of the DataElem to the given string,
   * allowing to create HTML docs from DataElems.
   * @param string the HTML text the representation will be added to
   */
  abstract public void HTML_ShowDoc(StringBuffer string);
}

