package gov.nist.nlpir.irf.de;

import java.io.Serializable;
import java.util.Vector;

import gov.nist.nlpir.irf.InfoObj;
import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.modality.IndexingModalities;


/**
 *  Representation of a double as a DataElem. This class will allow the
 *  use doubles as data elements, and to index them so that queries can 
 *  take them into account. As this class is concrete, all abstract 
 *  methods of DataElem are implemented here.
 *
 *  @version $Revision: 1.3 $ 	
 *   @author
 *  This software was produced by NIST, an agency of the U.S. government,  
 *  and by statute is not subject to copyright in the United States.        
 *  Recipients of this software assume all responsibilities associated      
 *  with its operation, modification and maintenance.  
 */
public class DeDouble extends InfoObj implements Serializable, DataElem {


  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  
  /** The double this DeDouble represents. 
   * @serial */
  private double value;
  


  /**
   *   Builds a DeDouble that represents the given value.
   *   @param v the value that the DeDouble will represent.
   */
  public DeDouble(double v) {
    value = v;
  }



  /**
   *   Builds a DeDouble that represents the given value.
   *   @param v the value that the DeDouble will represent - starting with
   *   a string
   */
  public DeDouble(String v) {
    if (v != null && v != "")
      value = new Double(v).doubleValue();
    value = 0;
  }
  


  /**
   * Returns the value of the represented double.
   *  @return the value of the represented double.
   */
  public final double getValue() { return value; }



  /**
   * Sets the value of the represented double
   * @param v the new value the DeDouble represents.
   */
  public final void setValue(double v) { value = v; }
    


  /**
   *  Displays the represented value surrounded by spaces..
   */
  public void present() { System.out.print(" " + value + " "); }
    


  /**
   *  Same as present(), but with no space.
   */
  public void presentBrief() { System.out.print(value); }
    


  /**
   * Matches the score of this DeDouble with another Object. For the 
   * method to succeed, the object must be a DeDouble also.
   * @param de the DeDouble this is compared with.
   * @return the score of the matching process.
   */
  public float match(DataElem de) { return (float)(exactMatch(de)?1.0:0.0); }



  /**
   * Compares this DeDouble to another. Definition would be 
   *   ExactMatch(DeDouble) if java allowed it.
   * @param de the DeDouble this is compared with.
   * @return true if the values are similar, <br>
   * false otherwise or if de is not a DeDouble.
   */
  public boolean exactMatch(DataElem de) { 
    return ((de instanceof DeDouble) && ((DeDouble)de).value == value);
  }



  /**
   * Extract indexing features from this data element according to
   * the indexing modalities
   * @param s source document of data element.
   * @param im indexing modalities of index in which data element will 
   * be stored.
   * @param field the document field from which this DE came
   * @return Vector of indexing features for this.
   */
  public Vector getIndexingFeatures(ProxyDocument source, 
				    IndexingModalities im, 
				    DataElem field) {
    Vector col = new Vector(1);
    col.addElement(new ProxyIndexingFeature(new ProxyDeDouble(this),
					    source, (ProxyDataElem)field));
    return col;
  }



  /**
   * Returns a string representation of the double
   * @return string representation of the double
   */
  public final String getValueAsString() {
    return Double.toString(value);
  }



  /**
   * Returns string representation of the double
   * @return string representation of teh double
   */
  public final String toString() {
    return Double.toString(value);
  }



  /**
   * Compares this DE-Double and the given DeDouble.
   * @return the score as an int. It's the integer part of the difference
   * between the two values. The result is Integer.MAX_VALUE if o2 is not
   * a DeDouble.
   * @param o the DeDouble this is compared to.
   */
  public int compareTo(DataElem o) {
    if (o instanceof DeDouble)
      return (int)(value - ((DeDouble)o).getValue());

    Support.errorMessage( 
			 "DeDouble.compareTo():" +  
			 "Incompatible Object. Expected DeDouble. Was passed " + 
			 o.getClass().getName() );
    return Integer.MAX_VALUE;
  }



  /**
   * Compares DeDoubles as compareTo(), but gives a boolean result.
   * @param o the DeDouble this is compared with.
   * @return true if their values are equal, of course,<br>
   * false otherwise or if o2 is not a DeDouble.
   */
  public boolean equals(Object o) {
    if (o instanceof DeDouble)
      return (value == ((DeDouble)o).getValue());

    Support.errorMessage("DeDouble.equals(): Incompatible Object. Expected " +
			 "DeDouble. Was passed " + o.getClass().getName() );
    return false;
  }



  /**
   * Redefines the default object.hashCode(). The resulting hash code 
   * is based on the one of the represented double value.
   *  @return hash code of DeDouble, used for Index, and Hashtable
   * @see java.util.Hashtable
   * @see java.lang.Object
   */
  public int hashCode() { 
    return (new Double(value)).hashCode();
  }
    


  /**
   * Adds the value to the HTML text given.
   * @param string the HTML text the value will be appended to.
   * @param i unused.
   * @param form unused.
   */
  public void addHTML(StringBuffer baseString, int i, boolean form) { 
    baseString.append(value); 
  }



  /**
   * Appends the value to the given string in order to create an HTML doc.
   * @param baseString the string which the value is added to.
   */
  public void HTML_ShowDoc(StringBuffer baseString) {
    baseString.append(value);
  }
}
