package gov.nist.nlpir.irf.de;

import java.io.Serializable;
import java.util.Vector;
import java.util.StringTokenizer;

import gov.nist.nlpir.irf.modality.IndexingModalities;
import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.irfman.IrfGlobal;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.feature.ProxyIdxFeatWordInText;

/**
 * This class is subclass of DeString specialized in its methods to
 * handle HTML-formatted text.
 * <br><br>
 *
 * @author 
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 * @version $Revision: 1.3 $
 */
public class DeHtml extends DeString implements Serializable {


  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

 
  // No additional instance data


  /**
   * Creates a DeHtml object given a string
   * @param v the string, which may content HTML tags.
   */
  public DeHtml(String v) {
    super(v);
  }
    


  /**
   * Creates a DeHtml object, assuming the given array of chars is 
   * an HTML String.
   * @param v the array of char corresponding to an HTML string.
   */
  public DeHtml(char v[]) {
    super(v);
  }
  


  /**
   * Computes the score given by the comparison between this and the given 
   * DeHtml
   * @return the score as an int. It's the difference between the HTML 
   * string values
   * @param o the DeHtml this is compared to
   */
  public int compareTo(DataElem o) {
    if (o instanceof DeHtml)
      return value.compareTo(((DeHtml)o).getValue());
    
    Support.errorMessage("DeHtml.compareTo():" +  
			 "Incompatible Object. Expected DeHtml." + 
			 " Was passed " + o.getClass().getName() );
    return 0; // avoids annoying compiler warning. Needs a convention
  }
  


  /**
   * Does the same comparison as compareTo(), but gives a boolean result.
   * @param o the DeHtml this is compared with
   * @return true if their values are equal,<br>
   * false otherwise or if o is not a DeHtml
    */
  public boolean equals(Object o) {
    if (o instanceof DeHtml)
      return (value.compareTo(((DeHtml)o).getValue()) == 0);
    
    Support.errorMessage("DeHtml.equals(): Incompatible Object. Expected DeHtml." +
			 " Was passed " + o.getClass().getName() );
    return false;
  }
  

  
  /**
   * Extracts Indexing Features from  DeHtml.
   * @param source object reference of source document.
   * @param im indexing modalities to be used with this DeString.
   * @param field the document field from which this DE came
   * @return vector of IndexingFeatures.
   */
  public Vector getIndexingFeatures(ProxyDocument source, 
				    IndexingModalities im,
				    DataElem field) {
    // Indexing Modalities; hard coded for now:
    Vector col = new Vector(30, 50);
    String word;
    boolean isHTML = false;
    
    HtmlAddrCount counter = new HtmlAddrCount(); // Counts words & paragraphs
    
    if (IrfGlobal.gTextStatistics == null) 
      IrfGlobal.gTextStatistics = new TextStatistics();

    StringTokenizer st = new StringTokenizer(value, DELIMITERS);
    while (st.hasMoreTokens()) {
      word = st.nextToken();
      isHTML = (word.charAt(0) == '<' && 
		word.charAt(word.length() - 1) == '>') ? true : false;
      if (isHTML)
	counter.newTag(word);
      else {
	counter.newWord();
	if (stopWordTable.containsWord(word))
	  IrfGlobal.gTextStatistics.newStopWord(); 
	else
	  if (word.charAt(0) == '<' || Character.isDigit(word.charAt(0)))
	    ; // Ignore tags and words starting with a digit 
	  else {			
	    word = stemmer.stemWord(word);
	    ProxyIndexingFeature anIF = 
	      new ProxyIdxFeatWordInText(new ProxyDeString(new DeString(word)),
				     source,  (ProxyDataElem)field,
				     counter.getWordCount(),
				     counter.getParagraphCount());
	    col.addElement(anIF);
	  }
      }
      IrfGlobal.gTextStatistics.newWord();
    }
    IrfGlobal.gTextStatistics.newDoc();
    return col;
  }
  


  /**
   *  Adds the current DeHtml value to the HTML text given  
   *  @param baseString the HTML text the value will be appended to
   *  @param maxNrOfChar maximum number of characters appended. 
   *  If 0, no limit to this number
   *  @param form if true, the string will be used to create a form, 
   *  a common text otherwise
   *  @return the new HTML string
   */
  public void addHTML(StringBuffer baseString, int maxNrOfChar,boolean form) {
    if (form)
      baseString.append("<pre>               <TEXTAREA COLS=63 ROWS=4 " +
			"<INPUT NAME=\"matchstr\" ></TEXTAREA><p><pre>");
    else if (maxNrOfChar == 0)
      baseString.append(value);
    else 
      baseString.append(value.substring(0, Math.min(maxNrOfChar,value.length())));
    
  }
  


  /**
   * Appends the DeHtml to the given string.
   * @param baseString the HTML text this  will be added to.
   */
  public void HTML_ShowDoc(StringBuffer baseString) {
    baseString.append(value);
  }
}

