package gov.nist.nlpir.irf.de;

import java.io.Serializable;
import java.util.Vector;

import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.modality.IndexingModalities;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.InfoObj;
import gov.nist.nlpir.irf.util.Support;

/**
 * Representation of a integer as a DataElem. <br><br>
 * This class will allow the use of integers as data elements and indexing 
 * them, so that queries can take them into account. As this class is 
 * concrete, all abstract methods of DataElem are implemented here.
 *  
 * @version $Revision: 1.3 $ 	
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */                      

public class DeInteger extends InfoObj implements Serializable, DataElem {

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;


  /** The value represented by this DeInteger. 
   * @serial */
  private int value;
  


  /**
   * Creates a DeInteger that represents the given integer.
   * @param v the value of the represented integer.
   */
  public DeInteger(int v) {
    value = v;
  }
  


  /**
   * Creates a DeInteger that represents the given integer.
   * @param v the value of the integer as a string.
   */
  public DeInteger(String v) {
    if (v != null &&
	v != "")
      value = new Integer(v).intValue();
    value = 0;
  }
  


  /**
   * Returns the value of the represented integer
   * @return the value of the represented integer.
   */
  public final int getValue() { return value; }
  


  /**
   * Sets the value of the represented integer
   * @param v the new value of the represented integer.
   */
  public final void setValue(int v) { value = v; }
    


  /**
   * Prints the value of the represented integer.
   */
  public void present() {
    System.out.print(value);
  }
  


  /**
   * Exactly the same thing as present(): prints the value.
   */
  public void presentBrief() {
    System.out.print(value);
  }


  
  /**
   * Returns a string representation of the integer
   * @return string representation of the integer. Same as getAsString().
   */ 
  public final String getValueAsString() {
    return Integer.toString(value); 
  }
  


  /**
   * Returns a string representation of the integer
   * @return string representation of the integer.
   */
  public final String toString() {
    return Integer.toString(value); 
  }
  


  /**
   * Gets the (actually unique) indexing feature of this DeInteger.
   * @param source the Document from which the integer was read.
   * @param im the controling indexing modalities 
   * @param field the document field from which this DE came
   * @return a vector with one element: an indexing feature linking a String
   * representing the value of this to an internal address linking this to
   * the given Document.
   */
  public Vector getIndexingFeatures(ProxyDocument source, IndexingModalities im,
				    DataElem field) {
    Vector col = new Vector(1);
    col.addElement(new ProxyIndexingFeature(new ProxyDeInteger(this),
					    source, (ProxyDataElem)field));
    return col;
  }
  


  /**
   * Compares the values of this and the given DeInteger and sets the result.
   * @param de the DeInteger this is compared to.
   * @return true if this and de have the same value,<br>
   * false otherwise or if de is not a DeInteger.
   */
  public boolean exactMatch(DataElem de) {
    return ((de instanceof DeInteger) && 
	    ((DeInteger)de).value == value);
  }
  


  /**
   * Computes the matching score between this and the given DeInteger. 
   * Actually, it's 1 if both DeIntegers have the same value, 0 otherwise.
   * @param de the DeInteger this is compared to.
   * @return 1 if this and de have the same value,<br>
   * 0 otherwise.
   */
  public float match(DataElem de) {
    return (float)(exactMatch(de)?1.0:0.0);
  }
  


  /**
   * Computes the difference between this and the given DeInteger.
   * The result is the mathematically defined difference.
   * @param o the DeInteger this is compared to.
   * @return the difference between this and compareTo, Integer.MAX_VALUE
   * if compareTo is not a DeInteger.
   */
  public int compareTo(DataElem o) {
    if (o instanceof DeInteger)
      return (value - ((DeInteger)o).getValue());
    
    Support.errorMessage("DeInteger.compareTo():" +  
			 "Incompatible Object. Expected DeInteger." + 
			 " Was passed " + o.getClass().getName() );
    return Integer.MAX_VALUE;
  }
  


  /**
   * Tests whether two DeIntegers have the same value.
   * The result is true if both DeIntegers have the same value.
   * @param compareToObj the DeInteger this is compared to.
   * @return true if this and compareTo have the same value,<br>
   * false otherwise.
   */
  public boolean equals(Object o) { // for index
    if (o instanceof DeInteger)
      return (value == ((DeInteger)o).getValue());
    
    Support.errorMessage("DeInteger.equals():" +  
			 "Incompatible Object. Expected DeInteger. Was passed " + 
			 o.getClass().getName() );
    return false;
  }
  


  /**
   * Redefines the default Object.hashCode().
   * @return hash code of DeInteger, used for Index, and Hashtable.
   * @see java.util.Hashtable
   * @see java.lang.Object
   */
  public final int hashCode() {
    return value;
  }
  


  /**
   * Appends the string representation of this to the given HTML text.<br>
   * Caution: string is modified by this function.
   * @param baseString the text to be extended by this. Caution: string 
   * itself is modified.
   * @param i not used.
   * @param form not used.
   * @return the new string.
   */
  public void addHTML(StringBuffer baseString, int i, boolean form) {
    baseString.append(value); 
  }
  


  /**
   * Appends a string representation of this to the given string.
   * @param baseString the string to be completed.
   */
  public void HTML_ShowDoc(StringBuffer baseString) {
    baseString.append(value); 
  }
}
