package gov.nist.nlpir.irf.de;

import java.io.Serializable;
import java.util.Vector;

import gov.nist.nlpir.irf.modality.IndexingModalities;
import gov.nist.nlpir.irf.InfoObj;
import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;

/**
 * Instances of this class represent a person name. As with all other 
 * DataElems, comparison, exact matching and presentation features are 
 * provided.
 *  
 *  @version $Revision: 1.3 $ 	
 *  @author
 *  This software was produced by NIST, an agency of the U.S. government,  
 *  and by statute is not subject to copyright in the United States.        
 *  Recipients of this software assume all responsibilities associated      
 *  with its operation, modification and maintenance.  
 */

public class DePersonName extends InfoObj implements Serializable, DataElem {

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  
  /** First name of the person name represented 
   * @serial  */
  private String firstName;


  /** Last name of the person name represented
   * @serial */ 
  private String surname;
  


  /**
   * Builds a DePersonName assuming the only string given is the person's
   * surname. The first name will be null.
   * @param sur the surname of the person.
   */
  public DePersonName(String sur) {
    surname = sur;
  }
  


  /**
   * Builds a DePersonName considering the only array of characters given
   * as the String representing the person's surname. The first name will be null.
   * @param sur considered as a string, will be the last name of the person.
   */
  public DePersonName(char sur[]) {
    surname = new String(sur);
  }
  


  /**
   * Assigns first and last name.
   * @param first person's first name.
   * @param sur person's surname.
   */
  public DePersonName(String first, String sur) {
    firstName = first;
    surname = sur;
  }
    


  /**
   * Treats the two arrays of characters as Strings and creates the 
   * DePersonName associated.
   */
  public DePersonName(char first[], char sur[]) {
    firstName = new String(first);
    surname = new String(sur);
  }



  /**
   * Retrieves the first name as a string.
   */
  public final String getFirstName() {
    return firstName;
  }
  


  /**
   * Retrieves the surname as a string.
   */
  public final String getSurname() {
    return surname;
  }
    


  /**
   * Sets the two main variables of the DePersonName.
   */
  public final void setValue(String first, String sur) {
    firstName = first;
    surname = sur;
  }


  
  /**
   * Prints a presentation of the DePersonName like this:<pre>
   * Name: myname
   * FirstName: myfirstName</pre>
   */
  public void present() {
    System.out.println("Name: " + surname +
		       "\nFirst name: " + firstName);
  }
  


  /**
   * Here, the presentation is briefer than with present():<pre>
   * myfirstName myname, </pre>
   */  

  public void presentBrief() {
    System.out.print(firstName + " " + surname + ", ");
  }
  


  /**
   * Compares this DePersonName to another.
   * @param de the DePersonName this is compared with.
   * @return true if both names and first names are the same,<br>
   * false otherwise or if de is not a DePersonName.
   */
  public boolean exactMatch(DataElem de) {
    return (de instanceof DePersonName &&
	    ((DePersonName)de).surname.compareTo(surname) == 0 && 
	    ((DePersonName)de).firstName.compareTo(firstName) == 0);
  }
  


  /**
   * Scores the match of this DePersonName with another Object. 
   * For the function to succeed, the object must be a DePersonName also. 
   * Actually the result is 1 if the two objects exactly match, 0 otherwise.
   * @param de the DePersonName this is compared with.
   * @return the score of the matching process.
   */
  public float match(DataElem de) {
    return (float)(exactMatch(de)?1.0:0.0);
  }
  


  /**
   * Overrides Object.hashCode(), based on the hashCoding of the surname.
   * @return hash code of DeDouble, used for Index, and Hashtable.
   * @see java.util.Hashtable
   * @see java.lang.Object
   */
  public int hashCode() {
    return surname.hashCode();
  }
  


  /**
   * Scores the comparison between this and the given DePersonName
   * @return the score as an int:<br>
   * If surnames are not the same, returns the comparison between them.<br>
   * If they are, returns the comparison between first names.<br>
   * In every case, if o is not a DePersonName, returns 0.
   * @param o the DePersonName this is compared to.
   * @see String#compareTo
   */
  public int compareTo(DataElem o) {
    if (o instanceof DePersonName) {
      int sur = surname.compareTo(((DePersonName)o).getSurname());
      if (sur != 0) // sur != 0 : the two surnames not equal
	return sur;
      else
	return firstName.compareTo(((DePersonName)o).getFirstName());
    }
    Support.errorMessage("DePersonName.compareTo():" +  
			 "Incompatible Object. Expected DePersonName. " + 
			 " Was passed " + o.getClass().getName() );
    return 0; // Avoids annoying compiler warning
  }
  


  /**
   * Does the same comparison as compareTo(), but gives a boolean result.
   * @param o the DePersonName this is compared with.
   * @return true if both their surnames and first names are equal,<br>
   * false otherwise or if o is not a DePersonName.
   */
  public boolean equals(Object o) {
    if (o instanceof DePersonName)
      if (surname.compareTo(((DePersonName)o).getSurname()) != 0)
	return false;
      else
	return(firstName.compareTo(((DePersonName)o).getFirstName()) == 0);
    
    Support.errorMessage("DePersonName.equals():" +  
			 "Incompatible Object. Expected DePersonName. " +
			 " Was passed " + o.getClass().getName() );
    return false;
  }
  


  /**
   * Extract indexing features from this person name
   * @param source source document of data element .
   * @param im indexing modalities of index in which data element will be 
   * stored.
   * @param field the document field from which this DE came 
   * @return collection of indexing features for data element.
   */
  
  public Vector getIndexingFeatures(ProxyDocument source, 
				    IndexingModalities im,
				    DataElem field) {
    Vector col = new Vector(1);
    col.addElement(new ProxyIndexingFeature(new ProxyDePersonName(this),
					    source, (ProxyDataElem)field));
    return col;
  }
  


  /**
   * Gives the same result as toString().
   * @return a string representation of teh person name.
   */
  public final String getValueAsString() {
    return surname + ", " + firstName;
  }
 


  /**
   * Returns a string representation of the person name.
   * @return string representation of object.
   * <pre>surname, firstname</pre>
   */
  public final String toString() {
    return surname + ", " + firstName;
  }
  


  /**
   * Adds the name and surname to the HTML text given. If it is to be read
   * in a form, the HTML code will allow the entry of values.
   * @param baseString the HTML text the value will be appended to.
   * @param maxNrOfChar maximum number of characters added, used only 
   * if form is false.
   * @param form true if the HTML code must be used in a form.
   * @return the new HTML string.
   */
  public void addHTML(StringBuffer baseString, int maxNrOfChar, boolean form) {
    if (form) {
      baseString.append("<strong>First Name:</strong><INPUT NAME=\"firstName\" SIZE=15>\n");
      baseString.append("<strong>Surname:</strong><INPUT NAME=\"surname\" SIZE=21> </pre><p>\n");
    } else if (maxNrOfChar == 0)
      baseString.append(firstName + " " + surname);
    else {
      String returnString = new String(firstName + " " + surname);
      baseString.append(returnString.substring(0, Math.min(maxNrOfChar,returnString.length())));
    }
  }



  /**
   * Adds the name and surname of the person to the HTML string given
   * (in this order).
   * @param baseString the String the two names will be added.
   */
  public void HTML_ShowDoc(StringBuffer baseString) {
    baseString.append(firstName + " " + surname);
  }
}
