package gov.nist.nlpir.irf.de;

import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.modality.IndexingModalities;
import gov.nist.nlpir.irf.InfoObj;
import gov.nist.nlpir.irf.de.normalize.Stemmer;
import gov.nist.nlpir.irf.irfman.IrfGlobal;
import gov.nist.nlpir.irf.de.normalize.StopWordTable;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.feature.ProxyIdxFeatWordInString;
import java.util.Vector;
import java.util.StringTokenizer;

/**
 * Representation of a string as a DataElem. <br><br>
 * This class will allow the use of strings as data elements and indexing 
 * them so that queries can take them into account. As this class is concrete,
 * all abstract methods of DataElem are implemented here.
 * This class is mainly composed of a String. It is the string that
 * a DeString object will refer to.
 *
 * @version $Revision: 1.2 $ 	
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public class DeString extends InfoObj implements DataElem {
  

  /** The string which this class models */
  protected String value;



  /** Working local alias for global stemmer to avoid repeated method calls */
  protected static final Stemmer stemmer = IrfGlobal.gStemmer;


  /** Working constant for use in tokenizing strings */
  protected static final String DELIMITERS = 
    " /.,;:'[]{}_-+=|\\~`!@#$%^&*()\"\n\t";
  

  /** Working local alias for global stop word table to avoid repeated method 
   * calls */
  protected static final StopWordTable stopWordTable = 
    IrfGlobal.gStopWordTable;



  /**
   * Creates a DeString that refers to the given String.
   */
  public DeString(String v) {
    value = v;
  }
  


  /**
   * Creates a DeString that refers to the given array of characters
   * considered as a String.
   */
  public DeString(char v[]) {
    value = new String(v);
  }
  


  /**
   * Retrieves the value of the represented String.
   */  
  public final String getValue() {
    return value;
  }
  


  /**
   * Sets the value of the represented String.
   * @param v the new value of the String.
   */  
  public final void setValue(String v) {
    value = v;
  }
  


  /**
   * Sets the value of the represented String with the given array of chars.
   * @param v the new value of the String.
   */  
  public final void setValue(char v[]) {
    value = new String(v);
  }
  


  /**
   * Returns the value of the represented string
   * @return the value of the represented string.
   */  
  public final String getValueAsString() {
    return value;
  }
  


  /**
   * Same as getValueAsString().
   */
  public final String toString() {
    return value;
  }
  


  /**
   * Prints the represented String with a space before and a comma
   * after.
   */
  public void present() {
    System.out.print(" " + value + "," ); 
  }
  


  /**
   * Prints the value of the String.
   */
  public void presentBrief() {
    System.out.print(value); 
  }
  


  /**
   * Compares this DeString to another.
   * @param de the DeString this is compared with.
   * @return true if both represented Strings are the same,<br>
   * false otherwise or if de is not a DeString.
   */
  
  public boolean exactMatch(DataElem de) {
    return ((de instanceof DeString) &&
	    (((DeString)de).value.compareTo(value) == 0));
  }
  
  

  /**
   * Scores the match of this DeString with another Object. 
   * For the function to succeed, the object must be a DeString also. 
   * Actually the result is 1 if the two objects exactly match, ie the 
   * Strings they represent are exactly the same, 0 otherwise.
   * @param de the DeString this is compared with.
   * @return the score of the matching process.
   */
  public float match(DataElem de) {
    return exactMatch(de)? 1.0f : 0.0f;
  }
  


  /**
   * Scores the comparison between this and the given DeString
   * @return the score as an int: the result is the comparison between the o2
   * String value and the value of this. Returns Integer.MAX_VALUE if o2 is
   * not a DeString.
   * @param o the DeString this is compared to.
   * @see String#compareTo
   */
  public int compareTo(DataElem o) {
    if (o instanceof DeString)
      return value.compareTo(((DeString)o).getValue());
    
    Support.errorMessage("DeString.compareTo():" +  
			 "Incompatible Object. Expected DeString. Was passed " + 
			 o.getClass().getName());
    
    return Integer.MAX_VALUE; /* how do you represent the fact that the
				 objects are incompatible? */
  }
  


  /**
   * Does the same comparison as compareTo(), but gives a boolean result.
   * @param o the DeString this is compared with.
   * @return true if both Strings are equal,<br>
   * false otherwise or if o is not a DeString.
   */
  public boolean equals(Object o) {
    if (o instanceof DeString)
      return (value.compareTo(((DeString)o).getValue()) == 0);
    
    Support.errorMessage("DeString.equals(): Incompatible Object." +
			 " Expected DeString. Was passed " + 
			 o.getClass().getName() );
    return false;
  }
  


  /**
   * Computes a hash code for the DE_string based on the hashCode of
   * the represented String.
   * @return hash code for use in java.util.Hashtable.
   */
  public final int hashCode() {
    return value.hashCode();
  }
  


  /**
   * Extracts Indexing Features from DeString.
   * @param source object reference of source document.
   * @param im indexing modalities to be used with this DeString.
   * @param field the document field from which this DE came
   * @return vector of IndexingFeatures.
   */
  public Vector getIndexingFeatures(ProxyDocument source, 
				    IndexingModalities im,
				    DataElem field) {
    // Indexing Modalities; hard coded for now:
    Vector col = new Vector(6, 12);
    String word;
    int wordCount = 0;
    
    // This is using java.util.StringTokenizer for now... It's much 
    // simpler than the method used before.
    StringTokenizer st = new StringTokenizer(value, DELIMITERS);
    
    while (st.hasMoreTokens()) {
      word = st.nextToken();
      wordCount++;
      if (!stopWordTable.containsWord(word))
	if (!Character.isDigit(word.charAt(0))) {
	  word = stemmer.stemWord(word);
	  ProxyIndexingFeature anIF = 
	    new ProxyIdxFeatWordInString(new ProxyDeString(new DeString(word)),
				     source, (ProxyDataElem)field, wordCount);
	  col.addElement(anIF);
	}
    }
    return col;
  }
  


  /**
   *  Adds the string to the HTML text given. If it is to be read
   *  in a form, the HTML code will allow the entry of values.
   *  @param baseString the HTML text the value will be appended to.
   *  @param maxNumberOfChar maximum number of characters added, 
   *  used only if form is false.
   *  @param form true if the HTML code must be used in a form, 
   *  false otherwise.
   *  @return the new HTML string.
   */
  public void addHTML(StringBuffer baseString, int maxNumberOfChar, boolean form){
    if (form == false)
      if (maxNumberOfChar == 0) 
	baseString.append(value);
      else
	baseString.append(value.substring(0, Math.min(maxNumberOfChar, value.length())));
    else
      baseString.append("      <INPUT NAME=\"matchstr\" SIZE=64> <p>");
  }
  


  /**
   *  Adds the value of the string between H1 tags to the given string.
   */
  public void HTML_ShowDoc(StringBuffer baseString) {
    baseString.append("\n<H1>" + value + "</H1>\n");
  }
}
