package gov.nist.nlpir.irf.de.normalize;
import gov.nist.nlpir.irf.de.normalize.Stemmer;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Implementation of the Porter stemming algorithm documented in:
 * Porter M.F. "An Algorithm For Suffix Stripping," Program 14 (3),
 * July 1980, pp. 130-137.<br><br>
 * 
 * Adapted to Java by Willie Rogers
 * from original C version by B. Frakes and C. Cox, 1986<br><br>
 
 *  Notes: 
 * <ul>
 * <li>This code will make little sense without the Porter article.
 * <li>The stemming function converts its input to lower case.
 * </ul>
 
 *  @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.
 * @version $Revision: 1.1 $
 */

public class PorterStemmer extends Stemmer {
  private Vector step1aRules;
  private Vector step1bRules;
  private Vector step1b1Rules;
  private Vector step1cRules;
  private Vector step2Rules;
  private Vector step3Rules;
  private Vector step4Rules;
  private Vector step5aRules;
  private Vector step5bRules;
  private static String lambda = new String("");
  private static final char EOS = '\0';
  private static int end;
  
  /**
   * Inner class to describle a single rule
   */
  
  class PorterRule {
    public int id;                 /* returned if rule fired */
    public String oldEnd;          /* suffix replaced */
    public String newEnd;          /* suffix replacement */
    public int oldOffset;          /* from end of word to start of suffix */
    public int newOffset;          /* from beginning to end of new suffix */
    public int minRootSize;        /* min root word size for replacement */
    public Condition condition;    /* the replacement test function (class) */
    
    PorterRule(int id,             /* returned if rule fired */
	       String oldEnd,      /* suffix replaced */
	       String newEnd,      /* suffix replacement */
	       int oldOffset,      /* from end of word to start of suffix */
	       int newOffset,      /* from beginning to end of new suffix */
	       int minRootSize,    /* min root word size for replacement */
	       Condition condition /* the replacement test function */) {
      this.id = id;
      this.oldEnd = oldEnd;
      this.newEnd = newEnd;
      this.oldOffset = oldOffset;
      this.newOffset = newOffset;
      this.minRootSize = minRootSize;
      this.condition = condition;
    }
  }
  
  /** 
   * This constructor creates all the rules it will use when asked to stem
   * a word.
   */
  public PorterStemmer() {
    
    /*****************   Initialized Private Data Structures   ********************/
    step1aRules = new Vector(4);
    step1aRules.addElement(new PorterRule(101, "sses", "ss",   3,  1, -1, null));
    step1aRules.addElement(new PorterRule(102, "ies",  "i",    2,  0, -1, null));
    step1aRules.addElement(new PorterRule(103, "ss",   "ss",   1,  1, -1, null));
    step1aRules.addElement(new PorterRule(104, "s",    lambda, 0, -1, -1, null));
    
    step1bRules = new Vector(3);
    step1bRules.addElement(new PorterRule(105, "eed", "ee",   2,  1,  0,  null));
    step1bRules.addElement(new PorterRule(106, "ed",  lambda, 1, -1, -1,  new ContainsVowel()));
    step1bRules.addElement(new PorterRule(107, "ing", lambda, 2, -1, -1,  new ContainsVowel()));
    
    step1b1Rules = new Vector(15);
    step1b1Rules.addElement(new PorterRule(108, "at",   "ate", 1,  2, -1,  null));
    step1b1Rules.addElement(new PorterRule(109, "bl",   "ble", 1,  2, -1,  null));
    step1b1Rules.addElement(new PorterRule(110, "iz",   "ize", 1,  2, -1,  null));
    step1b1Rules.addElement(new PorterRule(111, "bb",   "b",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(112, "dd",   "d",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(113, "ff",   "f",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(114, "gg",   "g",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(115, "mm",   "m",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(116, "nn",   "n",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(117, "pp",   "p",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(118, "rr",   "r",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(119, "tt",   "t",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(120, "ww",   "w",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(121, "xx",   "x",   1,  0, -1,  null));
    step1b1Rules.addElement(new PorterRule(122, lambda, "e",  -1,  0, -1,  new AddAnE()));
    
    step1cRules = new Vector(1);
    step1cRules.addElement(new PorterRule(123, "y",  "i",   0,  0, -1,  new ContainsVowel()));
    
    step2Rules = new Vector(20);
    step2Rules.addElement(new PorterRule(203, "ational", "ate",   6,  2,  0,  null));
    step2Rules.addElement(new PorterRule(204, "tional",  "tion",  5,  3,  0,  null));
    step2Rules.addElement(new PorterRule(205, "enci",    "ence",  3,  3,  0,  null));
    step2Rules.addElement(new PorterRule(206, "anci",    "ance",  3,  3,  0,  null));
    step2Rules.addElement(new PorterRule(207, "izer",    "ize",   3,  2,  0,  null));
    step2Rules.addElement(new PorterRule(208, "abli",    "able",  3,  3,  0,  null));
    step2Rules.addElement(new PorterRule(209, "alli",    "al",    3,  1,  0,  null));
    step2Rules.addElement(new PorterRule(210, "entli",   "ent",   4,  2,  0,  null));
    step2Rules.addElement(new PorterRule(211, "eli",     "e",     2,  0,  0,  null));
    step2Rules.addElement(new PorterRule(213, "ousli",   "ous",   4,  2,  0,  null));
    step2Rules.addElement(new PorterRule(214, "ization", "ize",   6,  2,  0,  null));
    step2Rules.addElement(new PorterRule(215, "ation",   "ate",   4,  2,  0,  null));
    step2Rules.addElement(new PorterRule(216, "ator",    "ate",   3,  2,  0,  null));
    step2Rules.addElement(new PorterRule(217, "alism",   "al",    4,  1,  0,  null));
    step2Rules.addElement(new PorterRule(218, "iveness", "ive",   6,  2,  0,  null));
    step2Rules.addElement(new PorterRule(219, "fulness", "ful",   5,  2,  0,  null));
    step2Rules.addElement(new PorterRule(220, "ousness", "ous",   6,  2,  0,  null));
    step2Rules.addElement(new PorterRule(221, "aliti",   "al",    4,  1,  0,  null));
    step2Rules.addElement(new PorterRule(222, "iviti",   "ive",   4,  2,  0,  null));
    step2Rules.addElement(new PorterRule(223, "biliti",  "ble",   5,  2,  0,  null));
    
    step3Rules = new Vector(7);
    step3Rules.addElement(new PorterRule(301, "icate",   "ic",    4,  1,  0,  null));
    step3Rules.addElement(new PorterRule(302, "ative",   lambda,  4, -1,  0,  null));
    step3Rules.addElement(new PorterRule(303, "alize",   "al",    4,  1,  0,  null));
    step3Rules.addElement(new PorterRule(304, "iciti",   "ic",    4,  1,  0,  null));
    step3Rules.addElement(new PorterRule(305, "ical",    "ic",    3,  1,  0,  null));
    step3Rules.addElement(new PorterRule(308, "ful",     lambda,  2, -1,  0,  null));
    step3Rules.addElement(new PorterRule(309, "ness",    lambda,  3, -1,  0,  null));
    
    step4Rules = new Vector(20);
    step4Rules.addElement(new PorterRule(401, "al",     lambda,  1, -1,  1,  null));
    step4Rules.addElement(new PorterRule(402, "ance",   lambda,  3, -1,  1,  null));
    step4Rules.addElement(new PorterRule(403, "ence",   lambda,  3, -1,  1,  null));
    step4Rules.addElement(new PorterRule(405, "er",     lambda,  1, -1,  1,  null));
    step4Rules.addElement(new PorterRule(406, "ic",     lambda,  1, -1,  1,  null));
    step4Rules.addElement(new PorterRule(407, "able",   lambda,  3, -1,  1,  null));
    step4Rules.addElement(new PorterRule(408, "ible",   lambda,  3, -1,  1,  null));
    step4Rules.addElement(new PorterRule(409, "ant",    lambda,  2, -1,  1,  null));
    step4Rules.addElement(new PorterRule(410, "ement",  lambda,  4, -1,  1,  null));
    step4Rules.addElement(new PorterRule(411, "ment",   lambda,  3, -1,  1,  null));
    step4Rules.addElement(new PorterRule(412, "ent",    lambda,  2, -1,  1,  null));
    step4Rules.addElement(new PorterRule(423, "sion",   "s",     3,  0,  1,  null));
    step4Rules.addElement(new PorterRule(424, "tion",   "t",     3,  0,  1,  null));
    step4Rules.addElement(new PorterRule(415, "ou",     lambda,  1, -1,  1,  null));
    step4Rules.addElement(new PorterRule(416, "ism",    lambda,  2, -1,  1,  null));
    step4Rules.addElement(new PorterRule(417, "ate",    lambda,  2, -1,  1,  null));
    step4Rules.addElement(new PorterRule(418, "iti",    lambda,  2, -1,  1,  null));
    step4Rules.addElement(new PorterRule(419, "ous",    lambda,  2, -1,  1,  null));
    step4Rules.addElement(new PorterRule(420, "ive",    lambda,  2, -1,  1,  null));
    step4Rules.addElement(new PorterRule(421, "ize",    lambda,  2, -1,  1,  null));
    
    step5aRules = new Vector(2);
    step5aRules.addElement(new PorterRule(501, "e",     lambda,  0, -1,  1,  null));
    step5aRules.addElement(new PorterRule(502, "e",     lambda,  0, -1, -1,  new RemoveAnE()));
    
    step5bRules = new Vector(1);
    step5bRules.addElement(new PorterRule(503, "ll",    "l",     1,  0,  1,  null));
  }
  
  final static boolean IsVowel(char c) {
    return ('a' == c || 'e' == c || 'i' == c || 'o' == c || 'u' == c);
  }
  
  /********************   Private Function Declarations   **********************
   *		    
   * Returns: a weird count of word size in adjusted syllables
   *
   * Purpose: Count syllables in a special way:  count the number 
   *          vowel-consonant pairs in a word, disregarding initial 
   *          consonants and final vowels.  The letter "y" counts as a
   *          consonant at the beginning of a word and when it has a vowel
   *          in front of it; otherwise (when it follows a consonant) it
   *          is treated as a vowel.  For example, the wordSize of "cat" 
   *          is 1, of "any" is 1, of "amount" is 2, of "anything" is 3.
   *
   * Plan:    Run a DFA to compute the word size
   *
   * Notes:   The easiest and fastest way to compute this funny measure is
   *          with a finite state machine.  The initial state 0 checks
   *          the first letter.  If it is a vowel, then the machine changes
   *          to state 1, which is the "last letter was a vowel" state.
   *          If the first letter is a consonant or y, then it changes
   *          to state 2, the "last letter was a consonant state".  In
   *          state 1, a y is treated as a consonant (since it follows
   *          a vowel), but in state 2, y is treated as a vowel (since
   *          it follows a consonant.  The result counter is incremented
   *          on the transition from state 1 to state 2, since this
   *          transition only occurs after a vowel-consonant pair, which
   *          is what we are counting.
   */
  
  /**
   *  @param word the word whose size is to be calculated.
   */
  
  public final static int wordSize(StringBuffer word) {
    int result = 0;
    int state = 0;  /* current state in the automaton */
    int i = 0;
    int length = word.length();
    
    /* Run a DFA to compute the word size */
    while (i < length) {
      char current = word.charAt(i);
      switch (state) {
      case 0: 
	state = IsVowel(current) ? 1 : 2;
	break;
      case 1: 
	state = IsVowel(current) ? 1 : 2;
	if (2 == state) result++;
	break;
      case 2: 
	state = (IsVowel(current) || 
		 ('y' == current)) ? 1 : 2;
	break;
      }
      i++;
    }
    return result;
  }
  
  abstract class Condition {
    abstract boolean condition(StringBuffer word);
  }
  
  /*
   * Returns: True if the word parameter contains a vowel,
   *          False otherwise.
   *
   * Purpose: Some of the rewrite rules apply only to a root containing
   *          a vowel, where a vowel is one of "aeiou" or y with a
   *          consonant in front of it.
   *
   * Plan:    Obviously, under the definition of a vowel, a word contains
   *          a vowel iff either its first letter is one of "aeiou", or
   *          any of its other letters are "aeiouy".  The plan is to
   *          test this condition.
   */
  
  class ContainsVowel extends Condition {
    final boolean condition(StringBuffer word) {
      int length = word.length();
      if (length == 0)
	return false;
      else {
	for (int i = 0; i < length; i++) 
	  if (IsVowel(word.charAt(i)))
	    return true;
	return false;
      }
    }
  }
  
  
  /**
   * Some of the rewrite rules apply only to a root with this characteristic.
   * @param word Buffer with the word checked
   * @return true if the current word ends with a consonant-vowel-consonant combination
   * and the second consonant is not w, x or y,<br> false otherwise
   */
  
  static final boolean endsWithCVC(StringBuffer word) {
    int length = word.length();	/* for finding the last three characters */
    if (length <= 2)
      return false;
    else {
      end = length - 1;
      char lastChar = word.charAt(end);
      boolean c1Flag = (! (IsVowel(lastChar) ||
			   ('w' == lastChar) || 
			   ('x' == lastChar) || 
			   ('y' == lastChar)));
      if (!c1Flag) return false;
      end--;
      lastChar = word.charAt(end);
      boolean vFlag = (IsVowel(lastChar) || 
		       ('y' == lastChar) );  /* vowel */
      if (!vFlag) return false;
      end--;
      boolean c2Flag = (! IsVowel(word.charAt(end)));  /* consonant */
      return c2Flag; // (c1Flag && vFlag && c2Flag);
    }
  }
  
  
  /*
   * Returns: True if the current word meets special conditions
   *          for adding an e.
   *
   * Purpose: Rule 122 applies only to a root with this characteristic.
   *
   * Plan:    Check for size of 1 and a consonant-vowel-consonant ending.
   */
  
  class AddAnE extends Condition {
    final boolean condition(StringBuffer word) {
      return endsWithCVC(word) && (1 == wordSize(word));
    }
  }
  
  /*
   * Returns: True if the current word meets special conditions
   *          for removing an e.
   *  
   * Purpose: Rule 502 applies only to a root with this characteristic.
   *    
   * Plan:    Check for size of 1 and no consonant-vowel-consonant ending.
   */
  
  class RemoveAnE extends Condition {
    final boolean condition(StringBuffer word) {
      return !endsWithCVC(word) && (1 == wordSize(word));
    }
  }
  
  /*
   *
   * Plan:    Loop through the rule set until a match meeting all conditions
   *          is found.  If a rule fires, return its id, otherwise return 0.
   *          Conditions on the length of the root are checked as part of this
   *          function's processing because this check is so often made.
   *          
   *          Notes:   This is the main routine driving the stemmer.  It goes through
   *          a set of suffix replacement rules looking for a match on the
   *          current suffix.  When it finds one, if the root of the word
   *          is long enough, and it meets whatever other conditions are
   *          required, then the suffix is replaced, and the function returns.
   */
  
  /**
   * Apply a set of rules to replace the suffix of a word.
   * @param word buffer with the stemmed word.
   * @param ruleList vector containing the replacement rules.
   * @return the ID of the rule fired, 0 if none is fired.
   */
  
  static final int replaceEnd(StringBuffer word, Vector ruleList) {
    String endingBuffer;
    int numberOfRules = ruleList.size();
    int wordLength = word.length();
    for (int i = 0; i < numberOfRules; i++) {
      PorterRule rule = (PorterRule)ruleList.elementAt(i);

      /*      int ending = end - rule.oldOffset; // set to start of possible stemmed suffix
      if (0 <= ending) {
	if (ending < wordLength)
	  endingBuffer = word.substring(ending);
	else
	  endingBuffer = "";
	if (0 == endingBuffer.compareTo(rule.oldEnd)) {
	  char tmpEnd[];
	  if (endingBuffer.length() > 0) {
	    tmpEnd = new char[end - ending + 1];
	    word.getChars(ending, end + 1, tmpEnd, 0);
	    word.setLength(ending);
	  } else {
	    tmpEnd = new char[1];
	    tmpEnd[0] = '\0';
	  }
	  if (rule.minRootSize < wordSize(word))
	    if (rule.condition == null || rule.condition.condition(word)) {
	      if (rule.newEnd.compareTo(lambda) != 0)
		word.append(rule.newEnd);
	      end = ending + rule.newOffset;
	      break;
	    }
	  if (tmpEnd[0] != '\0')
	    word.append(tmpEnd);
	  return rule.id;
	}
	} */

      int ending = end - rule.oldOffset; // set to start of possible stemmed suffix
      if (0 <= ending)
	if (ending < wordLength) {
	  endingBuffer = word.substring(ending);
	  if (0 == endingBuffer.compareTo(rule.oldEnd)) {
	    int formerLength = wordLength;
	    word.setLength(ending);
	    if (rule.minRootSize < wordSize(word))
	      if (rule.condition == null || rule.condition.condition(word)) {
		if (rule.newEnd.compareTo(lambda) != 0)
		  word.append(rule.newEnd);
		end = ending + rule.newOffset;
		break;
	      }
	    word.setLength(formerLength);
	    return rule.id;
	  }	
	} else {
	  if ("" == rule.oldEnd) {
	    if (rule.minRootSize < wordSize(word))
	      if (rule.condition == null || rule.condition.condition(word)) {
		// if (rule.newEnd.compareTo(lambda) != 0) // always true, as oldEnd == ""
		word.append(rule.newEnd);
		end = ending + rule.newOffset;
		break;
	      }
	    return rule.id;
	  }
	}
    }
    return 0; // No rule has been fired.
  }
  
  /* ********************   Public Function Declarations   **********************
   * Returns: FALSE if the word contains non-alphabetic characters
   *          and hence is not stemmed, TRUE otherwise
   *
   * Purpose: Stem a word
   *
   * Plan:    Part 1: Check to ensure the word is all alphabetic
   *          Part 2: Run through the Porter algorithm
   *          Part 3: Return an indication of successful stemming
   *
   * Notes:   This function implements the Porter stemming algorithm, with
   *          a few additions here and there.  See:
   *
   * Porter, M.F., "An Algorithm For Suffix Stripping,"
   * Program 14 (3), July 1980, pp. 130-137.
   *
   * Porter's algorithm is an ad hoc set of rewrite rules with
   * various conditions on rule firing.  The terminology of
   * "step 1a" and so on, is taken directly from Porter's
   * article, which unfortunately gives almost no justification
   * for the various steps.  Thus this function more or less
   * faithfully refects the opaque presentation in the article.
   * Changes from the article amount to a few additions to the
   * rewrite rules;  these are marked in the RuleList data
   * structures with comments.
   */
  
  /**
   * Stem word using porter stemming rules
   * @param term the word to be stemmed
   * @return stemmed version of word, word is null then stemmer
   * returns null. 
   */
  
  public final String stemWord(String term) {
    int rule;    /* which rule is fired in replacing an end */
    
    if (term == null) return null;
    
    StringBuffer word = new StringBuffer(term.toLowerCase());
    end = term.length() - 1;
    /* Part 1: Check to ensure the word is all alphabetic */
    
    /*  Part 2: Run through the Porter algorithm */
    replaceEnd(word, step1aRules); 
    rule = replaceEnd(word, step1bRules); 
    if ((106 == rule) || (107 == rule))
      replaceEnd(word, step1b1Rules); 
    replaceEnd(word, step1cRules);  
    replaceEnd(word, step2Rules);  
    replaceEnd(word, step3Rules);  
    replaceEnd(word, step4Rules);  
    replaceEnd(word, step5aRules); 
    replaceEnd(word, step5bRules);
    
    /* Part 3: Return an indication of successful stemming */
    return word.toString(); 
  }
}
