package gov.nist.nlpir.irf.de.normalize;

import gov.nist.nlpir.irf.util.Support;

import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.HashSet;
import java.util.Iterator;

/**
 * This class implements a table of unique stop word strings 
 * <br><br>
 * NOTE: entries are lowercased before addition and words to
 * be looked up are lowercased first.
 * 
 * @version $Revision: 1.1 $ 	
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public class StopWordTable {
  
  /** Table of stop words */
  private HashSet table;


  /** Working copy of end of file marker value */
  private static int EOF = -1;         // end of file marker
  

  /**
   * Constructs an empty stopword table
   */
  public StopWordTable() {
    table = new HashSet(1000);
  }


  /**
   *  Creates the Stop Word Table from a file. The file must contain the list
   *  of terms. A term is a sequence of letters and/or digits. Terms must
   *  be separated by one or more chars which are not letters/digits.
   *  If file cannot be opened, constructor returns with empty
   *  stringtable (size() = 0)
   *  @param fileName the name of the file used.
   */ 
  public StopWordTable(String fileName) {
    BufferedReader f = null;
    table = new HashSet(1000); // initial capacity 1000; default loadfactor (0.75)
    int ch = 0;                 // temp storage for character from file
    
    // Handle the case that the file is not found
    try {
      f = new BufferedReader(new FileReader(fileName));
    } catch (FileNotFoundException error) {
      Support.errorMessage("StopWordTable::StopWordTable(): " + error +
			   ", Sorry, can't open file: " + fileName );
      return; // return using empty string table;
    }
    StringBuffer term = new StringBuffer(); // reuse for all terms
    
    // Read through the file finding words and saving them in the table
    try {
      ch = f.read();
      while (ch != EOF) { 
	
	// ----- Discard leading non-term chars ---------------
	// ch cannot be EOF at this point   
	do {
	  if (Character.isLetterOrDigit((char)ch))  
	    break;
	  ch = f.read();            
	} while (ch!= EOF);
	
      // ----- Gather term chars-----------------------------
	while (ch != EOF) {
	  if (!Character.isLetterOrDigit((char)ch))  
	    break;
	  term.append((char)ch);
	  ch = f.read();     
	}
	
	// ----- Discard trailing non-term chars---------------
	while (ch != EOF) {
	  if (Character.isLetterOrDigit((char)ch))  
	  break;
	  ch = f.read();     
	}
	
	// If we got a term, save it in the table
	if (term.length() > 0) {
	  table.add(new String(term).toLowerCase());
	  term.setLength(0); // reuse term so hashes for same strings are ==
	}                    // identical strings in different StringBufs
	// yield different hashes and multiple entries
	
      } // end of outer while (ch != EOF)
    } catch (IOException ioe) {
      System.err.println("IOException in reading the stop word table.");
    }
  }
  

  /**
   * Adds a lowercased stop word to the table
   * @parm word the word to be added
   */
  public final void addWord(String word) {
    table.add(word.toLowerCase());
  }


  /**
   * Removes a stop word from the table
   * @parm word the word to be removed
   */
  public final void removeWord(String word) {
    table.remove(word);
  }
  

  /**
   * Returns the number of words in the table
   * @return the number of words in the table
   */
  public final int size() {
    return table.size();
  }


  /**
   * Presents the contents of the table.
   */
  public void present() {
    System.out.println( size() + " Stop words in the table:");
    Iterator words = table.iterator();
    while (words.hasNext())
      System.out.println(words.next());     
  }
  


  /**
   *  Tests whether the given word is in the table or not.
   *  @return true if the word is in the table,<br>
   *  false otherwise.
   */
  public final boolean containsWord(String wordToLookUp)  {
    return table.contains(wordToLookUp.toLowerCase());
  }
}
