package gov.nist.nlpir.irf.document;

import gov.nist.nlpir.irf.modality.IndexingModalities;
import gov.nist.nlpir.irf.modality.IndexingModalityUnit;
import gov.nist.nlpir.irf.document.ProxyDocument;

import gov.nist.nlpir.irf.index.ProxyIndex;
import gov.nist.nlpir.irf.index.ProxyIdfIndex;
import gov.nist.nlpir.irf.index.ProxyKeyWordIndex;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.proxy.VirtualProxy;

import java.util.Dictionary;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/** 
 * This class models a document collection
 * <br><br>
 * A document collection contains documents and associated
 * information such: what converter and indexing modalities
 * were used to create the documents and index them, what
 * indexes were created and where they are located, etc.
 *
 * Note: The default indexing technique for all indexing features
 * is set here - see method setDefaultIndexingModalities().
 * 
 * @version $Revision: 1.7 $  
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */                      

public class DocCollection implements java.io.Serializable {

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  /** The documents that make up this collection
   *  @serial  */
  private Vector             documents = new Vector();

  /** The collection's name
   * @serial  */
  private String             name = null;

  /** A brief description of the collection
   * @serial  */
  private String             description = null;

  /** The name of the class of documents in the collection
   * @serial  */
  private String             docClassName = null; 

  /** The name of the converter used to create the documents
   * @serial */
  private String             converterClassName = null;

  /** Information about the location of indexes for the collection
   * Must be set by constructor so existence of DocCollection implies
   * knowledge of index location information.
   * @serial  */
  private String             indexLocation = null;

  /** The indexing modalities that guide indexing of the documents
   * @serial   */
  private IndexingModalities indexingModalities = null; 

  /** The indexes created for this collection
   * @serial  */
  private Dictionary         indexes = new Hashtable(); // associated indexes

  /** Flag, true if refs to proxy docs and indexes can be added to 
   * the table of in-memory proxies
   * @serial */
  private boolean proxyRefsCounted = false; 

  /** Flag indicating whether this collection has been modified since it
   * was loaded */
  private transient boolean  modified = false;

  

  /** 
   * No-argument constructor
   */     
  public DocCollection() {
  }



  /** 
   * Constructor - initializes the collection with name, description, and 
   * index location information as specified
   * @param n the initial name for the collection
   * @param d the initial description for the collection
   * @param l information about the location of the collection index(s)
   */    
  public DocCollection(String n, String d, String l) {

    if (n == null)
      setName("");
    else
      setName(n);

    if (d == null)   
      setDescription("");
    if (d == null)   
      setDescription(d);

    if (l == null)
      setIndexLocation("");
    else
      setIndexLocation(l);
      
  }



  /** 
   * Tests whether the collection has been modified since last save
   * @return whether the collection has been modified
   */     
  final public boolean  getModified() {
    return modified;
  }



  /** 
   * Returns the number of documents in the collection
   * @return the number of documents in the collection
   */     
  final public int getNumDocs() {
    return documents.size();
  }



  /** 
   * Gets the name of the collection
   * @return The name of the collection
   */       
  final public String getName() {
    return name;
  }



  /** 
   * Sets the name of the collection
   */     
  final public void setName(String n) {
    name = n;
    modified = true;
  }



  /** 
   * Adds an index to the list of those associated with this collection
   * @param index the index to be added
   */     
 public void addIndex(ProxyIndex index) {
    if (proxyRefsCounted == true)
      {
      index.makePersistent();
      indexes.put(index.getName(),index.addRefToInMemoryProxiesByOid());
      }
    else
      {
      indexes.put(index.getName(), index);
      }

    modified = true;
  }



  /** 
   * Gets an enumeration to allow iteration over the available indexes
   * @return an enumeration allowing iteration over the available indexes
   */
  public Enumeration getIndexIter() {
    return indexes.elements();
  }


  /** 
   * Looks up an index by name in the dictionary of available indexes
   * @param indexname the name of the index to be looked up
   * @return the index corresponding to the specified name
   */
  public ProxyIndex findIndex(String indexname) {
    return (ProxyIndex)indexes.get(indexname);
  }



  /** 
   * Sets the documents
   * @param docs a vector of documents associated with this collection
   */     
  public void setDocuments(Vector docs) {
    int sizeDocs = 0;
    ProxyDocument proxyDoc = null;
    if (proxyRefsCounted == true)
      {
      sizeDocs = documents.size();
      for (int i = 0; i < sizeDocs; i++)
	{
        ((ProxyDocument)documents.elementAt(i)).
	  deleteRefFromInMemoryProxiesByOid();
	}
      }
    documents.removeAllElements();
    sizeDocs = docs.size();
    if (proxyRefsCounted == true)
      {
      for (int i = 0; i < sizeDocs; i++)
	{
	proxyDoc = ((ProxyDocument)documents.elementAt(i));
        proxyDoc.makePersistent();
	documents.addElement(proxyDoc.addRefToInMemoryProxiesByOid());
	}
      }
    else
      for (int i = 0; i < sizeDocs; i++)
	documents.addElement(docs.elementAt(i));

    modified = true;
  } 



  /** 
   * Adds a document to the collection - if the other documents
   * have been made persistent then the added document must be
   * made persistent before the reference to it is counted.
   * Register the reference to the proxy doc
   */     
  public void addDoc(ProxyDocument doc) {
    if (proxyRefsCounted == true)
      {
      doc.makePersistent();
      documents.addElement(doc.addRefToInMemoryProxiesByOid());
      }
    else
      {
      documents.addElement(doc);
      }
    modified = true;

  }



  /** 
   * Sets the description of the collection
   */     
  final public void setDescription(String desc) {
    description = desc;
    modified = true;
  }



  /** 
   * Gets the description of the collection
   * @return The description of the collection
   */     
  final public String getDescription() {
    return description;
  }



  /** 
   * Sets the name of the class for the document objects
   * in this collection
   */     
  final public void setDocClassName(String name) {
    docClassName = name;
    modified = true;
  }



  /** 
   * Gets the name of the class for the document objects
   * in this collection
   */     
  final public String getDocClassName() {
    return docClassName;
  }



  /** 
   * Sets the name of the class of the converter for this collection
   */     
  final public void setConverterClassName(String name) {
    converterClassName = name;
    modified = true;
  }



  /** 
   * Gets the name of the class for the document objects
   * in this collection
   */     
  final public String getConverterClassName() {
    return converterClassName;
  }



  /** 
   * Sets the collection-level location of the indexes for this collection
   * @param loc the index location info string
   */     
  final private void setIndexLocation(String loc) {
    if (loc == null || loc == "")
      indexLocation = "";
    else
      indexLocation = loc;
    modified = true;
  }



  /** 
   * Gets the collection-level location of the indexes for this collection
   */     
  final public String getIndexLocation() {
    return indexLocation;
  }



  /** 
   * Sets indexing modalities for this collection
   */     
  final public void setIndexingModalities(IndexingModalities IMs) {
    indexingModalities = IMs;
    modified = true;
  }



  /** 
   * Gets the indexing modalities for this collection
   * @return the indexing modalities associated with this collection
   */     
  final public IndexingModalities getIndexingModalities() {
    return indexingModalities;
  }



  /** 
   * Adds an indexing modality unit
   */     
  final public void addIndexingModalityUnit(IndexingModalityUnit modalityUnit) {
    indexingModalities.addModality(modalityUnit);
  }


  /** 
   * Sets the indexing modalities based on supplied sample (proxy) document 
   * using reflection.
   * Where an index does not exist for a particular feature, it is created. 
   * It is here that the type of the index is determined (e.g. IDF vs. 
   * KeyWord).
   *
   * @param r sample document
   */
  public void setDefaultIndexingModalities(DocumentIntrf r) {
    Vector fnames = r.getFeatureNames();
    int namesNumber = fnames.size();
    setIndexingModalities(new IndexingModalities());
    Document realR = null;
    if (r instanceof ProxyDocument)
      realR = (Document)((ProxyDocument)r).getRealObject();
    else
      realR = (Document)r;
    for (int i = 0; i < namesNumber; i++) {
      String fname = (String)fnames.elementAt(i);
      IndexingModalityUnit imu = new IndexingModalityUnit();
      imu.setFeatureName(fname);
      String indexname = fname + "index";
      ProxyIndex anIndex = findIndex(indexname);
      if (anIndex == null) { // index doesn't exist : create it!
	// Modify the following line in order to change the default indexing 
	// technique for all collections, e.g. 
	// anIndex = new ProxyKeyWordIndex(indexLocation,indexname);
	anIndex = new ProxyIdfIndex(indexLocation,indexname);
	anIndex.makePersistent();
	addIndex(anIndex);
      }
      imu.setIndex(anIndex);

      indexingModalities.addModality(imu);

    }

  }



  /**
   * Allows indexing modalities to be added/deleted by the end user
   * @param r sample document
   */
  public void editIndexingModalities(DocumentIntrf r) { 
    boolean done = false;
	
    while (! done) {
      indexingModalities.present();
      System.out.println("Options:\n" +
			 "  a   : add a new modality\n" +
			 "  d   : set default modalities");
      if (indexingModalities.getModalityCount() > 0)
	System.out.print("  x   : delete a modality\n");
		
      System.out.print("  0   : back to main menu\n" +
		       "Choose : ");
      String input = Support.readLine();
      // Ensure that only single character input is treated as such
      char option = '&';
      if( input.length() == 1 )
	option = input.charAt(0);

      switch (option) {

      //----------------------------------------------------
      case '0' : indexingModalities.present(); return; 

      //----------------------------------------------------
      case 'x' : {
	System.out.print("Give modality nr to delete (1-" + 
			 indexingModalities.getModalityCount() + ") : ");
	int d;
	d = Integer.parseInt(Support.readLine());
	if (d != 0 && d<=(indexingModalities.getModalityCount()))
	  {

	  // Get the name of the index associated with the IMU to be deleted
	  String targetIndexName = indexingModalities.getModalityAt(d-1).
	                             getIndexName();

	  // Delete the IMU from the modalities list
	  indexingModalities.removeModalityAt(d-1);

	  // If no other IMU is associated with the target IMU's index,
	  // then remove the latter index from the dictionary of indexes.
	  Vector modalities = indexingModalities.getModalities();
	  int modalitiesSize = modalities.size();
	  boolean found = false;
	  for (int i=0; i<modalitiesSize; i++)
	    {
	    if (targetIndexName == 
		((IndexingModalityUnit)modalities.elementAt(i)).
		getIndexName())
	      {
	      found = true;
	      break;
	      }
	    }
	  if (found == false)
	    indexes.remove(targetIndexName);
	  }

      } break;

      //----------------------------------------------------
      case 'a' : {
	System.out.print("Feature name for new modality: ");
	String name = Support.readLine();
	IndexingModalityUnit imu = new IndexingModalityUnit();
	imu.setFeatureName(name);

	System.out.print( "Index for new modality: " );
	name = Support.readLine();
	// Check dictionary to see if index by this name exists
	// If so, use it, else create a new one.
	ProxyIndex newIndex = (ProxyIndex)indexes.get(name);
	if (newIndex == null)
	  {
	  newIndex = new ProxyIdfIndex(indexLocation, name); 
	  newIndex.makePersistent();
	  addIndex(newIndex);
          }
	imu.setIndex(newIndex);
	indexingModalities.addModality(imu);
      } break;

      //----------------------------------------------------
      case 'd' : {
	setDefaultIndexingModalities(r);
      } break;

      //----------------------------------------------------
      default: {
	int opt = 0;
	try{
	  opt = Integer.parseInt( input );
	} catch( NumberFormatException e ) { 
	  System.out.print( "Please enter 0, a, d, or x\\n" );
	  continue;
	}
      }		
      
      } // end switch

      if (!indexingModalities.validate(r))
	Support.errorMessage("IndexingModalities::Edit()" +
			     "New modalities NOT validated");
    } // end while (! done) 

  }



  /** 
   * Removes any proxies duplicated during deserialization - replacing
   * them with references to a single instance of the proxy. Proxies
   * may be contained by the indexes dictionary and the indexing
   * modalities.
   */
  public void getFirstInstances() {
    proxyRefsCounted = true;
    for (int i = 0; i < documents.size(); i++)
      documents.setElementAt(((VirtualProxy)documents.elementAt(i)).
			     addRefToInMemoryProxiesByOid(),i);
    Enumeration keys = indexes.keys();
    while (keys.hasMoreElements()) {
      String currentKey = (String)keys.nextElement();
      ProxyIndex currentIndex = (ProxyIndex)indexes.get(currentKey);
      indexes.put(currentKey, currentIndex.getFirstInstance());
    } 
    Vector modalityList = indexingModalities.getModalities();
    for (int i = 0; i < modalityList.size(); i++)
      {
      ((IndexingModalityUnit)modalityList.elementAt(i)).getFirstInstances();
      }
  }
  
  

  /** 
   * Save the latest version of the documents, indexes, and indexing
   * modalities associated with this collection. Note that counting of
   * references to proxy documents and proxy indexes happens here since
   * this class contains (containers of) these references. The counting
   * of references to proxy indexes contained by the IMUs contained 
   * within the IndexingModalities contained by this class are handled 
   * by each IMU
   */    
  public void save() {
    if (proxyRefsCounted == false)
      {
      int numElems = documents.size();
      ProxyDocument doc = null;
      for (int i = 0; i < numElems; i++)
        {
        doc = (ProxyDocument)documents.elementAt(i);
        doc.makePersistent();
        doc.addRefToInMemoryProxiesByOid();
        }
      Enumeration e = indexes.elements();
      while (e.hasMoreElements()) 
	{
        ProxyIndex nextIndex = (ProxyIndex)e.nextElement();
        nextIndex.makePersistent();
        nextIndex.addRefToInMemoryProxiesByOid();
        nextIndex.replaceRealObject();
        }
      proxyRefsCounted = true;
      }

    if (indexingModalities != null) 
      indexingModalities.save();
  }



  /**
   * Displays info on contents of the document collection
   */
  public void present() {
    System.out.println("\nDocCollection:" +
		       "\n               Name: " + name +
		       "\n        Description: " + description +
		       "\nNumber of documents: " + documents.size() +
		       "\n          Modified?: " + modified);

    Enumeration e = indexes.elements();
    while (e.hasMoreElements()) {
      ProxyIndex proxyIndex = (ProxyIndex)e.nextElement();
      if (proxyIndex == null)
	System.out.println("proxyIndex == null");
      else {
	System.out.println(proxyIndex.getName());
	proxyIndex.presentStatistics();
      }
    }

    if (indexingModalities == null)
      System.out.println("Indexing modalities: none");
    else
      indexingModalities.present();
  }



  /**
   * Cleans up table of references to proxies prior to garbage collection
   *
   * If direct references to proxies or Java containers of references
   * have been registered, then unregister them.
   */
  protected void finalize() {
    if (proxyRefsCounted == true)
      {
      // References to proxy IR documents
      int documentsSize = documents.size();
      for (int i = 0; i < documentsSize; i++)
	((VirtualProxy)documents.elementAt(i)).
	  deleteRefFromInMemoryProxiesByOid();

      // References to proxy indexes from the dictionary of indexes
      Enumeration indexesEnum = indexes.elements();
      while (indexesEnum.hasMoreElements())
	((VirtualProxy)indexesEnum.nextElement()).
	  deleteRefFromInMemoryProxiesByOid();
      }
  }
}

