package gov.nist.nlpir.irf.document;

import gov.nist.nlpir.irf.modality.IndexingModalities;
import gov.nist.nlpir.irf.modality.IndexingModalityUnit;
import gov.nist.nlpir.irf.modality.RetrievalModalities;

import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.index.ProxyIndex;
import gov.nist.nlpir.irf.result.RetrievalResult;
import gov.nist.nlpir.irf.result.ResultForDocMatchingQuery;
import gov.nist.nlpir.irf.result.ResultForDocMatchingQueryModalityUnit;
import gov.nist.nlpir.irf.query.QueryRep;
import gov.nist.nlpir.irf.query.QueryModalityUnit;
import gov.nist.nlpir.irf.indexer.Indexer;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.query.Combinator;

import gov.nist.nlpir.irf.proxy.VirtualProxy;

import java.util.Vector;
import java.lang.reflect.Field;

/**
 * Parent class of all document classes.  To create a new type of
 * document, subclass this and create instance variables for
 * the fields you want to support in the document.  There is an
 * example of this in gov.nist.nlpir.irfapps.hci.HciDoc. (gov.nist.nlpir.irfapps.hci.HciDoc.java)
 *
 * @version $Revision: 1.6 $
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 *
 * @see gov.nist.nlpir.irf.index.Index
 * @see gov.nist.nlpir.irf.index.IdxIntern
 * @see gov.nist.nlpir.irfapps.hci.HciDoc
 */

public abstract class Document implements DocumentIntrf {
  
  /** Name of a system-supplied field */
  private static final String UID_String = "serialVersionUID";

  /** Flag indicating if references to contained proxies can be counted */
  private boolean proxyRefsCountedTrue = false;
  


  public void setProxyRefsCountedTrue() {
    proxyRefsCountedTrue = true;
  }
  


  /**
   * Gets feature names of document.
   * @return The collection of feature names
   */
  public Vector getFeatureNames() {  
    Class c = this.getClass();
    Field[] fields = c.getFields();
    int fieldsLength = fields.length;
    Vector col = new Vector(fieldsLength);
    for (int i = 0; i < fieldsLength; i++) {
      if (fields[i].getName().equals(UID_String))
	continue;
      col.addElement(fields[i].getName());
    }
    return col; 
  }
  


  /** 
   * Returns the name of an interface or primitive type
   * @return the name of an interface or primitive type, handling arrays.
   */
  private final static String typeName(Class t) {
    String brackets = "";
    while (t.isArray()) {
      brackets += "[]";
      t = t.getComponentType();
    }
    return t.getName() + brackets;
  }
    


  /**
   * Gets feature types of document. 
   * @return A collection of feature types as Strings.
   */
  public Vector getFeatureTypes() {
    Class c = this.getClass();
    Field[] fields = c.getFields();
    int fieldLength = fields.length;
    Vector col = new Vector(fieldLength);
    for (int i = 0; i < fieldLength; i++) {
      if (fields[i].getName().equals(UID_String))
	continue;
      col.addElement(typeName(fields[i].getType()));
    }
    return col; 
  }
  


  /**
   * Gets the values of the indexable features. Features can be indexed
   * if they are declared public.
   *
   * @return A collection of Data Elements (DataElem)
   */
  public Vector getFeatureValues() {
    Class c = this.getClass();
    Field[] fields = c.getFields();
    int fieldLength = fields.length;
    Vector col = new Vector(fieldLength);
    for (int i = 0; i < fieldLength; i++) {
      if (fields[i].getName().equals(UID_String))
	continue;
      try {
	col.addElement(fields[i].get(this));
      } catch (IllegalAccessException error) {
	System.err.print("Warning [" + error + "], the field " +
			 fields[i].getName() + " is private, it " +
			 "can't be indexed unless it's public."); 
      }
    }
    return col;
  }
    


  /**
   * Displays document in ascii form to screen
   */
  public void present() {     
    Vector names = getFeatureNames();
    Vector types = getFeatureTypes();
    Vector valuecol = getFeatureValues();
    
    int number = names.size();
    
    System.out.println("* Displaying a " + this.getClass().getName() + ": ");
    for (int i = 0; i < number; i++) {
      System.out.print(names.elementAt(i) + "(" + types.elementAt(i) + "): ");
      DataElem currentDE = (DataElem)valuecol.elementAt(i);
      if (currentDE == null)
	System.out.print("None");
      else
	currentDE.present();
      System.out.println();
    }
  }
    



  /**
   * Displays document on one line to standard out
   */
  public void presentBrief() { 
    Vector valuecol = getFeatureValues();
    int valueNumber = valuecol.size();
    for (int i = 0; i < valueNumber; i++) {
      DataElem currentValue = (DataElem)valuecol.elementAt(i);
      if (currentValue != null) {
	currentValue.presentBrief();
	System.out.print(", ");
      }
    }
    System.out.println();
  }
    


  /**
   * Indexes this document.
   *
   * @param im information to control indexing
   * @param proxyDoc proxy IR document needed for the indexing features
   */	
  public void index(IndexingModalities im, ProxyDocument proxyDoc) {
    Vector features = getIndexingFeatures(im, proxyDoc);
    Vector indexingModalityUnits = im.getModalities();
    int numberOfModalityUnits = indexingModalityUnits.size();
    
    for (int i = 0; i < numberOfModalityUnits; i++) {
      IndexingModalityUnit anIMU = 
	(IndexingModalityUnit)indexingModalityUnits.elementAt(i);
      ProxyIndex index = anIMU.getIndex();   // using modalities
      Vector indexingFeatures = (Vector)features.elementAt(i);
      // Create index entries in relevant index
      Indexer UIndexer = anIMU.getIndexer();	

      if (indexingFeatures != null)
	if (index != null && UIndexer != null)
	  index.addIndexingFeatures(indexingFeatures, UIndexer);
	else
	  Support.errorMessage("Document::index(): something wrong indexing " +
			       "an Document:\nindex = " + index +
			       "\nindexingFeatures = " + indexingFeatures +
			       "\nUIndexer = " + UIndexer + "\n");
    }
  }
    


  /**
   * Retrieves relevant documents to this one, as defined by the passed
   * retreival modalities
   *
   * @param rm Information controling retrieval
   * @param proxyDoc A proxy IR document needed for the results built
   * @return The retrieval result for the given modalities.
   */
  public RetrievalResult retrieve(RetrievalModalities rm,
					ProxyDocument proxyDoc) { 
    IndexingModalities im = rm.getIndexingModalities(); // temporary solution
    Vector allIndexingFeatures = getIndexingFeatures(im, proxyDoc);
    Vector imu = im.getModalities();
    int featureNumber = allIndexingFeatures.size();
    int imuNumber = imu.size();
    
    Combinator QStruct = rm.getCombinatorStructure(); 
    QueryRep query = new QueryRep(imu, QStruct);
    
    // Retrieve and get partial results, i.e. results for documents matching
    // the indexing features derived from a query attribute and fullfilling
    // the conditions specified in the attribute's query modality unit.
    Vector partialResults = new Vector();
    for (int i = 0; i < imuNumber; i++) {
      IndexingModalityUnit anIMU = (IndexingModalityUnit)imu.elementAt(i);
      ProxyIndex index = anIMU.getIndex();
      Vector indexingFeatures = (Vector)allIndexingFeatures.elementAt(i);
      if (indexingFeatures != null) {
	QueryModalityUnit qmu = query.getQueryModalityUnit(i);
	
	if (indexingFeatures.size() > 0) {
	  Vector tmp = index.retrieve(qmu, indexingFeatures);
	  
	  // add partial results to previous partial results
	  if (tmp != null) {
	    int PR_Number = tmp.size();
	    for (int k = 0; k < PR_Number; k++)
	      partialResults.addElement(tmp.elementAt(k));
	  }
	}
      }
    }
    
    // Combine the different partial document results
    Vector combinedResults = 
      makeResultsForDocMatchingQuery(query, partialResults);
    
    // Derive final result, an ordered collection of 
    // ResultForDocMatchingQuery objects sorted by their RSV values
    RetrievalResult finalResult = 
      new RetrievalResult(combinedResults);
    finalResult.sort();
    return finalResult;
  }
  

  
  /**
   * Returns indexing features as a Vector of Vectors: one inner
   * Vector for each modality unit (i.e., document field, index).
   *
   * @param im Information controling indexing
   * @param proxyDoc A proxy IR document need to build indexing features
   * @return Vector of vectors of indexing features from this document
   */
  public Vector getIndexingFeatures(IndexingModalities im, 
				    ProxyDocument proxyDoc) {
    Vector indexingModalityUnits = im.getModalities();
    int nbIMU = indexingModalityUnits.size();
    Vector toBeReturned = new Vector(nbIMU);
    
    for (int i = 0; i < nbIMU; i++) {
      IndexingModalityUnit anIMU = 
	(IndexingModalityUnit)indexingModalityUnits.elementAt(i);
      // Get the value of Doc attribute for the feature name of this IMU:
      DataElem aDE = anIMU.getFeatureValue(this);
      if (aDE != null) {
	// Get the indexing features for that DE:
	Vector indexingFeatures = 
	  anIMU.getIndexer().getIndexingFeatures(aDE, proxyDoc);
	toBeReturned.addElement(indexingFeatures);
      } else
	toBeReturned.addElement(null);
    }
    return toBeReturned;
  }
    


  /**
   * Gets indexing modalities for this document.
   * Should be overriden by subclass.
   * @return null always.
   */
  public IndexingModalities getIndexingModalities() { 
    return null;
  }
    


  /** 
   * Adds HTML to the given string buffer.
   *
   * @param baseString The StringBuffer the HTML text will be appended to.
   * @param unused Unused.
   * @param form Boolean set to true if you wish HTML document to be
   * a form document.
   */
  public void addHTML(StringBuffer baseString, int unused, boolean form) {
    baseString.append("<HTML>\n<TITLE>IRF Document</TITLE>\n" +
		      "<!-- Automatically generated by IRF application>\n" +
		      "<body bgcolor=\"#ffffff\">\n");
    Vector featureValues = getFeatureValues();
    Vector featureNames = getFeatureNames();
    
    if (form)
      baseString.append("<FORM METHOD=\"POST\" ACTION=\"http://www.nist-nlpir.gov:4096/cgi-bin/client\">" +
			"<FONT SIZE = \"+2\"> Query<p><HR>");
    else // not dealing with a form
      baseString.append("<FONT SIZE = \"+2\"> Displaying Document<p><HR>\n");
    
    int valueNumber = featureValues.size();
    for (int i = 0; i < valueNumber; i++) {
      DataElem feature = (DataElem)featureValues.elementAt(i);
      String fname = (String)featureNames.elementAt(i);
      baseString.append("<FONT SIZE = \"0\"><tt><strong>[" + fname +
			":]</strong></tt><FONT SIZE \"+0\"><br>");
      if (feature != null)
	feature.addHTML(baseString, 0, form);
      else
	baseString.append("NULL");
      baseString.append("<P>\n");
    }
    if (form == true) {
      baseString.append("<pre><INPUT TYPE=\"submit\" VALUE=\"Submit query\"" +
			"></pre><pre><INPUT TYPE=\"reset\" VALUE=\" New query \"></pre></FORM>");
    }
    baseString.append("<HR><IMG SRC=\"Icons/fire2.gif\" valign=top align=right alt=\"FIRE Logo\">\n");
  }
  


  /**
   * Makes a complete result for documents matching the supplied
   * query from the partial results supplied.
   *
   * @param query The query to match.
   * @param collPR A collection of partial Results.
   * @return A Vector of ResultForDocMatchingQuery. There is one of them by 
   * Document in the results, containing every Result..ModalityUnit linked 
   * to this document.
   */
  public Vector makeResultsForDocMatchingQuery(QueryRep query,
					       Vector collPartialResults) { 
    Vector combinedResults = new Vector();
    boolean found = false;
    int PR_Number = collPartialResults.size();
    for (int i = 0; i < PR_Number; i++) {
      ResultForDocMatchingQueryModalityUnit aPartialResult =
	(ResultForDocMatchingQueryModalityUnit)collPartialResults.elementAt(i);
      ProxyDocument doc1 = aPartialResult.getDocument();
      int CR_Number = combinedResults.size();
      for (int j = 0; j < CR_Number; j++) {
	ResultForDocMatchingQuery aCombinedResult =
	  (ResultForDocMatchingQuery)combinedResults.elementAt(j);
	ProxyDocument doc2 = aCombinedResult.getDocument();
	if (doc1 == doc2) {
	  // There is already a RFDMQuery for this document, let's add
	  // the current RFDMQModalityUnit to it.
	  aCombinedResult.addPartialResult(aPartialResult);
	  found = true;
	  break;
	}
      }
      if (found)
	found = false;
      else {
	// There was no RFDMQuery for this document in the result Vector, let's
	// add it initialised with the current RFDMQModalityUnit.
	Vector Coll = new Vector(1);
	Coll.addElement(aPartialResult);
	ResultForDocMatchingQuery aCombinedResult =
	  new ResultForDocMatchingQuery(query, doc1, Coll);
	combinedResults.addElement(aCombinedResult);
      }
    }

    //Calculate RSVs for the combined results once they have been fully assembled
    StringBuffer indent = new StringBuffer();
    int resultSize = combinedResults.size();
    for (int x = 0; x < resultSize; x++) {
      ResultForDocMatchingQuery cr =
	(ResultForDocMatchingQuery)combinedResults.elementAt(x);
      cr.calcRSV(query);
    }
    return combinedResults;
  }
  


  /**
   * Completes the recursion of getFirstInstance() calls
   *
   * @see VirtualProxy#getFirstInstance
   */
  public void getFirstInstances() {
    Class currentClass = getClass();
    Field[] currentFields = currentClass.getFields();
    for (int i = 0; i < currentFields.length; i++) {
      Field currentField = currentFields[i];
      try {
	Object field4This = currentField.get(this);
	if (field4This instanceof VirtualProxy)
	  currentField.set(this, ((VirtualProxy)field4This).getFirstInstance());
      } catch (IllegalAccessException iae) {
	System.err.println("Document.getFirstInstances(): " + iae +
			   "\nThe value could then not be set.");
      }
    }
  }



  /**
   * Build an IR document suitable for use as a query
   * @retunr An IR document suitable for use as a query
   */
  public abstract Document getQueryDocument();
  


  /**
   * Updates the counts of references to proxies for the contained
   * proxy data elements before this document is garbage collected
   */
  public void finalize() {
    if (proxyRefsCountedTrue) {
      Class currentClass = getClass();
      Field[] currentFields = currentClass.getFields();
      for (int i = 0; i < currentFields.length; i++) {
	Field currentField = currentFields[i];
	if (currentField.getName().equals(UID_String))
	  continue;
	try {
	  Object fieldForThis = currentField.get(this);
	  if (fieldForThis != null)
	    if (fieldForThis instanceof VirtualProxy)
	      ((VirtualProxy)fieldForThis).deleteRefFromInMemoryProxiesByOid();
	} catch (IllegalAccessException iae) {
	  System.err.println("Problem in Document.finalize()");
	}
      }
    }
  }
}
