package gov.nist.nlpir.irf.document;

import gov.nist.nlpir.irf.modality.RetrievalModalities;
import gov.nist.nlpir.irf.modality.IndexingModalities;

import gov.nist.nlpir.irf.result.RetrievalResult;
import gov.nist.nlpir.irf.query.QueryRep;
import gov.nist.nlpir.irf.proxy.VirtualProxy;

import java.util.Vector;

/**
 * Parent class of all IR document proxies.<br><br>  
 *
 * @version $Revision: 1.5 $
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 *
 */

public abstract class ProxyDocument extends VirtualProxy implements DocumentIntrf {

    
  // see API documentation on Reflection for more information on the
  // four methods below
    
  /** Gets feature names of document.
   * @return The collection of feature names
   */
  public Vector getFeatureNames() { 
    return ((Document)getRealObject()).getFeatureNames(); 
  }
    


  /**
   * Gets feature types of document. 
   * @return A collection of feature types as Strings.
   */
  public Vector getFeatureTypes() {
    return ((Document)getRealObject()).getFeatureTypes();
  }
    


  /**
   * Gets the values of the indexable features. Features can be indexed
   * if they are declared public.
   *
   * @return a collection of Data Elements (DataElem)
   */
  public Vector getFeatureValues() {
    // this should be defined to throws IllegalAccessException only
    // when none of the instance variables are visible.  (this may
    // already be true)
    return ((Document)getRealObject()).getFeatureValues();
  }
    


  /**
   * Displays document in ascii form to screen
   */
  public void present() {  
    ((Document)getRealObject()).present();
  }
   

 
  /**
   * Displays document on one line to screen
   */
  public void presentBrief() { 
    ((Document)getRealObject()).presentBrief();
  }
    


  /**
   * Indexes this document.
   *
   * @param im Indexing Modality list .
   */	
  public void index(IndexingModalities im, ProxyDocument proxyDoc) {
    ((Document)getRealObject()).index(im, proxyDoc);
  }


    
  /**
   * Returns this document as a query.
   *
   * @param rm Retrieval Modalities
   * @return The retrieval result for the given modalities.
   */
  public RetrievalResult retrieve(RetrievalModalities rm,
					ProxyDocument proxyDoc) { 
    return ((Document)getRealObject()).retrieve(rm, proxyDoc);      
    }
    
    

  /**
   * Returns indexing features as a Vector of Vectors: one inner
   * Vector for each modality unit.
   *
   * @param im Indexing Modalities
   * @return Vector containing vectors of indexing features
   */
  public Vector getIndexingFeatures(IndexingModalities im,
				    ProxyDocument proxyDoc) {
    return ((Document)getRealObject()).getIndexingFeatures(im, 
							      proxyDoc);     
  }
    


  /**
   * Gets indexing modalities for this document.
   * should be overriden by subclass.
   * @return null always.
   */
  public IndexingModalities getIndexingModalities() { 
    return ((Document)getRealObject()).getIndexingModalities();      
  }
    

    
  /** 
   * Adds HTML to the given string buffer.
   *
   * @param baseString The StringBuffer the HTML text will be appended to.
   * @param unused Unused.
   * @param form Boolean set to true if you wish HTML document to be
   * a form document.
   */
  public void addHTML(StringBuffer baseString, int unused, boolean form) {
    ((Document)getRealObject()).addHTML(baseString, unused, form);
  }
    


  /**
   * Makes a complete result for documents matching the supplied
   * query from the partial results supplied.
   *
   * @param query The  query to match.
   * @param collPR A collection of partial Results.
   * @return A Vector of ResultForDocMatchingQuery. There is one of them by Document
   * in the results, containing every Result..ModalityUnit linked to this document.
   */
  public Vector makeResultsForDocMatchingQuery(QueryRep query, Vector collPartialResults) { 
    
    Vector combinedResults =
      ((Document)getRealObject()).
      makeResultsForDocMatchingQuery(query,collPartialResults);
    return combinedResults;
  }


  
  /** Passes the message on to each component object which is
   *  a proxy - telling it to make itself persistent.
   *  Then make this object persistent and return its handle.
   */
  public void makePersistent() {  
    if (objectIdentifier == null) { // Only needed if real object not yet persistent
      Vector fields = getFeatureValues();
      int numberOfFields = fields.size();
      for (int i = 0; i < numberOfFields; i++) {
	VirtualProxy currentField = (VirtualProxy)fields.elementAt(i);
	if (currentField != null) {
	  currentField.makePersistent();
	  currentField.addRefToInMemoryProxiesByOid();
	}
      }
      ((Document)getRealObject()).setProxyRefsCountedTrue();
      super.makePersistent(); // Use the method in VirtualProxy
    }
  }
  


  /**
   * Returns a hashcode for an Document
   *
   * @return the hashcode for the Document
   */
  public int hashCode() {
    return getRealObject().hashCode();
  }



  /**
   * Tests whether two Documents have the same content
   *
   * @return the hashcode for the Document
   */
  public boolean equals(Object o) {
    if (o instanceof ProxyDocument)
      return ((Document)getRealObject()).equals(((ProxyDocument)o).getRealObject());
    else if (o instanceof Document)
      return ((Document)getRealObject()).equals(o);
    System.err.println("ProxyDocument.equals(): was expecting a " +
		       this.getClass().getName() + ", received a " +
		       o.getClass().getName() + ": " + o);
    return false;
  }



  /**
   * Enforce uniqueness of proxies by checking to see if a proxy for
   * this IR document already exists and returning that reference if
   * it exists; and see that analogous processing is carried out for
   * the contained proxy data elements.
   * <br><br>
   * Document potentially contains Proxies, so that its proxy if forced 
   * to override the default method in VirtualProxy to call getFirstInstance 
   * on all its members. This proxy must thus override the default method 
   * to call this cleaning function in Document.
   *
   * @return the reference to any pre-existing proxy for this one's real
   * object; if none pre-exists, return a reference to self.
   */
  public VirtualProxy getFirstInstance() {
    ProxyDocument definitive = (ProxyDocument)super.getFirstInstance();
    Document RO = (Document)definitive.getRealObject();
    RO.getFirstInstances();
    RO.setProxyRefsCountedTrue();
    return definitive;
  }



  /**
   * Creates an IR document that can be used as a query.
   *
   * @return a proxy Document suitable for use as a query
   */
  public abstract ProxyDocument getQueryDocument();



  /**
   * Makes this proxy lightweight (cut the link to its real object) 
   * and sees that the same is done for the contained proxy data
   * elements.
   */
  public void makeLightweight() {  
    if (!isLightweight()) {
      Vector fields = getFeatureValues();
      int numberOfFields = fields.size();
      for (int i = 0; i < numberOfFields; i++) {
	VirtualProxy currentField = (VirtualProxy)fields.elementAt(i);
	if (currentField != null) {
	  currentField.makeLightweight();
	}
      }
      super.makeLightweight();
    }
  }
}
