package gov.nist.nlpir.irf.feature;

import gov.nist.nlpir.irf.document.ProxyDocument;

import gov.nist.nlpir.irf.de.ProxyDataElem;
import gov.nist.nlpir.irf.proxy.VirtualProxy;

/**
 * Container for elements of an Indexing Feature (IF): the feature
 * itself (e.g., a term in a text document), the source document
 * that it was generated from and the field within the source from
 * which it was extracted. 
 *     
 * @version $Revision: 1.1 $
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 */

public class IndexingFeature {

  /** The feature itself, e.g., a term from a text document */
  private ProxyDataElem    feature;

  /** The document from which the feature was extracted */
  private ProxyDocument sourceDoc;

  /** The document field from which the feature was extracted */
  private ProxyDataElem    sourceField;

  /** The boolean to manage the inMemoryProxiesByOid table. 
   * True indicates that references to contained proxies can and
   * must be counted
   */
  private boolean proxyRefsCountedTrue = false;
    


  /**
   * Constructs a new indexing feature
   * @param f feature to be stored
   * @param s source document to be stored
   */
  public IndexingFeature(ProxyDataElem feature, ProxyDocument sourceDoc,
			 ProxyDataElem sourceField) {
    this.feature = feature;
    this.sourceDoc = sourceDoc;
    this.sourceField = sourceField;
  }


  /** 
   * Gets the IF's stored feature
   * @return feature reference
   */
  public final ProxyDataElem getFeature() {
    return feature;
  }



  /**
   * Gets this IF's source document
   * @return source reference
   */ 
  public final ProxyDocument getSourceDoc() {
    return sourceDoc;
  }



  /**
   * Sets this IF's feature  
   * @param f feature to be stored
   */
  public final void setFeature(ProxyDataElem feature) { 
    this.feature = feature;
  }



  /**
   * Sets this IF's source document 
   * @param s source document to be stored
   */

  public final void setSourceDoc(ProxyDocument sourceDoc) {
    this.sourceDoc = sourceDoc;
  }



  /**
   * Gets the field this IF was extracted from.
   * @return the field this IF was extracted from
   */
  public final ProxyDataElem getSourceField() {
    return sourceField;
  }



  /**
   * Sets the document field this IF was found in.
   */
  public final void setSourceField(ProxyDataElem sourceField) {
    this.sourceField = sourceField;
  }



  /**
   * Returns a string representation of indexing feature showing
   * feature and source.
   * @return string representation of indexing feature
   */
  public String toString() {
    return ("{IndexingFeature: feature = {" + feature +
	    "}, source = {" + sourceDoc + "}}");
  }



  
  /**
   * This method must be called if the IndexingFeature has registered its
   * references to its different parts.
   */
  public final void setProxyRefsCountedTrue() {
    proxyRefsCountedTrue = true;
  }
  


  /**
   * This method is mandatory to manage the inMemoryProxiesByOid table
   * in the case those references were counted for this object.
   */
  public void finalize() {
    if (proxyRefsCountedTrue) {
      feature.deleteRefFromInMemoryProxiesByOid();
      sourceDoc.deleteRefFromInMemoryProxiesByOid();
      sourceField.deleteRefFromInMemoryProxiesByOid();
    }
  }
}
