package gov.nist.nlpir.irf.feature;

import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.de.ProxyDataElem;
import gov.nist.nlpir.irf.feature.IndexingFeature;
import gov.nist.nlpir.irf.proxy.VirtualProxy;
import gov.nist.nlpir.irf.broker.PersistenceBroker;
import gov.nist.nlpir.irf.broker.braf.IdxFeatBrafBroker;

/**
 * Proxy class for Indexing Features.
 *
 * @version $Revision: 1.2 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated with
 * its operation, modification and maintenance.
 */

public class ProxyIndexingFeature extends VirtualProxy{

  /**
   * @param feature Feature to be stored.
   * @param sourceDoc Source document to be stored.
   * @param sourceField Field in the document from where this feature comes.
   */
  public ProxyIndexingFeature(ProxyDataElem feature, ProxyDocument sourceDoc,
			      ProxyDataElem sourceField) {
    setRealObject(new IndexingFeature(feature, sourceDoc, sourceField));
  }

  /**
   * No argument constructor for deserialization purpose.
   */
  public ProxyIndexingFeature() {
  }
  
  /**
   * @see gov.nist.nlpir.irf.feature.IndexingFeature#getFeature
   */
  public final ProxyDataElem getFeature() {
    return ((IndexingFeature)getRealObject()).getFeature();
  }
  
  /**
   * @see gov.nist.nlpir.irf.feature.IndexingFeature#getSourceDoc
   */
  public final ProxyDocument getSourceDoc() {
    return ((IndexingFeature)getRealObject()).getSourceDoc();
  }

  /**
   * @see gov.nist.nlpir.irf.feature.IndexingFeature#getSourceField
   */
  public final ProxyDataElem getSourceField() {
    return ((IndexingFeature)getRealObject()).getSourceField();
  }

  /**
   * @see gov.nist.nlpir.irf.feature.IndexingFeature#setFeature
   */
  public void setFeature(ProxyDataElem feature) {
    if (objectIdentifier != null) {
      ProxyDataElem current = getFeature();
      if (current != null)
	current.deleteRefFromInMemoryProxiesByOid();
      if (feature != null)
	feature.addRefToInMemoryProxiesByOid();
    }
    ((IndexingFeature)getRealObject()).setFeature(feature);
  }

  /**
   * @see gov.nist.nlpir.irf.feature.IndexingFeature#setSourceDoc
   */
  public void setSourceDoc(ProxyDocument sourceDoc) {
    if (objectIdentifier != null) {
      ProxyDocument current = getSourceDoc();
      if (current != null)
	current.deleteRefFromInMemoryProxiesByOid();
      if (sourceDoc != null)
	sourceDoc.addRefToInMemoryProxiesByOid();
    }
    ((IndexingFeature)getRealObject()).setSourceDoc(sourceDoc);
  }
  
  /**
   * @see gov.nist.nlpir.irf.feature.IndexingFeature#setSourceField
   */
  public void setSourceField(ProxyDataElem sourceField) {
    if (objectIdentifier != null) {
      ProxyDataElem current = getSourceField();
      if (current != null)
	current.deleteRefFromInMemoryProxiesByOid();
      if (sourceField != null)
	sourceField.addRefToInMemoryProxiesByOid();
    }
    ((IndexingFeature)getRealObject()).setSourceField(sourceField);
  }

  /**
   * Returns string representation of indexing feature showing
   * feature and source, preceded by <code>P </code>.
   * @return string representation of indexing feature
   */
  public String toString() {
    return "P " + getRealObject().toString();
  }
  

  /**
   * Override VirtualProxy's method with the same name. Passes this message
   * on to component proxies in the real IndexingFeature's source and
   * feature before calling the VirtualProxy's method to make this
   * ProxyIndexingFeature persistent.
   */          
  public void makePersistent() {
    if (objectIdentifier == null) { // Only needed if real object not yet persistent
      ProxyDocument sourceDoc = getSourceDoc();
      if (sourceDoc != null) {
	sourceDoc.makePersistent();
	sourceDoc.addRefToInMemoryProxiesByOid();
      }
      ProxyDataElem feature = getFeature();
      if (feature != null) {
	feature.makePersistent();
	feature.addRefToInMemoryProxiesByOid();
      }
      ProxyDataElem sourceField = getSourceField();
      if (sourceField != null) {
	sourceField.makePersistent();
	sourceField.addRefToInMemoryProxiesByOid();
      }
      ((IndexingFeature)getRealObject()).setProxyRefsCountedTrue();
      super.makePersistent();        // Makes this ProxyIndexingFeature persistent
    }
  }

  /**
   * The concrete proxy will override this method to get the broker
   * for the proxy object @return the broker for this object 
   */
  public PersistenceBroker createBroker() {
    /* alternative for this is to have one broker stored in the InfoServer
     * and retrieved in this method */
    return IdxFeatBrafBroker.instance();
  }
}
