package gov.nist.nlpir.irf.handle.braf;

import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;
import gov.nist.nlpir.irf.handle.Handle;
import gov.nist.nlpir.irf.irfman.IrfGlobal;
import gov.nist.nlpir.irf.util.Support;

import java.util.Vector;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;

/**
 * This class models everything common to handles used by
 * buffered random access file brokers.
 * <br><br>
 * A handle encapsulates information created and used solely
 * by a broker to manage some object in persistent storage
 *
 * @version $Revision: 1.2 $
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 *
 */

public abstract class BufferedRandomAccessFileHandle extends Handle {

  /** The name of the file containing the object to which this handle 
   * refers */
  private String file = (String)fileNames.elementAt(getFileCode());


  /** Byte offset of the stored object within the above file */
  private long startOfRecord  = 0;


  /** Byte length of the stored object */
  private int lengthOfRecord = 0;


  /** 
   * This vector contains the name of the different files that handles may 
   * be concerned with (i.e., the objects that handles are for are stored
   * in  those files) as soon as at least one handle is created. If a file 
   * "fileNames" containing this vector serialized exists, it is retrieved 
   * from there, otherwise it is created at this point and every new 
   * BRAF_Handle class appearing registers itself in this Vector thanks to 
   * the getCode() method. */
  public static Vector fileNames;


  /** This boolean will be false when the above Vector is read from disk. */
  private static boolean notReadFromDiskYet = true;


  /** The fully qualified name of the file where real objects are stored. */ 
  private static String fileNamesFileName = "DBfileNames";


  /** 
   * Static initializer method - gets path to brokered storage
   * from the InfoServer and prepends it to the constant file
   * name "DBfileNames" to form the name of the file which
   * contains the names of files used to hold brokered objects
   */
  static {
    // Find out the path (directory) for brokered storage
    String path = 
    IrfGlobal.gIrfManager.getInfoServer().getLocationOfBrokeredStorage();
    if ( path == null)
      path = "";
    else
      {
      // add file separator if needed
      if ( path.length() > 0 && !path.endsWith(Support.fileSeparator) )
	path = path + Support.fileSeparator;
      }
    fileNamesFileName = path + "DBfileNames";
  }



  /** Every subclass will override this static variable, initializing it 
   * with a call to getCode(). For example, the handle class for objects 
   * stored in a file named "ine" should contain the line
   * <pre>static int fileCode = getCode("ine");</pre> */
  protected abstract int getFileCode();



  /**
   * Returns the name of the file specified by this handle
   * @return the name of the file specified by this handle
   */
  public final String getFile() {
    return file;
  }



  /**
   * Sets the name of the file specified by this handle
   */
  public final void setFile(String fileName) {
    file = fileName;
  }



  /**
   * Returns the byte offset of the stored object
   * @return the byte offset of the stored object
   */
  public final long getStartOfRecord() {
    return startOfRecord;
  }



  /**
   * Sets the byte offset of the stored object
    @parm start the byte offset of the stored object
   */
  public final void setStartOfRecord(long start) {
    startOfRecord = start;
  }



  /**
   * Returns the length of the stored object
   * @return the length of the stored object
   */
  public final int getLengthOfRecord() {
    return lengthOfRecord;
  }



  /**
   * Sets the length of the stored object
   * @parm len the length of the store object
   */
  public final void setLengthOfRecord(int len) {
    lengthOfRecord = len;
  }
    


  /**
   * Sets the position of the object
   * @parm start byte offset of the stored object
   * @parm len byte length of the stored object
   */
  public final void setPositionOfRecord(long start, int len) {
    startOfRecord = start;
    lengthOfRecord = len;
  }





  /**
   * Returns a unique code for a file with the specified name.
   * As it is completely static, it shouldn't take too long to execute and 
   * it will save a lot a time when handles write themselves to disk because 
   * they will only write their file code instead of the complete name of 
   * the file.
   * @param fileName The name of the file where the objects referred to by 
   * this handle class will be stored.
   * @return A unique code for this fileName.
   */
  protected final static int getCode(String fileName) {
    if (notReadFromDiskYet)
      try {
	FileInputStream istream = new FileInputStream(fileNamesFileName);
	ObjectInputStream p = new ObjectInputStream(istream);
	fileNames = (Vector)p.readObject();
	p.close();
	istream.close();
	notReadFromDiskYet = false;
      } catch (FileNotFoundException fnfe) {
	fileNames = new Vector();
      } catch (ClassNotFoundException cnfe) {
	System.err.println("BRAF_Handle.getCode: Vector Class not found????");
	fileNames = new Vector();
      } catch (IOException ioe) {
	System.err.println("BRAF_Handle.getCode:" + ioe);
	fileNames = new Vector();
      }
    int result = fileNames.indexOf(fileName);
    if (result != -1)
      return result;
    fileNames.addElement(fileName);

    /* We are forced to write the Vector to a file again because it 
     * changed. The actual point to do it should in fact be the moment 
     * the class disappears but this kind of classFinalize() method doesn't 
     * exist. It's a little to pay to be able to get back the codes at next 
     * run. */
    try {
      FileOutputStream ostream = new FileOutputStream(fileNamesFileName);
      ObjectOutputStream p = new ObjectOutputStream(ostream);
      p.writeObject(fileNames);
      p.close();
      ostream.close();
    } catch (IOException ioe) {
      System.err.println("BRAF_Handle.getCode: Couldn't save the Vector...");
    }

    return fileNames.size() - 1;
  }



  /**
   * Writes an external representation of this object
   * @param out buffered random access file handle
   */
  public final void writeExt(BufferedRandomAccessFile out) {
    try {
      out.writeByte(getFileCode());
      out.writeLong(startOfRecord);
      out.writeLong((long)lengthOfRecord);
    } catch (java.io.IOException e) {
      System.err.println("Error writing BufferedRandomAccessFileHandle: "+e);
    }
  }



  /**
   * Writes the Handle in the given file in lieu of the already existing one,
   * if possible. It is impossible if the size of the new handle 
   * to write is bigger than the one of the former handle. In this case, the 
   * new handle is not written and the method returns false.
   * @param out The file where the new handle is supposed to be written.
   * @return <code>true</code> if the replacement occured correctly,<br>
   * <code>false</code> otherwise, which means writeExt must be called a a 
   * differentplace in the file.
   */
  public final boolean replaceExt(BufferedRandomAccessFile braf) {
    try {
      int formerFileCode = braf.readByte();
      if (formerFileCode == getFileCode()) {
	braf.writeLong(startOfRecord);
	braf.writeLong((long)lengthOfRecord);
	return true;
      } else
	return false;
    } catch (java.io.IOException e) {
      System.err.println("Error writing BufferedRandomAccessFileHandle: " + e);
    }
    return false;
  }




  /**
   * Read an external representation of this object
   * @param in buffered random access file handle
   */
  public final void readExt(BufferedRandomAccessFile in) {
    try {
      file = (String)fileNames.elementAt(in.readByte());
      startOfRecord = in.readLong();
      lengthOfRecord = (int)in.readLong();
    } catch (java.io.IOException e) {
      System.err.println("Error reading BufferedRandomAccessFileHandle: " + e);
    }
  }



  /**
   * Returns a representation of this object as a string.
   * @return a string representation of this object
   */
  public String toString() {
    return getClass().getName() + ">" + file + ":" + startOfRecord + ":" + lengthOfRecord + " ";
  }
}
