package gov.nist.nlpir.irf.index;

import gov.nist.nlpir.irf.index.braf.ProxyFeatureList;

import java.util.Enumeration;
import java.util.Vector;

/**
 * This class is the "heart" of the Index. It mainly contains two hash tables of
 * vectors. The first one, called valuesBySource, stores all indexing features
 * found, classified by source. It will allow user to retrieve all the features
 * of a given document, for example. The second one, called sourcesByValue, stores
 * the different values whatever their source is. It will allow user to retrieve
 * all the sources where a given feature appear.<br>
 * Each entry in a table corresponds to a vector of features. For the first table,
 * the vector contains all the features found for the source entry. In the second
 * table, one vector contains all the sources in which the entry feature can be
 * found.
 * <p>
 * An example of using the class is below
 * <pre>
 * setIndexingFeature = new DualKeyContainer(); 
 * 
 * // Add an element to DualKeyContainer using two keys
 * setIndexingFeature.put(source, key, value);
 * // More precise example
 * setIndexingFeature.put("doc1", "word1", "word1InDoc1");
 * setIndexingFeature.put("doc1", "word2", "word2InDoc1");
 * setIndexingFeature.put("doc2", "word1", "word1InDoc2");
 * // We now have:
 * // setIndexingFeatures.getSourceVector("word1") == {"word1InDoc1", "word1InDoc2"}
 * // setIndexingFeatures.getValuesVector("doc1") == {"word1InDoc1", "word2InDoc1"}
 * 
 * </pre>
 * </p>
 * 
 * @version $Revision: 1.2 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 * @see java.util.Hashtable
 */

public class DualKeyContainer {
  
  private IrfHashtable valuesBySource; // of Vectors of Objects.
  // Key: should be the Source where the Object was found, then Objects
  // with same Source will appear in the same Vector.
  
  private IrfHashtable sourcesByValue;   // of Vectors of Objects.
  // Key: should be the Feature of the Object, ie its content for example.
  // This way, Objects with the same content (but not necessarily the same
  // Source) will appear in the same Vector.
  
  private int valuesNumber;       // Number of values stored.
  private int sourcesNumber;      // Size of valuesBySource.
  private int uniqueValuesNumber; // Size of sourcesByValue.
  
  private static int SOURCE_VECTOR_INITIAL_CAPACITY = 30; // Usually a Source contains about 30 features, and then
  private static int SOURCE_VECTOR_CAPACITY_INCREMENT = 10; // the increment of 10 seems reasonnable
  private static int FEATURE_VECTOR_INITIAL_CAPACITY = 2; // Usually a feature appears only once or twice.
  private static int FEATURE_VECTOR_CAPACITY_INCREMENT = 10; // When it appears more, it can be really often.
  
  /**
   * Initializes both tables.
   */
  
  public DualKeyContainer() {
    valuesBySource = new IrfHashtable(1000);
    sourcesByValue = new IrfHashtable(1000);
    valuesNumber = 0;
    sourcesNumber = 0;
    uniqueValuesNumber = 0;
  }
  
  /**
   * Gives the number of sources that appear in the DualKeyContainer.
   *
   * @return The number of keys in the valuesBySource table.
   */
  
  public int valuesBySourceSize() {
    return valuesBySource.size();
  }
  
  /**
   * Gives the total number of features stored in the table.<br>
   *
   * @return The number of values in the table.
   */
  
  public int getValuesNumber() {
    return valuesNumber;
  }
  
  /**
   * Gives the number of features stored for the given source.
   *
   * @param source the source for which the number of values will be computed.
   * @return The number of values,<br>
   * 0 if the source doesn't appear.
   */
  
  public int getNumberOfValuesFor(Object source) {
    return ((Vector)valuesBySource.get(source)).size();
  }
  
  /**
   * Gives the number of sources containing the given feature.
   *
   * @param feature the feature for which the number of sources will be computed.
   * @return The number of sources,<br>
   * 0 if the feature doesn't appear.
   */
  
  public int getNumberOfSourcesFor(Object feature) {
    return ((Vector)sourcesByValue.get(feature)).size();
  }
  
  /**
   * @return True if no key table has been initialized.
   */
  
  public boolean isEmpty() {
    return (valuesBySource.isEmpty() && sourcesByValue.isEmpty());
  }
  
  /**
   * Returns all the features stored.
   *
   * @see DualKeyContainer#put
   */
  
  public Enumeration elements() {
    Vector toBeReturned = new Vector(uniqueValuesNumber);
    Enumeration kEnum = valuesBySource.elements();
    while (kEnum.hasMoreElements()) {
      Enumeration thisVector = ((Vector)kEnum.nextElement()).elements();
      while (thisVector.hasMoreElements())
	toBeReturned.addElement(thisVector.nextElement());
    }
    return toBeReturned.elements();
  }
  
  /**
   * Returns a Vector containing all the values stored in the DualKeyContainer.
   */
  
  public ProxyFeatureList getAllValues() {
    Vector toBeReturned = new Vector(uniqueValuesNumber);
    Enumeration elements = valuesBySource.elements();
    while (elements.hasMoreElements()) {
      Enumeration elements2 = ((Vector)elements.nextElement()).elements();
      while (elements2.hasMoreElements())
	toBeReturned.addElement(elements2.nextElement());
    }
    return new ProxyFeatureList("fl", toBeReturned);
  }
  
  /**
   * Returns the vector of values associated to the given source. The Vector
   * is empty if there is no value associated to source.
   *
   * @param source The source for which the values will be returned.
   * @return A vector of values.
   */
  
  public ProxyFeatureList getFeatureVector(Object source) {
    Vector toBeReturned = (Vector)valuesBySource.get(source);
    if (toBeReturned != null)
      return new ProxyFeatureList("fl", toBeReturned);
    else
      return new ProxyFeatureList();
  }
  
  /**
   * Returns the vector of sources that correspond to the given feature. The Vector
   * is empty if there is no value associated to source.
   *
   * @param feature The feature for which the sources will be returned.
   * @return A vector of values.
   */
  
  public ProxyFeatureList getSourceVector(Object feature) {
    Vector toBeReturned = (Vector)sourcesByValue.get(feature);
    if (toBeReturned != null)
      return new ProxyFeatureList("fl", toBeReturned);
    else
      return new ProxyFeatureList();
  }
  
  /**
   * @param source  Where the value was found.
   * @param feature The feature created with this value.
   * @param value   The object to be stored.
   */
  
  public void put(Object source, Object feature, Object value) {
    Vector vec = (Vector)valuesBySource.get(source);
    if (vec == null) {
      vec = new Vector(SOURCE_VECTOR_INITIAL_CAPACITY, SOURCE_VECTOR_CAPACITY_INCREMENT);
      valuesBySource.put(source, vec);
      sourcesNumber++;
    }
    vec.addElement(value);
    
    vec = (Vector)sourcesByValue.get(feature);
    if (vec == null) {
      vec = new Vector(FEATURE_VECTOR_INITIAL_CAPACITY, FEATURE_VECTOR_CAPACITY_INCREMENT);
      sourcesByValue.put(feature, vec);
      uniqueValuesNumber++;
    }
    vec.addElement(value);
    
    valuesNumber++;
  }
  
  /**
   * Clears both tables.
   */
  
  public void clear() {
    Enumeration kEnum = valuesBySource.elements();
    while (kEnum.hasMoreElements()) {
      ((Vector)kEnum.nextElement()).removeAllElements();
    } 
    valuesBySource.clear();
    kEnum = sourcesByValue.elements();
    while (kEnum.hasMoreElements()) {
      ((Vector)kEnum.nextElement()).removeAllElements();
    } 
    sourcesByValue.clear();
    valuesNumber = 0;
  }
  
  /**
   * Creates massive string representation of the DualKeyContainer.
   */
  
  public String toString() {
    StringBuffer str = new StringBuffer();
    /**** Values by Source *****************************/
    Enumeration keys2 = valuesBySource.keys();
    str.append("Values by source:\n");
    while (keys2.hasMoreElements()) {
      Object key2 = keys2.nextElement();
      str.append(key2 + "\t");
      Vector vec = (Vector)valuesBySource.get(key2);
      int n = vec.size();
      if (n > 0) {
	str.append(vec.elementAt(0));
	for (int i = 1; i < n; i++) 
	  str.append("\n\t\t" + vec.elementAt(i));
      }
      str.append("\n");	
    }
    /**** Unique Values ********************************/
    keys2 = sourcesByValue.keys();
    str.append("Unique Values:\n");
    while (keys2.hasMoreElements()) {
      Object key2 = keys2.nextElement();
      str.append(key2 + "\t");
      Vector vec = (Vector)sourcesByValue.get(key2);
      int n = vec.size();
      if (n > 0) {
	str.append(vec.elementAt(0));
	for (int i = 1; i < n; i++) 
	  str.append("\n\t\t" + vec.elementAt(i));
      }
      str.append("\n");	
    }
    str.append("Number of elements: " + valuesNumber + ".\n");
    return str.toString();
  }
  
  /**
   * Prints statistics about the DualKeyContainer, ie its size and the size of its
   * elements. Keys are also printed if their length is less than 15.
   *
   * @param depth 1, prints the size of the contained vectors for each hash table,<br>
   * 2, gives for each vector its key and the number of its elements,<br>
   * 3, gives the key of each vector and prints its content, each element truncated
   * to 25 characters if longer,<br>
   * 4 is for a special 2, ie only the size of the vectors are printed one behind
   * another.,br>
   * Any other value will result in an empty display.
   */
  
  public void showStatistics(int depth, int maxLengthIfDepth3) {
    if ((depth > 0) && (depth < 5)) {
      System.out.print("DualKeyContainer - Values by Source:");
      showStats(valuesBySource, depth, maxLengthIfDepth3);
      System.out.print("DualKeyContainer - Unique Values:");
      showStats(sourcesByValue, depth, maxLengthIfDepth3);
    }
  }
  
  /**
   * Prints statistics for ONE HVtable.
   *
   * @param table The hashtable of vectors to be presented.
   * @param depth Same as showStatistics() depth.
   */
  
  private void showStats(IrfHashtable table, int depth, int maxLengthIfDepth3) {
    if (depth < 4) {
      if (depth == 1) {
	System.out.println(" Sizes of vectors:");
	Enumeration keys2 = table.keys();
	while (keys2.hasMoreElements()) {
	  Object key2 = keys2.nextElement();
	  System.out.print(" " + ((Vector)table.get(key2)).size());
	}
	System.out.println("");
      } else if (depth == 2) {
	System.out.println("");
	Enumeration keys2 = table.keys();
	while (keys2.hasMoreElements()) {
	  Object key2 = keys2.nextElement();
	  int keySize = key2.toString().length();
	  String printKey = (keySize <= maxLengthIfDepth3) ?
	    key2.toString() :
	    (key2.toString().substring(0, maxLengthIfDepth3) + "/" + keySize);
	  System.out.println("    Vector of key [" + printKey +
			     "] in the given hash table has " +
			     ((Vector)table.get(key2)).size() +
			     " elements.");
	}
      } else if (depth == 3) {
	System.out.println("");
	Enumeration keys2 = table.keys();
	while (keys2.hasMoreElements()) {
	  Object key2 = keys2.nextElement();
	  int keySize = key2.toString().length();
	  String printKey = (keySize <= maxLengthIfDepth3) ?
	    key2.toString() :
	    (key2.toString().substring(0, maxLengthIfDepth3) + "/" + keySize);
	  System.out.print("  [" + printKey + "]->");
	  Vector toPrint = (Vector)table.get(key2);
	  int max = toPrint.size();
	  for(int i = 0; i < max; i++) {
	    String rep = toPrint.elementAt(i).toString();
	    int repSize = rep.length();
	    String printRep = (repSize < maxLengthIfDepth3) ?
	      ("[" + rep + "]") :
	      ("[" + rep.substring(0, maxLengthIfDepth3) + "/]");
	    System.out.print(printRep);
	  }
	  System.out.println("");
	}
      }
    } else if (depth == 4) {
      System.out.print(" H[" + table.size() + "]->");
      Enumeration keys2 = table.keys();
      while (keys2.hasMoreElements()) {
	Object key2 = keys2.nextElement();
	System.out.print(((Vector)table.get(key2)).size() + ",");
      }
      System.out.println("\b.");
    }
  }
  
  /**
   * Gives the number of sources in the DualKeyContainer.
   */
  
  public int getSourcesNumber() {
    return sourcesNumber;
  }
  
  /**
   * Returns the number of different values stored in the DualKeyContainer.
   */
  
  public int getUniqueValuesNumber() {
    return uniqueValuesNumber;
  }
  
  /**
   * Returns the enumeration of Objects used as sources in this DualKeyContainer.
   * Caution: these are the actual keys, ie an object used as a source may not be
   * returned if another object equal to the first one (considering hashCode()
   * and equals()) had already been used as a source for this DualKeyContainer.
   */
  
  public Enumeration getSources() {
    return valuesBySource.keys();
  }
  
  /**
   * Returns the enumeration of Objects used as features for this DualKeyContainer.
   * Caution: these are the actual keys, ie an object used as a feature may not be
   * returned if another object equal to the first one (considering hashCode()
   * and equals()) had already been used as a feature for this DualKeyContainer.
   * Thus, this object would be the one returned.
   */
  
  public Enumeration getValues() {
    return sourcesByValue.keys();
  }
  
  /**
   * When a value is stored, it appears in a Vector corresponding to its source (a
   * key of a hashtable). But the key in the hashtable may not be the object the value
   * was stored with. This method gives this object from the object used as a source.
   * This way, information about the Vector of sources (like its size, ...) can be
   * stored in the key used to access this Vector.
   *
   * @see IrfHashtable#getActualKey
   */
  
  
  public Object getActualSource(Object source) {
    return valuesBySource.getActualKey(source);
  }
  
  /**
   * When a value is stored, it appears in a Vector corresponding to its feature (a
   * key of a hashtable). But the key in the hashtable may not be the object the value
   * was stored with. This method gives this object from the object used as a feature.
   * Thus, information about the values stored for a feature can be stored in the key
   * (the feature).
   *
   * @see IrfHashtable#getActualKey
   */
  
  public Object getActualFeature(Object feature) {
    return sourcesByValue.getActualKey(feature);
  }
}
