package gov.nist.nlpir.irf.index;

import gov.nist.nlpir.irf.result.BasicRetrievalResult;
import gov.nist.nlpir.irf.result.ResultForDocMatchingQueryCondition;
import gov.nist.nlpir.irf.result.ResultForDocMatchingQueryModalityUnit;
import gov.nist.nlpir.irf.query.QueryModalityUnit;
import gov.nist.nlpir.irf.query.QueryCondition;

import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.de.ProxyDataElem;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.index.braf.ProxyFeatureList;
import gov.nist.nlpir.irf.index.braf.DeIntern;
import gov.nist.nlpir.irf.index.braf.IoAddrIntern;
import java.util.Vector;
import java.util.Enumeration;


/**
 * This class extends IdxIntern to support indexes based around the Inverse 
 * Document Frequency (IDF) term weighting measure.
 *
 * Note: The Math.log() function returns the natural log ("ln" or log to 
 * the base e) of its argument (see java.lang.Math.log()). We have used it 
 * here for the sake of performance, though traditionally IR applications 
 * use log base 2.
 * 
 * @version $Revision: 1.10 $ 	
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public class IdfIdxIntern extends IdxIntern {


  /**
   * Basic constructor with the Index location and name.
   * @param loc index location info
   * @param name index name
   */
  public IdfIdxIntern(String loc, String name) {
    super(loc,name);
  }



  /**
   * Returns a string representation of this index
   * @return a list of the features contained, with their description, the 
   * number of docs containing them and their score. If no feature is present,
   * returns <pre>no Features</pre>
   */
  public String toString() {
    Enumeration c1 = getFeatures();
    if (!c1.hasMoreElements())
      return "no Features";
    StringBuffer sBuf = new StringBuffer();
    while (c1.hasMoreElements()) {
      DeIntern aDE = (DeIntern)c1.nextElement();
      sBuf.append("  Feature " + aDE.getValueAsString() + 
		  ": # of docs containing feature = " +
		  aDE.getNrOfSources() + " IDF = " + aDE.getScore() + "\n");
    }
    return sBuf.toString();
  }
  


  /**
   * Updates the index after it has been created. In fact, an Index
   * isn't updated each time an object is added to so that efficiency is
   * better when several objects are added before using the index.
   */
  public void update() {
    super.update();
    updateFeatures();
    updateSources();
  }
  


  /** Number of threads to be started for calculation of feature scores */
  private final static int numCalcFeatureScoreThreads = 3;

  /**
   * Computes the scores of the stored features in one or more separate
   * threads. The scores are stored in the DeIntern keys. 
   */
  public void updateFeatures() {

    final int numFeatures = getNrOfFeatures();

    /** 
     * Manages the calculation of the "score" of features/DeInterns 
     * from the given enumeration of features and sets it in the 
     * DeIntern (key to the sources-by-value table). All instances
     * of the this thread class draw from the same enumeration of
     * features.
     */
    class CalcFeatureScoreThread extends Thread  
    {
      private Enumeration DeInternEnum =  null;


      CalcFeatureScoreThread(Enumeration DeInternEnum) {
	this.DeInternEnum = DeInternEnum;
      }


      public void run () {

	DeIntern aDE = null;
	int i = 0;
	int modulus = numFeatures / 100; 
	if (modulus == 0) modulus++;
	long lastStamp = Support.getTimeInMillis();
	long currentStamp = lastStamp;

	synchronized(DeInternEnum) // all threads share this enumeration
	  {
	  if (DeInternEnum.hasMoreElements())
	    aDE = (DeIntern)DeInternEnum.nextElement();
	  else
	    aDE = null;
          }
	  
	while (aDE != null) 
	  {
	  calcFeatureScore(aDE);
	  ((ProxyDataElem)aDE.getDE()).makeLightweight();
	  yield();

	  i++;
	  if ((i % modulus) == 0)   // message for each percent completed

      	    if (Support.DETAIL_TRACE <= Support.traceLevel)
	      {
	      currentStamp = Support.getTimeInMillis();
	      Support.traceMessage(i + "(" + 
				 (currentStamp - lastStamp) / 1000 + "s): " +
				 gov.nist.nlpir.irf.proxy.VirtualProxy.
				 getInMemoryProxyCounts());
	      lastStamp = currentStamp;
	      }

	  synchronized(DeInternEnum) // all threads share this enumeration
	    {
	    if (DeInternEnum.hasMoreElements())
	      aDE = (DeIntern)DeInternEnum.nextElement();
	    else
	      aDE = null;
	    }

	  }

      }
    }

    Thread[] threads = new Thread[numCalcFeatureScoreThreads];

    if (Support.COARSE_TRACE <= Support.traceLevel)
      Support.traceMessage(numFeatures +
			   " feature scores are going to be computed;");

    Enumeration c = getFeatures();

    for (int j=0; j<numCalcFeatureScoreThreads; j++)
      {
      threads[j] = new CalcFeatureScoreThread(c);
      threads[j].start();
      }
    for (int j=0; j<numCalcFeatureScoreThreads; j++)
      {
      try{threads[j].join();}
      catch(InterruptedException ie){System.out.println("Not sure what to do given that " + ie);}
      }

  }
  


  /** Number of threads to be started for calculation of document scores */
  private final static int numCalcDocumentScoreThreads = 3;

  /**
   * Computes the scores of the stored documents in one or more separate
   * threads. The scores are stored in the IoAddrIntern keys. 
   */
  public void updateSources() {

    final int numSources = getNrOfSources();

    /** 
     * Manages the calculation of the "score" of documents/IoAddrInterns 
     * from the given enumeration of features and sets it in the 
     * IoAddrIntern (key to the values-by-source table). All instances
     * of the this thread class draw from the same enumeration of
     * features.
     */
    class CalcDocumentScoreThread extends Thread  
    {
      private Enumeration IoAddrInternEnum =  null;


      CalcDocumentScoreThread(Enumeration IoAddrInternEnum) {
	this.IoAddrInternEnum = IoAddrInternEnum;
      }


      public void run () {

	IoAddrIntern anIoAddr = null;
	int i = 0;
	int modulus = numSources / 100; 
	if (modulus == 0) modulus++;
	long lastStamp = Support.getTimeInMillis();
	long currentStamp = lastStamp;

	synchronized(IoAddrInternEnum) // all threads share this enumeration
	  {
	  if (IoAddrInternEnum.hasMoreElements())
	    anIoAddr = (IoAddrIntern)IoAddrInternEnum.nextElement();
	  else
	    anIoAddr = null;
          }
	  
	while (anIoAddr != null) 
	  {
	  calcDocumentScore(anIoAddr);
	  anIoAddr.getDoc().makeLightweight();
	  yield();

	  i++;
	  if ((i % modulus) == 0)   // message for each percent completed

      	    if (Support.DETAIL_TRACE <= Support.traceLevel)
	      {
	      currentStamp = Support.getTimeInMillis();
	      Support.traceMessage(i + "(" + 
				 (currentStamp - lastStamp) / 1000 + "s): " +
				 gov.nist.nlpir.irf.proxy.VirtualProxy.
				 getInMemoryProxyCounts());
	      lastStamp = currentStamp;
	      }

	  synchronized(IoAddrInternEnum) // all threads share this enumeration
	    {
	    if (IoAddrInternEnum.hasMoreElements())
	      anIoAddr = (IoAddrIntern)IoAddrInternEnum.nextElement();
	    else
	      anIoAddr = null;
	    }

	  }

      }
    }

    Thread[] threads = new Thread[numCalcDocumentScoreThreads];

    if (Support.COARSE_TRACE <= Support.traceLevel)
      Support.traceMessage(numSources +
			 " document scores are going to be computed;");

    Enumeration c = getSources();

    for (int j=0; j<numCalcDocumentScoreThreads; j++)
      {
      threads[j] = new CalcDocumentScoreThread(c);
      threads[j].start();
      }
    for (int j=0; j<numCalcDocumentScoreThreads; j++)
      {
      try{threads[j].join();}
      catch(InterruptedException ie)
	{System.out.println("Not sure what to do given that " + ie);}
      }

  }


  
  /**
   * Calculates the score of the given feature/DeIntern and sets it 
   * in the DeIntern (key to the sources-by-value table). 
   * The score is the Inverse Document Frequency (IDF), after Sprck-Jones
   * (1972):
   *            IDF(t) = log(N/n) +1
   * where
   *            N == the number of documents in a collection, and
   *            n == the number of documents in the collection which 
   *                 contain term t.
   * See note above on the nature of Math.log(). 
   *
   * @param aDEI the DeIntern whose score is to be computed.
   */
  public void calcFeatureScore(DeIntern aDEI) {
    // Find out how many different documents contain the passed feature
    int i = getNrOfUniqueSources(aDEI);
    if (i > 0) {
      aDEI.setNrOfSources(i);

      // Get the total number of different documents in the collection
      int j = getSourcesNumber();

      // Calculate the score (i.e. the IDF)
      double idf = (Math.log(j / i)) + 1;
     
      // Save the score
      aDEI.setScore(idf);

    } else
      Support.errorMessage("IdfIdxIntern.calcFeatureScore(aDEI) has been "+
			   "called with aDEI that doesn't exist in the " +
			   "IdxIntern: " + aDEI);
  }
  


  /**
   * Computes the document score and sets it in the passed 
   * IoAddrIntern. The document score is the sum over all the document's 
   * features (e.g., terms in a text doc) of the squared weight of each 
   * feature. The weight of a feature is 1 + the log of the frequency of 
   * the feature in the document.

   * @param an_IOAddr the IoAddrIntern of the document whose score
   *                  is to be calculated
   */
  public void calcDocumentScore(IoAddrIntern anIO_Addr) {

    // Get the indexing features (IFs) derived from the document
    ProxyFeatureList theIFs = getIFsWithSource(anIO_Addr);

    // Make a temporary, modifiable vector of the Data Elements (DEs)
    // contained in the IFs
    Vector theFeatures = (Vector)theIFs.getDE_Features();

    // The real IFs are no longer needed
    theIFs.makeLightweight();

    // Loop through the features/DEs (e.g., terms in a text doc) calculating
    // the weight of each DE and incorporating the weight in the
    // score for the document
    double sigma = 0;
    int numberOfFeatures = theFeatures.size();
    for (int i = 0; i < numberOfFeatures; i++) {

      // Get a (proxy) DE
      ProxyDataElem currentFeature = (ProxyDataElem)theFeatures.elementAt(i);

      // If the current feature/DE has not been counted on an earlier pass...
      if (currentFeature != null) { // null means already counted

	// Get the DeIntern which is the index key for the DE; it's
	// more efficient to do the counting in terms of the DeInterns
        DeIntern aDEI = getFeature(currentFeature);

	// Count any addictional occurrences of this feature in the list
	// of features for this document
        int frequInDoc = aDEI.getFrequency(theFeatures, i + 1);
 
	// Calculate the weight and add it squared into the sum so far
        double weightInDoc = 1 + Math.log(frequInDoc);
        sigma += weightInDoc * weightInDoc;
      }
    }
    anIO_Addr.setScore(sigma); // save the score
  }
  


  /**
   * Positions fields of aResult to allow the computation of Retrieval
   * Status Value (RSV).
   * @param aResult the Result... used.
   */
  public void calcInputForRSV(ResultForDocMatchingQueryCondition aResult) {
    calcCombinedScore(aResult);
    calcWeightOfFeatureInDocument(aResult);
    calcWeightOfFeatureInQuery(aResult);
  }
  


  /**
   * Compensates for differences in the scores of retrieved features, e.g. 
   * in the case where the retrieved features are not exact matches for the 
   * query feature. For now, this compensation takes the form of calculating 
   * the average score of the retrieved features.
   * Sets the combinedScore in the passed parameter to this value.
   * @param aResult partial result on which to calculate the combined score
   */
  public void calcCombinedScore(ResultForDocMatchingQueryCondition aResult) {
    /* The indexing features matching a single condition may have different
       scores, e.g. IDF's, when an approximate matching has been applied. Thus
       some combination of the values is needed. At the moment, the 
       combination is done by calculating the mean value. A more flexible 
       approach would be nice! */
    double i = 0.0;
    double sigmaScore = 0.0;
    if (Support.getTraceLevel() >= Support.DETAIL_TRACE) 
      System.out.println(aResult);
    Enumeration aEnum = aResult.getBasicResults().elements();
    while (aEnum.hasMoreElements()) {
      BasicRetrievalResult aBRR = (BasicRetrievalResult)aEnum.nextElement();
      sigmaScore += aBRR.getScore();
      i += 1.0;
    }
    if ( i > 0.0 )
	aResult.setCombinedScore(sigmaScore / i);
    else aResult.setCombinedScore( 0.0 );
  }
  


  /**
   * Computes the weight of the given feature for the documents it appears in
   * and sets it in the result.
   * The weight is equal to the log of the frequency of the term in the 
   * document plus 1.
   * Note: The term frequency *ought* never to be zero (giving a log value 
   * of -Infinity) because IdfIdxIntern.calcInputForRSV() - the method that 
   * calls this one - is not invoked by 
   * IdxIntern.makeResultForDocMatchingQueryCondition() unless one
   * BasicRetrievalResult has been returned for the query condition.
   * @param aResult partial result for which to compute the weight...
   */
  public void calcWeightOfFeatureInDocument(ResultForDocMatchingQueryCondition aResult) {
    int freq = aResult.getBasicResults().size();
    double weight = 1 + Math.log(freq);	
    aResult.setWeightOfFeatureInDocument(weight);
  }
  


  /**
   * Calculates the weight of the feature in the query and sets it in the 
   * result. The weight is equal to the combined score (see 
   * calcCombinedScore()) * (( log of the frequency of the term in the query)
   * + 1).
   * Note: The query frequency *ought* never to be zero (giving a log value 
   * of -Infinity) because IdfIdxIntern.calcInputForRSV() - the method that 
   * calls this one - is not invoked by 
   * IdxIntern.makeResultForDocMatchingQueryCondition() unless one
   * BasicRetrievalResult has been returned for the query condition.
   * @param aResult partial result on which to calculate the weight...
   */
  public void calcWeightOfFeatureInQuery(ResultForDocMatchingQueryCondition aResult) {
    int freq = aResult.getQueryCondition().getFreqOfFeatureInQuery();
    double weight = aResult.getCombinedScore() * (1 + Math.log(freq)); 
    aResult.setWeightOfFeatureInQuery(weight);
  }
  


  /**
   * Calculates the Retrieval Status Value (RSV) of a single modality unit of
   * the result of a query. This is the score attained by a document for a 
   * single modality of the query.
   *
   * The RSV == 
   *    sigma(1 to n) (wiq * wid) / sqrt( sigma(1 to p)(wiq') * docscore)
   *    where:
   *                 n == the number of query terms ocurring in the document
   *               wiq == combinedScore * (1 + log(qf))
   *               wid == 1 + log(tf)
   *                 p == the number of terms in the query
   *              wiq' == IDF * (1 + log(qf))
   *          docscore == sigma(1 to q)(1 + log(tf))
   *                 q == the number of unique terms in the document
   *                qf == the frequency of a term in the query
   *                tf == the frequency of a term in the document
   * Note: In practice wiq and wiq' take the same value, as the method for
   * producing the combinedScore returns the IDF.
   * @param aResult partial result for which to calculate the RSV
   */
  public double calcRSV(ResultForDocMatchingQueryModalityUnit aResult) {
    Enumeration elmEnum = aResult.getPartialResults().elements();
    double Wiq;               // weight of feature in query;
    double Wid;               // weight of feature in document;
    double sigmaWiqWid = 0.0; // sigma(Wiq * Wij)
    double sigmaWiq2 = 0.0;   // sigma(Wiq * Wiq)
    ResultForDocMatchingQueryCondition aPR;

    while (elmEnum.hasMoreElements()) {
      aPR = (ResultForDocMatchingQueryCondition)elmEnum.nextElement();
      Wiq = aPR.getWeightOfFeatureInQuery();
      Wid = aPR.getWeightOfFeatureInDocument();
      sigmaWiqWid += Wiq * Wid;
    }	
    QueryModalityUnit aQM = aResult.getQueryModalityUnit();
    Enumeration condEnum = aQM.getConditions().elements();
    while (condEnum.hasMoreElements()) {
      QueryCondition aCondition = (QueryCondition)condEnum.nextElement();
      DeIntern actualFeature = getFeature(aCondition.getFeature());
      double idf = 0.0;
      if (actualFeature != null)
	idf = actualFeature.getScore();
      int f = aCondition.getFreqOfFeatureInQuery();
      Wiq = idf * (1 + Math.log((double)f));
      sigmaWiq2 += Wiq * Wiq;
    }
    double RSV = sigmaWiqWid / Math.sqrt(sigmaWiq2 * aResult.getDocScore());
    aResult.setRSV(RSV);
    return RSV;
  }
}
