package gov.nist.nlpir.irf.index;

import gov.nist.nlpir.irf.query.QueryCondition;
import gov.nist.nlpir.irf.query.QueryModalityUnit;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.query.IdxWeightsWithMatchDegree;
import gov.nist.nlpir.irf.result.BasicRetrievalResult;
import gov.nist.nlpir.irf.result.ResultForDocMatchingQueryCondition;
import gov.nist.nlpir.irf.result.ResultForDocMatchingQueryModalityUnit;

import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.de.ProxyDataElem;
import gov.nist.nlpir.irf.matcher.Matcherbp;
import gov.nist.nlpir.irf.index.braf.PersistentDualKeyContainer;
import gov.nist.nlpir.irf.index.braf.ProxyFeatureList;
import gov.nist.nlpir.irf.index.braf.PersistentIrfHashtable;
import gov.nist.nlpir.irf.index.braf.DeIntern;
import gov.nist.nlpir.irf.index.braf.IoAddrIntern;
import gov.nist.nlpir.irf.util.Support;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Dictionary;
import java.io.Serializable;

/** 
 * Base class for the internal representation of an Index. <br><br>
 * Objects of this class contain several sets that together provide the 
 * structure necessary to support an Index. It mainly contains a 
 * PersistentDualKeyContainer
 * for this purpose. Subclasses of this class may override  methods for 
 * calculating retrieval status values (RSV), etc.
 *   
 * @version $Revision: 1.10 $ 	
 *   
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 *   
 * @see gov.nist.nlpir.irf.index.Index
 * @see gov.nist.nlpir.irf.index.IdfIdxIntern
 * @see gov.nist.nlpir.irf.index.KeyWordIdxIntern
 * @see gov.nist.nlpir.irf.index.braf.DeIntern
 * @see gov.nist.nlpir.irf.index.braf.IoAddrIntern
 * @see gov.nist.nlpir.irf.feature.IndexingFeature 
 * @see gov.nist.nlpir.irf.de.DataElem
 * @see gov.nist.nlpir.irf.index.braf.PersistentDualKeyContainer
 */

public abstract class IdxIntern implements Serializable {

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;
    
  /** This flag is set when the Index is up-to-date - does not
   * need for Update to be run. 
   * @serial  */
  private boolean flagUpToDate;

  /** Main part of the IdxIntern. 
   *  @serial */
  private PersistentDualKeyContainer indexingFeatures; // keys: IoAddrIntern, DeIntern



  /** 
   * Constructor which allocates the initial data structures for IdxIntern.
   * @param location information about the location of the PDKC
   * @param name name of the index being supported
   */
  public IdxIntern(String location, String name) {
    indexingFeatures = new PersistentDualKeyContainer(location, name);
    flagUpToDate = false;
  }



  /**
   * Only here for serializarion's needs.
   */
  public IdxIntern() {
  }



  /** 
   * Initializes this instance of IdxIntern.
   */
  public void init() {
  }
    


  /** 
   * Displays an ascii representation of the Class.
   */
  public void present() {
    System.out.println(this);
  }
    


  /** 
   * Displays an ascii representation of the features
   * in the Index (members of features).
   */
  public void presentFeatures() {
    System.out.println("Presentating the features of a " + getClass().getName());
    Enumeration c1 = getFeatures();
    while (c1.hasMoreElements()) {
      DeIntern aDE = (DeIntern)c1.nextElement();
      System.out.print("  Feature " + aDE.getValueAsString() +
		       ": # of docs containing feature = " +
		       aDE.getNrOfSources() +
		       " IDF = " + aDE.getScore() + "\n");
    }
  }
    


  /** 
   * Displays an ascii representation of the sources.
   */
  public void presentSources() {
    System.out.println("Sources of a " + getClass().getName());
    Enumeration c = getSources();
    while (c.hasMoreElements()) {
      IoAddrIntern anIOAI = (IoAddrIntern)c.nextElement();
      System.out.println("  Address " + anIOAI);
    }
  }
    


  /** 
   * Generates a report on the number of sources, features and indexing
   * features.
   */
  public void presentStatistics() {
    System.out.println("  Nr of sources: " + indexingFeatures.getSourcesNumber() +
		       "\n  Nr of features: " + indexingFeatures.getUniqueValuesNumber() +
		       "\n  Nr of indexing features: " + indexingFeatures.getValuesNumber());
    //	       "\n  -- Tables efficiency --"); 
    //indexingFeatures.showHashtableStatistics();
  }  
    
    
  /**
   *  Calculates number of sources in index and return that value.
   *  @return number of sources in index.
   */
  public final int calcNrOfSources() {
    return indexingFeatures.getSourcesNumber();
  }
    


  /** 
   * Generates a set of query conditions.
   *
   * @param qm Query Modality Unit.
   * @param idxFeats Collection  of indexing features.
   * @return set of query conditions.
   */ 
  public Vector makeQueryConditions(QueryModalityUnit qm, Vector idxFeats) {
    Vector setQC = new Vector();
    int numberOfIF = idxFeats.size();
    for (int i = 0; i < numberOfIF; i++) {
      ProxyIndexingFeature anIF = (ProxyIndexingFeature)idxFeats.elementAt(i);
      DataElem aDE = anIF.getFeature();
      if (aDE != null) {
	int freqOfFeatureInQuery = 1;
	anIF.setFeature(null);
	for (int j = 0; j < numberOfIF; j++) {
	  anIF = (ProxyIndexingFeature)idxFeats.elementAt(j);
	  if (anIF.getFeature() != null && aDE.equals(anIF.getFeature())) {
	    anIF.setFeature(null);
	    freqOfFeatureInQuery++;
	  }
	}
	QueryModalityUnit qm2 = new QueryModalityUnit(qm.getIndexName(), qm.getMatcher(), qm.getCombiner(), 
				  qm.getDocType(), qm.getAttributeName(), qm.getThresholdMD());
	QueryCondition cond = new QueryCondition(qm2);
	cond.setFeature(aDE);	
	cond.setFreqOfFeatureInQuery(freqOfFeatureInQuery);
	setQC.addElement(cond);
      }
    }
    return setQC;
  }
    


  /** 
   * Evaluates a Query Modality Unit.
   *
   * @param qmu Query Modality Unit
   * @param IFs Collection of IndexingFeatures
   * @return set of query conditions
   */
  public Vector evalQueryModalityUnit(QueryModalityUnit qmu, Vector IFs) {
    Vector partialResults = new Vector();
    Vector setQC = makeQueryConditions(qmu, IFs);
    int numberOfQC = setQC.size(); 
	
    // For each query condition
    for (int i = 0; i < numberOfQC; i++) {
      QueryCondition cond = (QueryCondition)setQC.elementAt(i); 
      Vector tmp = evalQueryCondition(cond);
      int sizeTmp = tmp.size();
      for (int j = 0; j < sizeTmp; j++)
	partialResults.addElement(tmp.elementAt(j));	
    }
    int i = IFs.size();
    qmu.setConditions(setQC);
    return makeResultsForDocMatchingQueryModalityUnit(qmu, partialResults, i, numberOfQC); 
  }
    
  /** 
   * Evaluates a query condition. The result is a vector of 
   * ResultForDocMatchingQueryCondition.
   *
   * @param condition The Query Condition.
   * @return Vector of ResultForDocMatchingQueryCondition.
   */
  public Vector evalQueryCondition(QueryCondition condition) {
    Vector results;
    Vector setIndexingFeatures = new Vector();
    DataElem queryFeature = condition.getFeature(); 
    QueryModalityUnit theQMU = condition.getQueryModalityUnit();
    Matcherbp matcher = theQMU.getMatcher();
    double threshold = theQMU.getThresholdMD();
    String dt = theQMU.getDocType();
    if (dt == null)
      dt = "*any*";
    String an = theQMU.getAttributeName();
    if (an == null)
      an = "*any*";
    String m = matcher.getClass().getName();
    if ((m.compareTo("NullMatcherbp") == 0) || (threshold == 1.0)) {
      // exact matching 
      DeIntern aDEI = getFeature(queryFeature);	
      if (aDEI != null) {
	// get all IndexingFeatures which have the same feature as
	// IndexingFeature of QueryCondition.  In addition, the document
	// type and the attribute name must match.
	setIndexingFeatures = getIFsWithFeature(aDEI, dt, an);
      }
    } else {
      // approximate matching
      // iterate over features 
      Enumeration c = getFeatures();
      while (c.hasMoreElements()) {
	DeIntern aDEI = (DeIntern)c.nextElement();
	// check whether feature matches the feature of IndexingFeature of
	// QueryCondition with a sufficiently high matching degree.
	if ((matcher.match(aDEI.getDE(), queryFeature)) > threshold) {
	  // get corresponding IndexingFeatures
	  Vector foo = getIFsWithFeature(aDEI, dt, an);
	  int sizeOfFoo = foo.size();
	  for (int i = 0; i < sizeOfFoo; i++)
	    setIndexingFeatures.addElement(foo.elementAt(i));
	}
      }
    }
    if (setIndexingFeatures.size() > 0) {
      results = evalQueryCondition(condition, setIndexingFeatures);
    } else {
      results = new Vector();
    }
    return results;
  }
    


  /** 
   * Generates a vector of results that match the features
   * in the specified query condition.
   *
   * @param condition Query Condition.
   * @param setIndexingFeatures Vector of IndexingFeatures.
   * @return Vectorof ResultForDocMatchingQueryCondition.
   */ 
  public Vector evalQueryCondition(QueryCondition condition, 
				   Vector setIndexingFeatures)  {
    // Get the details of query condition
    DataElem qFeature = condition.getFeature();
    QueryModalityUnit theQMU = condition.getQueryModalityUnit();
    Matcherbp matcher = theQMU.getMatcher();
    IdxWeightsWithMatchDegree combiner = theQMU.getCombiner();
    double threshold = theQMU.getThresholdMD();
    String dt = theQMU.getDocType();
    String an = theQMU.getAttributeName();
    // determine evaluation mode
    int flagXMatch = 0;
    if ((matcher.getClass().getName().compareTo("NullMatcherbp") == 0) || (threshold == 1.0))
      flagXMatch = 1;
    // iterate over given set of IndexingFeatures
    Hashtable tmp = new Hashtable();
    int numberOfIF = setIndexingFeatures.size();
    for (int i = 0; i < numberOfIF; i++) {
      ProxyIndexingFeature anIF = (ProxyIndexingFeature)setIndexingFeatures.elementAt(i);
      double md;
      // calculate matching degree if needed
      if (flagXMatch == 1)
	md = 1.0;
      else
	md = matcher.match(qFeature, anIF.getFeature());
      if (md > threshold) {
	// get score (e.g. IDF) of indexing feature
	DeIntern theDEI = getFeature(anIF.getFeature());
	double score = theDEI.getScore();
	BasicRetrievalResult aBRR = new BasicRetrievalResult(anIF, md, score);
	anIF.makeLightweight();
	// store the result temporarily in a set 
	tmp.put(aBRR, aBRR);
      }		
    }
    return makeResultForDocMatchingQueryCondition(condition, tmp);
  }
  
  

  /** 
   * Generates a vector of results for a given query condition
   *
   * @param condition Query Condition.
   * @param setBRR Basic Retrieval Result.
   * @return a vector
   */ 
  public Vector makeResultForDocMatchingQueryCondition(QueryCondition condition, 
						       Hashtable setBRR) {
    /*-------------------------------------------------------------------
     * Iterate over the BasicRetrievalResults passed, grouping
     * them by the document which is the source for the feature.
     *------------------------------------------------------------------*/
    Vector setResult = new Vector();
    Hashtable tmp = new Hashtable();
    Vector collBRR = null;

    // Create an array of the BRRs in setBRR
    Enumeration bc = setBRR.elements();
    int arrBRR_Size = setBRR.size();
    BasicRetrievalResult[] arrBRR = new BasicRetrievalResult[arrBRR_Size];
    for (int i = 0; i < arrBRR_Size; i++)
      arrBRR[i] = (BasicRetrievalResult)bc.nextElement();
    bc = null;

    for (int i = 0; i < arrBRR_Size; i++) {

      BasicRetrievalResult curBRR = arrBRR[i];

      if (curBRR == null)  // if this BRR has been processed, skip it
	continue;

      ProxyIndexingFeature anIF = curBRR.getIndexingFeature();
      ProxyDocument targetDoc = anIF.getSourceDoc();
      /*---------------------------------------------------------------
       * This BRR will be for a document we have not seen before,
       * so add it immediately to the collBRR we are building. The
       * collBRR will contain all the BRRs from setBRR which refer to 
       * the same source document. Mark this entry in the array of BRRs
       * to be processed, as complete
       *--------------------------------------------------------------*/
      collBRR = new Vector(1);
      collBRR.addElement(curBRR);
      arrBRR[i] = null;  // mark as processed

      /*---------------------------------------------------------------
       * Now look for additional BRRs from setBRR which refer to the
       * same document as the one just added.
       *--------------------------------------------------------------*/
      for (int j=i+1; j<arrBRR_Size; j++) {
	curBRR = arrBRR[j];
	if (curBRR == null)
	  continue;

	// --- Compare references to ProxyDocuments assuming uniqueness
	if ( curBRR.getIndexingFeature().getSourceDoc() == targetDoc ) {
	  collBRR.addElement(curBRR);
	  arrBRR[j] = null;  // mark as processed
	} 

      } // end (int j=i+1; j<arrBRR_Size; j++)

      tmp.clear();

      ResultForDocMatchingQueryCondition aResult = 
	new ResultForDocMatchingQueryCondition(condition, targetDoc, collBRR); 
      calcInputForRSV(aResult);

      setResult.addElement(aResult);

    } // end for (int i=0; i<arrBRR_Size; i++)

    return setResult;
  }
    

    
  /** 
   * Generates a vector of results for a given query modality unit
   * 
   * This method takes a vector of ResultForDocMatchingQueryConditions
   * (called partial results) and groups these by source document to form 
   * a combined result vector of ResultForDocMatchingQueryModalityUnits 
   * (called combined results), exactly one of the latter combined Results 
   * for each source referred to in the partial results of* the incoming 
   * vector.
   *
   * @param qmu Query Modality Unit
   * @param collPR Partial Retrieval Result.
   * @return a vector
   */
  public Vector makeResultsForDocMatchingQueryModalityUnit(QueryModalityUnit qmu,
							   Vector collPR, 
							   int numQueryFeatures,
							   int numUniqueQueryFeatures)  {
    double maxRSV = 0.0;
    Vector combinedResults = new Vector();
    boolean found = false;
    ResultForDocMatchingQueryModalityUnit aCR; 
    int numberOfPR = collPR.size();
    /*------------------------------------------------------------------
     * Loop through the input partial results. These are 
     * ResultForDocMatchingQueryConditions. There may be multiple ones
     * for a given source.
     *------------------------------------------------------------------*/
    for (int i = 0; i < numberOfPR; i++) {
      ResultForDocMatchingQueryCondition aPR =
	(ResultForDocMatchingQueryCondition)collPR.elementAt(i);
      int numberOfCR = combinedResults.size();
      ProxyDocument doc1 = aPR.getDocument();
      /*--------------------------------------------------------------
       * Loop through the combined results - these are 
       * ResultForDocMatchingQueryModalityUnits in the vector of combined 
       * results being built for output - to see if a combined Result 
       * exists to for the source referred to  by the current incoming 
       * partial result.
       * If one exists, add the incoming partial result to it; otherwise,
       * create a new combined Result, add the current incoming partial 
       * result to it, and add it to the output vector.
       *--------------------------------------------------------------*/
      for (int j = 0; j < numberOfCR; j++) {
	aCR = (ResultForDocMatchingQueryModalityUnit)combinedResults.elementAt(j);
	ProxyDocument doc2 = aCR.getDocument();
	// --- Compare references to ProxyDocuments assuming uniqueness
	if (doc1 ==  doc2) {   
	  aCR.addPartialResult(aPR);  
	  found = true;
	  break;
	}
      }
      if (found)       
	found = false; 
      else {
	/*-----------------------------------------------------------
	 * A combined result for the current incoming partial Result's
	 * source does not exist in the output vector, so create one
	 * to include  a new partial result based on the current incoming
	 * partial result and add the new combined result to the output
	 * vector.
	 *----------------------------------------------------------*/
	Vector coll = new Vector();
	coll.addElement(aPR);
	IoAddrIntern aSrc = getActualSource(doc1);
	if (aSrc != null ) {
	  aCR = new ResultForDocMatchingQueryModalityUnit(qmu, doc1, 
							  aSrc.getScore(),
							  coll, maxRSV);
	  aCR.setNumUniqueQueryFeatures(numUniqueQueryFeatures);
	  aCR.setNumQueryFeatures(numQueryFeatures);
	  combinedResults.addElement(aCR);
	} else
	  Support.errorMessage("DOC1 NOT FOUND in indexingFeatures: " + doc1
			       + "\nhashCode(): " + doc1.hashCode() + "\n");
      }
    } 	
    
    // Iterate over set of ResultForDocMatchingQueryModalityUnit objects
    // and calculate RSV based on corresponding partial results.
    
    int numberOfCR = combinedResults.size();
    for (int i = 0; i < numberOfCR; i++) {
      aCR = (ResultForDocMatchingQueryModalityUnit)combinedResults.elementAt(i);
      double currentRSV = calcRSV(aCR); // This method will set the computed RSV in
      if (currentRSV > maxRSV)          // the ResultForDocMatchingQueryModalityUnit
	maxRSV = currentRSV;	  
    }
    for (int i = 0; i < numberOfCR; i++) {
      aCR = (ResultForDocMatchingQueryModalityUnit)combinedResults.elementAt(i);
      aCR.setMaxRSV(maxRSV);
    }
    return combinedResults;
  }
    


  /** 
   * Calculates the retrieval status value 
   *
   * @param aResult result for document matching query condition.
   * @see IdfIdxIntern
   * @see KeyWordIdxIntern
   */ 
  public abstract void calcInputForRSV(ResultForDocMatchingQueryCondition aResult);
    


  /** 
   * Calculates the retrieval status value
   *
   * @param aResult result for document matching query modality unit.
   * @return Retrieval Status Value
   * @see IdfIdxIntern
   * @see KeyWordIdxIntern
   */
    
  public abstract double calcRSV(ResultForDocMatchingQueryModalityUnit aResult);


    
  /** 
   * Exposes the inner structure of the IdxIntern 
   * It is mainly used to pass methods to the 
   * PersistentDualKeyContainer for probe purpose:
   * showing statistics for example, or getting a display. Should never 
   * be used to give instructions to the PDKC as this class is supposed 
   * to wrap it entirely.
   */
  public PersistentDualKeyContainer getIndexingFeatures() {
    return indexingFeatures;
  }
  


  /** 
   * Gets the number of features.
   * @return number of features
   */
  public final int getNrOfFeatures() {
    return indexingFeatures.getUniqueValuesNumber();
  }
    


  /** 
   * Gets the number of sources.
   * @return number of sources
   */
  public final int getNrOfSources() {
    return indexingFeatures.getSourcesNumber();
  }
    


  /** 
   * Gets the number of unique sources for a feature.
   * @param aFeature a feature
   * @return number of unique sources
   */    
  public int getNrOfUniqueSources(DeIntern aFeature) {
    /* This method uses the fact that calling put on a hashtable with an
       already existing key replaces the object the key was associated to.
       It makes it easy to calculate the number of different objects there
       are in the feature vector, each object creating a new entry in the
       hashtable only if it wasn't there before. */

    ProxyFeatureList someIF = getIFsWithFeature(aFeature);
    int numberOfIF = someIF.size();
    Hashtable someSources = new Hashtable(numberOfIF); 
    for (int i = 0; i < numberOfIF; i++) {
      ProxyIndexingFeature anIF = (ProxyIndexingFeature)someIF.elementAt(i);
      // We actually don't care about the value, only the number of values 
      // will be used:
      someSources.put(anIF.getSourceDoc().getObjectIdentifier(), anIF);
    }
    someIF.makeLightweight(); // We don't need this list anymore.
    return someSources.size();
  }
   
 
  
  /** 
   * Gets the number of unique sources for a feature.
   * @param aFeature a feature
   * @return number of unique sources
   */   
  public int getNrOfUniqueSources0(DeIntern aFeature) {
    ProxyFeatureList someIF = getIFsWithFeature(aFeature);
    int numberOfIF = someIF.size();
    int sourceCount = 0;
    ProxyDocument sources[] = new ProxyDocument[numberOfIF];
    for (int i = 0; i < numberOfIF; i++) 
      sources[i] = ((ProxyIndexingFeature)someIF.elementAt(i)).getSourceDoc();
    someIF = null;
    for (int i = 0; i < numberOfIF; i++) {
      ProxyDocument curSource = sources[i];
      if (curSource != null) {
	sourceCount++;
	for (int j = i + 1; j < numberOfIF; j++)
	  if (curSource == sources[j])
	    sources[j] = null;
      }
    }
    return sourceCount;
  }
    


  /** 
   * Gets the number of unique features for a source.
   * @param aSource a source
   * @return number of unique features
   */
  public int getNrOfUniqueFeatures(IoAddrIntern aSource) {
    Hashtable someFeatures = new Hashtable();
    ProxyFeatureList someIF = getIFsWithSource(aSource);
    int numberOfIF = someIF.size();
    for (int i = 0; i < numberOfIF; i++) {
      ProxyIndexingFeature anIF = (ProxyIndexingFeature)someIF.elementAt(i);
      // We actually don't care about the value, only the number of
      // different values will be used:
      someFeatures.put(anIF.getFeature().getObjectIdentifier(), anIF);
    }
    someIF.makeLightweight(); // We don't need this list anymore.
    return someFeatures.size();
  }


    
  /** 
   * Gets the number of sources for a feature.
   *
   * @param aFeature a feature
   * @return number of sources
   */
  public final int getNrOfSources(DeIntern aFeature) {
    return getIFsWithFeature(aFeature).size();
  }
    

  /** 
   * Gets the number of features for a source.
   *
   * @param aSource a source
   * @return number of features
   */
  public final int getNrOfFeatures(IoAddrIntern aSource) {
    return getIFsWithSource(aSource).size();
  }


    
  /** 
   * Gets the feature for a DataElement.
   *
   * @param aDE a DataElement.
   * @return The feature if it exists, null otherwise.
   */          
  public final DeIntern getFeature(DataElem aDE) {
    return (DeIntern)indexingFeatures.getActualFeature(aDE);
  }
    


  /** 
   * Gets the feature for a DeIntern.
   *
   * @param aDeI a DeIntern
   * @return the feature. It is the real DeIntern stored for this
   * feature (aDE should be another one created from the same
   * DataElem).
   */
  public final DeIntern getFeature(DeIntern aDeI) {
    return (DeIntern)indexingFeatures.getActualFeature(aDeI);
  }
    


  /** 
   * Gets the indexing features with this feature.
   *
   * @param aFeature a feature
   * @return the indexing features if there are some,<br>
   * an empty Vector otherwise.
   */ 
  public final ProxyFeatureList getIFsWithFeature(DeIntern aFeature) {
    return indexingFeatures.getSourceVector(aFeature);
  }
   


  /** 
   * Gets the indexing features with this feature and the correct
   * docType and attributeName.
   *
   * @param aFeature a feature
   * @param docType the type of the document.
   * @param attributeName name of the attribute.
   * @return the indexing features
   */
  public Vector getIFsWithFeature(DeIntern aFeature,
				  String docType, String attributeName) {
    return indexingFeatures.getSourceVector(aFeature).getAsVector();
  }
  


  /** 
   * Gets the indexing features of a source.
   *	
   * @param aSource a source
   * @return The indexing features corresponding to the source as a Vector.
   * The result is empty if no indexing feature exists for this source.
   */
  public final ProxyFeatureList getIFsWithSource(IoAddrIntern aSource) {
    return indexingFeatures.getFeatureVector(aSource);
  }
    


  /** 
   * Gets the number of indexing features.
   * @return the number indexing features
   */
  public final int getNrOfIndexingFeatures() {
    return indexingFeatures.getValuesNumber();
  }


    
  /**
   * Creates an indexing feature in the Index Intern. Update()
   * has to be called before the IdxIntern is stable.
   *
   * @param a set of Indexing Features
   * @see gov.nist.nlpir.irf.feature.IndexingFeature
   */          
  public void addIndexingFeatures(Vector idxFeats) {
    flagUpToDate = false;
    int numberOfIF = idxFeats.size();
    IoAddrIntern theSource = null;
    ProxyDocument lastSource = null;
    if (numberOfIF > 0) {
      lastSource = ((ProxyIndexingFeature)idxFeats.elementAt(0)).getSourceDoc();
      theSource = new IoAddrIntern(lastSource);
    }

    for (int i = 0; i < numberOfIF; i++) {
      ProxyIndexingFeature anIF = (ProxyIndexingFeature)idxFeats.elementAt(i);

      ProxyDocument currentSource = anIF.getSourceDoc();
      if (lastSource != currentSource) {
	theSource = new IoAddrIntern(currentSource);
	lastSource = currentSource;
      }

      /* The following code has been written to ensure the unicity of equal proxies:
	 if two features contain references to proxies that are equal regarding
	 hashCode() and equals(), the code below makes the added feature point to
	 the same ProxyDataElem than all the other ones, and, most of all, the
	 DeIntern used as a key for this feature. This way, the former ProxyDataElem
	 it was pointing to can be collected. */
      ProxyDataElem sourceFeature = anIF.getFeature();
      DeIntern theFeature = (DeIntern)indexingFeatures.getActualFeature(sourceFeature);
      if (theFeature == null)
	theFeature = new DeIntern(sourceFeature);
      else
	anIF.setFeature((ProxyDataElem)theFeature.getDE());

      indexingFeatures.put(theSource, theFeature, anIF);
    }
  }



  /**
   * Returns number of indexing features for a source.
   *
   * @param aSource source document address
   * @return number of Indexing Features for a source.
   *
   */
  public final int existsSource(ProxyDocument aSource) {
    return getIFsWithSource(getActualSource(aSource)).size();
  }


    
  /**
   * Updates the index. This is a default implementation that does nothing
   * at all except setting the flagUpToDate to true. It must be redefined
   * to manage all particular aspects of specialized IdxIndexes.
   */
  public void update() {
    flagUpToDate = true;
  }


 
  /**
   * Resets data structures
   */
  public void clear() {
    indexingFeatures.clear();
    flagUpToDate = false;
  }


    
  /**
   * Gets the Indexing Features containing all of the
   * sources which have the supplied feature.
   *
   * @param aFeature feature to use as key.
   * @return a hashtable of indexing features
   */
  public Hashtable getSources(DeIntern aFeature) {
    Hashtable someSources = new Hashtable(); 
    ProxyFeatureList someIF = getIFsWithFeature(aFeature);
    int numberOfIF = someIF.size();
    for (int i = 0; i < numberOfIF; i++) {
      ProxyIndexingFeature anIF = (ProxyIndexingFeature)someIF.elementAt(i);
      someSources.put(anIF.getSourceDoc(), anIF);
    }
    return someSources;
  }
    


  /**
   * Gets features that are present in supplied source
   * @param aSource reference to source document
   * @return collection of features
   */
  public Vector getFeatures(IoAddrIntern aSource) {
    ProxyFeatureList someIF = getIFsWithSource(aSource);
    Enumeration e = someIF.elements();
    return getFeaturesFromEnumeration(e);
  }



  /**
   * Gest features out of a vector of IndexingFeatures
   * @param someIF collection of IndexingFeatures
   * @return collection of features
   */
  public final Vector getFeatures(ProxyFeatureList someIF) {
    return getFeaturesFromEnumeration(someIF.elements());
  }


    
  /**
   * Gets features out of a collection of IndexingFeatures which
   * is a subclass of Dictionary (Hashtable, PolyKeyHashtable).
   * @param someIF collection of IndexingFeatures
   * @return collection of features
   */
  public final Vector getFeatures(Dictionary someIF) {
    return getFeaturesFromEnumeration(someIF.elements());
  }


    
  /**
   * Factorizes the common code of getFeatures(*).
   *
   * @param theEnum The enumeration of IndexingFeatures.
   * @return The result of getFeatures(*)
   */
  private static final Vector getFeaturesFromEnumeration(Enumeration theEnum) {
    Vector someFeatures = new Vector();
    while (theEnum.hasMoreElements()) {
      ProxyIndexingFeature anIF = (ProxyIndexingFeature)theEnum.nextElement();
      someFeatures.addElement(anIF);
    }
    return someFeatures;
  }


    
  /**
   * Returns the features for which indexing features have been stored.
   * @return The Enumeration of Objects (should be DeIntern) used as
   * features in this IdxIntern.
   */
  public final Enumeration getFeatures() {
    return indexingFeatures.getValues();
  }
    


  /**
   * Returns the sources for which indexing features have been stored.
   * @return The Enumeration of Objects (should be IoAddrIntern) used as
   * sources in this IdxIntern.
   */
  public final Enumeration getSources() {
    return indexingFeatures.getSources();
  }



  /**
   * Returns the value of the flag uptodate.
   * @return true if the IdxIntern is uptodate, false otherwise.
   */
  public final boolean getFlagUpToDate() {
    return flagUpToDate;
  }



  /**
   * Returns the number of sources that have been stored
   * for this IdxIntern.
   */
  public final int getSourcesNumber() {
    return indexingFeatures.getSourcesNumber();
  }


    
  /**
   * Returns the number of features that have been stored
   * (NOT Indexing Features) in this IdxIntern.
   */
  public final int getFeaturesNumber() {
    return indexingFeatures.getUniqueValuesNumber();
  }



  /**
   * In order to retrieve the IoAddrIntern associated to
   * the given document, this method should be called.
   *
   * @see gov.nist.nlpir.irf.index.braf.PersistentIrfHashtable#getActualKey
   */
  public IoAddrIntern getActualSource(ProxyDocument source) {
    IoAddrIntern result =
      (IoAddrIntern)indexingFeatures.getActualSource(source);
    return (result == null) ? new IoAddrIntern(source) : result;
  }
}
