package gov.nist.nlpir.irf.index;

import gov.nist.nlpir.irf.query.QueryModalityUnit;

import gov.nist.nlpir.irf.InfoObj;
import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.indexer.Indexer;
import gov.nist.nlpir.irf.util.Support;

import java.util.Enumeration;
import java.util.Vector;


/** 
 * Provides some basic index information and access to the index features.
 * It also provides basic functionality for using the index. This class is
 * aimed to extension in order to provide specific kinds of index: with
 * Inversed Document Frequency, for example (see IdfIndex) or with Key
 * word (see KeyWordIndex).
 * 
 * @version $Revision: 1.5 $
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 *                             
 * @see InfoObj
 * @see IdfIndex
 * @see KeyWordIndex
 */

public abstract class Index extends InfoObj implements IndexIntrf, java.io.Serializable {
    

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  /** the name of the index 
   *  @serial */
  protected String indexName;      

  /** the name of the indexer used to create this index 
   *  @serial */
  protected String appliedIndexer;
  
  /** the index itself 
   *  @serial */
  /*
   * This IdxIntern has to be specialized by extending classes. If it is not
   * the entire class won't work as IFs is the most important part of Index.
   */
  protected IdxIntern IFs;
  

  /** 
   * Creates an Index with the supplied name and sets the appliedIndexer to 
   * "none". May be called by subclasses. In every case, these subclasses 
   * will have to set the IdxIntern in the Index.
   * 
   * @param name index name.
   */
  public Index(String name) {
    indexName = name;
    appliedIndexer = "none";
  }
  
  /**
   * No argument constructor only used by deserialization mechanism.
   */
  public Index() {
    indexName = "Unnamed Index";
    appliedIndexer = "none";
  }
  
  /** 
   * Gets the index name.
   * 
   * @return The index name.
   */
  public String getName() {
    return indexName;
  }
  
  
  /**
   * Returns the features for which indexing features have been stored.
   * @return The Enumeration of Objects (should be DeIntern) used as
   * features in this Index
   */

  public Enumeration getFeatures() {
    return IFs.getFeatures();
  }
    
  /**
   * Returns the sources for which indexing features have been stored.
   * @return The Enumeration of Objects (should be IoAddrIntern) used as
   * sources in this Index
   */

  public Enumeration getSources() {
    return IFs.getSources();
  }

  /** 
   * Updates the index and its features.
   */
  public void update() {
      if (Support.COARSE_TRACE <= Support.traceLevel)
	  Support.traceMessage("Index::update for " + indexName);
    IFs.update();
  }
  
  /** 
   * Clears the index and its features.
   */
  public void clear() {
    appliedIndexer = "none";
    IFs.clear();	
  }
  
  /** 
   * Executes a search and retrieves documents.
   * 
   * @param qm Query Modality Unit
   * @param setIF Vector of indexing features
   * @return Results or null if submitted to the wrong index.
   */
  public Vector retrieve(QueryModalityUnit qm, Vector setIF) {
    // check given index name
    String name = qm.getIndexName();
    if (0 != name.compareTo(indexName)) {
      Support.errorMessage("Index::Retrieve: " +
			   "wrong Index specified in QueryCondition");
      return null;
    }
    // check whether index is uptodate
    if (! IFs.getFlagUpToDate()) {
      Support.errorMessage("Index, " + indexName + ", is not uptodate!");
      return null;
    }
    return IFs.evalQueryModalityUnit(qm, setIF);
  }
  
  /** 
   *Presents the status of the index and its features.
   */
  public void present() {
    if (!IFs.getFlagUpToDate()) {
      System.out.println("Index::present, " + indexName +
			 ", is not uptodate!\n");
      return;
    }
    System.out.println("Index: " + indexName + ":");
    IFs.presentFeatures();	
  }
    
  /** 
   * Presents the status of the index and its feature statistics.
   */
  public void presentStatistics() {
    IFs.presentStatistics();
  }
  
  /**
   * Gives a String Representaiton of the Index.
   */
  public String toString() {
    return("Index {indexName: " + indexName + 
	   ", appliedIndexer: " + appliedIndexer +
	   ", IFs:" + IFs + "}");
  }
  
  /** 
   * Verifies that the same indexer is applied consistantly.
   * 
   * @param an_Indexer an indexer
   * @return true if the Index is well applied, false otherwise.
   */
  
  public boolean checkIndexer(Indexer anIndexer) {
    String indexerName = anIndexer.getClass().getName();
    if (0 == appliedIndexer.compareTo("none"))
      appliedIndexer = indexerName;
    else if (0 != appliedIndexer.compareTo(indexerName)) {
      Support.errorMessage("ERROR: indexer not the same as used before, it should be " +
			   appliedIndexer + "." );
      return false;
    }
    return true;
  }
  
  /** 
   * Add indexing features.
   * 
   * @param indexingFeatures Vector of indexing features.
   * @param anIndexer an indexer.
   * @return true if the Indexer exists and is correct,<br>
   * false otherwise
   */
    
  public boolean addIndexingFeatures(Vector indexingFeatures, Indexer anIndexer) {
    if (!checkIndexer(anIndexer))
      return false;
    IFs.addIndexingFeatures(indexingFeatures);
    return true;   
  }
  
  /**
   * statistics about the index.
   *
   * @param deep depth of the statistics in the intern index
   * @param maxLengthIf3 maximum number of characters printed.
   * @see gov.nist.nlpir.irf.index.braf.PersistentDualKeyContainer#showStatistics
   */

  public void showStatistics(int depth, int maxLengthIf3) {
    IFs.getIndexingFeatures().showStatistics(depth, maxLengthIf3);
  }
}
