package gov.nist.nlpir.irf.index;

import java.util.Enumeration;
import java.util.Dictionary;
import java.io.IOException;
import java.io.NotActiveException;
import java.io.OptionalDataException;
import java.util.NoSuchElementException;

/**
 * IrfHashtable entry. It associates a key to a value, keeping also
 * track of the hashcode so that it doesn't have to be recomputed each
 * time. The projection between the hashcode space and the main array
 * of the IrfHashtable may result result in conflicts: different keys
 * may have the same hashcode, and hashcode may project in the same place
 * in the main array. That's why there is a <code>next</code> pointer in
 * the entry, it allows to have collision lists, ans the hashcode is
 * there because in one given collision list, several hashcodes may be
 * present.
 */
class IrfHashtableEntry {
  int hash;
  Object key;
  Object value;
  IrfHashtableEntry next; // in the collision list.
  
  protected Object clone() {
    IrfHashtableEntry entry = new IrfHashtableEntry();
    entry.hash = hash;
    entry.key = key;
    entry.value = value;
    entry.next = (next != null) ? (IrfHashtableEntry)next.clone() : null;
    return entry;
  }
}

/**
 * This class is a classic implementation of an hashtable class. It adds
 * an important feature usual Hashtables don't provide: as storing in 
 * the table only uses <code>hashCode()</code> and <code>equals()</code>,
 * and that these methods can  be refined in each class to have a specific
 * behavior, it is actually  possible to access an object in the hashtable
 * with different keys, provided these keys have the same hashCode and
 * return <code>true</code> when they are compared with <code>equals()
 * </code>. This is used in DualKeyContainer and in IdxIntern. Thus, there
 * is a need for a method <code>getActualKey()</code> that returns the
 * object actually stored as a key when you access the hashtable with a 
 * false key - a key not actually stored.<br><br>
 *
 * Inheritance couldn't be used to add this feature to java.util.Hashtable
 * because the core of the Hashtable, ie the array of entries is private
 * and not protected, so that extending Hashtable doesn't allow addition
 * of this kind of feature.<br>
 * 
 * It also contains a method that reveals the length of the base table
 * array.
 *
 * @version $Revision: 1.3 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 */

public class IrfHashtable extends Dictionary
  implements Cloneable, java.io.Serializable {

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;
  
  /** The hashtable data -protected so that this class can be extended. */
  protected transient IrfHashtableEntry table[];

  /** The total number of entries in the hashtable. */
  private transient int count;

  /** @serial Rehashes the table when count exceeds this threshold. */
  private int threshold;

  /** @serial The load factor for the hashtable. */
  private float loadFactor;



  /**
   * Constructs a new, empty hashtable with the specified initial 
   * capacity and the specified load factor. 
   *
   * @param      initialCapacity   the initial capacity of the hashtable.
   * @param      loadFactor        a number between 0.0 and 1.0.
   * @exception  IllegalArgumentException  if the initial capacity is less
   *               than or equal to zero, or if the load factor is less than
   *               or equal to zero.
   */
  public IrfHashtable(int initialCapacity, float loadFactor) {
    if ((initialCapacity <= 0) || (loadFactor <= 0.0))
      throw new IllegalArgumentException();
    this.loadFactor = loadFactor;
    table = new IrfHashtableEntry[initialCapacity];
    threshold = (int)(initialCapacity * loadFactor);
  }

  

  /**
   * Constructs a new, empty hashtable with the specified initial capacity
   * and default load factor.
   *
   * @param   initialCapacity   the initial capacity of the hashtable.
   */
  public IrfHashtable(int initialCapacity) {
    this(initialCapacity, 0.75f);
  }
  
  
  
  /**
   * Constructs a new, empty hashtable with a default capacity and load
   * factor. 
   */
  public IrfHashtable() {
    this(101, 0.75f);
  }
  

  
  /**
   * Returns the number of keys in this hashtable.
   *
   * @return  the number of keys in this hashtable.
   */
  public int size() {
    return count;
  }

  
  
  /**
   * Returns the size of the base array
   *
   * @return  the length of the base array for this hashtable.
   */
  public int length() {
    return table.length;
  }
  
  /**
   * Tests if this hashtable maps no keys to values.
   *
   * @return  <code>true</code> if this hashtable maps no keys to values;
   *          <code>false</code> otherwise.
   */
  public boolean isEmpty() {
    return count == 0;
  }
  
  
  
  /**
   * Returns an enumeration of the keys in this hashtable.
   *
   * @return  an enumeration of the keys in this hashtable.
   * @see     java.util.Enumeration
   * @see     #elements()
   */
  public synchronized Enumeration keys() {
    return new IrfHashtableEnumerator(table, true);
  }
  
  
  
  /**
   * Returns an enumeration of the values in this hashtable.
   * Use the Enumeration methods on the returned object to fetch the elements
   * sequentially.
   *
   * @return  an enumeration of the values in this hashtable.
   * @see     java.util.Enumeration
   * @see     #keys()
   */
  public synchronized Enumeration values() {
    return new IrfHashtableEnumerator(table, false);
  }
  
  
  /**
   * @return an enumeration of the elements in the IrfHashtable
   */  
  public synchronized Enumeration elements() {
    return new IrfHashtableEnumerator(table, false);
  }
  
  

  /**
   * Tests if some key maps into the specified value in this hashtable.
   * This operation is more expensive than the <code>containsKey</code>
   * method.
   *
   * @param      value   a value to search for.
   * @return     <code>true</code> if some key maps to the
   *             <code>value</code> argument in this hashtable;
   *             <code>false</code> otherwise.
   * @exception  NullPointerException  if the value is <code>null</code>.
   * @see        #containsKey(java.lang.Object)
   */
  public synchronized boolean contains(Object value) {
    if (value == null)
      throw new NullPointerException();
    IrfHashtableEntry entries[] = table;
    int entryNumber = entries.length;
    for (int i = 0; i < entryNumber; i++)
      for (IrfHashtableEntry e = entries[i]; e != null; e = e.next)
	if (e.value.equals(value))
	  return true;
    return false;
  }



  /**
   * Tests if the specified object is a key in this hashtable.
   * 
   * @param   key   possible key.
   * @return  <code>true</code> if the specified object is a key in this
   *          hashtable; <code>false</code> otherwise.
   * @see     #contains(java.lang.Object)
   */
  public synchronized boolean containsKey(Object key) {
    IrfHashtableEntry entries[] = table;
    int hash = key.hashCode();
    int index = (hash & 0x7FFFFFFF) % entries.length;
    for (IrfHashtableEntry e = entries[index]; e != null; e = e.next)
      if ((e.hash == hash) && e.key.equals(key))
	return true;
    return false;
  }



  /**
   * Returns the value to which the specified key is mapped in this 
   * hashtable.
   *
   * @param   key   a key in the hashtable.
   * @return  the value to which the key is mapped in this hashtable;
   *          <code>null</code> if the key is not mapped to any value in
   *          this hashtable.
   * @see     #put(java.lang.Object, java.lang.Object)
   */
  public synchronized Object get(Object key) {
    IrfHashtableEntry entries[] = table;
    int hash = key.hashCode();
    int index = (hash & 0x7FFFFFFF) % entries.length;
    for (IrfHashtableEntry e = entries[index]; e != null; e = e.next)
      if ((e.hash == hash) && e.key.equals(key))
	return e.value;
    return null;
  }



  /**
   * Rehashes the contents of the hashtable into a hashtable with a 
   * larger capacity. This method is called automatically when the 
   * number of keys in the hashtable exceeds this hashtable's capacity 
   * and load factor. 
   */
  protected void rehash() {
    IrfHashtableEntry oldTable[] = table;
    int oldCapacity = oldTable.length;

    int newCapacity = oldCapacity * 2 + 1;
    IrfHashtableEntry newTable[] = new IrfHashtableEntry[newCapacity];

    threshold = (int)(newCapacity * loadFactor);
    table = newTable;

    for (int i = 0; i < oldCapacity; i++) {
      for (IrfHashtableEntry old = oldTable[i]; old != null; ) {
	IrfHashtableEntry e = old;
	old = old.next;

	int index = (e.hash & 0x7FFFFFFF) % newCapacity;
	e.next = newTable[index];
	newTable[index] = e;
      }
    }
  }



  /**
   * Maps the specified <code>key</code> to the specified 
   * <code>value</code> in this hashtable. Neither the key nor the 
   * value can be <code>null</code>. 
   * <p>
   * The value can be retrieved by calling the <code>get</code> method 
   * with a key that is equal to the original key. Equal here means that
   * it has the same hash code and returns true to the method equals().
   *
   * @param      key     the hashtable key.
   * @param      value   the value.
   * @return     the previous value of the specified key in this hashtable,
   *             or <code>null</code> if it did not have one.
   * @exception  NullPointerException  if the key or value is
   *               <code>null</code>.
   * @see     java.lang.Object#equals(java.lang.Object)
   * @see     #get(java.lang.Object)
   */
  public synchronized Object put(Object key, Object value) {
    if (value == null)
      throw new NullPointerException();

    // Makes sure the key is not already in the hashtable.
    IrfHashtableEntry entries[] = table;
    int hash = key.hashCode();
    int index = (hash & 0x7FFFFFFF) % entries.length;
    for (IrfHashtableEntry e = entries[index]; e != null; e = e.next)
      if ((e.hash == hash) && e.key.equals(key)) {
	Object old = e.value;
	e.value = value;
	return old;
      }

    if (count >= threshold) {
      // Rehash the table if the threshold is exceeded
      rehash();
      return put(key, value);
    } 

    // Creates the new entry.
    IrfHashtableEntry e = new IrfHashtableEntry();
    e.hash = hash;
    e.key = key;
    e.value = value;
    e.next = entries[index];
    entries[index] = e;
    count++;
    return null;
  }



  /**
   * Removes the key (and its corresponding value) from this 
   * hashtable. This method does nothing if the key is not in the hashtable.
   *
   * @param   key   the key that needs to be removed.
   * @return  the value to which the key had been mapped in this hashtable,
   *          or <code>null</code> if the key did not have a mapping.
   */
  public synchronized Object remove(Object key) {
    IrfHashtableEntry entries[] = table;
    int hash = key.hashCode();
    int index = (hash & 0x7FFFFFFF) % entries.length;
    for (IrfHashtableEntry e = entries[index], prev = null;
	 e != null; prev = e, e = e.next)
      if ((e.hash == hash) && e.key.equals(key)) {
	if (prev != null)
	  prev.next = e.next;
	else
	  entries[index] = e.next;
	count--;
	return e.value;
      }
    return null;
  }



  /**
   * Clears this hashtable so that it contains no keys. 
   */
  public synchronized void clear() {
    IrfHashtableEntry entries[] = table;
    int entryNumber = entries.length;
    for (int i = 0; i < entryNumber; i++)
      entries[i] = null;
    count = 0;
  }



  /**
   * Creates a shallow copy of this hashtable. The keys and values 
   * themselves are not cloned. 
   * This is a relatively expensive operation.
   *
   * @return  a clone of the hashtable.
   */
  public synchronized Object clone() {
    try { 
      IrfHashtable t = (IrfHashtable)super.clone();
      t.table = new IrfHashtableEntry[table.length];
      
      IrfHashtableEntry entries[] = table;
      int entryNumber = entries.length;
      for (int i = 0; i < entryNumber; i++)
	t.table[i] = (entries[i] != null) ?
	  (IrfHashtableEntry)entries[i].clone() : null;
      return t;
    } catch (CloneNotSupportedException e) { 
      throw new InternalError();
    }
  }



  /**
   * Returns a rather long string representation of this hashtable.
   *
   * @return  a string representation of this hashtable.
   */
  public synchronized String toString() {
    int max = size() - 1;
    StringBuffer buf = new StringBuffer();
    Enumeration k = keys();
    Enumeration e = elements();
    buf.append("{");

    for (int i = 0; i <= max; i++) {
      String s1 = k.nextElement().toString();
      String s2 = e.nextElement().toString();
      buf.append(s1 + "=" + s2);
      if (i < max) {
	buf.append(", ");
      }
    }
    buf.append("}");
    return buf.toString();
  }



  /**
   * WriteObject is called to save the state of the hashtable to a stream.
   * Only the keys and values are serialized since the hash values may be
   * different when the contents are restored.
   * iterate over the contents and write out the keys and values.
   * @exception IOException Various problems writing to a file
   */
  private synchronized void writeObject(java.io.ObjectOutputStream s)
    throws IOException {
    // Write out the length, threshold, loadfactor
    s.defaultWriteObject();
    
    // Write out length, count of elements and then the key/value objects
    s.writeInt(table.length);
    s.writeInt(count);
    IrfHashtableEntry entries[] = table;
    int entryNumber = entries.length;
    for (int i = 0; i < entryNumber; i++) {
      IrfHashtableEntry entry = entries[i];
      
      while (entry != null) {
	s.writeObject(entry.key);
	s.writeObject(entry.value);
	entry = entry.next;
      }
    }
  }
  
  
  
  /**
   * readObject is called to restore the state of the hashtable from
   * a stream.  Only the keys and values are serialized since the
   * hash values may be different when the contents are restored.
   * Read count elements and insert into the hashtable. 
   * @exception ClassNotFoundException Deserialization can't create object
   * @exception IOException Various problems reading from a file
   */
  private synchronized void readObject(java.io.ObjectInputStream s)
    throws IOException, ClassNotFoundException {
    // Read in the length, threshold, and loadfactor
    s.defaultReadObject();
    
    // Read the original length of the array and number of elements
    int origlength = s.readInt();
    int elements = s.readInt();

    // Compute new size with a bit of room 5% to grow but
    // No larger than the original size.  Make the length
    // odd if it's large enough, this helps distribute the entries.
    // Guard against the length ending up zero, that's not valid.
    int length = (int)(elements * loadFactor) + (elements / 20) + 3;
    if (length > elements && (length & 1) == 0)
      length--;
    if (origlength > 0 && length > origlength)
    length = origlength;
    
    table = new IrfHashtableEntry[length];
    count = 0;
    
    // Read the number of elements and then all the key/value objects
    for (; elements > 0; elements--) {
      Object key = s.readObject();
      Object value = s.readObject();
      put(key, value);
    }
  }
  

  
  /**
   * Returns the object actually stored as a key for this possible key.
   * The returned object will thus have the same hash code as key, and
   * they will return true if equals is called on them.
   * 
   * @param   key   possible key.
   * @return  <code>true</code> if the specified object is a key in this
   *          hashtable; <code>false</code> otherwise.
   * @see     #contains(java.lang.Object)
   */
  public synchronized Object getActualKey(Object key) {
    IrfHashtableEntry entries[] = table;
    int hash = key.hashCode();
    int index = (hash & 0x7FFFFFFF) % entries.length;
    for (IrfHashtableEntry e = entries[index]; e != null; e = e.next)
      if ((e.hash == hash) && e.key.equals(key))
	return e.key;
    return null;
  }
}

/**
 * A hashtable enumerator class.  This class should remain opaque 
 * to the client. It will use the Enumeration interface. 
 */
class IrfHashtableEnumerator implements Enumeration {
  boolean keys;
  int index;
  IrfHashtableEntry table[];
  IrfHashtableEntry entry;
  
  IrfHashtableEnumerator(IrfHashtableEntry table[], boolean keys) {
    this.table = table;
    this.keys = keys;
    this.index = table.length;
  }
  
  public boolean hasMoreElements() {
    if (entry != null)
      return true;
    while (index-- > 0)
      if ((entry = table[index]) != null)
	return true;
    return false;
  }
  
  public Object nextElement() {
    if (entry == null)
      while ((index-- > 0) && ((entry = table[index]) == null));
    if (entry != null) {
      IrfHashtableEntry e = entry;
      entry = e.next;
      return keys ? e.key : e.value;
    }
    throw new NoSuchElementException("IrfHashtableEnumerator");
  }
}
