package gov.nist.nlpir.irf.index.braf;

import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.de.ProxyDataElem;
import gov.nist.nlpir.irf.feature.IndexingFeature;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.index.braf.ProxyFeatureList;

import java.util.Vector;
import java.io.IOException;
import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

/**
 * Internal representation of a Data Element, this class, unlike DataElem,
 * won't be inherited to describle every internal data element that may be
 * encountered. This class is part of the PersistentDualKeyContainer
 * scheme. It manages its own persistence mechanism with <code>writeTo()
 * </code> and <code>readFrom()</code>. It could still be used with
 * another scheme as it provides methods used by indexing and retrieving,
 * for computation and storage of scores mainly.
 *
 * @version $Revision: 1.3 $ 	
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 *
 * @see gov.nist.nlpir.irf.de.DataElem
 * @see gov.nist.nlpir.irf.index.braf.PersistentDualKeyContainer
 */

public class DeIntern {
 
  /** Number of documents in which equivalent DataElements appear. */
  private int nrOfSources;

  /**
   * Score of the data element, calculated with different methods
   * provided by the IdxIntern subclass used.
   * @see gov.nist.nlpir.irf.index.IdxIntern
   */
  private double score;

  /** A DataElem equivalent to all the DataElems this DeInterns represents. */
  private DataElem de;
    
  /** A dirty bit for persistence purpose. */
  private transient boolean dirty;

  /**
   * This constructor initializes the DeIntern to default values and
   * positions the reference to the DataElem the DeIntern is linked to.
   */
  public DeIntern(DataElem de) {
    this.de = de;
    nrOfSources = 0;
    score = 0;
    dirty = false;
    ((ProxyDataElem)de).makePersistent();
    ((ProxyDataElem)de).addRefToInMemoryProxiesByOid();
  }



  /**
   * This constructor is only used for persistence reasons. It is used
   * by <code>readFrom()</code> to recreate the DeIntern from the file
   * given as a parameter.
   */  
  private DeIntern(BufferedRandomAccessFile in) throws IOException {
    de = gov.nist.nlpir.irf.proxy.PersistentObjectManager.buildDE(in);
    nrOfSources = in.readInt();
    score = in.readDouble();
    dirty = false;
  }


  
  /**
   * @return only the representation of the internal data element.<br>
   * caution: it is not the same function as toString().
   */
  public String getValueAsString() {
    return de.getValueAsString();
  }


  
  /**
   * @return the representation of both the internal data element
   * and the data element represented itself.
   */
  public String toString() {
    return "DEinDEI:" + de;
  }


  
  /**
   * Compares this DeIntern to another object. If the given object is
   * a DeIntern, the function returns true if the represented DataElem
   * are equals. If it is a data element, it returns true if this data
   * element is the one represented by this DeIntern.
   *
   * @return true if the represented DataElem are equals
   * @see ProxyDataElem#equals
   */
  public boolean equals(Object o) {
    if (o instanceof DeIntern)
      return (de.equals(((DeIntern)o).getDE()));
    else if (o instanceof DataElem)
      return ((ProxyDataElem)de).equals((DataElem)o);
    Support.errorMessage("DeIntern.equals was called with a " +
			 o.getClass().getName() + " when a DataElem " +
			 "or a DeIntern was expected.");
    return false;
  }


  
  /**
   * Retrieves the represented data element.
   */
  public final DataElem getDE() {
    return de;
  }


  
  /**
   * Sets the represented data element.
   */
  public final void setDE(ProxyDataElem de) {
    this.dirty = true;
    if (this.de != null)
      ((ProxyDataElem)this.de).deleteRefFromInMemoryProxiesByOid();
    if (de != null)
      de.addRefToInMemoryProxiesByOid();
    this.de = de;
  }


  
  /**
   * Gives the number of sources the internal data element appears in.
   */
  public final int getNrOfSources() {
    return nrOfSources;	
  }


  
  /**
   * Sets the number of sources to the value given.
   */
  public final void setNrOfSources(int i) {
    this.dirty = true;
    nrOfSources = i;
  }


  
  /**
   * Gives the score of the represented data element.
   */
  public final double getScore() {
    return score;
  }


  
  /**
   * Sets the score of the data element linked to this.
   */
  public final void setScore(double s) {
    this.dirty = true;
    score = s;
  }


  
  /**
   * Gives the number of times this appears in the given dictionary
   * considered as a collection of DeIntern. Uses the values
   * and doesn't compare the instances.
   * @param features vector of indexing features.
   * @param startIndex The index in the vector where the computation
   * should start.
   * @return frequency of this feature in supplied vector of
   * indexing features.
   */
  public final int getFrequency(Vector features, int startIndex) {
    int frequ = 1; /* See IdfIdxIntern.calcDocumentScore(): we can
		      avoid one comparison in this loop by assuming
		      the first is always true. */
    int featuresSize = features.size();
    for (int i = startIndex; i < featuresSize; i++) {
      ProxyDataElem aDE = (ProxyDataElem)features.elementAt(i);
      // if (aDE != null) // null indicates it's been counted
      // Got rid of this test because it is useless,
      // we can't have (aDE == null == de).
      if (aDE == de) {
	frequ++;
	features.setElementAt(null, i); // don't count this one again
      }
    }
    return frequ;
  }


  
  /**
   * Gives a hash code based on the value represented.
   */
  public final int hashCode() {
    return de.hashCode();
  }


  
  /**
   * Dematerializing method for the tuned serialization scheme used
   * mainly by HashBlock.
   */
  public synchronized void writeTo(BufferedRandomAccessFile out)
    throws IOException {
    gov.nist.nlpir.irf.proxy.PersistentObjectManager.writeDE((ProxyDataElem)de, out);
    out.writeInt(nrOfSources);
    out.writeDouble(score);
  }


  
  /**
   * The symmetric method of <code>writeTo()</code>.
   */
  public static DeIntern readFrom(BufferedRandomAccessFile in)
    throws IOException {
    return new DeIntern(in);
  }



  /**
   * Gets the dirty bit. Mainly used to know whether the DeIntern must
   * be saved when it is dumped out of memory or not.
   */
  public final boolean isDirty() {
    return dirty;
  }



  /**
   * As a DeIntern maintains a reference to a ProxyDataElem, it needs
   * a finalize() method to unregister from this Proxy.
   */
  public void finalize() {
    ((ProxyDataElem)de).deleteRefFromInMemoryProxiesByOid();
  }
}

