package gov.nist.nlpir.irf.index.braf;

import gov.nist.nlpir.irf.proxy.VirtualProxy;
import gov.nist.nlpir.irf.proxy.Oid;

import gov.nist.nlpir.irf.index.braf.IoAddrIntern;
import gov.nist.nlpir.irf.index.braf.DeIntern;

import java.io.IOException;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

/**
 * HashBlock entry. It associates a key to a value, keeping also track
 * of the hashcode so that it doesn't have to be recomputed each time.
 * The projection between the hashcode space and the main array of the
 * HashBlock may result result in conflicts: different keys may have the
 * same hashcode, and hashcode may project in the same place in the main
 * array. That's why there is a <code>next</code> pointer in the entry,
 * it allows to have collision lists, ans the hashcode is there because
 * in one given collision list, several hashcodes may be present.
 */
class HashBlockEntry {
  /** Key hashcode. Thus, it doesn't have to be recalculated each time. */
  int hash;
  /** Key the value is maped to. */
  Object key;
  /** Can only be a ProxyFeatureList or its Oid
      if it hasn't been materialized yet. */
  Object value;
  /** Links the next entry in the collision list. */
  HashBlockEntry next;

  /**
   * Overrides the Object method, used by HashBlock.clone().
   */
  protected Object clone() {
    HashBlockEntry entry = new HashBlockEntry();
    entry.hash = hash;
    entry.key = key;
    entry.value = value;
    entry.next = (next != null) ? (HashBlockEntry)next.clone() : null;
    return entry;
  }
}


/**
 * This class is a classic implementation of a hash table, except that it
 * adds an important feature usual hashtables don't provide: as storing in
 * the table only uses <code>hashCode()</code> and <code>equals()</code>,
 * and that these methods can be refined in each class to have a specific
 * behavior, it is actually possible to access an object in the hashtable
 * with different keys, provided these keys have the same hashCode and
 * return <code>true</code> when they are compared with <code>equals()
 * </code>. This is used in PersistentDualKeyContainer and in IdxIntern.
 * Thus, there is a need for a method getActualKey that returns the object
 * actually stored as the key when you access the hashblock with a false
 * key -a key not actually stored.<br>
 * 
 * It also contains a method that reveals the length of the base table
 * array.
 *
 * @version $Revision: 1.3 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance.
 */
public class HashBlock implements Cloneable {
  
  /** The hash table data. Is protected so that this class can be extended. */
  protected transient HashBlockEntry table[];
  
  /** The total number of entries in the hashblock. */
  private transient int count;
  
  /** Rehashes the hashblock when count exceeds this threshold. */
  private int rehashThreshold;
  
  /** The load factor for the hashblock. */
  private float loadFactor;
  
  /** This class object will be used for materialization of lists
      stored in the block. */
  static Class proxyFeatureListClass = gov.nist.nlpir.irf.index.braf.ProxyFeatureList.class;
  
  /** Dirty bit. Only reflects the fact a ProxyFeatureList has been
      added. To see whether a key has changed or not, going through
      every key is currently the only possibility. */
  private boolean dirty;



  /**
   * Constructs a new, empty hashblock with the specified initial 
   * capacity and the specified load factor. 
   *
   * @param      initialCapacity   the initial capacity of the hashtable.
   * @param      loadFactor        a number between 0.0 and 1.0.
   * @exception  IllegalArgumentException  if the initial capacity is less
   *               than or equal to zero, or if the load factor is less than
   *               or equal to zero.
   */
  public HashBlock(int initialCapacity, float loadFactor) {
    if ((initialCapacity <= 0) || (loadFactor <= 0.0))
      throw new IllegalArgumentException();
    this.loadFactor = loadFactor;
    this.dirty = false;
    this.count = 0;
    table = new HashBlockEntry[initialCapacity];
    rehashThreshold = (int)(initialCapacity * loadFactor);
  }
  


  /**
   * Constructs a new, empty hashtable with the specified initial capacity
   * and default load factor.
   *
   * @param   initialCapacity   the initial capacity of the hashtable.
   */
  public HashBlock(int initialCapacity) {
    this(initialCapacity, 0.75f);
  }
  


  /**
   * Constructs a new, empty hashtable with a default capacity and load
   * factor. 
   */
  public HashBlock() {
    this(101, 0.75f);
  }



  /**
   * private constructor only used by the persistence mechanism, when coming
   * back from disk.
   */
  private HashBlock(HashBlockEntry table[], int count, int rehashThreshold,
		    float loadFactor) {
    this.table = table;
    this.count = count;
    this.rehashThreshold = rehashThreshold;
    this.loadFactor = loadFactor;
    this.dirty = false;
  }



  /**
   * @return  the number of keys in this hashtable.
   */
  public final int size() {
    return count;
  }
  


  /**
   * @return  the length of the base array for this hashtable.
   */
  public final int length() {
    return table.length;
  }


  
  /**
   * Tests if this hashtable maps no keys to values.
   *
   * @return <code>true</code> if this hashtable maps no keys to values;
   *         <br><code>false</code> otherwise.
   */
  public final boolean isEmpty() {
    return count == 0;
  }


  
  /**
   * Returns an enumeration of the keys in this hashtable.
   *
   * @return  an enumeration of the keys in this hashtable.
   * @see     java.util.Enumeration
   * @see     gov.nist.nlpir.irf.index.braf.HashBlock#elements()
   */
  public final synchronized Enumeration keys() {
    return new HashBlockEnumerator(table, true);
  }


  
  /**
   * Returns an enumeration of the values in this hashtable.
   * Use the Enumeration methods on the returned object to
   * fetch the elements sequentially.
   *
   * @return  an enumeration of the values in this hashtable.
   * @see     java.util.Enumeration
   * @see     gov.nist.nlpir.irf.index.braf.HashBlock#keys()
   */
  public final synchronized Enumeration elements() {
    return new HashBlockEnumerator(table, false);
  }
  


  /**
   * Tests if some key maps into the specified value in this
   * HashBlock. This operation is more expensive than the
   * <code>containsKey()</code> method.
   *
   * @param      value   a value to search for.
   * @return     <code>true</code> if some key maps to the
   *             <code>value</code> argument in this hashtable;
   *             <code>false</code> otherwise.
   * @exception  NullPointerException  if the value is <code>null</code>.
   * @see        gov.nist.nlpir.irf.index.braf.HashBlock#containsKey(java.lang.Object)
   */
  public synchronized boolean contains(Object value) {
    if (value == null)
      throw new NullPointerException();

    HashBlockEntry entries[] = table;
    int entryNumber = entries.length;
    
    for (int i = 0; i < entryNumber; i++)
      for (HashBlockEntry e = entries[i]; e != null; e = e.next)
	if (e.value.equals(value))
	  return true;
    return false;
  }


  
  /**
   * Tests if the specified object is a key in this hashtable.
   * 
   * @param   key   possible key.
   * @return  <code>true</code> if the specified object is a key in this
   *          hashtable; <code>false</code> otherwise.
   * @see     gov.nist.nlpir.irf.index.braf.HashBlock#contains(java.lang.Object)
   */
  public synchronized boolean containsKey(Object key) {
    HashBlockEntry entries[] = table;
    int hash = key.hashCode();
    int index = (hash & 0x7FFFFFFF) % entries.length;
    for (HashBlockEntry e = entries[index]; e != null; e = e.next)
      if ((e.hash == hash) && e.key.equals(key))
	return true;
    return false;
  }
  


  /**
   * Returns the value to which the specified key is mapped in this
   * hashtable.
   *
   * @param   key   a key in the hashtable.
   * @return  the value to which the key is mapped in this hashtable;
   *          <code>null</code> if the key is not mapped to any value in
   *          this hashtable.
   * @see     gov.nist.nlpir.irf.index.braf.HashBlock#put(java.lang.Object, int,
   *                                  java.lang.Object)
   */
  public synchronized Object get(Object key, int hash) {
    HashBlockEntry entries[] = table;
    int index = hash % entries.length;
    for (HashBlockEntry e = entries[index]; e != null; e = e.next)
      if ((e.hash == hash) && e.key.equals(key)) {
	/* On demand materialization:
	   When a HashBlock comes back from disk, a large part of the
	   ProxyFeatureLists it contains may never be accessed before
	   the block is dumped to disk again. That's why the actual
	   construction of the proxy only happens when it is needed,
	   here for get. The Oid, that was considered the value, is
	   then exchanged for the Proxy, so that the proxy will come
	   up without having to be built in subsequent calls to get
	   with the same key. */
	Object value = e.value;
	if (value instanceof ProxyFeatureList)
	  return value;
	else {
	  ProxyFeatureList realValue = (ProxyFeatureList)VirtualProxy.
	    getProxyFor(proxyFeatureListClass, (Oid)value);
	  e.value = realValue;
	  return realValue;
	}
      }
    return null;
  }


  
  /**
   * Rehashes the contents of the HashBlock into a HashBlock with a
   * larger capacity. This method is called automatically when the
   * number of keys in the HashBlock exceeds this HashBlock's capacity
   * and load factor. 
   */
  protected synchronized void rehash() {
    dirty = true;
    HashBlockEntry entries[] = table;
    int oldCapacity = entries.length;
    
    int newCapacity = oldCapacity * 2 + 1;
    HashBlockEntry newTable[] = new HashBlockEntry[newCapacity];
    
    rehashThreshold = (int)(newCapacity * loadFactor);
    
    for (int i = 0; i < oldCapacity; i++) {
      HashBlockEntry old = entries[i];
      while (old != null) {
	int index = old.hash % newCapacity;
	HashBlockEntry next = old.next;
	old.next = newTable[index];
	newTable[index] = old;
	old = next;
      }
    }

    table = newTable;
  }


    
  /**
   * Maps the specified <code>key</code> to the specified 
   * <code>value</code> in this hashtable. Neither the key nor the 
   * value can be <code>null</code>. 
   * <p>
   * The value can be retrieved by calling the <code>get</code> method 
   * with a key that is equal to the original key. Equal here means that
   * it has the same hash code and returns true to the method equals().
   *
   * @param      key     the hashblock key. Must either be an IoAddrIntern
   *                     or a DeIntern.
   * @param      hash    hashCode of the key. As it has already been computed,
   *                     it saves a few CPU cycles.
   * @param      value   the value. During put, it should be ProxyFeatureList
   *                     even if its Oid is actually enough.
   * @return     the previous value of the specified key in this hashtable,
   *             or <code>null</code> if it did not have one.
   * @exception  NullPointerException  if the key or value is
   *             <code>null</code>.
   * @see java.lang.Object#equals(java.lang.Object)
   * @see gov.nist.nlpir.irf.index.braf.HashBlock#get(java.lang.Object, int)
   */
  public synchronized Object put(Object key, int hash, Object value) {
    dirty = true;
    // Make sure the value is not null
    if (value == null)
      throw new NullPointerException();
    
    // Persistence management.
    if (value instanceof VirtualProxy) {
      ((VirtualProxy)value).makePersistent();
      ((VirtualProxy)value).addRefToInMemoryProxiesByOid();
    }

    // Makes sure the key is not already in the hashtable.
    HashBlockEntry entries[] = table;
    int index = hash % entries.length;
    for (HashBlockEntry e = entries[index]; e != null; e = e.next)
      if ((e.hash == hash) && e.key.equals(key)) {
	Object old = e.value;
	if (old instanceof VirtualProxy)
	  ((VirtualProxy)old).deleteRefFromInMemoryProxiesByOid();
	e.value = value;
	return old;
      }
    
    if (count >= rehashThreshold) {
      // Rehash the table if the threshold is exceeded
      rehash();
      return put(key, hash, value);
    } 
    
    // Creates the new entry.
    HashBlockEntry e = new HashBlockEntry();
    e.hash = hash;
    e.key = key;
    e.value = value;
    e.next = entries[index];
    entries[index] = e;
    count++;
    return null;
  }
  


  /**
   * Removes the key (and its corresponding value) from this 
   * hashtable. This method does nothing if the key is not in the hashtable.
   *
   * @param   key   the key that needs to be removed.
   * @return  the value to which the key had been mapped in this hashtable,
   *          or <code>null</code> if the key did not have a mapping.
   */
  public synchronized Object remove(Object key) {
    dirty = true;
    HashBlockEntry entries[] = table;
    int hash = key.hashCode();
    int index = (hash & 0x7FFFFFFF) % entries.length;
    for (HashBlockEntry e = entries[index], prev = null; e != null;
	 prev = e, e = e.next)
      if ((e.hash == hash) && e.key.equals(key)) {
	if (prev != null) 
	  prev.next = e.next;
	else
	  entries[index] = e.next;
	count--;
	if (e.value instanceof VirtualProxy)
	  ((VirtualProxy)e.value).deleteRefFromInMemoryProxiesByOid();
	return e.value;
      }
    return null;
  }


  
  /**
   * Clears this hashblock so that it contains no keys. 
   */
  public synchronized void clear() {
    dirty = true;
    HashBlockEntry entries[] = table;
    int entryNumber = entries.length;
    for (int i = 0; i < entryNumber; i++)
      entries[i] = null;
    count = 0;
  }
  


  /**
   * Creates a clone of this hashblock. The keys and values 
   * themselves are not cloned. 
   *
   * @return  a clone of the hashtable.
   */
  public synchronized Object clone() {
    try { 
      HashBlock t = (HashBlock)super.clone();
      HashBlockEntry entries[] = table;
      int entryNumber = entries.length;
      t.table = new HashBlockEntry[entryNumber];
      for (int i = 0; i < entryNumber; i++)
	t.table[i] = (entries[i] != null) ?
	  (HashBlockEntry)entries[i].clone() : null;
      return t;
    } catch (CloneNotSupportedException e) { 
      throw new InternalError();
    }
  }


  
  /**
   * Returns a string representation of the hashblock.
   *
   * @return  a string representation of the hashblock.
   */
  public synchronized String toString() {
    int max = size() - 1;
    StringBuffer buf = new StringBuffer();
    Enumeration k = keys();
    Enumeration e = elements();
    buf.append("{");
    
    for (int i = 0; i <= max; i++) {
      String s1 = k.nextElement().toString().substring(0,20);
      Object nextEl =  e.nextElement();
      String s2 = "";
      if (nextEl instanceof Oid)
	s2 = nextEl.toString();
      else
	s2 = ((ProxyFeatureList)nextEl).getObjectIdentifier().toString();
      buf.append(s1 + "=" + s2);
      if (i < max)
	buf.append(", ");
    }
    buf.append("}");
    return buf.toString();
  }
  


  /**
   * Returns the object actually stored as a key for this possible key.
   * The returned object will thus have the same hash code as key, and
   * they will return true if equals is called on them.
   * 
   * @param   key   possible key.
   * @return  <code>true</code> if the specified object is a key in this
   *          hashtable; <code>false</code> otherwise.
   * @see     gov.nist.nlpir.irf.index.braf.HashBlock#contains(java.lang.Object)
   */
  public synchronized Object getActualKey(Object key, int hash) {
    HashBlockEntry entries[] = table;
    int index = hash % entries.length;
    for (HashBlockEntry e = entries[index]; e != null; e = e.next)
      if ((e.hash == hash) && e.key.equals(key))
	return e.key;
    return null;
  }



  /**
   */  
  public boolean isDirty() {
    if (dirty)
      return true;

    /* Let's see if no key has changed since last materialization/creation. */

    // First, let's find out if we store IoAddrInterns or DeInterns:
    boolean storedIOAI = false;
    HashBlockEntry entries[] = table;
    int entryNumber = entries.length;
    for (int i = 0; i < entryNumber; i++)
      if (entries[i] != null) {
	storedIOAI = (entries[i].key instanceof IoAddrIntern);
	break;
      }
    
    if (storedIOAI)
      for (int i = 0; i < entryNumber; i++) {
	HashBlockEntry entry = entries[i];
	while (entry != null) {
	  if (((IoAddrIntern)entry.key).isDirty())
	    return true;
	  entry = entry.next;
	}
      }
    else
      for (int i = 0; i < entryNumber; i++) {
	HashBlockEntry entry = entries[i];
	while (entry != null) {
	  if (((DeIntern)entry.key).isDirty())
	    return true;
	  entry = entry.next;
	}
      }
    
    return false;
  }



  /**
   * Writes this block to the file given in parameter.
   */
  public synchronized void writeTo(BufferedRandomAccessFile out) 
    throws IOException {
    out.writeInt(rehashThreshold);
    out.writeFloat(loadFactor);
    out.writeInt(table.length);
    out.writeInt(count);

    /* Find the first non-null key and set the boolean telling
     * whether we are storing IoAddrInterns or DeInterns. */
    int i = 0;
    while (table[i] == null)
      i++;
    boolean storedIOAI = (table[i].key instanceof IoAddrIntern);
    out.writeBoolean(storedIOAI);

    for (int index = 0; index < table.length; index++) {
      HashBlockEntry entry = table[index];
      
      while (entry != null) {
	if (storedIOAI)
	  ((IoAddrIntern)entry.key).writeTo(out);
	else
	  ((DeIntern)entry.key).writeTo(out);
	out.writeInt(entry.hash);
	if (entry.value instanceof Oid)
	  ((Oid)entry.value).writeTo(out);
	else
	  (((VirtualProxy)entry.value).getObjectIdentifier()).writeTo(out);
	entry = entry.next;
      }
    }
  }



  /**
   * Reads a HashBlock back from the file it was previously written
   * with <code>writeTo()</code>.
   */
  public synchronized static HashBlock readFrom(BufferedRandomAccessFile in)
    throws IOException {
    int rehashThreshold = in.readInt();
    float loadFactor = in.readFloat();
    int origLength = in.readInt();
    int count = in.readInt();
    
    // Are we reading IoAddrInterns or DeInterns ?
    boolean storedIOAI = in.readBoolean();

    // Compute new size with a bit of room 5% to grow but
    // No larger than the original size.  Make the length
    // odd if it's large enough, this helps distribute the entries.
    // Guard against the length ending up zero, that's not valid.
    int length = (int)(count * loadFactor) + (count / 20) + 3;
    if (length > count && (length & 1) == 0)
      length--;
    if (origLength > 0 && length > origLength)
      length = origLength;
    
    // This will be the actual table in the materialized HashBlock.
    HashBlockEntry table[] = new HashBlockEntry[length];

    Object key = null;
    
    // Read the number of elements and then all the key/value objects
    for (int i = 0; i < count; i++) {
      if (storedIOAI)
	key = IoAddrIntern.readFrom(in);
      else
	key = DeIntern.readFrom(in);
      int hash = in.readInt();
      Object valueOid = Oid.readFrom(in);
      
      HashBlockEntry current = new HashBlockEntry();
      current.key = key;
      current.hash = hash;
      current.value = valueOid;
      
      int index = hash % length;   // Because of the "projection", this operation may lead
      current.next = table[index]; // to a slot already used in the table for objects with
      table[index] = current;      /* a different hashcode. That's why we have to link the
				      "may-be-there" list to the new one. (line 2 / 3) */
    }

    return new HashBlock(table, count, rehashThreshold, loadFactor);
  }



  /**
   * As this HashBlock contains Proxies and has counted the references
   * it has to them, it needs to remove those references before being
   * finalized, and this is done in this method.
   */
  public void finalize() {
    HashBlockEntry entries[] = table;
    int entryNumber = entries.length;
    for (int i = 0; i < entryNumber; i++)
      for (HashBlockEntry e = entries[i]; e != null; e = e.next)
	if (e.value instanceof VirtualProxy)
	  ((VirtualProxy)e.value).deleteRefFromInMemoryProxiesByOid();
  }
}

/**
 * A HashBlock enumerator class.  This class should remain opaque 
 * to the client. It will use the Enumeration interface. 
 */
class HashBlockEnumerator implements Enumeration {
  boolean keys; /* true if the enumerator concerns the keys, false if it is
		 * for the values. */
  int index;
  HashBlockEntry table[];
  HashBlockEntry entry; /* Except right after creation, points to the
			   next object to be returned by nextElement()
			   and not to the last one. */
  
  HashBlockEnumerator(HashBlockEntry table[], boolean keys) {
    this.table = table;
    this.keys = keys;
    this.index = table.length;
  }
  
  public boolean hasMoreElements() {
    if (entry != null)
      return true;

    while (index-- > 0)
      if ((entry = table[index]) != null)
	return true;

    return false;
  }
  
  public Object nextElement() {
    if (entry == null)
      while ((index-- > 0) && ((entry = table[index]) == null));

    if (entry != null) {
      HashBlockEntry e = entry;
      entry = e.next;
      if (keys)
	return e.key;
      else {
	// Usual on-demand materialization management:
	Object value = e.value;
	if (value instanceof ProxyFeatureList)
	  return value;
	else {
	  ProxyFeatureList realValue =
	    (ProxyFeatureList)VirtualProxy.getProxyFor(HashBlock.proxyFeatureListClass, (Oid)value);
	  e.value = realValue;
	  return realValue;
	}
      }
    }
    throw new NoSuchElementException("HashBlockEnumerator");
  }
}
