package gov.nist.nlpir.irf.index.braf;

import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.util.Support;

import java.io.IOException;
import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

/**
 * This class is the internal representation of a source, ie a document.
 * It will be used as a key in the different indexes, but it also embeds
 * information only used by indexes for retrieval: for instance, the
 * score of the document, calculated by every index in a manner given by
 * the index (IDF, keyword, ...), and the number of features this source
 * refers to (in a text application, for example, the number of indexed
 * terms found in this source for the considered fields -depending on the
 * index). This class has been designed to be used in close conjunction
 * with PersistentDualKeyContainer; it manages its own persistence
 * mechanism, exactly like DeIntern. It may still be used in a different
 * scheme in order to take advantage of the already implemented methods
 * in IdxIntern and its subclasses without having to change them.
 *
 * @version $Revision: 1.2 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance.
 *
 * @see gov.nist.nlpir.irf.index.braf.DeIntern
 * @see gov.nist.nlpir.irf.index.braf.PersistentDualKeyContainer
 */  
public class IoAddrIntern {

  /** Number of features found in the considered source. */    
  private int nrOfFeatures;
  /** Score of this source, computed with an index-dependant heuristic. */
  private double score;
  /** The actual source this IoAddrIntern deals with. */
  private ProxyDocument doc;
  /** Dirty bit. */
  private transient boolean dirty;


    
  /** 
   * Constructor which initializes the source to doc, nrOfFeatures to 0
   * and score to 0.0.
   * 
   * @param anIO_Addr a ProxyDocument.
   */
  public IoAddrIntern(ProxyDocument doc) {
    nrOfFeatures = 0;
    score = 0.0;
    this.doc = doc;
    this.dirty = false;
    doc.makePersistent();
    doc.addRefToInMemoryProxiesByOid();
  }



  /**
   * This constructor is only called by <code>readFrom()</code>.
   */
  private IoAddrIntern(BufferedRandomAccessFile in) throws IOException {
    doc = gov.nist.nlpir.irf.proxy.PersistentObjectManager.buildDoc(in);
    nrOfFeatures = in.readInt();
    score = in.readDouble();
    dirty = false;
  }



  /** 
   * Returns the score of the IoAddrIntern.
   * 
   * @return the score.
   */
  public final double getScore() {
    return score;
  }


  
  /** 
   * Returns the number of features of the IoAddrIntern.
   * 
   * @return the number of features
   */
  public final int getNrOfFeatures() {
    return nrOfFeatures;
  }


    
  /** 
   * Sets the number of features of the IoAddrIntern to the
   * supplied value.
   * 
   * @param n number of features.
   */
  public final void setNrOfFeatures(int n) {
    dirty = true;
    nrOfFeatures = n;
  }
  


  /** 
   * Sets the score of the IoAddrIntern to the supplied value.
   *
   * @param s the score
   */
  public final void setScore(double s) {
    dirty = true;
    score = s;
  }


  
  /** 
   * Returns the ProxyDocument of the IoAddrIntern.
   * 
   * @return the doc
   */
  public final ProxyDocument getDoc() {
    return doc;
  }


  
  /** 
   * Sets the ProxyDocument of the IoAddrIntern.
   */
  public final void setDoc(ProxyDocument doc) {
    dirty = true;
    if (this.doc != null)
      this.doc.deleteRefFromInMemoryProxiesByOid();
    this.doc = doc;
    if (this.doc != null)
      this.doc.addRefToInMemoryProxiesByOid();
  }


  
  /** 
   * Generates the string representation of this IoAddrIntern.
   * 
   * @return A string containing the number of features, score,
   * doc and addr fields.
   */
  public String toString() {
    return ("{IoAddrIntern: " + "{# of features: " + nrOfFeatures + "}, {score: "
	    + score + "}, " + "{doc: " + doc.toString().substring(6, 20) + "...}}");
  }


  
  /**
   * Gives a hashcode based on the source document, otherwise indexing
   * would encounter problems (features of the same source not indexed
   * as being in the same source).<br>
   * Note: The default hashCode can only be accessed via super.hashCode().
   */
  public int hashCode() {
    return doc.hashCode();
  }


  
  /**
   * Completely overrides the equals in Object so that it is the
   * one called by Hashtable for example.
   *
   * @return true only if the given object is a IoAddrIntern and
   * concerns the same document (instance comparison, assumes there
   * is only one instance for a given document).
   */
  public boolean equals(Object o) {
    if (o instanceof IoAddrIntern)
      return doc == ((IoAddrIntern)o).getDoc();
    else if (o instanceof ProxyDocument)
      return doc == o;
    Support.errorMessage("IoAddrIntern.equals expects an IoAddrIntern. A " +
			 o.getClass().getName() + " was passed.");
    return false;
  }


  
  /**
   * Tuned serialization mechanism. Uses the IdxFeatBrafBroker methods to
   * write the correct information.
   */
  public synchronized void writeTo(BufferedRandomAccessFile out)
    throws IOException {
    gov.nist.nlpir.irf.proxy.PersistentObjectManager.writeDoc(doc, out);
    out.writeInt(nrOfFeatures);
    out.writeDouble(score);
  }



  /**
   * Tuned deserialization mechanism. Also needs the IdxFeatBrafBroker
   * mechanism to be able to read a IoAddrIntern back from disk.
   */
  public static IoAddrIntern readFrom(BufferedRandomAccessFile in)
    throws IOException {
    return new IoAddrIntern(in);
  }


  
  /**
   * Returns the dirty bit status.
   */
  public final boolean isDirty() {
    return dirty;
  }



  /**
   * As this class contains a reference to a proxy, it needs a finalize
   * method to complete housekeeping about references to proxies.
   */
  public void finalize() {
    doc.deleteRefFromInMemoryProxiesByOid();
  }
}
