package gov.nist.nlpir.irf.index.braf;

import java.io.IOException;
import java.util.Vector;

import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.util.Support;

import gov.nist.nlpir.irf.proxy.VirtualProxy;
import gov.nist.nlpir.irf.proxy.Oid;
import gov.nist.nlpir.irf.handle.braf.FeatureListBrafHandle;
import gov.nist.nlpir.irf.proxy.PersistentObjectManager;
import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

/**
 * Pool of multiple linked lists.
 *
 * @version $Revision: 1.5 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 */
public class LinkedListPool {

  /** @serial As a pool is only contained in one file, only one offset has 
   * to be managed. */
  long nextOffset;

  /** @serial Unlike some different mechanisms, this is directly the 
   * filename, not just a prefix. */
  String poolFileName;


  /** The file in which every feature list will be stored. */
  BufferedRandomAccessFile poolFile = null;

  /** The local equivalent of <code>null</code> */
  public static long END_MARKER = -1;
  


  /**
   * @param poolFilePrefix prefix of poolFile file name.
   * @param initialize true if building a new listpool
   */
  public LinkedListPool(String poolFileSpec) {
    poolFileName = poolFileSpec;
    openPoolFile();
  }
  


  /**
   * @param initialize true if building a new listpool
   * called if poolFile file needs to be opened, usually after the
   * object has been materialized through serialization.
   */
  private void openPoolFile()  {
    try {
      poolFile = new BufferedRandomAccessFile(poolFileName,
					      BufferedRandomAccessFile.READ +
					      BufferedRandomAccessFile.WRITE);
      nextOffset = poolFile.getLength();
    } catch (IOException ioe) {
      System.err.println("Error opening poolFile: " + ioe);
      poolFile = null; 
    }
  }



  /**
   * @param start offset in poolFile file when list starts.
   * @return vector of accessible objects
   */
  public Vector getList(FeatureListBrafHandle handle) {
    FeatureList vec = new FeatureList(1);
    vec.setProxyRefsCountedTrue();
    long offset = handle.getValue();
    ProxyIndexingFeature value = null;
    try {
      synchronized(poolFile) {
	while (offset != END_MARKER) {
	  poolFile.seek(offset);
	  value = PersistentObjectManager.buildIF(poolFile);
	  vec.addElement(value);
	  offset = poolFile.readLong();
	}
      }
    } catch (IOException ioe) {
      System.err.println("Error seeking poolFile: " + ioe + 
			 ", offset tried is " + offset +
			 ", handle was " + handle + ".\n" +
			 "Succeeded in retrieving " + vec);
      return null;
    }
    return vec;
  }



  /**
   * Stores an object which conforms to the ProxyIndexingFeature class.
   * @return new offset to beginning of list 
   */

  public void put(FeatureListBrafHandle handle, ProxyIndexingFeature value) {
    try {
      synchronized (poolFile) {
	if (handle.getValue() != END_MARKER) {
	  poolFile.seek(handle.getOffsetToLink());
	  poolFile.writeLong(nextOffset);
	} else
	  handle.setValue(nextOffset);
	poolFile.seek(nextOffset);
	PersistentObjectManager.writeIF(value, poolFile);
	handle.setOffsetToLink(poolFile.getFilePointer());
	poolFile.writeLong(END_MARKER);
	nextOffset = poolFile.getFilePointer();
      }
    } catch (Exception e) {
      System.err.println("Error writing value to poolFile: " + e);
    }
  }





  /**
   * Needs to be called before the class disappears, otherwise the file will
   * be corrupted. As this class registered with the InfoServer thanks to the
   * <code>registerForShutdown()</code> method, this InfoServer will invoke
   * the current method when it receives itself a shutdown signal, so none
   * has to call it if the InfoServer mechanism is used.
   */
  public void shutdown() {
    try {
      poolFile.close();
    } catch (IOException ioe) {
      Support.errorMessage("LinkedListPool.shutdown(): Can't close poolFile.");
    }
  }
}
