package gov.nist.nlpir.irf.index.braf;

import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.util.Support;

import java.util.Vector;
import java.io.Serializable;
import java.util.Enumeration;

/**
 * This class is the "heart" of the Index. It mainly contains two
 * PersistentIrfHashtables of feature lists. The first one, called
 * valuesBySource, stores all the indexing features found, classified
 * by source. It will allow user to retrieve all the features of a given
 * document, for example. The second one, called sourcesByValue, stores
 * the different values whatever their source is. It will allow user to
 * retrieve all the sources where a given feature appear.<br><br>
 *
 * Each entry in a table corresponds to a list of features. For the
 * first table, the vector contains all the features found for the
 * source entry. In the second table, one vector contains all the
 * sources in which the entry feature can be found. The features
 * are shared by the lists, ie each feature belongs to two lists.
 * <p>
 * An example of using the class is below
 * <pre>
 * setIndexingFeature = new PersistentDualKeyContainer("myFile", true); 
 * 
 * // Add an element to DualKeyContainer using two keys
 * setIndexingFeature.put(source, key, value);
 * // More precise example
 * setIndexingFeature.put("doc1", "word1", "word1InDoc1");
 * setIndexingFeature.put("doc1", "word2", "word2InDoc1");
 * setIndexingFeature.put("doc2", "word1", "word1InDoc2");
 * // We now have:
 * // setIndexingFeatures.getSourceVector("word1") == {"word1InDoc1", "word1InDoc2"}
 * // setIndexingFeatures.getValuesVector("doc1") == {"word1InDoc1", "word2InDoc1"}
 *
 * // But you will actually have to define more precise objects:
 * // the keys have to be DeIntern and the sources have to be
 * // IoAddrIntern, the values have to be ProxyIndexingFeature.
 * </pre>
 * </p>
 * The comments of this class are going to define a <b>Contract</b> that
 * any class being used to replace this one must conform to. One may wish
 * to get rid of the current PersistentDualKeyContainer class because it
 * embeds its own persistence mechanism. Thus, instead of modifying it,
 * the easiest may be to write completely another class using the same
 * interface. If parts of PersitentDualKeyContainer are to be reused, the
 * ways those parts are currently managed will have to be studied closely
 * (management of references count, collection, shutdown issues) in order
 * to avoid any strange side effect, even if most of them are commented.
 * The contract is only defined for public methods, obviously, because
 * the interface doesn't constrain to a certain type of inner mechanisms.
 *
 * @version $Revision: 1.8 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance.
 *
 * @see gov.nist.nlpir.irf.index.DualKeyContainer
 */
public class PersistentDualKeyContainer implements Serializable {

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  /** Table of features accessed by document
      @serial */
  private PersistentIrfHashtable sourcesByValue;

  /** Table of features accessed by feature
      @serial */
  private PersistentIrfHashtable valuesBySource;

  /** Directory containing files that comprise the PDKC 
      @serial */
  private String DB_Directory;

  /** Name of the pool file for sources-by-value 
      @serial */
  private String poolNameSbV;

  /** Name of the pool file for values-by-source 
      @serial */
  private String poolNameVbS;

  /** Number of values stored
      @serial */
  private int valuesNumber;      

  /** Size of sourcesByValue
      @serial */
  private int sourcesNumber;    

  /** Size of valuesBySource
      @serial */
  private int uniqueValuesNumber;
  

  private static int
    SOURCE_VECTOR_INITIAL_CAPACITY = 5,    // Usually a source contains 30 features, and
    SOURCE_VECTOR_CAPACITY_INCREMENT = 10,//**  // then the increment of 10 seems reasonnable.
    FEATURE_VECTOR_INITIAL_CAPACITY = 2,    // Usually a feature appears only once or twice.
    FEATURE_VECTOR_CAPACITY_INCREMENT = 5;//** // When it appears more, it can be really often.


  //    TRANSIENT_FEATURE_LIST_THRESHOLD = 3;   // Feature lists are made light weight when they
  //                                           are smaller than this.

  /*  private final static int
    SOURCE_VECTOR_INITIAL_CAPACITY = 1,    // Now that feature lists are always lightweight,
    SOURCE_VECTOR_CAPACITY_INCREMENT = 1,  // we can save space by making their "ghost" as
    FEATURE_VECTOR_INITIAL_CAPACITY = 1,   // small as possible.
    FEATURE_VECTOR_CAPACITY_INCREMENT = 1,*/
      //    COMMON_FEATURE_THRESHOLD = 8;	   // If a feature appears more often
  // than that, its associated sources vector won't be made lightweight.


  /**
   * Constructor for PersistentDualKeyContainer
   * <b>Contract:</b> This constructor is only used by IdxIntern. Thus,
   * its signature may change if you need to provide it with more
   * paramaters, as the switch of class will lead to a change in the
   * client class (here, IdxIntern). As a classic contructor, this
   * method is responsible for initializing the entire structure of
   * the PDKC.
   * @param DB_Directory the name of the directory
   * @param indexName  name of the index this PDKC supports
   * or the radical of the names for the DB files.
   * @see gov.nist.nlpir.irf.index.IdxIntern
   */
  public PersistentDualKeyContainer(String DB_Directory, String indexName) {

    /* Mainly used during indexing, it should be enough to have only one 
     * slot as most accesses to this table should concern the previous 
     * source. */

    /*---------------------------------------------------------------------*
     * The passed index location information is interpreted by the PDKC as
     * a fully qualified directory path (i.e. from the root). We assure
     * here that the path ends in the "slash" appropriate for the platform.
     * If the path is null then it is changed to the null string.
     *---------------------------------------------------------------------*/
    String slash = Support.fileSeparator;
    if (DB_Directory == null)
      DB_Directory = "";
    else
      if (DB_Directory.length() > 0 && !DB_Directory.endsWith(slash))
	DB_Directory = DB_Directory+slash;

    sourcesByValue = 
      new PersistentIrfHashtable(DB_Directory, 
				 indexName + "sBv", 1, 1, 100000, .75f);
    valuesBySource = 
      new PersistentIrfHashtable(DB_Directory, 
				 indexName + "vBs", 1, 1, 2000, .70f);
    
    poolNameSbV= DB_Directory +"DB.pool." +indexName.substring(0,5) +".SbV";
    poolNameVbS= DB_Directory +"DB.pool." +indexName.substring(0,5) +".VbS";
      
    this.DB_Directory = DB_Directory;
    valuesNumber = 0;
    sourcesNumber = 0;
    uniqueValuesNumber = 0;
  }	
  


  /**
   * Returns the Vector associated to the parameter key.<br>
   * <b>Contract:</b> The returned ProxyFeatureList must contain every
   * ProxyIndexingFeature that was stored in the PDKC with a featureKey
   * matching the one given in this method's parameter. A key is
   * considered to match another one if they both return the same result
   * to <code>hashcode()</code> and return <code>true</code> when the
   * method equals() is called on each other.
   * @param featureKey value used to search the container.
   * @return a list of lightweight proxies for indexing features. The
   * list will never be null. If no feature matches the value, the
   * returned list is empty.
   */
  public final ProxyFeatureList getSourceVector(Object featureKey) {
    ProxyFeatureList result = (ProxyFeatureList)sourcesByValue.get(featureKey);
    if (result != null)
      return result;
    return new ProxyFeatureList(poolNameSbV, 0);
  }
  


  /**
   * <b>Contract:</b> The returned ProxyFeatureList must contain every
   * ProxyIndexingFeature that was stored in the PDKC with a sourceKey
   * matching the one given in this method's parameter. A key is
   * considered to match another one if they both return the same result
   * to <code>hashcode()</code> and return <code>true</code> when this
   * method is called on each other.
   * @param sourceKey source used to search the container.
   * @return a list of proxies for  indexing features. As
   * for getSourceVector, the list will never be null.
   */
  public final ProxyFeatureList getFeatureVector(Object sourceKey) {
    ProxyFeatureList result = (ProxyFeatureList)valuesBySource.get(sourceKey);
    if (result != null)
      return result;
    return new ProxyFeatureList(poolNameVbS, 0);
  }


  
  /** Caching mechanism objects: This cache is only used by put().
      It dramatically improves performance. */
  private transient ProxyFeatureList lastValues = null;
  ///  private transient IoAddrIntern lastSource = null;
  private transient Object lastSource = null;



  /**
   * Puts a ProxyIndexingFeature instance in the container. It may then
   * be retrieved either by source with an equivalent IoAddrIntern or
   * (Proxy)Document or by value with an equivalent DeIntern or
   * (Proxy)DataElem. For equivalence of DEs/DeInterns and
   * IoAddrIntern/Documents, see those classes.<br>
   * <b>Contract:</b> If ProxyFeatureLists are used beneath that, they
   * can be made lightweight right after they are created, thus not
   * loading memory (their mechanism allows them to grow without being
   * in memory unless you need to read them, not just extend them). This
   * method must take care of three variables: <code>sourcesNumber
   * </code>, <code>uniqueValuesNumber</code> and <code>valuesNumber
   * </code>. The first two are incremented by one if <code>sourceKey
   * </code> and <code>featureKey</code> respectively are encountered
   * for the first time. The third variable has to be increased at every
   * call, because each of them will result in an extra feature in the
   * PersistentDualKeyContainer.
   * @param sourceKey an IOAddrIntern
   * @param featureKey a DeIntern
   * @param proxyObject Must be a ProxyIndexingFeature. Declared as
   * an Object only for compliance with DualKeyContainer.
   * @see gov.nist.nlpir.irf.index.braf.DeIntern
   * @see gov.nist.nlpir.irf.index.braf.IoAddrIntern
   * @see gov.nist.nlpir.irf.document.Document
   * @see gov.nist.nlpir.irf.de.DataElem
   */
  //  public void put(IoAddrIntern sourceKey, DeIntern featureKey,
  public void put(Object sourceKey, Object featureKey,
		  ProxyIndexingFeature proxyObject) {
    ProxyFeatureList values;
    if ((lastSource != null) && (sourceKey.equals(lastSource)))
      // Cache hit !! The previous key used was the same as this one.
      values = lastValues;
    else {
      //if (lastValues != null)
      //lastValues.makeLightweight(); // May not be used for a long time
      values = (ProxyFeatureList)valuesBySource.get(sourceKey);
      if (values == null) { // No list was stored for this source yet.
	values = new ProxyFeatureList(poolNameVbS,
				      SOURCE_VECTOR_INITIAL_CAPACITY);
	valuesBySource.put(sourceKey, values); // This makes values persistent.
	values.makeLightweight(); // Shouldn't be materialized before update.
	sourcesNumber++;
      }
      lastValues = values;
      lastSource = sourceKey;
    }
    values.addElement(proxyObject);
    
    ProxyFeatureList sources = (ProxyFeatureList)sourcesByValue.get(featureKey);
    if (sources == null) { // No list was stored for this feature yet.
      sources = new ProxyFeatureList(poolNameSbV,
				     FEATURE_VECTOR_INITIAL_CAPACITY);
      sourcesByValue.put(featureKey, sources); // This makes sources persistent.
      sources.makeLightweight(); // Shouldn't be materialized before update.
      uniqueValuesNumber++;
    }
    sources.addElement(proxyObject);
    // if (sources.size() < COMMON_FEATURE_THRESHOLD)
    // could be size % A == 0. If soft refs work, no need for this line : make it lightweight each time.
    // sources.makeLightweight();
    proxyObject.makeLightweight();
    valuesNumber++;
  }


  
  /**
   * Classic representation method.
   */
  public String toString() {
    StringBuffer str = new StringBuffer();
    /**** Values by Source ********************************/
    Enumeration keys2 = valuesBySource.keys();
    str.append("\nVALUES BY SOURCE:\n");
    while (keys2.hasMoreElements()) {
      Object key2 = keys2.nextElement();
      str.append(key2 + "  >>>>  ");
      ProxyFeatureList vec = (ProxyFeatureList)valuesBySource.get(key2);
      str.append(vec.size() + ":" + vec + "\n");	
    }
    /**** Sources by Value ********************************/
    keys2 = sourcesByValue.keys();
    str.append("\nSOURCES BY VALUE:\n");
    while (keys2.hasMoreElements()) {
      Object key2 = keys2.nextElement();
      str.append(key2 + "  >>>>  ");
      ProxyFeatureList vec = (ProxyFeatureList)sourcesByValue.get(key2);
      str.append(vec.size() + ":" + vec + "\n");	
    }
    str.append("Number of elements: " + valuesNumber + ".\n");
    return str.toString();
  }


  
  /**
   * Gives the total number of features stored in the table.<br>
   * <b>Contract:</b> Each time a feature is added to the PDKC, the
   * <code>valuesNumber</code> variable must be increased (see <code>
   * put()</code>). This method returns the value of this variable.
   *
   * @return The number of values in the table.
   */
  public final int getValuesNumber() {
    return valuesNumber;
  }
 


  /**
   * @return <code>true</code> if no key table has been initialized.<br>
   * <b>Contract:</b> Returns true if no feature is stored in the PDKC,
   * i.e. it has just been created or emptied.
   */
  public final boolean isEmpty() {
    return uniqueValuesNumber == 0;
  }
  


  /**
   * Returns all the features stored.<br>
   * <b>Contract:</b> The Enumeration returned contains all the
   * ProxyIndexingFeatures stored in the PDKC. The order isn't important.
   *
   * @see gov.nist.nlpir.irf.index.DualKeyContainer#put
   */
  public final Enumeration elements() {
    return getAllValues().elements();
  }
  


  /**
   * Returns a Vector containing all the values stored in the
   * PersistentDualKeyContainer.<br>
   * <b>Contract:</b> This Vector is the concatenation of all Vectors or
   * Lists of ProxyIndexingFeatures that may be found in the PDKC.
   */
  public Vector getAllValues() {
    Vector toBeReturned = new Vector(uniqueValuesNumber);
    Enumeration elements = valuesBySource.elements();
    while (elements.hasMoreElements()) {
      Vector elements2 = ((ProxyFeatureList)elements.nextElement()).getAsVector();
      int nbElts2 = elements2.size();
      for (int i = 0; i < nbElts2; i++)
	toBeReturned.addElement(elements2.elementAt(i));
    }
    return toBeReturned;
  }


  
  /**
   * Prints statistics about the PersistentDualKeyContainer, ie its size
   * and the size of its elements. Keys are also printed if their length
   * is less than 15.
   *
   * @param depth
   * 1, prints the size of the contained vectors for each hash table,<br>
   * 2, gives for each vector its key and the number of its elements,<br>
   * 3, gives the key of each vector and prints its content, each element
   * truncated to 25 characters if longer,<br>
   * 4 is for a special 2, ie only the size of the vectors are printed
   * one behind another.<br>
   * Any other value will result in an empty display.
   */
  public void showStatistics(int depth, int maxLengthIfDepth3) {
    if ((depth > 0) && (depth < 5)) {
      System.out.print("PersistentDualKeyContainer - Values by Source:");
      showStats(valuesBySource, depth, maxLengthIfDepth3);
      System.out.print("PersistentDualKeyContainer - Unique Values:");
      showStats(sourcesByValue, depth, maxLengthIfDepth3);
    }
  }


  
  /**
   * Prints statistics for ONE HVtable.
   *
   * @param table The hashtable of vectors to be presented.
   * @param depth Same as showStatistics() depth.
   */
  private void showStats(PersistentIrfHashtable table, int depth, int maxLengthIfDepth3) {
    if (depth == 1) {
      System.out.println(" Sizes of vectors:");
      Enumeration lists = table.values();
      while (lists.hasMoreElements()) {
	ProxyFeatureList currentPFL = (ProxyFeatureList)lists.nextElement();
	System.out.print(" " + currentPFL.size());
	currentPFL.makeLightweight();
      }
      System.out.println("");
    } else if (depth == 2) {
      Enumeration keys2 = table.keys();
      while (keys2.hasMoreElements()) {
	Object key2 = keys2.nextElement();
	int keySize = key2.toString().length();
	String printKey = (keySize <= maxLengthIfDepth3) ?
	  key2.toString() :
	  (key2.toString().substring(0, maxLengthIfDepth3) + "/" + keySize);
	ProxyFeatureList currentPFL = (ProxyFeatureList)table.get(key2);
	System.out.print("\n    PFL of key [" + printKey +
			 "] in the given hash table has " +
			 currentPFL.size() + " elements.");
	currentPFL.makeLightweight();
      }
    } else if (depth == 3) {
      Enumeration keys2 = table.keys();
      while (keys2.hasMoreElements()) {
	Object key2 = keys2.nextElement();
	int keySize = key2.toString().length();
	String printKey = (keySize <= maxLengthIfDepth3) ?
	  key2.toString() :
	  (key2.toString().substring(0, maxLengthIfDepth3) + "/" + keySize);
	System.out.print("\n  [" + printKey + "]->");
	ProxyFeatureList toPrint = (ProxyFeatureList)table.get(key2);
	int max = toPrint.size();
	for(int i = 0; i < max; i++) {
	  String rep = toPrint.elementAt(i).toString();
	  int repSize = rep.length();
	  String printRep = (repSize < maxLengthIfDepth3) ?
	    ("[" + rep + "]") :
	    ("[" + rep.substring(0, maxLengthIfDepth3) + "/]");
	  System.out.print(printRep);
	}
	toPrint.makeLightweight();
      }
    } else if (depth == 4) {
      System.out.print(" H[" + table.size() + "]->");
      Enumeration lists = table.values();
      while (lists.hasMoreElements()) {
	ProxyFeatureList currentPFL = (ProxyFeatureList)lists.nextElement();
	System.out.print(currentPFL.size() + ",");
	currentPFL.makeLightweight();
      }
      System.out.println("\b.");
    }
  }



  /**
   * To retrieve the statistics concerning the hashtables inside the PDKC.
   */
  public void showHashtableStatistics() {
    sourcesByValue.showStatistics();
    valuesBySource.showStatistics();
  }


  
  /**
   * Gives the number of sources in the DualKeyContainer.<br>
   * <b>Contract:</b> Each time a source is added to the PDKC, the
   * <code>sourcesNumber</code> variable must be increased (see <code>
   * put()</code>). This method returns the value of this variable.
   * "A source is added" means a ProxyIndexingFeature is given for
   * storage with a sourceKey that cannot be found yet as being a
   * source for at least one ProxyIndexingFeature.
   */
  public final int getSourcesNumber() {
    return sourcesNumber;
  }



  /**
   * Returns the number of different values stored in the
   * PersistentDualKeyContainer.<br>
   * <b>Contract:</b> The name of this method is a bit ambiguous. It
   * must return the <code>uniqueValuesNumber</code> variable. This
   * variable counts the number of different features there is in the
   * pDKC, ie the number of different <code>featureKeys put()</code>
   * has been called with.
   */
  public final int getUniqueValuesNumber() {
    return uniqueValuesNumber;
  }


  
  /**
   * Returns the enumeration of Objects used as sources in this
   * PersistentDualKeyContainer. Caution: these are the actual keys, ie
   * an object used as a source may not be returned if another object
   * equal to the first one (considering <code>hashCode()</code> and
   * <code>equals()</code>) had already been used as a source for this
   * PersistentDualKeyContainer. Basically, it's an enumeration of
   * IoAddrInterns.<br>
   * <b>Contract:</b> Every <i>different</i> IoAddrIntern used as a
   * source for <code>put()</code> must be present in the Enumeration
   * result. "different" for two IoAddrInterns means they return <code>
   * false</code> when called one another for <code>equals()</code>.
   */
  public final Enumeration getSources() {
    return valuesBySource.keys();
  }
  


  /**
   * Returns the enumeration of Objects used as features for this
   * PersistentDualKeyContainer.<br>
   * Caution: these are the actual keys, ie an object used as a feature
   * may not be returned if another object equal to the first one
   * (considering hashCode() and equals()) had already been used as a
   * feature for this DualKeyContainer. Thus, this object would be the
   * one returned. It is an enumeration of DeInterns.<br>
   * <b>Contract:</b> Same as <code>getSources()</code>, except it
   * returns an Enumeration of DeInterns and not IoAddrInterns.
   */
  public final Enumeration getValues() {
    return sourcesByValue.keys();
  }
  


  /**
   * Returns the length of the base array in the
   * PersistentIrfHashtable by source.<br>
   * @return count of bins valuesBySouce table
   */
  public final int getSourceBinCount() {
    return valuesBySource.length();
  }
  


  /**
   * Returns the length of the base array in the
   * PersistentIrfHashtable by feature.<br>
   * @return count of bins sourcesByValue table
   */
  public final int getFeatureBinCount() {
    return sourcesByValue.length();
  }



  /**
   * When a value is stored, it appears in a FeatureList corresponding
   * to its source (a key of a hashtable). But the key in the hashtable
   * may not be the object the value was stored with. This method gives
   * this object from the object used as a source. This way, information
   * about the FeatureList of sources (like its size, ...) can be stored
   * in the key used to access this list.<br>
   * <b>Contract:</b> This method incarnates the problem addressed
   * several times: two IoAddrInterns may return <code>true</code> but
   * still be different objects. The only condition for this is that
   * they "represent" the same IRF_Document. But IoAddrInterns contains
   * more data than just a reference to a document. Thus, to gather
   * information in the IoAddrIntern concerning a document, this method
   * must be called to ensure the correct IO_AdrIntern is actually used.
   *
   * @see PersistentIrfHashtable#getActualKey
   */
  public final Object getActualSource(Object source) {
    return valuesBySource.getActualKey(source);
  }


  
  /**
   * When a value is stored, it appears in a FeatureList corresponding
   * to its feature (a key of a hashtable). But the key in the hashtable
   * may not be the object the value was stored with. This method gives
   * this object from the object used as a feature. Thus, information
   * about the values stored for a feature can be stored in the key (the
   * feature).<br>
   * <b>Contract:</b> The problem to understand here is the same as in
   * <code>getActualSource()</code>: DeInterns and DEs can be considered
   * equals as soon as they represent the same data, not necessarily in
   * the same place. The storage in the PDKC must takes this in account.
   *
   * @see PersistentIrfHashtable#getActualKey
   */
  public final Object getActualFeature(Object feature) {
    return sourcesByValue.getActualKey(feature);
  }



  /**
   * Clears both tables.
   */
  public void clear() {
  }



  /**
   * Gives the number of features stored for the given source.
   *
   * @param source the source for which the number of values will be computed.
   * @return The number of values,<br> 0 if the source doesn't appear.
   */
  public final int getNumberOfValuesFor(Object source) {
    return ((ProxyFeatureList)valuesBySource.get(source)).size();
  }



  /**
   * Gives the number of sources containing the given feature.
   *
   * @param feature the feature for which the number of sources will be computed.
   * @return The number of sources,<br> 0 if the feature doesn't appear.
   */
  public final int getNumberOfSourcesFor(Object feature) {
    return ((ProxyFeatureList)sourcesByValue.get(feature)).size();
  }
}
