package gov.nist.nlpir.irf.index.braf;

import gov.nist.nlpir.irf.index.DualKeyContainer;
import gov.nist.nlpir.irf.util.Support;

import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

import java.io.Serializable;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * Persistent version of IrfHashtable. This hashtable has been designed to 
 * remain only partly in memory, separating the hash space in slices that 
 * can be or not be in memory at one point. The coming to/back from disk of 
 * those slices is managed via a LFU-like algorithm thanks to <i>slots</i> 
 * receiving <i>blocks</i>.
 *
 * @version $Revision: 1.3 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the U.S. government, 
 * and by statute is not subject to copyright in the United States.  
 * Recipients of this software assume all responsibilities associated with 
 * its operation, modification and maintenance.
 *
 * @see gov.nist.nlpir.irf.index.DualKeyContainer
 * @see java.util.Hashtable
 */

public class PersistentIrfHashtable implements Serializable {

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  /** This array is composed of slots. Each of them may contain a block. */
  private transient HashBlock[] blocks;
  
  /** Size of the previous array. The more slots, the less swap but 
   * the more memory usage. 
   * @serial */
  private int numberOfSlots;

  /** Total number of blocks composing the hashtable. 
   * @serial */
  int numberOfBlocks;

  /**Size of each block computed knowing the number of blocks 
   * and the maximum number of hashcodes susceptible of being coded 
   * (Integer.MAX_VALUE).  
   * @serial */
  private int blockSize;

  /** For each block number, gives the slot number where this block is, 
   * or -1 if the block isn't in memory. */
  private transient int[] blockMap;

  /** for each slot number, gives the number of the block this slot 
   * contains, -1 if none. */
  private transient int[] inversedBlockMap; 
					       
  /** For each slot, gives the number of times it has been used since it 
   * was filled with a block. Used to implement a kind of LFU algorithm. */
  private transient int[] blocksUsed; 

  /** The blocks will be stored on disk in files in the directory
   * fully specified by filePath  
   * @serial */
  private String filePath;

  /** The blocks will be stored on disk in files named filePrefixOO 
   * where OO is a number from 0 to (numberOfBlocks - 1).  
   * @serial */
  private String filePrefix;

  /** Total number of elements stored in the blocks.  
   * @serial */
  private int count;
  
  /** Blocks will be created with this parameter. 
   *  @serial */
  private int blockInitialCapacity;

  /** Blocks will be created with this parameter.  
   * @serial*/
  private float blockLoadFactor;

  /** Statistics on the Hashtable efficiency. */
  private transient int accesses = 0;
  private transient int hits = 0;



  /**
   * Builds a PersistentIrfHashtable with the given charateristics:
   *
   * @param filePath the path from root to the file(s)
   * @param filePrefix the prefix that will be used for the names of the
   *        files where the differentblocks of the hashtable will be stored.
   * @param numberOfSlots the more slots there is, the less swapping should 
   *        occur but the more memory used.
   * @param numberOfBlocks number of chunks for the hashtable. The more 
   *        numerous they will be, the smaller and the faster to load, but 
   *        again it may mean several swapping phases.
   * @param blockInitialCapacity - as in @ref java.util.Hashtable
   * @param blockLoadFactor - as in @ref java.util.Hashtable
   */

  public PersistentIrfHashtable(String filePath,
				  String filePrefix, 
				  int numberOfSlots, 
				  int numberOfBlocks, 
				  int blockInitialCapacity, 
				  float blockLoadFactor) {

    this.filePath = filePath;
    this.numberOfSlots = numberOfSlots;
    this.numberOfBlocks = numberOfBlocks;
    this.blockSize = Integer.MAX_VALUE / numberOfBlocks + 1;
    this.blockInitialCapacity = blockInitialCapacity;
    this.blockLoadFactor = blockLoadFactor;
    this.count = 0;
    blocks = new HashBlock[numberOfSlots];
    blocksUsed = new int[numberOfSlots];
    inversedBlockMap = new int[numberOfSlots];
    for (int i = 0; i < numberOfSlots; i++) {
      blocksUsed[i] = 0;        // No slot has ever been used,
      inversedBlockMap[i] = -1; // nor contains a block.
    }
    blockMap = new int[numberOfBlocks];
    for (int i = 0; i < numberOfBlocks; i++)
      blockMap[i] = -1; // No block is in memory or even exists.
    this.filePrefix = "DB" + filePrefix;
  }



  /**
   * This put is the equivalent of the one in Hashtable. The only restriction
   * is that as it has been hardwired for IRF purposes, only 
   * ProxyFeatureLists can be stored with this method, so this persistent 
   * version of the hashtable can only store ProxyFeatureLists.
   * But keys can be of whatever type you'd like. Just don't forget to have 
   * tuned hashCode() and equals() methods if you may need them.
   *
   * @see java.util.Hashtable#put
   * @see gov.nist.nlpir.irf.index.braf.ProxyFeatureList
   */
  public synchronized Object put(Object key, Object value) {
    int hash = key.hashCode() & 0x7FFFFFFF; // This & is to get rid of the sign problem.
    int blockNr = hash / blockSize;         // hashCodes are equally shared between blocks.
    HashBlock currentBlock = getBlock(blockNr);
    Object result = currentBlock.put(key, hash, value);
    if (result == null) // only possible if there was no object matching this
      count++;          // (hashCode, equals) couple in the HashBlock.
    return result;
  }
  


  /**
   * The comment for put is also valid here.
   *
   * @see java.util.Hashtable#get
   */
  public synchronized Object get(Object key) {
    int hash = key.hashCode() & 0x7FFFFFFF; // To get rid of the sign problem.
    int blockNr = hash / blockSize;         // mapping of blocks using the hashCode.
    HashBlock currentBlock = getBlock(blockNr);
    return currentBlock.get(key, hash);
  }
  


  /**
   * This method is only used by the PersistentIrfHashtable itself and by 
   * its enumerator. It would be private if it was possible. It implements 
   * the question "Where is this block of the hashtable I need ?".
   */
  synchronized HashBlock getBlock(int blockNr) {
    accesses++; // For stats.

    int slot = blockMap[blockNr]; // Is the block already in memory ?
    if (slot == -1)               // No, it wasn't, load it and tell me 
                                  //where it is now.
      slot = load(blockNr);
    else
      hits++;
    
    blocksUsed[slot] += 2;                  // slot usage management: all slot usages
    for (int i = 0; i < numberOfSlots; i++) // are reduced by 1 except the one of the
      blocksUsed[i]--;                      // retrieved slot that grows by 1.
    
    return blocks[slot]; // Now the queried block is in memory.
  }



  /**
   * This method finds out where to load a block with a LFU-like algorithm, 
   * loads the block and returns the slot number where it just loaded the 
   * block.
   */
  private synchronized int load(int blockNr) {
    /* The following code computes the slot where the block should be loaded.
     *  It is the slot that has been the less used recently. This information
     *  is known via the blocksUsed[] array. */
    int whereToLoad = 0;
    int lessUsedBlock = blocksUsed[0];
    for (int i = 1; i < numberOfSlots; i++) {
      int currentUsage = blocksUsed[i];
      if (currentUsage <= lessUsedBlock) {
	whereToLoad = i;
	lessUsedBlock = currentUsage;
      }
    }
    
    /* Now, whereToLoad is the indice of the slot where the block is to be 
     * loaded. All data structure regarding the slot management have to be 
     * updated to reflect this change. */
    int previous = inversedBlockMap[whereToLoad]; // Which block is in the 
                                                  // slot currently ?
    if (previous > -1) {                   // There is a block, let's push 
                                           // it out of memory...
      if (blocks[whereToLoad].isDirty())   // but only if it is needed.
	writeBlock(blocks[whereToLoad], previous);
      blockMap[previous] = -1;
    }
    
    blocks[whereToLoad] = readBlock(blockNr); // Loads the block where it is supposed to be.
    blocksUsed[whereToLoad] = numberOfSlots; /* Seems a good usage to start with so that the
						block isn't trashed too quickly. */
    blockMap[blockNr] = whereToLoad;         // Let's keep track of where this block is stored,
    inversedBlockMap[whereToLoad] = blockNr; // and of the fact this slot now contains this block.
    
    return whereToLoad; // Because now the name of this variable could be whereItHasBeenLoaded
  }

  /**
   * As the name tells, this method writes a block to disk, using a tuned serialization
   * scheme with a BufferedRandomAccessFile.
   * @param toWrite the block that is to be written.
   * @param blockNr the actual BLOCK number, NOT the SLOT number. Be careful.
   */
  
  private synchronized void writeBlock(HashBlock toWrite, int blockNr) {
    String fileName = filePath + filePrefix + blockNr; // Builds the file name "filePrefix$$".
    try {
      BufferedRandomAccessFile out =
	new BufferedRandomAccessFile(fileName, BufferedRandomAccessFile.CREATE);
      toWrite.writeTo(out);
      out.close();
    } catch(IOException ioe) {
      Support.errorMessage("PersistentIrfHashtable: " + ioe);
    }
  }

  /**
   * This method is the symetric of writeBlock(), as it reads a HashBlock back from disk.
   * The special behavior is that if the file supposed to contain the block isn't found,
   * this method assumes the block needs to be created and then returns a new empty HashBlock.
   * This method also uses a tuned serialization mechanism with a BufferedRandomAccessFile.
   * @param blockNr the number of the block to be loaded.
   * @return the block read from disk, a new one if the block didn't already exist. Never
   * <code>null</code> unless an IOException has occured.
   */

  private synchronized HashBlock readBlock(int blockNr) {
    String fileName = filePath + filePrefix + blockNr;
    try {
      BufferedRandomAccessFile in =
	new BufferedRandomAccessFile(fileName, BufferedRandomAccessFile.READ);
      HashBlock read = HashBlock.readFrom(in);
      in.close();
      return read;
    } catch (FileNotFoundException fnfe) { // The block doesn't exist: create it !
      return new HashBlock(blockInitialCapacity, blockLoadFactor);
    } catch (IOException ioe) {
      Support.errorMessage("PersistentIrfHashtable-ioe: " + ioe);
    }
    return null;
  }
  
  /**
   * Returns the keys available in the hashtable for the get method.
   *
   * @see java.util.Hashtable#keys
   */
  public final synchronized Enumeration keys() {
    return new PersistentIrfHashtableEnumerator(this, true);
  }

  /**
   * Returns the values stored in the hashtable, exactly like the elements() 
   * method.
   *
   * @see java.util.Hashtable#elements
   */
  public final synchronized Enumeration values() {
    return new PersistentIrfHashtableEnumerator(this, false);
  }

  /**
   * Number of values stored in the persistent hashtable.
   */
  public final int size() {
    return count;
  }
  


  /**
   * Exact same result as size().
   */
  public final int length() {
    return count;
  }



  /**
   * With special hashCode() and equals() methods, you may want to know at 
   * one point what is the key you are really working with if you call 
   * <code>get(thisKey)</code> or <code>put(thisKey, something)</code>. 
   * This method returns the actual object that will be used as a key if 
   * you give thisKey as a parameter to one of the two methods.
   */
  public synchronized Object getActualKey(Object key) {
    int hash = key.hashCode() & 0x7FFFFFFF;
    int blockNr = hash / blockSize;
    HashBlock currentBlock = getBlock(blockNr);
    return currentBlock.getActualKey(key, hash);
  }
  


  /**
   * Exact same result as values().
   *
   * @see PersistentIrfHashtable#values
   */
  public final synchronized Enumeration elements() {
    return new PersistentIrfHashtableEnumerator(this, false);
  }
  


  /**
   * The serializable interface method for storage.
   */
  private synchronized void writeObject(java.io.ObjectOutputStream out) 
    throws IOException {
    for (int i = 0; i < numberOfSlots; i++)     // Out of all slots,
      if (blocks[i] != null)                    // Some of them contain blocks
	if (blocks[i].isDirty())                // that may need to be saved,
	  writeBlock(blocks[i], inversedBlockMap[i]); // so let's save them right now.
    out.defaultWriteObject(); // Only a few things are not transient, but this should be fast.
  }
  


  /**
   * The serializable interface method for matrialization.
   */
  private synchronized void readObject(java.io.ObjectInputStream in)
    throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    blocks = new HashBlock[numberOfSlots];
    blocksUsed = new int[numberOfSlots];
    inversedBlockMap = new int[numberOfSlots];
    for (int i = 0; i < numberOfSlots; i++) {
      blocksUsed[i] = 0;        // No slot has ever been used,
      inversedBlockMap[i] = -1; // nor contains a block.
    }
    blockMap = new int[numberOfBlocks];
    for (int i = 0; i < numberOfBlocks; i++)
      blockMap[i] = -1; // No block is present in memory.
  }



  /**
   * Displays statistics about the table and its usage with the given format:
   * <pre>
   * PersistentIrfHashtable tableName: x hits for y accesses, ie z% hits.
   * </pre>
   */
  public void showStatistics() {
    System.out.println("PersistentIrfHashtable " + filePrefix + ": " + hits +
		       " hits for " + accesses + " accesses, ie " +
		       (hits * 100 / (accesses + 1)) + "% hits.");
  }
}



/**
 * The PersistentIrfHashtable enumerator class. This class should remain 
 * opaque to the client. It will use the Enumeration interface. 
 */
class PersistentIrfHashtableEnumerator implements Enumeration {
  boolean keys;
  HashBlock currentBlock;
  Enumeration currentEnum;
  int currentNumber;

  PersistentIrfHashtable table; // Here is why:
  /* Due to the nature of this class, we need a reference to the object we're
   * working with so that the block management can still happen. An other
   * solution would have been to load every object from every block, and then
   * the Enumeration could have been given a definitve time stamp. Here,
   * different threads may modify the PersistentHashtable, thus modifying the
   * enumeration while it is used. Well, the Sun hashtable also behaves with
   * way.
   * Summary: this enumeration CAN'T be used as a static representation of the
   * table at one moment. It will evolve with the table, so a code like:<pre>
   *    Enumeration enum = PJHT.elements();
   *    while (enum.hasMoreElements())
   *      PJHT.put(aFunctionOf(enum.nextElement());
   * </pre>is probably doomed to fail.
   */
  PersistentIrfHashtableEnumerator(PersistentIrfHashtable table, boolean keys) {
    currentBlock = table.getBlock(0);
    currentEnum = keys ? currentBlock.keys() : currentBlock.elements();
    
    this.table = table;
    this.keys = keys;
    this.currentNumber = 0;
  }
  


  public boolean hasMoreElements() {
    while (!currentEnum.hasMoreElements() && currentNumber < (table.numberOfBlocks - 1)) {
      currentNumber++;
      currentBlock = table.getBlock(currentNumber);
      currentEnum = keys ? currentBlock.keys() : currentBlock.elements() ;
    }
    return currentEnum.hasMoreElements();
  }
  


  public Object nextElement() {
    if (currentEnum.hasMoreElements())
      return currentEnum.nextElement();
    else
      do {
	currentNumber++;
	currentBlock = table.getBlock(currentNumber);
	currentEnum = keys ? currentBlock.keys() : currentBlock.elements() ;
      } while (currentNumber < table.numberOfBlocks && !currentEnum.hasMoreElements());
    return currentEnum.nextElement();
  }
}
