package gov.nist.nlpir.irf.index.braf;

import java.util.Vector;
import java.util.Enumeration;
import gov.nist.nlpir.irf.irfman.IrfGlobal;
import gov.nist.nlpir.irf.irfman.InfoServer;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.proxy.VirtualProxy;
import gov.nist.nlpir.irf.broker.PersistenceBroker;
import gov.nist.nlpir.irf.broker.braf.FeatureListBrafBroker;
import gov.nist.nlpir.irf.handle.braf.FeatureListBrafHandle;
import gov.nist.nlpir.irf.proxy.Oid;
import gov.nist.nlpir.irf.util.Support;

/**
 * Proxy class for FeatureLists.
 *
 * @version $Revision: 1.4 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance.
 */
public class ProxyFeatureList extends VirtualProxy {

  private transient String poolFileName;

  /**
   * Creates a ProxyFeatureList with no real object associated. Should only 
   * be used by VirtualProxy.getProxyFor() via the newInstance() method.
   */
  public ProxyFeatureList() {
    /* If this method sets the real object, Feature Lists can't come back 
     * from disk until makeLightweight() is called on this proxy. Meanwhile, 
     * this proxy would stand for a empty list whereas its real object 
     * actually exists ! */
    getPersistenceBroker(); // Kind of initialization.
  }
  


  /**
   * Creates a ProxyFeatureList that refers to the given Vector and stores
   * the list in the pool file given.
   */
  public ProxyFeatureList(String poolFileName, Vector theVector) {
    this.poolFileName = poolFileName;
    int sizeV = theVector.size();
    FeatureList FL = new FeatureList(sizeV);
    for (int i = 0; i < sizeV; i++)
      FL.setElementAt(theVector.elementAt(i), i);
    setRealObject(FL);
    getPersistenceBroker(); // Kind of initialization.
  }
  


  /**
   * Creates a ProxyFeatureList with the given capacity.
   */
  public ProxyFeatureList(String poolFileName, int capacity) {
    this.poolFileName = poolFileName;
    setRealObject(new FeatureList(capacity));
    getPersistenceBroker(); // Kind of initialization.
  }
  

  /**
   * @return string representation of the Vector
   */
  public String toString() {
    return "P " + getRealObject().toString();
  }
  


  /**
   * Equivalent to the method in Vector.
   * 
   * @see Vector#addElement
   */ 
  public synchronized void addElement(Object toAdd) {
    ((VirtualProxy)toAdd).makePersistent();
    if (!isLightweight()) {
      ((VirtualProxy)toAdd).addRefToInMemoryProxiesByOid();
      ((Vector)getRealObject()).addElement(toAdd);
    }
    if (objectIdentifier != null) {
      FeatureListBrafHandle currentHandle =
	(FeatureListBrafHandle)iS.getHandlesByOid().getHandleByOid(objectIdentifier);
      ((FeatureListBrafBroker)broker).addElement((ProxyIndexingFeature)toAdd, currentHandle);
      iS.getHandlesByOid().replaceHandleByOid(objectIdentifier, currentHandle);
    }
  }



  /**
   * Equivalent to the method in Vector.
   * 
   * @see Vector#size
   */
  public final int size() {
    return ((Vector)getRealObject()).size();
  }
  


  /**
   * Equivalent to the method in Vector.
   * 
   * @see Vector#elementAt
   */
  public final synchronized Object elementAt(int range) {
    return ((Vector)getRealObject()).elementAt(range);
  }
  


  /**
   * The concrete proxy will override this method to get the broker
   * for the proxy object @return the broker for this object 
   */
  public PersistenceBroker createBroker() {
    /** alternative for this is to have one broker stored in the
	InfoServer and retrieved in this method */
    return FeatureListBrafBroker.instance();
  }
  


  public final Enumeration elements() {
    return ((Vector)getRealObject()).elements();
  }
  


  /**
   * Overriding of the standard method in VirtualProxy
   */
  public synchronized void makePersistent() {
    if (objectIdentifier == null) {
      FeatureListBrafHandle currentHandle =
	new FeatureListBrafHandle(poolFileName,
				  LinkedListPool.END_MARKER);
      objectIdentifier = Oid.getNextOid();
      FeatureList realVector = (FeatureList)getRealObject();
      if (realVector != null) {
	int numberOfValues = realVector.size();
	for (int i = 0; i < numberOfValues; i++) {
	  ProxyIndexingFeature theVP = (ProxyIndexingFeature)realVector.elementAt(i);
	  theVP.makePersistent();
	  theVP.addRefToInMemoryProxiesByOid();
	  ((FeatureListBrafBroker)broker).addElement(theVP, currentHandle);
	}
      } else
	Support.errorMessage("ProxyFeatureList.putRealObject(): no real"
			     + " object to dematerialize.");
      handles.addHandleByOid(objectIdentifier, currentHandle);
      realVector.setProxyRefsCountedTrue();
    }
  }



  /**
   * Rewrites the object after it has been changed. May never be used,
   * now that put() directly upgrades the storage.
   */  
  public synchronized void replaceRealObject() {
    FeatureListBrafHandle currentHandle = null;
    Vector values = (Vector)getRealObject();
    int size = values.size();
    for (int i = 0; i < size; i++) {
      ProxyIndexingFeature theVP = (ProxyIndexingFeature)values.elementAt(i);
      theVP.replaceRealObject();
    }
    if (currentHandle != null)
      handles.addHandleByOid(objectIdentifier, currentHandle);
  }
  


  /**
   * At the time of the writing of this class, lots of clients were using it
   * as a proxy for a Vector and not for a FeatureList. That's why this 
   * method remains.
   */
  public final synchronized Vector getAsVector() {
    return (Vector)getRealObject();
  }
  


  /**
   * Returns the value DEs contained in the IndexingFeatures in the list.
   * Optimization needed by the
   * <code>IdfIdxIntern.updateSources()</code> method.
   */
  public final synchronized Vector getDE_Features() {
    Vector theRO = (Vector)getRealObject();
    int size = theRO.size();
    Vector result = new Vector(size);
    for (int i = 0; i < size; i++)
      result.addElement(((ProxyIndexingFeature)theRO.elementAt(i)).getFeature());
    return result;
  }
  


  /**
   * Like for every other Proxy, this method cuts the link between the
   * real  object and the proxy so that the first one can be collected.
   * Here, this  method acts the fact that every contained proxy is
   * also made lightweight This way, garbage collection will be faster
   * and, most of all, will happen for every contained proxy, otherwise
   * a reference to a contained  proxy may stay around and prevent its
   * real object from being collected.<br>
   */
  /* In case soft references are used in VirtualProxy, the call to the
   * Precursor method must disappear and this comment can be added:
   * This version of makeLightweight doesn't call its precursor in
   * VirtualProxy because it doesn't currently need the soft reference
   * mechanism, it is even incompatible with the FeatureList mechanism.
   * Thus there could also be a special <tt>getRealObject()</tt>, but
   * the soft reference mechanism doesn't interfere with this class
   * (softRef is always null, therefore not used).
   */
  public synchronized void makeLightweight() {
    if (!isLightweight()) {
      Vector theRO = (Vector)getRealObject();
      int size = theRO.size();
      for (int i = 0; i < size; i++)
	((ProxyIndexingFeature)theRO.elementAt(i)).makeLightweight();
      super.makeLightweight();
    }
  }
}
