package gov.nist.nlpir.irf.irfman;

import gov.nist.nlpir.irf.document.DocCollection;
import gov.nist.nlpir.irf.util.Support;

import java.util.Dictionary;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.Serializable;
import gov.nist.nlpir.irf.proxy.HandlesByOid;
//import gov.nist.nlpir.irf.proxy.AnySizeHandlesByOid;
import gov.nist.nlpir.irf.proxy.OneSizeHandlesByOid;
import gov.nist.nlpir.irf.proxy.VirtualProxy;
import java.lang.reflect.Method;

/** 
 * The InfoServer is responsible for keeping track of available indexes,
 * document collections, and indirectly of available indexers and 
 * query-document matchers.
 * <br><br>
 * 
 * @version $Revision: 1.5 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.                          
 */                      
public class InfoServer implements Serializable {
 

 /** serial version universal id - put here so Java does not insert one 
  * which may change due to revisions and make it impossible to deserialize
  * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

 /** Handles by Oid table 
  *  @serial */
  private HandlesByOid handlesByOid;


  /** Available document collections. 
   *  @serial */
  private Dictionary docColls;


  /** Names of the used converters if they actually exist. 
   * @serial */
  private Vector converters;


  /** High-level information for use by brokers in locating persistent
      storage - transient here; saved by collection in DocCollection. */
  private transient String locationOfBrokeredStorage;


  /** Manager for available indexers. */
  private transient IndexerManager indexerMan;


  /** Manager for available query-doc matchers. */
  private transient MatcherManager matcherMan;


  /** The Oid class can't keep track itself of the last Oid given away 
   *  because of the <code>static</code> keyword vs Serialization behavior. 
   *  We must then store it here.
   *  It is static, so the basic serialization can't take care of it. 
   *  It doesn't have any static initalizer for the same reason. Only the 
   *  initialize() method can set it at first, and then the deserialization 
   *  process will assure we can get new Oids during subsequent runs.
   *  @see gov.nist.nlpir.irf.proxy.Oid
   */
  public static int lastUsedOidAtStart;



  public InfoServer() {
    initialize();
    indexerMan = new IndexerManager();
    matcherMan = new MatcherManager();
    handlesByOid = new OneSizeHandlesByOid();
    //handlesByOid = new AnySizeHandlesByOid();
    VirtualProxy.setIS(this);
  }
 

  /** 
   * Initialize the Infoserver dictionary of document collections, etc.
   */
  private void initialize() {
    docColls = new Hashtable();
    locationOfBrokeredStorage = null;
    converters = new Vector(3, 1);
    lastUsedOidAtStart = 0;
  }


  /** 
   * @return the handles-by-Oid table
   */
  public HandlesByOid getHandlesByOid() { 
    return handlesByOid; 
  }


  /**
   * If one wants to change the size of the cache used to match Handles to 
   * Oids, this function * is for this purpose.
   */
  public void setMaxHandlesByOidCacheEntries(int maxHandlesByOidCacheEntries) {
    /* As it is now a write-back algorithm, this won't work... */
    //    this.maxHandlesByOidCacheEntries = maxHandlesByOidCacheEntries;
    //handlesByOidCache.setMaxEntries(maxHandlesByOidCacheEntries);
  }


  /** 
   * Set the the locationOfBrokeredStorage
   */
  public void setLocationOfBrokeredStorage(String location) { 
    locationOfBrokeredStorage = location; 
  }


  /** 
   * @return the locationOfBrokeredStorage
   */
  public String getLocationOfBrokeredStorage() { 
    return locationOfBrokeredStorage; 
  }


  /** 
   * Print out a list of available resources
   */
  public void present() {
    Enumeration it = null;

    System.out.println("DocCollections ------------------------------------ ");
    it = docColls.elements();
    while (it.hasMoreElements())
      ((DocCollection)it.nextElement()).present();
        
    System.out.print("Handles by Oid ------------------------------------ \n" + 
		     handlesByOid.getHandlesByOidSize() + 
		     " Oid keys in handlesByOid table " +
		     "inMemoryProxiesByOid ---------------------------\n" +
		     //		     VirtualProxy.getInMemoryProxiesByOidSize() +
		     " Oid keys in inMemoryProxiesByOid table\n" +
		     //VirtualProxy.getInMemoryProxiesByOidBinCount() +
		     " bins in inMemoryProxiesByOid table\n" +
		     //VirtualProxy.getInMemoryProxyCounts() +
		     handlesByOid.getStatistics());
  }
  

  /** 
   * Add a document collection to the dictionary of available document 
   * collections
   * @param doccoll the document collection to be added
   */
  public void addDocColl(DocCollection doccoll) {
    docColls.put(doccoll.getName(), doccoll);
  }


  /** 
   * Get an enumeration to allow interation over the available document
   * collections
   * @return an enumeration allowing iteration over the available document
   * collections
   */
  public Enumeration getDocCollIter() {
    return docColls.elements();
  }


  /** 
   * Look up a document collection by name in the dictionary of available 
   * collections
   * @param doccollname the name of the document collection to be looked up
   * @return the document collection corresponding to the specified name
   */
  public DocCollection findDocColl(char[] doccollname) {
    return (DocCollection)docColls.get(new String(doccollname));
  }


  /** 
   * Look up a document collection by name in the dictionary of available 
   * collections
   * @param doccollname the name of the document collection to be looked up
   * @return the document collection corresponding to the specified name
   */
  public DocCollection findDocColl(String doccollname) {
    return (DocCollection)docColls.get(doccollname);
  }


  /** 
   * Find the document collections that have been modified 
   * @return the modified document collections
   */
  public Vector getModifiedDocColls() {
    Vector v = new Vector();
    Enumeration e = docColls.elements();
    while (e.hasMoreElements()) {
      DocCollection docColl = (DocCollection)e.nextElement();
      if (docColl.getModified())
	v.addElement(docColl);
    }
    return v;
  }


  /** 
   * Get the IndexerManager which is a component of this InfoServer
   * @return the IndexerManager which is a component of this InfoServer
   */
  public IndexerManager getIndexerManager() {
    return indexerMan;
  }


  /** 
   * Get the MatcherManager which is a component of this InfoServer
   * @return the MatcherManager which is a component of this InfoServer
   */
  public MatcherManager getMatcherManager() {
    return matcherMan;
  }


  public void addConverter(String converterName) {
    converters.addElement(converterName);
  }


  public Vector getConverters() {
    return converters;
  }


  /**
   * For proxy management purpose.
   *
   * @see gov.nist.nlpir.irf.proxy.VirtualProxy#getFirstInstance
   */
  public void getFirstInstances() {
    VirtualProxy.setIS(this);
    Enumeration keys = docColls.keys();
    while (keys.hasMoreElements()) {
      String currentKey = (String)keys.nextElement();
      DocCollection currentDocColl = (DocCollection)docColls.get(currentKey);
      currentDocColl.getFirstInstances();
    } 
  }


  /**
   * This method must be called before this InfoServer disappears, otherwise 
   * there will likely be synchronization problems, like Handles not being 
   * saved.
   */
  public void shutdown() {
    handlesByOid.shutdown();
    Enumeration registeredClasses = classesToShutdown.elements();
    while (registeredClasses.hasMoreElements()) {
      Class current = (Class)registeredClasses.nextElement();
      Class nullArray[] = new Class[0];
      try {
	Method shutdownMethod = current.getDeclaredMethod("shutdown" , nullArray);
	shutdownMethod.invoke(null, nullArray);
      } catch(NoSuchMethodException nse) {
	Support.errorMessage("Class " + current.getName() +
			     " should expose a public static shutdown() method" +
			     " before being registered in InfoServer.");
      } catch(IllegalAccessException iae) {
	Support.errorMessage("InfoServer.shutdown: (Is " + current.getName() +
			     ".shutdown public?) " + iae);
      } catch(java.lang.reflect.InvocationTargetException ite) {
	Support.errorMessage("InfoServer.shutdown: (Is " + current.getName() +
			     ".shutdown static?) " + ite);
      }
    }
  }
  
  /** This static vector will contain every class that registered itself with
      the <code>registerForShutdown()</code> method. When the InfoServer shuts
      down, each of these classes will be notified. */
  private static Vector classesToShutdown = new Vector(5);
  
  
  /**
   * This method must be called from a static initializer by every class that
   * wishes to be notified when the InfoServer shuts down. When this happens,
   * the InfoServer will invoke the <code>shutdown()</code> method on every
   * class that registered with the current method. This method will have to
   * be <code>static public</code>. A class may wish to be notified this way
   * if, for instance, it uses a file that may not be saved if the class isn't
   * aware of the shutdown process.<br>
   * <b>Example</b>:<br:
   * The class <code>FileClient</code> uses a file (variable name <code>theFile
   * </code>) and thus it wishes to be notified when the application is about
   * to leave so that it can close this file. It must then register with
   * InfoServer using a static initializer like this one:<pre>
   * static {
   *   theInfoServer().
   *     registerForShutdown(FileClient.class);
   * }
   * </pre>This way the given InfoServer is aware of the existence of this
   * class. It will call <code>shutdown()</code> on FileClient before it
   * disappears. Thus FileClient must provide this shutdown method:<pre>
   * public static void shutdown() {
   *   try {
   *     theFile.close();
   *   } catch (IOException ioe) {
   *     Support.errorMessage("FileClient.shutdown(): Can't close theFile.");
   *   }
   * }
   * </pre>
   */
  public static void registerForShutdown(Class toBeRegistered) {
    classesToShutdown.add(toBeRegistered);
  }
}
