package gov.nist.nlpir.irf.irfman;

import gov.nist.nlpir.irf.irfman.IrfGlobal;
import gov.nist.nlpir.irf.irfman.InfoServer;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.proxy.Oid;
import gov.nist.nlpir.irf.proxy.HandlesByOid;

import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import java.util.Enumeration;
import java.util.Vector;

/** 
 * Information Retrieval Manager
 * <br><br>

 * The IrfManager is a central, globally accessible control object through
 * which document collections, indexes, and other information can be accessed.
 * The IrfManager is responsible for the global thread of control, although
 * this is mostly delegated to other parts of the framework.
 *
 * @version $Revision: 1.3 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.                           
 */                      

public class IrfManager {

  private InfoServer infoServer; 
  private String directoryForInfoServerSerialization; // for use at shutdown


  public IrfManager(String serializationDirectory, boolean createIfNotFound)
    throws FileNotFoundException {
    if (IrfGlobal.gIrfManager != null ) // should only have one such object!
      Support.warningMessage("IrfManager.IrfManager() gIrfManager already exists!");
    IrfGlobal.gIrfManager = this;
    directoryForInfoServerSerialization = serializationDirectory;

    try { // Deserialize the InfoServer
      // Let's take care of the Oid problem.
      BufferedRandomAccessFile fileOid = 
	new BufferedRandomAccessFile(makeLastOid_FileName(), BufferedRandomAccessFile.READ);
      Oid.init(fileOid.readInt());
      fileOid.close();
      
      FileInputStream fileInStream = 
	new FileInputStream(makeInfoServerSerializationFileName());
      GZIPInputStream gZipInStream = new GZIPInputStream(fileInStream);
      ObjectInputStream in = new ObjectInputStream(gZipInStream);
      infoServer = (InfoServer)in.readObject();
      infoServer.getFirstInstances();
      infoServer.getHandlesByOid().openHandlesByOid(serializationDirectory);
      in.close();
      if (Support.COARSE_TRACE <= Support.traceLevel)
	if (serializationDirectory == null)
	  Support.traceMessage("IrfManager: Using existing InfoServer from " 
			   + System.getProperty("user.dir"));
	else  
	  Support.traceMessage("IrfManager: Using existing InfoServer from " 
			   + serializationDirectory);
    } catch (IOException e) {
      if (e instanceof FileNotFoundException) {
	if (createIfNotFound) {
	  infoServer = new InfoServer();
	  infoServer.getHandlesByOid().openHandlesByOid(serializationDirectory);
	  if (Support.COARSE_TRACE <= Support.traceLevel){
	      if (serializationDirectory == null)
		 Support.traceMessage("IrfManager: Creating new InfoServer in "
				      + System.getProperty("user.dir"));
	      else
		 Support.traceMessage("IrfManager: Creating new InfoServer in "
				      + serializationDirectory);
	  }
	} else
          throw new FileNotFoundException(e.getMessage());
      } else
        Support.errorMessage(e.toString());
    } catch (ClassNotFoundException e) {
      Support.errorMessage(e.toString());
    }
  }
  

  /** 
   * Get the InfoServer that is part of this IrfManager
   * @return the InfoServer which is a component of this IRManager
   */
  public InfoServer getInfoServer() { 
    return infoServer; 
  }

  /** 
   * Set the directory for serializing the InfoServer iff none is set
   * @param directory directory in which InfoServer is to be serialized
   */
  public boolean setDirectoryForInfoServerSerialization (String directory) { 
    if (directoryForInfoServerSerialization == null) {
      directoryForInfoServerSerialization = directory;
      return true;
    } else
      return false;
  }
  
  /** 
   * Compose the fully qualified name of the InfoServer serialization file
   * from the directoryForInfoServerSerialization and the fixed name of the
   * file "DB.Info".
   * @return name fully qualified file name for InfoServer serialization
   */
  public String makeInfoServerSerializationFileName() {
    return makeSerializationNameWith("DB.Info");
  }

  public String makeLastOid_FileName() {
    return makeSerializationNameWith("DB.Oid");
  }

  private String makeSerializationNameWith(String serializedInfoServerFileName) {
    String separator = System.getProperty("file.separator");
    if (directoryForInfoServerSerialization != null)
      if (!directoryForInfoServerSerialization.endsWith(separator)) 
	return directoryForInfoServerSerialization + separator + 
	  serializedInfoServerFileName;
      else
	return directoryForInfoServerSerialization + 
	  serializedInfoServerFileName;
    else
      return "." + separator + serializedInfoServerFileName;
  }
  
  /** 
   * Print out information on the contents of the IrfManager
   */
  public void present() {
    System.out.print("============================\n" +
		     "|  Contents of IR manager: |\n" +
		     "============================\n");
    infoServer.present();
  }
  
  /** 
   * Find the document collections that have been modified 
   * @return the modified document collections
   */
  public Vector getModifiedDocColls() {
    return infoServer.getModifiedDocColls();
  }
  
  /** 
   * Serialize the IrfManager to a known location
   */
  public void shutdown() { 
    infoServer.shutdown();
    try { // Serialize the info server after all its indexes have been stored.
      FileOutputStream fileOutStream = 
	new FileOutputStream(makeInfoServerSerializationFileName());
      
      GZIPOutputStream gZipOutStream = new GZIPOutputStream(fileOutStream);
      ObjectOutputStream out = new ObjectOutputStream(gZipOutStream);
      out.writeObject(infoServer);
      out.flush();
      out.close();

      BufferedRandomAccessFile fileOid =
	new BufferedRandomAccessFile(makeLastOid_FileName(),
				     BufferedRandomAccessFile.CREATE);
      fileOid.writeInt(Oid.getLastUsedOid());
      fileOid.close();
    } catch (IOException e) {
      Support.errorMessage(e.toString());
    }
  }
}
