package gov.nist.nlpir.irf.matcher;

import gov.nist.nlpir.irf.feature.IndexingFeature;
import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.util.Support;

/** 
 * The Matcherbp class is responsible for matching one set of indexing
 * features against another (e.g. a set from a document used as a query
 * against a set derived from a document and kept in an index). Different
 * kinds of matching can be defined since features are complex objects
 * containing other objects.<br><br>
 * Note: All these functions return a double result which can only be 1.0
 * or 0.0 and not a boolean. This is because another type of Matcher could
 * be defined, returning more precise results.
 * 
 * @see IndexingFeature
 * 
 * @version $Revision: 1.1 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */                      

public class Matcherbp {
    
  /** 
   * Determines whether two indexing features match
   * @param if1 First indexing feature
   * @param if2 Second indexing feature
   * @return 1.0 if they match, 0.0 otherwise.
   */
  public double match(IndexingFeature if1, IndexingFeature if2) {
    if (!sameAttributeName(if1, if2))
      return 0.0;
    DataElem de1 = if1.getFeature();
    DataElem de2 = if2.getFeature();
    return match(de1, de2);
  }
    


  /** 
   * Determines if two data elements match
   * @param de1 First data element 
   * @param de2 Second data element
   * @return 1.0 if they match, 0.0 otherwise.
   */
  public double match(DataElem de1, DataElem de2) {
    String name1 = de1.getClass().getName();
    String name2 = de2.getClass().getName();
    if (name1.compareTo(name2) != 0) {
      Support.errorMessage("ERROR: debp's are not of the same class.");
      return 0.0;
    }
    if (de1.match(de2) != 0.0)
      return 0.0;
    return 1.0;
  }
    

    
  /** 
   * Determines if two data elements match exactly
   * @param de1 First data element 
   * @param de2 Second data element
   * @return true if they match exactly, ie are similar, false otherwise.
   */ 
  public boolean exactMatch(DataElem de1, DataElem de2) {
    String name1 = de1.getClass().getName();
    String name2 = de2.getClass().getName();
    if (name1.compareTo(name2) != 0) {
      Support.errorMessage("ERROR: debp's are not of the same class.");
      return false;
    }
    if ((de1.getValueAsString()).compareTo(de2.getValueAsString()) != 0)
      return false;
    return true;
  }
    


  /** 
   * Determines if two indexing features match exactly
   * @param if1 First indexing feature
   * @param if2 Second indexing feature
   * @return true if they match, false otherwise.
   */
  public boolean exactMatch(IndexingFeature if1, IndexingFeature if2) {
    if (!sameAttributeName(if1, if2))
      return false;
    DataElem de1 = if1.getFeature();
    DataElem de2 = if2.getFeature();
    return exactMatch(de1, de2);
  }
    


  /** 
   * Determines if two indexing features have the same attribute name and 
   * match
   * @param if1 First indexing feature
   * @param if2 Second indexing feature
   * @return true if they match, false otherwise.
   */
  public boolean extendedExactMatch(IndexingFeature if1, 
				    IndexingFeature if2) {
    if (!exactMatch(if1, if2))
      return false;
    return sameAddress(if1, if2);
  }
    


  /** 
   * Determines if two indexing features have the same attribute name and 
   * match
   * @param if1 First indexing feature
   * @param if2 Second indexing feature
   * @return true if they do, false otherwise.
   */
  public boolean sameAddress(IndexingFeature if1, IndexingFeature if2) {
    if (!sameAttributeName(if1, if2))
      return true;
    DataElem de1 = if1.getFeature();
    DataElem de2 = if2.getFeature();
    return exactMatch(de1, de2);
  }
    


  /** 
   * Determine if two indexing features have the same attribute name
   * @param if1 First indexing feature.
   * @param if2 Second indexing feature.
   * @return true if they do or if one or both of them is null, 
   * false otherwise.
   */
  public boolean sameAttributeName(IndexingFeature if1, IndexingFeature if2) {
    /*	String attributeName1 = if1.getAttributeName();
	String attributeName2 = if2.getAttributeName();
	
	if (attributeName1 != null && attributeName2 != null)
	if (attributeName1.compareTo(attributeName2) != 0)
	return false;
	return true;
	}*/
      return true;
    }
}
