package gov.nist.nlpir.irf.modality;

import gov.nist.nlpir.irf.document.Document;
import gov.nist.nlpir.irf.document.DocumentIntrf;

import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.index.ProxyIndex;
import gov.nist.nlpir.irf.indexer.Indexer;
import gov.nist.nlpir.irf.indexer.UniversalIndexer;
import gov.nist.nlpir.irf.util.Support;

import java.lang.reflect.Field;

/**
 * Container for information relating an index, the feature (i.e.,
 * document field) on which it is built, the indexer used, etc.
 * <br><br> 
 * @version $Revision: 1.2 $
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 *
 * @see gov.nist.nlpir.irf.index.Index 
 * @see IndexingModalities
 * @see gov.nist.nlpir.irf.irfman.InfoServer
 */

public class IndexingModalityUnit extends Object implements java.io.Serializable{
 
  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  /** Name of feature (i.e. document field) to be indexed
   * @serial  */
  private String featureName = null;         

  /** A an instance of Field class for the document field being indexed. 
   * not to be serialized with this IMU */
  private transient Field managedField = null; 

  /** Index for this feature, i.e., document field
   * @serial  */
  private ProxyIndex index = null;            

  /** An instance of the indexer for this IMU
   * @serial */
  private Indexer indexer = null;             

  /** Flag indicating refs to proxy indexes have been added to the table 
   * of in-memory proxies 
   * @serial */
  private boolean proxyRefsCounted = false; 


 
  /**
   * Constructor: sets a default indexer
   */
  public IndexingModalityUnit() {
    indexer = new UniversalIndexer();
  }
  


  /**
   * Gets the feature name of this modality unit.
   * @return The current feature name.
   */
  final public String getFeatureName() {
    return featureName;
  }
    


  /**
   * Sets feature name 
   * @param name new feature name
   */
  final public void setFeatureName(String name) {
    featureName = name;
  }


 
  /** 
   * Gets indexer for this modality unit
   * @return current indexer
   */
  final public Indexer getIndexer() {
    return indexer;
  }



  /** 
   * Sets indexer for this modality unit
   * @param ixr indexer 
   */
  final public void setIndexer(Indexer ixr ) {
    indexer = ixr;
  }

 

  /** 
   * Gets index
   * @return index  
   */
  final public ProxyIndex getIndex() {
    return index;
  }
 

 
  /** 
   * Sets index
   * @param new index
   */
  public void setIndex(ProxyIndex index ) {
    if (proxyRefsCounted == true)
      {
      if (this.index != null)
        this.index.deleteRefFromInMemoryProxiesByOid();
      index.makePersistent();
      index.addRefToInMemoryProxiesByOid();
      }    
    this.index = index;
  }


 
  /** 
   * Gets index name
   * @return index name 
   */
  public String getIndexName() {
    return index.getName();
  }


 
  /**
   * Returns the data elements from 'fvals' corresponding to the
   * featureName of this IMU.  Matches this Indexing Modality Unit's
   * featureName against (Doc) featureNames in 'fnames'.
   *
   * @param fnames collection of feature names 
   * @param fvals collection of feature values
   * @return The feature values corresponding to the feature name of this IMU.
   */
  public DataElem getFeatureValue(Document theDoc) {
    if (managedField == null)
      try {
	managedField = theDoc.getClass().getField(featureName);
      } catch (NoSuchFieldException nsfe) {
	Support.errorMessage("IndexingModalityUnit::getFeatureValue() " + nsfe);
      }
    DataElem toBeReturned = null;
    try {
      toBeReturned = (DataElem)managedField.get(theDoc);
    } catch (IllegalAccessException iae) {
      Support.errorMessage("IndexingModalityUnit::getFeatureValue() " + iae);
    }
    return toBeReturned;
  }
  


  /**
   * If references from this object to its index have not
   * been added to the table of in-memory proxies, then 
   * add the reference and mark it as added.
   */

  public void save() {
    if (proxyRefsCounted == false)
      {
      index.makePersistent();
      index.addRefToInMemoryProxiesByOid();
      proxyRefsCounted = true;
      }
  }

 

  /** 
   * Prints representation of indexing modality unit to standard
   * output showing featureName, indexer, and indexName. 
   */
  public void present() {
    System.out.print(featureName != null ? featureName : "[NULL]" + "\t");
    System.out.print(indexer != null ? " \t"+indexer.getClass().getName() : "[NULL]" + "\t");
    System.out.println(getIndexName() != null ? " \t"+getIndexName() : "[NULL]");
  }
   
 
  /** 
   * Returns string representation of indexing modality unit showing
   * featureName, indexer, and indexName.
   * @return string representation of indexing modality unit.
   */
  public String toString() {
    return ("featureName: " + featureName + ", indexer: " +
      (indexer != null ? indexer.getClass().getName() : "[null]") +
      ", IndexName: " + getIndexName());
  }



  /** 
   * Validates this supplied document against this modality unit
   * @return  true always currently.
   */
  public boolean validate(DocumentIntrf r) {
    return true;
  }
 


  /** 
   * Removes proxy duplicated during deserialization - replace
   * with references to a single instance of the proxy.
   */
  public void getFirstInstances() {
    proxyRefsCounted = true;
    index = (ProxyIndex)index.getFirstInstance();
  }

 

  /** 
   * Performs clean-up just before being garbage collected
   */
  protected void finalize() {
    // If the IMU refers to an index and this reference
    // has been counted; then decrement the count.
    if (this.index != null && proxyRefsCounted == true)
      {
      this.index.deleteRefFromInMemoryProxiesByOid();
      }
  }

}
