package gov.nist.nlpir.irf.proxy;

import gov.nist.nlpir.irf.proxy.Oid;
import gov.nist.nlpir.irf.handle.Handle;

import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.util.HashingCache;

import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

import java.io.IOException;



/** 
 * The HandlesByOid table is responsible for keeping track of the
 * handle of each persistent object by its object identifier (Oid).
 * Oids are assigned by IRF and handles may vary according to the
 * persistence mechanism used.
 * <br><br>
 * @see gov.nist.nlpir.irf.proxy.Oid
 * @see gov.nist.nlpir.irf.handle.Handle

 * 
 * @version $Revision: 1.3 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.                          
 */                      
public class AnySizeHandlesByOid extends HandlesByOid {

  /** The file used to keep track of handles by Oid. */
  private static BufferedRandomAccessFile handlesByOidIndexFile = null;

  /** Used to find the file associating Handles to Oids. */
  private static String handlesByOidIndexFileLocation = null;

  /** In the Index file, every entry is only the offset of the handle in
      the data file, so it's only a long that has to be stored. */
  private static int handlesByOidIndexEntrySize = 8;



  /** 
   * Opens the handles-by-oid files
   * @param directory directory in which to open the files
   */
  public void openHandlesByOid(String directory) {
    super.openHandlesByOid(directory);
    String locationForHandlesByOidIndex = null;
    if (directory != null)
      if (!directory.endsWith(fileSeparator))
	locationForHandlesByOidIndex = directory + fileSeparator + "DB.HanI";
      else
	locationForHandlesByOidIndex = directory +  "DB.HanI";
    else
      locationForHandlesByOidIndex = "." + fileSeparator + "DB.HanI";
    try {
      handlesByOidIndexFile =
	new BufferedRandomAccessFile(locationForHandlesByOidIndex,
				     BufferedRandomAccessFile.WRITE |
				     BufferedRandomAccessFile.READ);
    } catch (IOException ioe) {
      Support.errorMessage("InfoServer: can't open HandlesByOid file(s)");
    }
  }
  


  /** 
   * Sets the location of the handlesByOidIndexFile
   * @param location location of the handlesByOidIndexFile
   */
  final public void setHandlesByOidIndexFileLocation(String location) { 
    handlesByOidIndexFileLocation = location; 
  }



  /**
   * Writes the given Handle in the data file, and keeps track of its
   * starting offset in the index file. This method will always write
   * a Handle at the end of the data file. It will never attempt to
   * write it in place if a former Handle had already been written for
   * OID. Thus, this class will work even with handles changing size for
   * a given OID, and there are two ways to optimize it:<ul>
   * <li> Trying to rewrite a Handle in place. This can be done by
   * remembering the size of the former Hande somewhere.
   * <li> If one can assume that the Handle associated to a given OID
   * will never change in size, then Handles can always ne written in
   * place.
   * </ul>
   */
  protected void writeHandleForOid(Oid oid, Handle handle) {
    synchronized (handlesByOidDataFile) {
      try {
	/* Get the length of the data file to which the handle will
	 * write itself; this is the offset of the new data.
	 * Position the cursor to that offset.
	 * Have the handle write itself to the data file. */
	long positionToReach = ((long)oid.getValue() - 1) *
	  handlesByOidIndexEntrySize;
	handlesByOidIndexFile.seek(positionToReach);
	long nextHandleOffset = handlesByOidDataFile.getLength();
	handlesByOidIndexFile.writeLong(nextHandleOffset);
	handlesByOidDataFile.seek(nextHandleOffset);
	PersistentObjectManager.writeHandle(handle, handlesByOidDataFile);
      } catch (java.io.IOException ioe) {
	Support.errorMessage("InfoServer.writeHandleForOid(): " +
			     "can't add handle " + handle +
			     "to handlesByOidDataFile for oid " + oid);
      }
    }
  }



  /**
   * Reads a Handle back in memory from disk. From the OID, it computes
   * the offset in the index file where the offset of the handle in the
   * data file is written. It then accesses the index file, retrieves
   * the offset of the handle, and read it from the datafile.
   */
  protected Handle readHandleForOid(Oid oid) {
    synchronized (handlesByOidDataFile) {
      /* Seek to the location of the index file entry for this oid.
       * Assume there is an entry for this Oid.
       * Read the entry, the offset into the data file at which the handle
       * was stored. */
      try {
	long positionToReach = ((long)oid.getValue() - 1) * handlesByOidIndexEntrySize;
	handlesByOidIndexFile.seek(positionToReach);
	long handleOffset = handlesByOidIndexFile.readLong();
	handlesByOidDataFile.seek(handleOffset);
	return PersistentObjectManager.buildHandle(handlesByOidDataFile);
      } catch (java.io.IOException ioe) {
	Support.errorMessage("HandlesbyOid.getHandleByOid: can't get from handlesByOidDataFile");
	throw new Error();
      }
    }
  }
                                                                             


  /** 
   * Gets the size of the table 
   * @return the number of keys (Oids) in the table
   */
  public long getHandlesByOidSize() {
    long size = 0;
    try  {
      size = handlesByOidIndexFile.getLength() / handlesByOidIndexEntrySize;
    } catch (java.io.IOException ioe) {
      Support.errorMessage("InfoServer.getHandlesByOidSize:" + ioe);
    }
    return size;
  }


  /**
   * Makes sure modifications to cache entries are reflected in their
   * persistent analogues.
   */
  public void shutdown() {
    super.shutdown();
    try {
      handlesByOidIndexFile.close();
    } catch (IOException ioe) {
      Support.errorMessage("AnySizeHandlesByOid.shutdown(): " + ioe);
    }
  }
}
