package gov.nist.nlpir.irf.proxy;

import gov.nist.nlpir.irf.proxy.Oid;
import gov.nist.nlpir.irf.handle.Handle;

import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.util.HashingCache;

import java.io.Serializable;
import java.io.IOException;

import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

/** 
 * The HandlesByOid table is responsible for keeping track of the
 * handle of each persistent object by its object identifier (Oid).
 * Oids are assigned by IRF and handles may vary according to the
 * persistence mechanism used.<br><br>
 *
 * This class is abstract. Thus, there may be several implementations,
 * matching different families of handles. If all handles have the same
 * size when they're written to disk, then the OneSizeHandlesByOid class
 * can be used to get an actual instance of HandlesByOid. If they can be
 * different, then AnySizeHandlesByOid will work, as it uses a more
 * generic mechanism to match Handles to Oids.
 *
 * @see gov.nist.nlpir.irf.proxy.Oid
 * @see gov.nist.nlpir.irf.handle.Handle
 * 
 * @version $Revision: 1.6 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.                          
 */                      
abstract public class HandlesByOid implements Serializable {

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  /** The file used to keep track of handles by Oid. */
  protected static BufferedRandomAccessFile handlesByOidDataFile = null;

  /** Used to find the file associating Handles to Oids. */
  private static String handlesByOidDataFileLocation = null;

  /** Initial size of the cache used to match handles to Oids. */
  private static int maxHandlesByOidCacheEntries = 10000;

  /** Cache for handles by Oid */
  private static HashingCache handlesByOidCache = 
    new HashingCache(maxHandlesByOidCacheEntries);

  /** Local copy of System file separator */
  protected final static String fileSeparator = 
    System.getProperty("file.separator");
  
  /** Statistics on handlesByOid table and cache. */
  private static transient long handlesByOidAdds = 0;
  private static transient long handlesByOidRemoves = 0;
  private static transient long handlesByOidUpdates = 0;
  private static transient long handlesByOidGets = 0;
  private static transient long handlesByOidGetsFromCache = 0;
 

  /** 
   * Opens the handles-by-oid files
   * @param directory directory in which to open the files
   */
  public void openHandlesByOid(String directory) {
    String locationForHandlesByOidData = null;
    if (directory != null)
      if (!directory.endsWith(fileSeparator))
	locationForHandlesByOidData = directory + fileSeparator + "DB.HanD";
      else
	locationForHandlesByOidData = directory +  "DB.HanD";
    else
      locationForHandlesByOidData = "." + fileSeparator + "DB.HanD";
    try {
      handlesByOidDataFile =
	new BufferedRandomAccessFile(locationForHandlesByOidData,
				     BufferedRandomAccessFile.WRITE |
				     BufferedRandomAccessFile.READ);
    } catch (IOException ioe) {
      Support.errorMessage("InfoServer: can't open HandlesByOid file(s)");
    }
  }
  


  /** 
   * Sets the location of the handlesByOidDataFile
   * @param location location of the handlesByOidDataFile
   */
  final public void setHandlesByOidDataFileLocation(String location) { 
    handlesByOidDataFileLocation = location; 
  }



  /** 
   * Returns the maximum number of entries in the cache
   * @return the maximum number of entries in the cache
   */
  final public int getMaxHandlesByOidCacheEntries() { 
    return maxHandlesByOidCacheEntries;
  }



  /** 
   * Gets the location of the handlesByOidDataFile;
   * @return the location of the handlesByOidDataFile
   */
  final public String getHandlesByOidDataFileLocation() { 
    return handlesByOidDataFileLocation;
  }



  /**
   * Returns a String that contains statistics about this table and the
   * cache it embeds.
   */
  public String getStatistics() {
    long percentCacheHits = 0;
    if (handlesByOidGets > 0)
      percentCacheHits = handlesByOidGetsFromCache
	* 100 / handlesByOidGets;
    
    return ("\nAdds    = " + handlesByOidAdds+
	    "\nRemoves = " + handlesByOidRemoves+
	    "\nUpdates = " + handlesByOidUpdates+
	    "\nGets    = " + handlesByOidGets+
	    "\nMax cache size  = " + 
	    maxHandlesByOidCacheEntries+
	    "\nGets from cache = " + (percentCacheHits+" %\n"));
  }



  /** 
   * Adds a handle to the table of handles by Oid
   * @param handle handle to be saved
   * @param Oid object identified to be associated with the handle
   */
  public void addHandleByOid(Oid oid, Handle handle) {
    handlesByOidAdds++;
    Object[] result = handlesByOidCache.add(oid, handle);
    if (result != null)
      writeHandleForOid((Oid)result[0], (Handle)result[1]);
  }



  /**
   * Associates the given Handle to the given Oid in the file 
   * pointed by handlesByOidDataFile. It will allow this matching to live 
   * even after shutdown.
   * For the cache mechanism, see the class below:
   * @see gov.nist.nlpir.irf.util.HashingCache
   * @param oid Oid for which a handle is to be written
   * @param handle Handle to be associated with the specified Oid
   */
  protected abstract void writeHandleForOid(Oid oid, Handle handle);



  /** 
   * Replaces the entry in table of handles by Oid for the specified Oid
   * with a new one using the same Oid the new handle passed. If no entry 
   * to replace exists, write an error message, but add the new entry and 
   * continue.
   * @param Oid for desired handle
   * @param newHandle the new handle to replace the existing one
   */
  public void replaceHandleByOid(Oid oid, Handle newHandle) {
    handlesByOidUpdates++;
    Object[] result = handlesByOidCache.add(oid, newHandle);
    if (result != null)
      writeHandleForOid((Oid)result[0], (Handle)result[1]);
  }


  
  /** 
   * Gets a handle from the table of handles by Oid
   * @param Oid for desired object/handle
   * @return the handle associated with the specified Oid
   */
  public Handle getHandleByOid(Oid oid) {
    handlesByOidGets++;
    Handle handle = (Handle)handlesByOidCache.find(oid);
    if (handle != null)
      handlesByOidGetsFromCache++;    
    else {
      handle = readHandleForOid(oid);
      Object[] result = handlesByOidCache.put(oid, handle);
      if (result != null)
	writeHandleForOid((Oid)result[0], (Handle)result[1]);
    }
    return handle;
  }


  /**
   * Reads the handle from the disk. This method is gonna specify the
   * way files are used for storing Handles (index, no index, ...)
   */
  protected abstract Handle readHandleForOid(Oid oid);


  /** 
   * Gets the size of the table 
   * @return the number of keys (Oids) in the table
   */
  abstract public long getHandlesByOidSize();


  /**
   * Changes the size of the cache used to match Handles to Oids
   * @param maxHandlesByOidCacheEntries maximum number of entries
   * allowed in handles by Oid cache
   */
  public void setMaxHandlesByOidCacheEntries(int maxHandlesByOidCacheEntries) {
    /* As it is now a write-back algorithm, this won't work... */
    //    this.maxHandlesByOidCacheEntries = maxHandlesByOidCacheEntries;
    //handlesByOidCache.setMaxEntries(maxHandlesByOidCacheEntries);
      /*    Object[] OidHandleCouplesToSave = handlesByOidCache.getDirtyValues();
    int max = OidHandleCouplesToSave.length;
    for (int i = 0; i < max; i += 2)
      writeHandleForOid((Oid)OidHandleCouplesToSave[i], 
      (Handle)OidHandleCouplesToSave[i + 1]);*/
  }



  /**
   * Makes sure modifications to cache entries are reflected in their
   * persistent analogues.
   */
  public void shutdown() {
    Object[] OidHandleCouplesToSave = handlesByOidCache.getDirtyValues();
    int max = OidHandleCouplesToSave.length;
    for (int i = 0; i < max; i += 2)
      writeHandleForOid((Oid)OidHandleCouplesToSave[i], 
			(Handle)OidHandleCouplesToSave[i + 1]);
    try {
      handlesByOidDataFile.close();
    } catch (IOException ioe) {
      Support.errorMessage("HandlesByOid.shutdown(): " + ioe);
    }
  }
}
