package gov.nist.nlpir.irf.proxy;

import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

import java.io.IOException;

/** 
 * This class encapsulates the Object Identifier concept - a value which is
 * unique to each object.<br><br>
 *
 * Right now, this class is just an int wrapper.
 *
 * @version $Revision: 1.2 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public class Oid implements java.io.Serializable {


  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;


  /** @serial Unique component of the Oid, its integer value. */
  private int value;


  /** Private lock for initialization of the class. */
  private static boolean initialized = false;
  

  /** used to assign Oids to proxies only. */
  private static int lastUsedOid = 0;


  /** 
   * Private constructor used by getNextOid() and deserialization mechanism
   * with readFrom(). 
   */
  private Oid(int value) {
    this.value = value;
  }



  /** 
   *Gets the integer value oh this Oid. 
   */
  public final int getValue() {
    return value;
  }



  /** 
   * Returns a string representation of this object 
   * @return a string representation of this object
   */
  public final String toString() {
    return String.valueOf(value);
  }



  /** 
   * Gets the next Oid. At initialization, starts at 1, and then gives them 
   * away 1 by 1. Doesn't use any mechanism like MS GUIDs or Corba UIDs, but 
   * the mechanism used for (de)serialization prevents the objects from 
   * being used outside IRF, so there is no problem with that.
   * @return The next object identifier
   */
  public static final synchronized Oid getNextOid() {
    return new Oid(++lastUsedOid); 
  }



  /**
   * Returns the value of the last used Oid
   * It may be interesting to know how many Oids have already been given. 
   * @return the last used Oid
   */
  public static final synchronized int getLastUsedOid() {
    return lastUsedOid;
  }



  /**
   * Allows an exterior class to initialize the counter of given 
   * Oids. This way, even after shutting down the application, Oids can be 
   * given in a consistent way, provided this method is called by an object 
   * that went through the shutdown remembering the last Oid given. This 
   * method can only be called once per run.
   *
   * @param pLastUsedOid the value to be used to initialize the lastUsedOid
   */
  public static final synchronized void init(int pLastUsedOid) {
    if (initialized)
      Support.errorMessage("Tried to initialize the Oid counter twice.");
    else
      lastUsedOid = pLastUsedOid;
    initialized = true;
  }



  /** Returns a hashcode based on the value only. */
  public final int hashCode() {
    return value;
  }
  


  /**
   * Overrides the Object method.
   *
   * @return True only if obj is an Oid with the same value (ie IS the same,
   * but possibly a different object because materialization doesn't 
   * ensure unicity of Oids.
   */
  public final boolean equals(Object obj) {
    if ((obj != null) && (obj instanceof Oid))
      return value == ((Oid)obj).getValue();
    return false;
  }
  


  /**
   * Reads the external representation of this object in the given file 
   * beginning where its pointer currently is.
   *
   * @return null if the Oid stored was null, the Oid otherwise.
   * @param in random access file where the Oid must be read from.
   */
  public static Oid readFrom(BufferedRandomAccessFile in) {
    int value = 0;
    try {
      value = in.readInt();
    } catch (IOException ioe) {
      Support.errorMessage("VirtualProxy.readFrom(): " + ioe);
    }
    return (value == 0) ? null : new Oid(value);
  }
 

 
  /**
   * Writes external representation of this object
   * @param out random access file handle
   */
  public void writeTo(BufferedRandomAccessFile out) {
    try {
      out.writeInt(getValue());
    } catch (java.io.IOException ioe) {
      Support.errorMessage("VirtualProxy.writeTo(): " + ioe);
    }
  }



  /**
   * Writes a null Oid to file
   * This method is called by Oid clients when a optional Proxy they 
   * contain isn't present and thus doesn't have an Oid. dTthe readFrom() 
   * method will restore this Oid as <code>null</code>.
   */
  public static void writeNullOid(BufferedRandomAccessFile out) {
    try {
      out.writeInt(0);
    } catch (java.io.IOException ioe) {
      Support.errorMessage("VirtualProxy.writeNullOid(): " + ioe);
    }
  }
}
