package gov.nist.nlpir.irf.proxy;

import gov.nist.nlpir.irf.proxy.Oid;
import gov.nist.nlpir.irf.handle.Handle;

import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.util.HashingCache;

import java.io.IOException;

/** 
 * This particular implementation of HandlesByOid assumes all handles
 * take the same space when they are written to disk. This size currently
 * is 17, but if it was to change, this class can easily be extended just
 * redefining this value (not forgetting that 1 must be added to the size
 * to take care of the code written on one byte before every handle.
 * @version $Revision: 1.4 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.                          
 */                      
public class OneSizeHandlesByOid extends HandlesByOid {

  /* 
   * Size in bytes of an entry in the index file: 
   *   size of byte = 1 => handle type code
   *
   * and then, for regular handles:
   *   size of byte = 1 => fileCode
   *   size of long = 8 => record start
   *   size of long = 8 => record length
   *
   * whereas for feature list handles:
   *   size of byte = 1 => fileCode
   *   size of long = 8 => start
   *   size of long = 8 => offsetToLink
   *                 --- 
   *		     18 in both cases. Hopefully.
   */

  /** Assumed constant entry size for serialized Handle.
      It is just protected so that one may inherit this class and
      just redefine it if he had longer/shorter handles. */
  protected static int handlesByOidDataEntrySize = 18;

  

  /**
   * Directly writes the handle at the correct offset in the data file.
   * The offset is easily computed with a multiplication of the
   * handlesByOidDataEntrySize variable by the OID given in parameter.
   */
  protected final void writeHandleForOid(Oid oid, Handle handle) {
    synchronized (handlesByOidDataFile) {
      try {
	handlesByOidDataFile.seek(((long)oid.getValue() - 1) * handlesByOidDataEntrySize);
	PersistentObjectManager.writeHandle(handle, handlesByOidDataFile);
      } catch (java.io.IOException ioe) {
	Support.errorMessage("InfoServer.writeHandleForOid: can't add to handlesByOidDataFile");
      }
    }
  }



  /**
   * Reads the Handle back from the data file. The offset at which it
   * is written is calculated by multiplying the OID by the size of a
   * "serialized" handle, as this size is unique.
   */
  protected Handle readHandleForOid(Oid oid) {
    synchronized (handlesByOidDataFile) {
      try {
	handlesByOidDataFile.seek(((long)oid.getValue() - 1) * handlesByOidDataEntrySize);
	return PersistentObjectManager.buildHandle(handlesByOidDataFile);
      } catch (java.io.IOException ioe) {
	Support.errorMessage("HandlesbyOid.getHandleByOid: can't get from handlesByOidDataFile");
	throw new Error();
      }
    }
  }

  

  /** 
   * Gets the size of the table 
   * @return the number of keys (Oids) in the table
   */
  public long getHandlesByOidSize() {
    long size = 0;
    try  {
      size = handlesByOidDataFile.getLength() / handlesByOidDataEntrySize;
    } catch (java.io.IOException ioe) {
      Support.errorMessage("InfoServer.getHandlesByOidSize:" + ioe);
    }
    return size;
  }
}
