package gov.nist.nlpir.irf.proxy;

import gov.nist.nlpir.irf.handle.Handle;
import gov.nist.nlpir.irf.handle.braf.FeatureListBrafHandle;
import gov.nist.nlpir.irf.handle.braf.IdxFeatTextBrafHandle;
import gov.nist.nlpir.irf.handle.braf.IdxFeatStringBrafHandle;
import gov.nist.nlpir.irf.handle.braf.DeHtmlBrafHandle;
import gov.nist.nlpir.irf.handle.braf.DeStringBrafHandle;
import gov.nist.nlpir.irf.handle.braf.DeIntegerBrafHandle;
import gov.nist.nlpir.irf.handle.braf.DePersonNameBrafHandle;
import gov.nist.nlpir.irf.handle.braf.DeDoubleBrafHandle;
import gov.nist.nlpir.irf.handle.braf.IdxFeatBrafHandle;
import gov.nist.nlpir.irfapps.trec.cran.CranfieldDocumentBrafHandle;
import gov.nist.nlpir.irfapps.hci.HciDocBrafHandle;
import gov.nist.nlpir.irfapps.trec.fbis.FbisDocumentBrafHandle;
import gov.nist.nlpir.irf.handle.braf.IndexBrafHandle;

import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.de.DataElem;
import gov.nist.nlpir.irf.de.ProxyDataElem;
import gov.nist.nlpir.irf.de.ProxyDeString;
import gov.nist.nlpir.irf.de.ProxyDePersonName;
import gov.nist.nlpir.irf.de.ProxyDeHtml;
import gov.nist.nlpir.irf.de.ProxyDeDouble;
import gov.nist.nlpir.irf.de.ProxyDeInteger;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;
import gov.nist.nlpir.irf.feature.ProxyIdxFeatWordInString;
import gov.nist.nlpir.irf.feature.ProxyIdxFeatWordInText;
import gov.nist.nlpir.irf.document.ProxyDocument;

import gov.nist.nlpir.irfapps.hci.ProxyHciDoc;
import gov.nist.nlpir.irfapps.trec.fbis.ProxyFbisDocument;
import gov.nist.nlpir.irfapps.trec.cran.ProxyCranfieldDocument;

import gov.nist.nlpir.misc.io.BufferedRandomAccessFile;

import java.util.Vector;
import java.io.IOException;
import java.io.FileNotFoundException;

/** 
 * This class manages the two kinds of objects that may be written to
 * disk in a very short form: Proxies and Handles. Every kind of proxy
 * that must be stored in files, plus every kind of handle that has to
 * be stored in a file, must be registered with this class. This way,
 * the <code>PersistentObjectManager</code> gives them a unique code
 * which ensures an instance of the class will be possible to recreate.
 * Before this class existed, codes had to be given by hand, forcing the
 * IRF user to modify the equivalent class in the framework. Now, an
 * application defining new types of proxies/handles just has to extend
 * this class and registers the extra classes by simply calling
 * <code>addXClassName()</code>, <code>X</code> being either Proxy or
 * Handle, with the class name as a parameter.
 * 
 * @version $Revision: 1.3 $ 	
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public class PersistentObjectManager {

  /** Contains the descendants of VirtualProxy class names. */
  static protected Vector proxyClassNameVector;
  /** Contains the VirtualProxy children Class objects, in order to call
      <code>isInstance()</code> and <code>newInstance()</code>. */
  static protected Vector proxyClassVector;
  /** Size of the two previous vectors, used several times. */
  static protected int proxyClassNumber;

  /** Contains classes that just need a code, and do not
      necessarily comply to a given class or interface. */
  static protected Vector handleClassNameVector;
  /** Contains the Class objects, in order to call
      <code>isInstance()</code> and <code>newInstance()</code>. */
  static protected Vector handleClassVector;
  /** Size of the two previous vectors, used several times. */
  static protected int handleClassNumber;


  /** Working instance of ProxyDE... class */
  private static Class proxyDeHtml_Class = ProxyDeHtml.class;

  /** Working instance of ProxyDE... class */
  private static Class proxyDeStringClass = ProxyDeString.class;

  /** Working instance of ProxyDE... class */
  private static Class proxyDePersonNameClass = ProxyDePersonName.class;

  /** Working instance of ProxyDE... class */
  private static Class proxyDeDoubleClass = ProxyDeDouble.class;

  /** Working instance of ProxyDE... class */
  private static Class proxyDeIntegerClass = ProxyDeInteger.class;

  /** Working instance of ProxyDE... class */
  private static Class proxyHciDocClass = ProxyHciDoc.class;

  /** Working instance of ProxyDE... class */
  private static Class proxyFbisDocumentClass = ProxyFbisDocument.class;

  /** Working instance of ProxyDE... class */
  private static Class proxyCranfieldDocumentClass = ProxyCranfieldDocument.class;

  private static Class proxyIndexingFeatureClass = ProxyIndexingFeature.class;
  private static Class proxyIdxFeatWordInStringClass = ProxyIdxFeatWordInString.class;
  private static Class proxyIdxFeatWordInTextClass = ProxyIdxFeatWordInText.class;



  /**
   * This constructor allows to perfectly tune the classes managed,
   * specifying all of them and not just using the default ones provided
   * by the no argument constructor.
   */
  static protected void personalizedInit(Vector proxyClassNames,
					       Vector handleClassNames) {
    proxyClassNameVector = proxyClassNames;
    proxyClassNumber = proxyClassNames.size();
    handleClassNameVector = handleClassNames;
    handleClassNumber = handleClassNames.size();
  }



  /**
   */
  static public void defaultInit() {
    proxyClassNameVector = new Vector(8);
    proxyClassNumber = 0;
    addProxyClassName("gov.nist.nlpir.irf.de.ProxyDeHtml");
    addProxyClassName("gov.nist.nlpir.irf.de.ProxyDeString");
    addProxyClassName("gov.nist.nlpir.irf.de.ProxyDePersonName");
    addProxyClassName("gov.nist.nlpir.irf.de.ProxyDeDouble");
    addProxyClassName("gov.nist.nlpir.irf.de.ProxyDeInteger");
    addProxyClassName("gov.nist.nlpir.irfapps.hci.ProxyHciDoc");
    addProxyClassName("gov.nist.nlpir.irfapps.trec.fbis.ProxyFbisDocument");
    addProxyClassName("gov.nist.nlpir.irfapps.trec.cran.ProxyCranfieldDocument");

    handleClassNameVector = new Vector(12);
    handleClassNumber = 0;
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.FeatureListBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.IdxFeatTextBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.IdxFeatStringBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.DeHtmlBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.DeStringBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.DeIntegerBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.DePersonNameBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.DeDoubleBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.IdxFeatBrafHandle");
    addHandleClassName("gov.nist.nlpir.irfapps.trec.cran.CranfieldDocumentBrafHandle");
    addHandleClassName("gov.nist.nlpir.irfapps.hci.HciDocBrafHandle");
    addHandleClassName("gov.nist.nlpir.irf.handle.braf.IndexBrafHandle");
  }



  /**
   * Registers a new proxy class with this manager.
   * @param proxyClassName The complete name of the proxy class to register
   */
  static public void addProxyClassName(String proxyClassName) {
    proxyClassNameVector.addElement(proxyClassName);
    proxyClassNumber++;
  }



  /**
   * Registers a new handle class with this manager.
   * @param handleClassName The complete name of the handle class to register
   */
  static public void addHandleClassName(String handleClassName) {
    handleClassNameVector.addElement(handleClassName);
    handleClassNumber++;
  }



  /**
   * This method assigns a code to every class, and reorganizes the name
   * vectors so that indexes of names in them match the given code.
   */
  static public void init() {
    proxyClassVector = new Vector(proxyClassNumber);
    for (int i = 0; i < proxyClassNumber; i++) {
      Class proxyClassToAdd = null;
      try {
	proxyClassToAdd = Class.forName((String)proxyClassNameVector.elementAt(i));
      } catch (ClassNotFoundException cnfe) {
	System.out.println("CNFE: " + cnfe);
	System.exit(-1);
      }
      boolean added = false;
      for (int j = 0; j < proxyClassVector.size(); j++) {
	if (((Class)proxyClassVector.elementAt(j)).isAssignableFrom(proxyClassToAdd)) {
	  proxyClassVector.add(j, proxyClassToAdd);
	  added = true;
	  break;
	}
      }
      if (!added)
	proxyClassVector.addElement(proxyClassToAdd);
    }

    // recreate proxyClassNameVector:
    for (int i = 0; i < proxyClassNumber; i++)
      proxyClassNameVector.setElementAt(((Class)proxyClassVector.elementAt(i)).getName(), i);

    // Create the handle class vector.
    handleClassVector = new Vector(handleClassNumber);
    for (int i = 0; i < handleClassNumber; i++)
      try {
	handleClassVector.
	  addElement(Class.forName((String)handleClassNameVector.elementAt(i)));
      } catch (ClassNotFoundException cnfe) {
	Support.errorMessage("CNFE: " + cnfe);
	System.exit(-1);
      }
  }



  /**
   * Writes the given proxy in the given file. This class ensures that
   * <code>buildProxy()</code> called at the same offset in the file will
   * recreate this proxy.
   */
  static public void writeProxy(VirtualProxy theProxy, BufferedRandomAccessFile braf)
    throws IOException {
    braf.writeByte(getProxyType(theProxy));
    theProxy.getObjectIdentifier().writeTo(braf);
  }
  


  /**
   * Returns the assigned code to the class of this proxy.
   */
  static private int getProxyType(VirtualProxy theProxy) {
    Class theClass = theProxy.getClass();
    for (int i = 0; i < proxyClassNumber; i++)
      // if (((Class)proxyClassVector.elementAt(i)).isInstance(theProxy))
      // Optimized in:
      if (proxyClassVector.elementAt(i) == theClass)
	return i;
    Support.errorMessage("PersistentObjectManager.getProxyType: " +
			 theClass.getName() +
			 " class wasn't registered properly.");
    return -1;
  }
  
  
  
  /**
   * Builds the proxy found in the given file at its current file pointer.
   */  
  static public VirtualProxy buildProxy(BufferedRandomAccessFile braf)
    throws IOException {
    int code = braf.readByte();
    try {
      return VirtualProxy.getProxyFor(((Class)proxyClassVector.elementAt(code)),
				      Oid.readFrom(braf));
    } catch (ArrayIndexOutOfBoundsException aioobe) {
      Support.errorMessage("POM.buildProxy: Code not found - " + aioobe);
    }
    return null;
  }



  /**
   * Writes the given handle in the given file where the file
   * pointer is currently.
   */
  static public void reflectiveWriteHandle(Handle theHandle,
						 BufferedRandomAccessFile braf)
    throws IOException {
    braf.writeByte(getHandleType(theHandle));
    theHandle.writeExt(braf);
  }



  /**
   * Returns the code assigned to the class of the given handle. This code
   * will allow the manager to know which type of handle it was when it has
   * to rebuild one that was stored on disk.
   */  
  static private int reflectiveGetHandleType(Handle theHandle) {
    Class theClass = theHandle.getClass();
    for (int i = 0; i < handleClassNumber; i++)
      // if (((Class)handleClassVector.elementAt(i)).isInstance(theHandle))
      // Optimized in:
      if (handleClassVector.elementAt(i) == theClass)
	return i;
    Support.errorMessage("PersistentObjectManager.getHandleType:" +
			 theClass.getName() +
			 " class was not registered properly.");
    return -1;
  }



  /**
   * Builds the handle which is supposed to be found in the given file at
   * its current file pointer.
   */
  static public Handle reflectiveBuildHandle(BufferedRandomAccessFile braf)
    throws IOException {
    int code = braf.readByte();
    try {
      return (Handle)((Class)handleClassVector.elementAt(code)).newInstance();
    } catch (InstantiationException ie) {
      Support.errorMessage("POM.buildHandle: " + ie);
    } catch (IllegalAccessException iae) {
      Support.errorMessage("POM.buildHandle: " + iae);
    } catch (ArrayIndexOutOfBoundsException aioobe) {
      Support.errorMessage("POM.buildHandle: Code not found - " + aioobe);
    }
    return null;
  }



  /**
   * Presents the different codes assigned to the proxy/handle classes.
   */  
  static public void present() {
    System.out.println("Proxy classes by code:");
    for (int i = 0; i < proxyClassNumber; i++)
      System.out.println(i + ": " +
			 ((Class)proxyClassVector.elementAt(i)).getName());

    System.out.println("Handle classes by code:");
    for (int i = 0; i < handleClassNumber; i++)
      System.out.println(i + ": " +
			 ((Class)handleClassVector.elementAt(i)).getName());
  }


  /**
   * Returns a code indicating the type of the data element passed
   * @arg de DataElem to categorize
   * @return code
   */
  public static int getDE_Type(DataElem de) {
    // Test for DeHtml must precede that for parent class DeString !
    if      (de instanceof ProxyDeHtml)       return 1;
    else if (de instanceof ProxyDeString)     return 2;
    else if (de instanceof ProxyDePersonName) return 3;
    else if (de instanceof ProxyDeDouble)     return 4;
    else if (de instanceof ProxyDeInteger)    return 5;
    else {
      Support.errorMessage("IdxFeatBrafBroker.getDE_Type found unknown DE:" + de);
      return 0;
    }
  }



  /**
   * Returns a code indicating the type of proxy IR document passed
   * @return document type code
   */
  public static int getDocType(ProxyDocument ird) {
    if      (ird instanceof ProxyHciDoc)           return 1;
    else if (ird instanceof ProxyFbisDocument)     return 2;
    else if (ird instanceof ProxyCranfieldDocument) return 3;
    else {
      Support.errorMessage("IdxFeatBrafBroker.getDocType found unknown Doc:" + 
			   ird);
      return 0;
    }
  }
  /**
   * Writes the given proxy in the given file. This class ensures that
   * <code>buildProxy()</code> called at the same offset in the file will
   * recreate this proxy.
   */
  static public void writeDE(ProxyDataElem theProxy, BufferedRandomAccessFile braf)
    throws IOException {
    braf.writeByte(getDE_Type(theProxy));
    theProxy.getObjectIdentifier().writeTo(braf);
  }
  
  /**
   * Writes the given proxy in the given file. This class ensures that
   * <code>buildProxy()</code> called at the same offset in the file will
   * recreate this proxy.
   */
  static public void writeDoc(ProxyDocument theProxy, BufferedRandomAccessFile braf)
    throws IOException {
    braf.writeByte(getDocType(theProxy));
    theProxy.getObjectIdentifier().writeTo(braf);
  }
  
  
  /**
   * Returns a Proxy Data Element of the type associated with the value
   * in the byte[] passed
   * @arg value code indicating which ProxyDE was stored
   * @return DataElem
   */
  public static ProxyDataElem buildDE(BufferedRandomAccessFile braf)
    throws IOException {
    switch (braf.readByte()) {
    case 1: 
      return (ProxyDataElem)VirtualProxy.
	getProxyFor(proxyDeHtml_Class, Oid.readFrom(braf));
    case 2: 
      return (ProxyDataElem)VirtualProxy.
	getProxyFor(proxyDeStringClass, Oid.readFrom(braf));
    case 3:
      return (ProxyDataElem)VirtualProxy.
	getProxyFor(proxyDePersonNameClass, Oid.readFrom(braf));
    case 4:
      return (ProxyDataElem)VirtualProxy.
	getProxyFor(proxyDeDoubleClass, Oid.readFrom(braf));
    case 5:
      return (ProxyDataElem)VirtualProxy.
	getProxyFor(proxyDeIntegerClass, Oid.readFrom(braf));
    default:
      braf.unread();
      Support.errorMessage("IdxFeatBrafBroker.buildDE found unknown DE code: " + 
			   braf.readByte());
      return null;
    }
  }
  


  /**
   * Builds a proxy IR document from serialized data in the specified file
   * @arg braf buffered random access file containing the serialized document
   * @return the deserialized proxy IR document
   */
  public static ProxyDocument buildDoc(BufferedRandomAccessFile braf)
    throws IOException {
    int value = braf.readByte();
    switch (value) {
    case 1:
      return (ProxyHciDoc)VirtualProxy.
	getProxyFor(proxyHciDocClass, Oid.readFrom(braf));
    case 2:
      return (ProxyFbisDocument)VirtualProxy.
	getProxyFor(proxyFbisDocumentClass, Oid.readFrom(braf));
    case 3:
      return (ProxyCranfieldDocument)VirtualProxy.
	getProxyFor(proxyCranfieldDocumentClass, Oid.readFrom(braf));
    default:
      Support.errorMessage("IdxFeatBrafBroker.buildDoc found unknown Doc code:" + 
			   value);
      return null;
    }
  }



  public static Handle buildHandle(BufferedRandomAccessFile braf)
    throws IOException {
    Handle handle = null;
    switch (braf.readByte()) {
    case 1: handle = new DeDoubleBrafHandle(); break;
    case 2: handle = new DeHtmlBrafHandle(); break;
    case 3: handle = new DeIntegerBrafHandle(); break;
    case 4: handle = new DePersonNameBrafHandle(); break;
    case 5: handle = new DeStringBrafHandle(); break;
    case 6: handle = new FbisDocumentBrafHandle(); break;
    case 7: handle = new HciDocBrafHandle(); break;
    case 8: handle = new IdxFeatBrafHandle(); break;
    case 9: handle = new IdxFeatStringBrafHandle(); break;
    case 10: handle = new IdxFeatTextBrafHandle(); break;
    case 11: handle = new IndexBrafHandle(); break;
    case 12: handle = new FeatureListBrafHandle(); break;
    case 13: handle = new CranfieldDocumentBrafHandle(); break;
    default:
      braf.unread();
      Support.errorMessage("PersistentObjectManager.buildHandle: unknown code = " 
			   + braf.readByte() + ": cannot retrieve handle object "
			   + "supposed to start at " + braf.getFilePointer() + ".");
      return null;
    }
    handle.readExt(braf);
    return handle;
  }



  /**
   * Returns a unique code for every kind of Handle that may be 
   * encountered.
   * @param handle Handle for which a code is to be returned
   * @return the code for the specified handle
   */
  public static int getHandleType(Handle handle) {
    if      (handle instanceof IdxFeatTextBrafHandle)             return 10;
    else if (handle instanceof DeHtmlBrafHandle)           return 2;
    else if (handle instanceof FeatureListBrafHandle) return 12;
    else if (handle instanceof IdxFeatStringBrafHandle)             return 9;
    else if (handle instanceof DeStringBrafHandle)          return 5;
    else if (handle instanceof DeDoubleBrafHandle)          return 1;
    else if (handle instanceof DePersonNameBrafHandle)      return 4;
    else if (handle instanceof FbisDocumentBrafHandle)      return 6;
    else if (handle instanceof HciDocBrafHandle)            return 7;
    else if (handle instanceof IdxFeatBrafHandle)                return 8;
    else if (handle instanceof CranfieldDocumentBrafHandle)  return 13;
    else if (handle instanceof DeIntegerBrafHandle)         return 3;
    else if (handle instanceof IndexBrafHandle)              return 11;
    else {     
      Support.errorMessage("PersistentObjectManager.getHandleType: unknown handle of type " +
			   handle.getClass().getName());
      return 0;
    }
  }

  public static void writeHandle(Handle handle, BufferedRandomAccessFile braf)
    throws IOException {
    braf.writeByte(getHandleType(handle));
    handle.writeExt(braf);
  }

  
  /**
   * This method builds an IndexingFeature from the BufferedRandomAccessFile
   * used. It reads the code assigned to the IF, and then creates the IF with
   * the needed class.
   */
  public static ProxyIndexingFeature buildIF(BufferedRandomAccessFile braf)
    throws java.io.IOException {
    switch (braf.readByte()) {
    case 1:
      return (ProxyIndexingFeature)VirtualProxy.
	getProxyFor(proxyIndexingFeatureClass, Oid.readFrom(braf));
    case 2:
      return (ProxyIndexingFeature)VirtualProxy.
	getProxyFor(proxyIdxFeatWordInStringClass, Oid.readFrom(braf));
    case 3:
      return (ProxyIndexingFeature)VirtualProxy.
	getProxyFor(proxyIdxFeatWordInTextClass, Oid.readFrom(braf));
    default:
      braf.unread();
      System.err.println("PersistentObjectManager.buildIF: unknown code " +
			 braf.readByte() + " accessing " + braf);
    }
    return null;
  }
  

  /**
   * Writes the given proxy in the given file. This class ensures that
   * <code>buildProxy()</code> called at the same offset in the file will
   * recreate this proxy.
   */
  static public void writeIF(ProxyIndexingFeature theProxy,
			     BufferedRandomAccessFile braf)
    throws IOException {
    braf.writeByte(getIF_Type(theProxy));
    theProxy.getObjectIdentifier().writeTo(braf);
  }
  


  /**
   * This method is a perfect candidate for a general ObjectProvider class
   * that would centralize them. Many different classes may want to store IndexingFeatures.
   *
   * @return an arbitrary code for the given IF. But this code will allow to completely
   * rebuild the IF.
   */
  public static int getIF_Type(ProxyIndexingFeature IF) {
    /* !!! Take care to the order of comparisons : A ProxyIdxFeatWordInText
     * **IS** a ProxyIndexingFeature !!! */
    if      (IF instanceof ProxyIdxFeatWordInText)   return 3;
    else if (IF instanceof ProxyIdxFeatWordInString) return 2;
    else if (IF instanceof ProxyIndexingFeature) return 1;
    else
      System.err.println("LinkedListPool.getIF_Type: unknown IF type for " + IF);
    return 0;
  }

}
