package gov.nist.nlpir.irf.proxy;

import gov.nist.nlpir.irf.irfman.InfoServer;
import gov.nist.nlpir.irf.util.Support;

import gov.nist.nlpir.irf.proxy.HandlesByOid;
import gov.nist.nlpir.irf.broker.PersistenceBroker;
import gov.nist.nlpir.irf.handle.Handle;

import java.util.Vector;
//import java.lang.ref.SoftReference;

/** 
 * This is an abstract class which represents what's common to all
 * proxies for support of on-demand (de)materialization of the real objects
 * for which they stand.
 * <br><br>
 * 
 * @version $Revision: 1.13 $ 	
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */
public abstract class VirtualProxy implements java.io.Serializable {

  /** Tracks in-memory proxies for persistent objects to prevent
   * duplication. key = objectIdentifier; value = VirtualProxy */
  private static transient Vector inMemoryProxiesByOid = new Vector(10000, 1);
  // The 1 allows the Vector to grow at a different rate than (* 2) each time.
  
  /** The persistence broker for this proxy's real object */  
  protected transient PersistenceBroker broker = null; 
  
  /** The real object for which this object is a proxy, a handle,
   * a smart reference */
  private transient Object realObject = null; 

  /*  A soft reference to the real object in case it has been made
      persistent. May allow to get it back without having to
      materialize it. 
  */
  /* This soft reference is currently commented out because it leads to a
     OutOfMemoryException (sic). There must be a problem either in the API
     or in the JVM. It looks like thi mechanism can't handle too many objects.
  */
  //private transient SoftReference softRef = null;

  /** serial version universal id - put here so Java does not insert one 
   * which may change due to revisions and make it impossible to deserialize
   * earlier versions of serialized objects */
  static final long serialVersionUID = 0;

  /** Unique identifier for persistent object 
   * @serial */
  protected Oid objectIdentifier = null; 

  /** The info server - here to save getting it every time its needed. 
   * This assumes the IrfManager is created before this class is loaded. */
  protected static InfoServer iS;

  /** The table of handles by Oid. Here for ease of access */
  protected static HandlesByOid handles;

  /** Gets this proxy's Oid 
   * @return this proxy's Oid
   */
  public Oid getObjectIdentifier() {
    return objectIdentifier;
  }



  /** Sets the info server reference */
  public static void setIS(InfoServer pIS) {
    iS = pIS;
    handles = pIS.getHandlesByOid();
  }



  /**
   * Makes the object persistent IFF not already persistent
   * Assign it an oid if it has none; save its handle by Oid; and
   * save the Oid in the proxie
   */          
  public void makePersistent() {
    if (objectIdentifier == null) {   // if not persistent yet
      objectIdentifier = Oid.getNextOid();
      if (realObject != null)
	handles.addHandleByOid(objectIdentifier, 
			       dematerializeRealObject(realObject));
      else
	Support.errorMessage("VirtualProxy: No real object to put");
    }
  }
  


  /**
   * Replaces the realobject associated with this proxy.
   * Call it when the object has changed. Change only makes sense
   * if there is a real object in memory, different from the one
   * on disk. Therefore this method does nothing if there is no
   * realObject reference. (Moral: makeLightweight() makes 
   * replaceRealObject() impossible. Replace before making lightweight.)
   */
  public void replaceRealObject() {
    if (realObject != null) {
      Handle newHandle = 
	dematerializeRealObject(handles.
				getHandleByOid(objectIdentifier), realObject);
      // Assume handle changed
      handles.replaceHandleByOid(objectIdentifier, newHandle); 
    }
  }
  


  /** 
   * Returns the real object for which this proxy object stands
   * @return The real object for which this proxy stands
   */          
  public final Object getRealObject() {
    if (realObject == null)
      /* The following code uses softRef, and that's why it is commented out.
      if (softRef != null) {        // This test is necessary for proxies
	realObject = softRef.get(); // that just came back from disk.
	if (realObject == null)
	  realObject = materializeRealObject();
      } else */ 
	realObject = materializeRealObject();
    return realObject;
  }
  

  /** 
   * Make this proxy object lightweight - cut the link to its real object
   */          
  public void makeLightweight() {
    if (objectIdentifier == null)
      Support.errorMessage("ProxyDataElem.makeLightweight: Real object" +
			   " not persistent and so not recoverable.");
    // This code uses softRef, and thus is commented out.
    //    if (realObject != null)
    // softRef = new SoftReference(realObject);
    realObject = null;
  }
  

  /** Sets the broker field 
   * @param the broker for this proxy
   */
  protected final void setBroker(PersistenceBroker b) {
    broker = b;
  }


  /** Returns the broker for this proxy 
   * @return the broker for this proxy
   */
  protected final PersistenceBroker getBroker() {
    return broker;
  }


  /** Sets the real object field 
   * @param o the real object for this proxy
   */
  protected final void setRealObject(Object o) {
    realObject = o;
  }


  /**
   * Tests whether this proxy is lightweight or not. For this method, a
   * proxy is considered lightweight if its realObject reference is null.
   * @return true if the realObject reference is null.
   */
  // When softRef is used, this function's name should be changed to
  // realObjectReachable or something like that,
  // BECAUSE OF THE SIDE EFFECT RESTORING THE REAL OBJECT REFERENCE.
  protected final boolean isLightweight() {
    /* Code using softRef.
    if (realObject == null) {
      if (softRef != null)          // This test is necessary for proxies
	realObject = softRef.get(); // that just came back from disk.
      return (realObject == null);
    } else
      return false; */
    return (realObject == null);
  }
  
  
  /* Sets the object identifier for this proxy.
   *  Package visibility only - for use by brokers materializing proxies
   * @param the Oid
   */
  final void setObjectIdentifier(Oid objectIdentifier) {
    this.objectIdentifier = objectIdentifier;
  }



  /** 
   * Get the broker for this proxy object
   * @return the broker for this object
   */          
  protected final PersistenceBroker getPersistenceBroker() {
    if (broker == null)       
      broker = createBroker();
    return broker;
  }



  /** 
   * Materializes the real object for which this proxy object stands
   * @return the real object for which this proxy stands
   */          
  private final Object materializeRealObject() {
    return getPersistenceBroker().
      objectWith(handles.getHandleByOid(objectIdentifier));
  }

  
  
  /** 
   * Dematerializes the real object for which this proxy object stands
   * @param obj the object to be dematerialized/deserialized
   * @return Handle of the designated object
   */          
  private final Handle dematerializeRealObject(Object obj) {
    return getPersistenceBroker().handleOf(obj);
  }


  
  /**
   * Dematerializes the updated real object for which this proxy object stands
   * @param oldHandle The handle the object used to have.
   * @return the new handle for the object. May be the same.
   */
  private final Handle dematerializeRealObject(Handle oldHandle, Object obj) {
    return getPersistenceBroker().handleOf(oldHandle, obj);
  }


  
  /**
   * Enforce uniqueness of proxies by checking to see if a proxy for
   * this IR document already exists and returning that reference if
   * it exists.
   * Default definition. Override it if the Proxy refers to a class 
   * potentially containing proxies to call getFirstIntances() on this class.
   * <br>
   * Principle: when a proxy comes back from disk, it may be redundant with 
   * one already existing. There is no easy way to test it from inside the 
   * proxy, so the client is responsible for the unicity of the proxy by 
   * calling this function on it right after it is created.
   *<br>
   * Example:
   * <pre>
   * Proxy1 myProxy = readProxy1FromDisk();
   * myProxy = myProxy.getFirstInstance();
   * </pre>
   * Of course, if at one point it is sure this instance of the proxy is
   * the first one to be materialized, then the call to this function can be
   * avoided.<br>
   * Note on Proxy-containers Objects: If a Proxy X refers to an object Y 
   * that may contain proxies, then Y must contain a <code>void 
   * getFirstInstances()</code> method, and an overriding of the current 
   * method in the proxy X must apply it to Y. If an object is a container 
   * of container of Proxies, it must define a <code>void getFirstInstances()
   * </code> that recursively calls the same method on the contained 
   * container, and so on.<br>
   * Note on non-Proxy Objects: On non-Proxy object, the method has no 
   * return value because it is just used to pass the recursive call.
   */
  public VirtualProxy getFirstInstance() {
    return addRefToInMemoryProxiesByOid();
  }
  


  /**
   * If there is already a proxy with the specified Oid, then return it
   * after recording another reference to said proxy. Otherwise create
   * a proxy (an instance of proxyClass), add it to the table and return
   * it.
   *
   * @param proxyClass Class of the proxy to be gotten
   * @param objectIdentifier Oid of the proxy to tobe gotten
   * @returns an existing instance or a new one
   */
  public final static VirtualProxy getProxyFor(Class proxyClass, 
					       Oid objectIdentifier) {   
    VirtualProxy theProxy = null;
    int oid = objectIdentifier.getValue();
    testInMemoryProxiesByOidTable(oid);
    
    IMP_Entry entry = (IMP_Entry)inMemoryProxiesByOid.elementAt(oid);
    if (entry == null) {
      try {
	theProxy = (VirtualProxy)proxyClass.newInstance();
	theProxy.setObjectIdentifier(objectIdentifier);
      } catch (InstantiationException ie) {
	System.err.println("VirtualProxy.getProxyFor: " + ie);
      } catch (IllegalAccessException iae) {
	System.err.println("VirtualProxy.getProxyFor: " + iae);
      }
      theProxy = theProxy.addRefToInMemoryProxiesByOid();
    } else
      theProxy = entry.proxy.addRefToInMemoryProxiesByOid();  // known to be in table
    
    return theProxy;
  }



  /** 
   * The concrete virtual proxy will create a broker for this proxy object
   * @return the broker for this object
   */          
  public abstract PersistenceBroker createBroker();


  /**
   * Returns the class instance as string.
   * @return the class instance as string
   */
  public String toString() {
    return "P " + getRealObject().toString();
  }


  
  /**
   * Reads external representation of this object
   * @param in random access file handle
   */
  public final void readExternal(gov.nist.nlpir.misc.io.BufferedRandomAccessFile in) {
    objectIdentifier = Oid.readFrom(in);
  }
 


  /**
   * Writes external representation of this object
   * @param out random access file handle
   */
  public final void writeExternal(gov.nist.nlpir.misc.io.BufferedRandomAccessFile out) {
    objectIdentifier.writeTo(out);
  }



  /*-------------------------------------------------------------------*
   * Code dealing with the table of in-memory proxies for persistent
   * objects
   *-------------------------------------------------------------------*/
  private class IMP_Entry {  // entry (value) in In-Memory Proxy table 
    public int refCount = 1;
    public VirtualProxy proxy = null;
    
    public IMP_Entry(int refCount, VirtualProxy proxy) {
      this.refCount = refCount;
      this.proxy = proxy;
    }
  }
  


  /**
   * Adds to the table of in-memory proxies a reference to the specified proxy
   * returning any existing one if one exists
   * 
   * @param Oid the object identifier of the proxy to be added
   */
  public final VirtualProxy addRefToInMemoryProxiesByOid() {
    if (objectIdentifier == null)
      throw new RuntimeException("VirtualProxy.addRef...: Can't add ref"
				 + " to proxy for non-persistent object");

    int oid = objectIdentifier.getValue();
    testInMemoryProxiesByOidTable(oid);
    
    IMP_Entry entry = (IMP_Entry)inMemoryProxiesByOid.elementAt(oid);

    if (entry == null) {
      entry = new IMP_Entry(1, this); /* ref to instance of VP needed 
				       * to invoke constructor ! */
      inMemoryProxiesByOid.setElementAt(entry, oid);
      //} else synchronized (entry) { // Shouldn't this synbchronized be LARGER ???
    } else synchronized (inMemoryProxiesByOid) {
      entry.refCount++;
    }
    return entry.proxy;
  }


  /**
   * Deletes from the table of in-memory proxies a reference to the 
   * specified proxy 
   *
   * @return virtual proxy
   */
  public final void deleteRefFromInMemoryProxiesByOid() {
    int oid  = objectIdentifier.getValue();
    testInMemoryProxiesByOidTable(oid);
    
    IMP_Entry entry = (IMP_Entry)inMemoryProxiesByOid.elementAt(oid);
    if (entry == null) {
      Handle handle = iS.getHandlesByOid().getHandleByOid(objectIdentifier); 
      Support.errorMessage("VirtualProxy.deleteRefFromInMemoryProxiesByOid: " +
			   "No entry for Oid = " + objectIdentifier + " Handle class = " +
			   ((handle != null) ? handle.getClass().getName() : " No handle"));
      //} else synchronized (entry) {
    } else synchronized (inMemoryProxiesByOid) {
      entry.refCount--;
      if (entry.refCount <= 0)
	inMemoryProxiesByOid.setElementAt(null, oid);
    }
  }
  
  /**
   * Gets the number of keys in the subject table
   * @return number of keys in inMemoryProxiesByOid table
   */
  public final static int getInMemoryProxiesByOidSize() {
    return inMemoryProxiesByOid.size();
  }



  /**
   * Gets the number of bins in the subject table
   * @return number of bins in inMemoryProxiesByOid table
   */
  public final static int getInMemoryProxiesByOidBinCount() {
    return inMemoryProxiesByOid.capacity();
  }



  /**
   * Returns information on the number of in-memory proxies for persistent
   * objects - lightweight versus heavyweight
   * @return a string containing the information
   */
  public static String getInMemoryProxyCounts() {
    int nrOfProxies = inMemoryProxiesByOid.size();
    // Accumulators for light and heavy proxies
    int ph = 0, pl = 0;
    
    // Accumulators for reference counts for light and heavy versions of 
    // each type of proxy
    int phC = 0, plC = 0;
    
    for (int i = 0; i < nrOfProxies; i++) {
      IMP_Entry current = (IMP_Entry)inMemoryProxiesByOid.elementAt(i);
      if (current != null)
	if (current.proxy.realObject == null) {
	  pl++;
	  plC += current.refCount;
	} else {
	  ph++;
	  phC += current.refCount;
	}
    }
    return ("In-memory proxy counts: Heavy " + ph + " counts " + phC +
	    " light " + pl + " counts " + plC);
  }
  
  

  /**
   * Writes the contents of the inMemoryProxiesByOid table to standard
   * out 
   */
  public static void presentInMemoryProxies() {
    String state = null;
    System.out.println("InMemoryProxies...");
    int nrOfProxies = inMemoryProxiesByOid.size();
    for (int i = 0; i < nrOfProxies; i++) {
      IMP_Entry current = (IMP_Entry)inMemoryProxiesByOid.elementAt(i);
      if (current != null) {
	if (current.proxy.realObject == null)
	  state = "L";
	else
	  state = "H";
	System.out.println("State = " + state + "  Refcount = " +
			   current.refCount + "  Proxy = " +
			   current.proxy.getClass().getName() +
			   "  Oid = " + current.proxy.getObjectIdentifier());
      }
    }
    System.out.println(getInMemoryProxyCounts());
  }
  

  /**
   * Tests whether the InMemoryProxiesByOid table is large enough to
   * accept this oid. If it's not, it extends the table. The current
   * extension of the table is an increment of 50, but this makes the
   * table allocated again and again. That's why a better solution
   * would be to multiply its size by a factor of 1.2.
   * There will be an int overflow problem if there are more than
   * 2^31 (current int maximum) oids to store.
   */
  private static void testInMemoryProxiesByOidTable(long oid) {
    if (oid >= inMemoryProxiesByOid.size()) {
      int newCapacity = (int)(inMemoryProxiesByOid.capacity() * 1.2);
      // But this may not be enough, chek again:
      if (oid >= newCapacity)
	newCapacity = (int)oid + 1; 
      inMemoryProxiesByOid.ensureCapacity(newCapacity);
      inMemoryProxiesByOid.setSize(newCapacity);
    }
  }
}
