package gov.nist.nlpir.irf.query;
import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.de.DataElem;
import java.util.Vector; 

/** 
 * Holds information on a query condition, i.e. a particular aspect of a query. This information includes the 
 * query feature itself, the number of times the feature occurs in the query, the way(s) in which the feature
 * is to be evaluated against the document index and the query document in which the query feature 
 * occurs.
 *
 * @version $Revision: 1.2 $
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public class QueryCondition {

    /** 
     * The document containing the query from which this query condition is derived.
     */
    private ProxyDocument queryDoc;

    /** 
     * The query document attribute.
     */
    private DataElem queryDocAttr;

    /** 
     * The query feature (e.g. a query term such as a word in a text-based retrieval system).
     */
    private DataElem feature;

    /** 
     * The QueryModalityUnit which describes the actions to be performed regarding this query condition.
     */
    private QueryModalityUnit QMU;

    /** 
     * The frequency of the query feature in the query.
     */
    private int freqOfFeatureInQuery;

    /**
     * The maximum frequency of any feature in the query.
     */
    private int maxFreqOfAnyFeatureInQuery;

    /** 
     * Creates a QueryCondition for the QueryModalityUnit and initializes freqOfFeatureInQuery 
     * and maxFreqOfAnyFeatureInQuery to 1.
     * 
     * @param unit A QueryModalityUnit.
     */
    public QueryCondition(QueryModalityUnit unit) {
	QMU = unit;      
	freqOfFeatureInQuery = 1;
	maxFreqOfAnyFeatureInQuery = 1;
    }

    /** 
     * Creates a QueryCondition for the supplied query, query document type,
     * feature, and QueryModalityUnit. In addition freqOfFeatureInQuery is initialized
     * to freq and maxFreqOfAnyFeatureInQuery is initialized to maxf
     * 
     * @param q Query document.
     * @param a Query document attribute.
     * @param f A feature.
     * @param unit A QueryModalityUnit.
     * @param freq Frequency of feature in query.
     * @param maxf Maximum frequency of any feature in query.
     */
    public QueryCondition(ProxyDocument q, DataElem a, DataElem f, 
		   QueryModalityUnit unit, int freq, int maxf) {
	queryDoc = q;
	queryDocAttr = a;
	feature = f;
	QMU = unit;	
	freqOfFeatureInQuery = freq;
	maxFreqOfAnyFeatureInQuery = maxf;
    }
    
    /** 
     * Sets the query document.
     * 
     * @param q Query document.
     */
    public void setQueryDoc(ProxyDocument q) {
	queryDoc = q;
    }
    

    /** 
     * Gets the query document.
     * 
     * @return The document.
     */
    public ProxyDocument getQueryDoc() {
	return queryDoc;
    }
    
    
    /** 
     * Sets QueryDocAttr to a.
     * 
     * @param a Query document attribute.
     */
    public void setQueryDocAttr(DataElem a) {
	queryDocAttr = a;
    }
    
    
    /** 
     * Gets the query document attribute.
     * 
     * @return The attribute.
     */
    public DataElem getQueryDocAttr() {
	return queryDocAttr;
    }


     /** 
      * Sets the query feature.
      * 
      * @param f a feature.
      */
    public void setFeature(DataElem f) {
	feature = f;
    }


    /** 
     * Gets the query feature.
     * 
     * @return feature
     */
    public DataElem getFeature() {
	return feature;
    }
    
    /** 
     * Gets the Query Modality Unit of the QueryCondition.
     * 
     * @return The Query Modality Unit.
     */
    public QueryModalityUnit getQueryModalityUnit() {
	return QMU;
    }

    /** 
     * Gets the frequency of the feature in the query.
     * 
     * @return The frequency.
     */
    public int getFreqOfFeatureInQuery() {
	return freqOfFeatureInQuery;
    }
    
    
    /** 
     * Sets the frequency of the feature in the query.
     * 
     * @param f Frequency of feature in query.
     */
    public void setFreqOfFeatureInQuery(int f) {
	freqOfFeatureInQuery = f;
    }
    
    
    /** 
     * Gets the maximum frequency of any feature in the query.
     * 
     * @return This maximum frequency.
     */
    public int getMaxFreqOfAnyFeatureInQuery() {
	return maxFreqOfAnyFeatureInQuery; 
    }
    
    
    /** 
     * Sets the maximum frequency of any feature in the query.
     * 
     * @param f Maximum frequency of any feature in query.
     */
    public void setMaxFreqOfAnyFeatureInQuery(int f) {
	maxFreqOfAnyFeatureInQuery = f;
    }
}
