package gov.nist.nlpir.irf.query;

import gov.nist.nlpir.irf.query.Combinator;
import gov.nist.nlpir.irf.de.DeInteger;
import gov.nist.nlpir.irf.de.DeDouble;
import gov.nist.nlpir.irf.de.DeString;
import gov.nist.nlpir.irf.modality.IndexingModalityUnit;
import gov.nist.nlpir.irf.matcher.Matcherbp;
import gov.nist.nlpir.irf.matcher.NullMatcherbp;
import gov.nist.nlpir.irf.util.Support;

import java.util.Vector;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * This class manipulates the representation of a query. The query can
 * either be structured or not. If it is, the queryCollection vector
 * inside of the QueryRep object starts with a DeString naming the
 * functor of the query, and contains then the two members of the query.
 * If it is not, this vector is only a collection of QueryModalityUnits.
 *
 * @version $Revision: 1.3 $     
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */

public class QueryRep {

  /** A vector of QueryModalityUnits and/or vectors of QueryModalityUnits 
   * from the query, representing the relationships between the parts of the
   * query.*/
  private Vector queryCollection;

    /** 
     * Creates a vector of QueryModalityUnits and/or vectors of 
     * QueryModalityUnits from the Query.
     * This structure represents the relationships between the parts of the
     * query.
     * 
     * @param IMs Indexing modalities of query document
     * @param Q An ordered collection for specifying the query structure.<br>
     * Description of the collection if the query is structured:<pre>
     * 0 1 2 3 4
     * | | | | `DeDouble: the weight of the second feature if it exists
     * | | | `--DeInteger: the number of the second feature in the IMs Vector
     * | | |                OR a Vector like this one, and then 4 is null.
     * | | `----DeDouble: the weight of the first feature if it exists
     * | `------DeInteger: the number of the second feature in the IMs Vector
     * |                    OR a Vector like this one, and then 2 is null.
     * `--------DeString: Functor name.
     * </pre>
     * If the query is unstructured:<pre>
     * 0 1 2 3 ad lib
     * | | | `-DeDouble:  Weight of the 2nd Indexing Modality
     * | | `---DeInteger: Number of the Modality
     * | `-----DeDouble:  Weight of the 1st Indexing Modality
     * `-------DeInteger: Number of the Modality
     * </pre>
     */
    public QueryRep(Vector IMs, Combinator Q) {
	Matcherbp matcher = new NullMatcherbp();
	IdxWeightsWithMatchDegree combiner = new IdxWeightsWithMatchDegree();
	DeInteger QM_Number;
	DeDouble featureWeight;
	IndexingModalityUnit anIMU;
	QueryModalityUnit qm;
	int number;

	queryCollection = new Vector();
	try {
	    if ((Q.elementAt(0) instanceof DeString) && (Q.size() == 5)) {
		// structured query
		queryCollection.addElement(Q.elementAt(0));	// functor
		if (Q.elementAt(1) instanceof DeInteger) {
		    // terminal node
		    QM_Number = (DeInteger)Q.elementAt(1);
		    featureWeight = (DeDouble)Q.elementAt(2);
		    number = QM_Number.getValue();
		    anIMU = (IndexingModalityUnit)IMs.elementAt(number);
		    // get attribute name & doc type!!!!
		    qm = new QueryModalityUnit(anIMU.getIndexName(), matcher, combiner, null, null, 0.0);
		    qm.setNumber(number);
		    qm.setWeight(featureWeight.getValue());
		    queryCollection.addElement(qm);	
		} else if (Q.elementAt(1) instanceof Vector) {
		    // non-terminal node
		    queryCollection.addElement(new QueryRep(IMs, (Combinator)Q.elementAt(1))); 
		} else {
		    Support.errorMessage("QueryRep.QueryRep(): 2nd element of Vector Q " +
					 "not recognized.");
		}
		if (Q.elementAt(3) instanceof DeInteger) {
		    // terminal node
		    QM_Number = (DeInteger)Q.elementAt(3);
		    featureWeight = (DeDouble)Q.elementAt(4);
		    number = QM_Number.getValue();
		    anIMU = (IndexingModalityUnit)IMs.elementAt(number);
		    // get attribute name & doc type!!!!
		    qm = new QueryModalityUnit(anIMU.getIndexName(), 
					       matcher, combiner, null, 
					       null, 0.0);
		    qm.setNumber(number);
		    qm.setWeight(featureWeight.getValue());
		    queryCollection.addElement(qm);	 
		    
		} else if (Q.elementAt(3) instanceof Vector) {
		    // non-terminal node
		    queryCollection.addElement(new QueryRep(IMs, 
(Combinator)Q.elementAt(3)));
		} else {
		    Support.errorMessage("QueryRep.QueryRep(): 4th element of Vector Q " +
					 "not recognized.");
		}

	    } else if ((Q.elementAt(0) instanceof DeInteger) &&
		       (((Q.size() / 2) * 2) == Q.size())) {
	      // unstructured query
	      int qSize =  Q.size();
	      int i = 0;
	      while (i < qSize) {
		QM_Number = (DeInteger)Q.elementAt(i++);
		featureWeight = (DeDouble)Q.elementAt(i++);
		number = QM_Number.getValue();
		anIMU = (IndexingModalityUnit)IMs.elementAt(number);
		// get attribute name & doc type!!!!
		qm = new QueryModalityUnit(anIMU.getIndexName(), matcher, 
					   combiner, null, null, 0.0);
		qm.setNumber(number);
		qm.setWeight(featureWeight.getValue());
		queryCollection.addElement(qm);	
		}
	    } else {
		Support.errorMessage("QueryRep.QueryRep(): Combinator Q not recognized.");
	    }
	} catch (ClassCastException e) {
	    Support.errorMessage("QueryRep.QueryRep(): One of the Vectors does " +
				 "not contain arguments of the expected type.");
	} catch (NoSuchElementException e) {
	    Support.errorMessage("QueryRep.QueryRep(): IMs doesn't contain the " +
				 "expected number of arguments.");
	} catch (ArrayIndexOutOfBoundsException e) {
	    Support.errorMessage("QueryRep.QueryRep(): The number of one of the " +
				 "QMUs doesn't match any IMU in Vector IMs.");
	}

    }

    

    /** 
     *Identifies a structured query.
     *
     *@return True/False
     *
     */
    public boolean isStructuredQuery() {
	return ((queryCollection.size() == 3)
		&& (queryCollection.elementAt(0) instanceof DeString));
    }


    
    /** 
     * Identifies a unstructured query.
     *
     * @return True/False
     */
    public boolean isUnstructuredQuery() {
	try {
	    return (queryCollection.elementAt(0) instanceof QueryModalityUnit);
	} catch (NoSuchElementException e) {
	    Support.errorMessage("QueryRep.isUnstructuredQuery(): query collection is empty.");
	    return false;
	}
    }



    /** 
     * Gets QueryModalityUnit number n.
     *
     * @param n Subscript of desired QueryModalityUnit.
     * @return The requested QueryModalityUnit or null if not found.
     */
    public QueryModalityUnit getQueryModalityUnit(int n) {
	int queryCollSize = queryCollection.size();
	for (int i=0; i<queryCollSize; i++) {
	    Object o = queryCollection.elementAt(i);
	    if ((o instanceof QueryModalityUnit) && 
		(((QueryModalityUnit)o).getNumber() == n))
		return (QueryModalityUnit)o;
	    if (o instanceof QueryRep) {
		QueryModalityUnit qm = ((QueryRep)o).getQueryModalityUnit(n);
		if (qm != null)
		    return qm;
	    }
	}
	return null;
    }



    /**
     * Returns a structured Vector of QueryRep and QueryModalityUnits. 
     *
     * See the description of the class for more information.
     * @return a structured vector of QueryRep and QueryModalityUnits.
     */

    public Vector getQueryCollection() {
	return queryCollection;
    }
}
