package gov.nist.nlpir.irf.query;

import gov.nist.nlpir.irf.de.DeDouble;
import gov.nist.nlpir.irf.de.DeInteger;
import gov.nist.nlpir.irf.de.DeString;
import gov.nist.nlpir.irf.modality.IndexingModalities;
import gov.nist.nlpir.irf.modality.IndexingModalityUnit;

import java.util.Vector;

/**
 * The StructuredCombinator
 *
 * @version $Revision: 1.1 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 */


public class StructuredCombinator extends Combinator {

    public StructuredCombinator(){
    }

  /**
   * Create a structured combinator.
   * @param size number of objects to be held by the structured combinator.
   */
    public StructuredCombinator(int size){
	super(size);

    }


  /**
   * Create and set structured combinator.
   *
   * @param combName which scoring function to apply.
   * @param imuIndexA index of an IndexingModalityUnit in IndexingModalities
   * @param imuWeightA weight of index of IndexingModalityUnit A
   * @param imuIndexB index of an IndexingModalityUnit in IndexingModalities
   * @param imuWeightB weight of index of IndexingModalityUnit B

   */
    public StructuredCombinator(DeString combName,
				int imuIndexA, double imuWeightA,
				int imuIndexB, double imuWeightB){

	this(5);                               // Set the size of this
	                                       // structured combinator to 5
	                                       // objects.
	addElement(combName);                  // Functon name.
	addElement(new DeInteger(imuIndexA)); // IMU index A
	addElement(new DeDouble(imuWeightA)); // IMU index A weight
	addElement(new DeInteger(imuIndexB)); // IMU index B
	addElement(new DeDouble(imuWeightB)); // IMU index B weight
}



  /**
   * Create and set combinator structure.
   *
   * @param combName which scoring function to apply.
   * @param scmb StructuredCombinator to be included in this combinator.
   * @param scmbWeight weight to be applied to scmb
   * @param imuIndex index of an IndexingModalityUnit in IndexingModalities
   * @param imuWeight weight to be applied IndexingModalityUnit at imuIndex
   */

    public StructuredCombinator(DeString combName,
				StructuredCombinator scmb,
				double scmbWeight,
				int imuIndex, double imuWeight){

	this(5);                               // Set the size of this
	                                       // structured combinator to 5
	                                       // objects.
	addElement(combName);                  // Functon name.
	addElement(scmb);                      // StructuerdCombinator to be
	                                       // included.
	addElement(new DeDouble(scmbWeight)); // Weight for scmb
	addElement(new DeInteger(imuIndex));  // IMU index 
	addElement(new DeDouble(imuWeight));  // IMU weight
}



  /**
   * Create and set combinator structure.
   * There must be at least 2 indexing modality units.
   *
   * @param function which scoring function to apply.
   * @param numberOfModalities number of IndexingModalities.
   * @param normalized weights used in scoring
   */

    public StructuredCombinator(String function,
				int numberOfModalities,
				double[] normalizedWeights){

	this(5);     // Set the size of this combinator to 5 objects.

	StructuredCombinator tmpVector = null;

	// Convert functon name to DeString.
	DeString combName = new DeString(function);
	
	// Create a structured combinator for the supplied function an IMU
	// index and weight as the left pair and an IMU index and weight on
	// the right.

	StructuredCombinator tmpVector2
	    = new StructuredCombinator(combName,
				       0, normalizedWeights[0],
				       1, normalizedWeights[1]);

	for (int i = 2; i < numberOfModalities; i++) {
	    // Create a structured combinator for the supplied function a
	    // structured combinator and weight as the left pair and an 
	    // IMU index and weight on the right.

	    tmpVector = new StructuredCombinator(combName,
						 tmpVector2, 1.0,
						 i, normalizedWeights[i]);
	    tmpVector2 = tmpVector;
	}

	// Add the elements of tempVector2 to this Structured Combinator
	// then it can be treated as the pure vector it superceeds.
	//
	// This isn't an ideal solution but should produce the same order as
	// tmpVector2 in this object, an extention of the vector class.
	// Which is the same order as the code in SampleApp produced in
	// the combinator vector.
	// This is a StructuredCombinator so it won't have more the 5 elements
	// to add so this shouldn't be overly expensive.
	
	for (int j = 0; j < tmpVector2.size(); j++)
	    addElement(tmpVector2.elementAt(j));
    }


  /**
   * display current structured combinator to standard output.
   * @param i width in characters of indentation
   * @param im current indexing modalities
   */
  
  
    public void present(int i, IndexingModalities im) {
	System.out.print("Structured Combinator\n");
	if (elementAt(0) instanceof DeString) { // structured
	    ((DeString)(elementAt(0))).present();
	    System.out.print("\n");
	    printSpaces(i);
	    System.out.print( ((DeDouble)(elementAt(2))).getValue() + " : ");
	    if (elementAt(1) instanceof  Vector) // structured
		((Combinator)(elementAt(1))).present(i+8, im);
	    else {
		presentCombinatorIMU((IndexingModalityUnit)
				     (im.getModalities().elementAt(((DeInteger)(elementAt(1))).getValue())));
	    } 
	    
	    printSpaces(i);
	    System.out.print(((DeDouble)(elementAt(4))).getValue() + " : ");
	    if (elementAt(3) instanceof Vector) // structured
		((Combinator)(elementAt(3))).present(i + 8, im);
	    else {
		presentCombinatorIMU((IndexingModalityUnit)
				     (im.getModalities().elementAt(((DeInteger)(elementAt(3))).getValue())));
	    } 
	} 
    }



    /**
     * print structured combinator with indexes instead of imus.
     *
     * @param i width in characters of indentation
     */

    public void presentBrief(int i) {
	if (elementAt(0) instanceof DeString) { // structured
	    ((DeString)(elementAt(0))).present();
	    System.out.print("\n");
	    printSpaces(i);
	    System.out.print( ((DeDouble)(elementAt(2))).getValue() + " : ");
	    if (elementAt(1) instanceof  Vector) // structured
		((StructuredCombinator)((elementAt(1)))).presentBrief(i+8);
	    else 
		System.out.println( ((DeInteger)(elementAt(1))).getValue() +
				    " : ");
	    
	    printSpaces(i);
	    System.out.print(((DeDouble)(elementAt(4))).getValue() + " : ");
	    if (elementAt(3) instanceof Vector) // structured
		((StructuredCombinator)((elementAt(3)))).presentBrief(i+8);
	    else
		System.out.println( ((DeInteger)(elementAt(3))).getValue() +
				    " : ");
	} 
    }


    /**
     * print structured combinator, v, with indexes instead of imus.
     * @param v a structured combinator
     * @param i width in characters of indentation
     */

    public void presentBrief(Vector v, int i) {
	if (v. elementAt(0) instanceof DeString) { // structured
	    ((DeString)(v.elementAt(0))).present();
	    System.out.print("\n");
	    printSpaces(i);
	    System.out.print( ((DeDouble)(v.elementAt(2))).getValue() +
			      " : ");
	    if (v.elementAt(1) instanceof  Vector) // structured
		((StructuredCombinator)((v.elementAt(1)))).presentBrief((Vector)v.elementAt(1),i+8);
	    else 
		System.out.println( ((DeInteger)(v.elementAt(1))).getValue() +
				    " : ");
	    
	    printSpaces(i);
	    System.out.print(((DeDouble)(v.elementAt(4))).getValue() + " : ");
	    if (v.elementAt(3) instanceof Vector) // structured
		((StructuredCombinator)((v.elementAt(3)))).presentBrief((Vector)v.elementAt(3),i+8);
	    else
		System.out.println( ((DeInteger)(v.elementAt(3))).getValue() +
				    " : ");
	} 
    }

}
