package gov.nist.nlpir.irf.result;

import gov.nist.nlpir.irf.query.QueryModalityUnit;
import gov.nist.nlpir.irf.query.QueryCondition;
import gov.nist.nlpir.irf.feature.ProxyIndexingFeature;

import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.de.DeString;
import gov.nist.nlpir.irf.util.Support;
import gov.nist.nlpir.irf.query.QueryRep;
import gov.nist.nlpir.irf.query.QueryModalityUnit;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Comparator;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/** 
 * The result of a search modelling the match between a document and a
 * query - at the level of a single document and all the information
 * needed to calculate the RSV for the document. 
 *<br><br> 

 *
 * @version $Revision: 1.2 $
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.
 */

public class ResultForDocMatchingQuery implements Comparator {
    

  /** The query associated with this result */
  private QueryRep   query;

  /** The retrieval status value for this result */
  private double      RSV;

  /** The document for which this result comprises the evidence */
  private ProxyDocument doc;		

  /** The lower-level results - these are per-field/index. 
   * Consists of ResultForDocMatchingQueryModalityUnit elements
   */
  private Vector      partialResults;	   		


  /** 
   * Creates an empty ResultForDocMatchingQuery
   */
  public ResultForDocMatchingQuery() {
  } 

  /** 
   * Creates a ResultForDocMatchingQuery for the supplied query, doc and 
   * results. The RSV is initialized to 0.0.
   * 
   * @param q A query.
   * @param d An Document
   * @param results Partial results (Vector of ResultForDocMatchingQueryModalityUnits)
   */
  public ResultForDocMatchingQuery(QueryRep q, ProxyDocument d, 
				   Vector results) { 
    partialResults = results;
    query = q;
    doc = d; 
    RSV = 0.0;
  } 
    


  /** 
   * Returns the document this result refers to.
   * @return the document.
   */
  public ProxyDocument getDocument() {
    return doc;
  }
    


  /** 
   * Returns the partialResults this result is a part of.
   * @return The partial results.
   */
  public Vector getPartialResults() {
    return partialResults;
  }
    


  /** 
   * Gets the Retrieval Status Value (RSV).
   * @return RSV
   */
  public double getRSV() {
    return RSV;
  }
    


  /**
   * Calculates the final Retrieval Status Value for a document by combining 
   * the scores obtained for each term in the query across all indexes.
   *
   * @return RSV - Retrieval Status Value
   */
  public double calcRSV(QueryRep aQueryRep) { 
    Vector aQColl = aQueryRep.getQueryCollection();
    // for avoiding errors in case query is empty
    if (aQColl.isEmpty())
      return 0.0;	

    double combinedRSV = 0.0;
    Enumeration nextQElem = aQColl.elements();
    Object arg0 = nextQElem.nextElement();
    ResultForDocMatchingQueryModalityUnit aPR;
	
    if ((arg0 instanceof DeString) && (aQColl.size() == 3)) {
      // structured query
      double v1 = 0.0;
      double v2 = 0.0;
      DeString functor = (DeString)arg0;
      Object arg1 = nextQElem.nextElement();
      Object arg2 = nextQElem.nextElement();

      if (arg1 instanceof QueryModalityUnit) 
	{          // arg1 is a terminal node:

	  // find the partial result achieved for given QueryModalityUnit
	  Enumeration i1 = partialResults.elements();
	  while (i1.hasMoreElements()) 
	    {
	    aPR = (ResultForDocMatchingQueryModalityUnit)i1.nextElement();
	    if (aPR.getQueryModalityUnit() == arg1)
	      {
	      // get RSV of partial result and multiply with weight
	      // This is the score for this document in this IMU
	      v1 = (aPR.getRSV() / aPR.getMaxRSV()) * 
		(aPR.getQueryModalityUnit().getWeight());
	      break;
	      }
	    }
	} else if (arg1 instanceof QueryRep) // arg1 is a non-terminal node:
	  {
	  v1 = this.calcRSV((QueryRep)arg1);
	  }

      if (arg2 instanceof QueryModalityUnit) 
	{          // arg2 is a terminal node:

	// find the partial result achieved for given QueryModalityUnit
	Enumeration i2 = partialResults.elements();
	while (i2.hasMoreElements()) 
	  {
	  aPR = (ResultForDocMatchingQueryModalityUnit)i2.nextElement();
	  if (aPR.getQueryModalityUnit() == arg2)
	    {
	    // get RSV of partial result and mulitply with weight
	    v2 = (aPR.getRSV() / aPR.getMaxRSV()) * 
	      (aPR.getQueryModalityUnit().getWeight());
	    break;
	    }
	  }
	} else if (arg2 instanceof QueryRep)   // arg2 is a non-terminal node:
	  {
	  v2 = this.calcRSV((QueryRep)arg2);
	  }

      // combine the two values
      if (0 == functor.getValueAsString().compareTo("lin"))
	// linear combination
	{
	combinedRSV = v1 + v2;
	}
      else if (0 == functor.getValueAsString().compareTo("max")){
	// maximum of the two values
	combinedRSV = (v1 > v2) ? v1 : v2;
      }
      else if (0 == functor.getValueAsString().compareTo("min")){
	// Minimum non-zero value
      }
    } 
    else if (arg0 instanceof QueryModalityUnit) 
      {
	// unstructured query consisting of QueryModalityUnit* objects
	Enumeration nextQElem2 = aQColl.elements();
	while (nextQElem2.hasMoreElements()) 
	  {
	  QueryModalityUnit aQueryModalityUnit = (QueryModalityUnit)nextQElem2.nextElement();
	  Enumeration i1 = partialResults.elements();
	  while (i1.hasMoreElements()) 
	    {
	    aPR = (ResultForDocMatchingQueryModalityUnit)i1.nextElement();
	    if (aPR.getQueryModalityUnit() == aQueryModalityUnit) 
	      {
	      // get RSV of partial result and mulitply with weight
	      // and add this to previous results
	      combinedRSV += ((aPR.getRSV() / aPR.getMaxRSV()) 
			      * (aPR.getQueryModalityUnit().getWeight()));
	      break;
	      }
	    }
	  }
      }
    RSV = combinedRSV;
    return RSV;
  } 
    


  /** 
   * Adds ResultForDocMatchingQueryModalityUnit to the partialResults.
   * 
   * @param aPR Result for doc matching query modality unit.
   */
  public void addPartialResult(ResultForDocMatchingQueryModalityUnit aPR) {
    partialResults.addElement(aPR);	
  }
    


  /** 
   * Compares two results in terms of Retrival Status Value
   * 
   * @param o2 another result of this type
   * @return 1 if this' results' RSV is smaller than o2's,<br>
   * -1 if it is greater<br>
   * and 0 if they are equal.
   */
  public int compare(Object o2) { 
    double difference = ((ResultForDocMatchingQuery)o2).getRSV() - RSV;
    if (difference > 0)
      return 1;
    else 
      return (difference < 0) ? -1 : 0;
  }



  /*----------------------------------------*
   * Implementation of Comparator interface
   *----------------------------------------*/
  /** 
   * Compares two results in terms of Retrival Status Value
   * NOTE: This compare reverses the standard sort sequence
   * so that index 1 of the list contains the highest RSV
   * 
   * @param o1 a result of this type
   * @param o2 another result of this type
   * @return 1 if o1's RSV < o2's RSV <br>
   *        -1 if o1's RSV > o2's RSV <br>
   *         0 if o1's RSV = o2's RSV
   */
  public int compare(Object o1, Object o2) { 
    double difference = ((ResultForDocMatchingQuery)o1).getRSV() - 
      ((ResultForDocMatchingQuery)o2).getRSV();
    if (difference > 0)
      return -1;
    else if (difference < 0)
      return  1;
    else
      return 0;
  }



  /** 
   * Gets results for these indexing features
   *
   * @return a vector of indexing features
   */
  public Vector getIndexingFeatures() { 
    Vector indexingFeatures = new Vector();
    Enumeration nextPR = partialResults.elements();
    int i = 1;
    while (nextPR.hasMoreElements()) {

      ResultForDocMatchingQueryModalityUnit aPR =
	(ResultForDocMatchingQueryModalityUnit)nextPR.nextElement();
      Vector tmp = aPR.getIndexingFeatures();
      Enumeration nextIF = tmp.elements();
      int j = 1;
      ProxyIndexingFeature anIF = null;
      while (nextIF.hasMoreElements()) {
	anIF = (ProxyIndexingFeature)nextIF.nextElement();
	indexingFeatures.addElement(anIF);
      }
    }
    return indexingFeatures;
  }
    


  /** 
   * Returns the query 
   * @return the query
     */
  public QueryRep getQuery() {
    return query;
  }
    


  /** 
   * Displays information about this result
   * 
   */
  public void present() { 
    System.out.print("Combined RSV = " + getRSV() + "\n");
    Enumeration prElements = partialResults.elements();
    while (prElements.hasMoreElements()) {
      ResultForDocMatchingQueryModalityUnit elm =
	(ResultForDocMatchingQueryModalityUnit)prElements.nextElement();
      elm.present();
    }
  }  
  
}
