package gov.nist.nlpir.irf.result;

import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.query.QueryCondition;
import java.util.Vector; 

/** 
 * The result of a search modelling the match between a document and a query
 * condition - at the level of a feature (e.g., a keyword in a text document 
 * query)
 * <br><br>
 * 
 * @version $Revision: 1.2 $
 *
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 *                            
 */
public class ResultForDocMatchingQueryCondition {
  

  /** The query condition met by this result */
  private QueryCondition condition;

  /** The document associated with this result */
  private ProxyDocument doc;

  /** A value reflecting some combination of the scores of features in the document
   *  which match features in the query
   */
  private double combinedScore;

  /** The weight of a feature in the document. This is usually derived as some function
   *  of the frequency of the feature in the document, e.g. the logarithm of the feature's
   *  frequency (see IdfIdxIntern.calcWeightOfFeatureInDocument())
   */
  private double weightOfFeatureInDocument;

  /** The weight of a feature in the query. This is usually derived as some function
   *  of the frequency of the feature in the query, e.g. the logarithm of the feature's
   *  frequency (see IdfIdxIntern.calcWeightOfFeatureInQuery())
   */
  private double weightOfFeatureInQuery;

  /** The collection of BasicRetrievalResults contributing to this result */
  private Vector basicResults;	
  


  /** 
   * Creates a ResultForDocMatchingQueryCondition for the supplied query 
   * condition, document and results. WeightOfFeatureInDocument, 
   * WeightOfFeatureInQuery and combinedScore are initialized to 0.0.
   *
   * @param c A query condition.
   * @param d A document.
   * @param results results.
   */
  public ResultForDocMatchingQueryCondition(QueryCondition condition, 
					    ProxyDocument doc, 
					    Vector basicResults) {
    this.condition = condition;
    this.doc = doc;
    this.basicResults = basicResults;	
    weightOfFeatureInDocument = 0.0;
    weightOfFeatureInQuery = 0.0;
    combinedScore = 0.0;  
  }
  


  /** 
   * Creates a string representation for a ResultForDocMatchingQueryCondition
   * @return The string representation. 
   */
  public String toString() {
    return("ResultForDocMatchingQueryCondition:\n" +
	   "{condition = " + condition + "," +
	   "Doc = " + doc + ",\n" +
	   "WeightOfFeatureInDocument = " +  weightOfFeatureInDocument + 
	   ",\n" +
	   "WeightOfFeatureInQuery = " +  weightOfFeatureInQuery + ",\n" +
	   "CombinedScore = " +  combinedScore + "\n" +
	   "BasicResults = " + basicResults + "}\n");
  }
  


  /** 
   * Sets the combinedScore to the supplied value.
   * @param idf The new combinedScore. 
   */
  public void setCombinedScore(double idf) {
    combinedScore = idf;
  }
  

  
  /** 
   * Gets the combinedScore
   * @return the CombinedScore.
   **/
  public double getCombinedScore() {
    return combinedScore;
  }  
  


  /** 
   * Gets the condition.
   * @return The condition. 
   */
  public QueryCondition getQueryCondition() {
    return condition;
  }
  

  
  /** 
   * Gets the document.
   * @return The document.
   */
  public ProxyDocument getDocument() {
    return doc;
  }
  


  /** 
   * Sets the weightOfFeatureInDocument to the supplied value.
   * @param w The new weight of the feature in the document. 
   */
  public void setWeightOfFeatureInDocument(double w) {
    weightOfFeatureInDocument = w;
  }
  

  
  /** 
   * Gets the weightOfFeatureInDocument.
   * @return Weight of the feature in the document.
   */
  public double getWeightOfFeatureInDocument() {
    return weightOfFeatureInDocument;
  }
  

  
  /** 
   * Sets the weightOfFeatureInQuery to the supplied value.
   * @param w The new weight of the feature in the query.
   */
  public void setWeightOfFeatureInQuery(double w) {
    weightOfFeatureInQuery = w;
  }
  

  
  /** 
   * Gets the weightOfFeatureInQuery.
   * @return Weight of the feature in the query.
   */
  public double getWeightOfFeatureInQuery() {
    return weightOfFeatureInQuery;
  }
  


  /** 
   * Gets the basicResults.
   * @return Basic results.
   */
  public Vector getBasicResults() {
    return basicResults;
  }
  


  /** 
   * Sets the basicResults to the supplied value.
   * @param results the new basic result.
   */
  public void setBasicResults(Vector results) {
    basicResults = results;
  }
}
