package gov.nist.nlpir.irf.result;

import gov.nist.nlpir.irf.query.QueryModalityUnit;
import gov.nist.nlpir.irf.query.QueryCondition;

import java.util.Vector;
import gov.nist.nlpir.irf.de.DeDouble;
import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.util.Support;

/** 
 * The result of a search modelling the match between a document and a
 * query modality unit - at the level of a single document field.
 * <br><br>
 * 
 * @version $Revision: 1.4 $
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 *                             
 */
public class ResultForDocMatchingQueryModalityUnit {
  

  /** Retrieval status value for the document at this level of result */
  private double RSV;

  /** The score of the document, computed with an index-dependent heuristic
   *   (derived from gov.nist.nlpir.irf.index.braf.IoAddrIntern.getScore()) */
  private double docScore;

  /** Number of features in the query */
  private int numQueryFeatures;         

  /** Number of unique features in the query */
  private int numUniqueQueryFeatures;

  /** The query modality unit associated with this result */
  private QueryModalityUnit QMU;

  /** The maximum Retrieval Status Value for any document in this modality */
  private double maxRSV;

  /** The document associated with this result */
  private ProxyDocument document;

  /** Collection of lower-level results (ResultForDocMatchingQueryCondition)
      comprised by this result */
  private Vector partialResults;
  


  /** 
   * Creates a ResultForDocMatchingQueryModalityUnit for the supplied 
   * query modality unit, document, score, results and maximum RSV. 
   * RSV is initialized to 0.0.
   *
   * @param q A query modality unit.
   * @param d A document.
   * @param s The document score,
   * @param results partial results. Must be a Vector of 
   * ResultForDocMatchingQueryConditions.
   * @param mr Maximum retrieval status value.
   */
  public ResultForDocMatchingQueryModalityUnit(QueryModalityUnit QMU, 
					       ProxyDocument document,
					       double docScore, 
					       Vector partialResults,
					       double maxRSV) {
    this.QMU = QMU;
    this.document = document; 
    this.docScore = docScore;
    RSV = 0.0;
    this.partialResults = partialResults;
    this.maxRSV = maxRSV;
  }
  


  /** 
   * Gets the query modality unit associated with this result
   * @return The query modality unit
   */
  public QueryModalityUnit getQueryModalityUnit() {
    return QMU;
  }
  


  /** 
   * Gets the document associated with this result.
   * @return The document .
   */
  public ProxyDocument getDocument() {
    return document;
  }
  


  /** 
   * Sets the RSV to the supplied value.
   * @param r The new RSV.
   */
  public void setRSV(double r) {
    RSV = r;
  }


  
  /** 
   * Gets the RSV.
   * @return The retrieval status value.
   */
  public double getRSV() {
    return RSV;
  }
  


  /** 
   * Adds the ResultForDocMatchingQueryCondition to the partialResults.
   * @param aPR result for doc matching query condition.
   */
  public void addPartialResult(ResultForDocMatchingQueryCondition aPR) {
    partialResults.addElement(aPR);
  }



  /** 
   * Gets number of entries in partialResults
   * @return count 
   */
  public int getCountOfPartialResults(){
    return partialResults.size();
  }
  


  /** 
   * Gets the docScore
   * @return document score.
   */
  public double getDocScore() {
    return docScore;
  }
  


  /** 
   * Sets the number of query features to the supplied value.
   * @param n The new number of query features.
   */
  public void setNumQueryFeatures(int n) {
    numQueryFeatures = n;
  }
  


  /** 
   * Gets the number of query features.
   * @return The number of query features.
   */
  
  public int getNumQueryFeatures() {
    return numQueryFeatures;
  }
  


  /** 
   * Sets the number of unique query features to the supplied value.
   * @param n The new number unique of query features.
   */
  public void setNumUniqueQueryFeatures(int n) {
    numUniqueQueryFeatures = n;
  }
  


  /** 
   * Gets the number of unique query features
   * @return The number of unique query features.
   */
  public int getNumUniqueQueryFeatures() {
    return numUniqueQueryFeatures;
  }
  


  /** 
   * Gets a vector of the indexing features comprised by this result
   * @return a vector of unique query features.
   */
  public Vector getIndexingFeatures() {
    Vector indexingFeatures = new Vector();
    int nbPR = partialResults.size();
    for (int i = 0; i < nbPR; i++) {
      ResultForDocMatchingQueryCondition aPR =
	(ResultForDocMatchingQueryCondition)partialResults.elementAt(i);
      Vector basicResults = aPR.getBasicResults();
      int nbBR = basicResults.size();
      for (int j = 0; j < nbBR; j++) {
	BasicRetrievalResult aBRR = 
	  (BasicRetrievalResult)basicResults.elementAt(j);
	indexingFeatures.addElement(aBRR.getIndexingFeature());
      }
    }
    return indexingFeatures;
  }
  


  /** 
   * Sets the maxRSV to the supplied value.
   * @param v The new maxRSV.
   */
  public void setMaxRSV(double v) {
    maxRSV = v;
  }
  


  /** 
   * Gets the maxRSV.
   * @return The maximum retrieval status value.
   */
  public double getMaxRSV() {
    return maxRSV;
  }
  

  
  /** 
   * Gets the partialResults.
   * @return The partial results.
   */
  public Vector getPartialResults() {
    return partialResults;
  }
  


  /** 
   * Displays information about this Result
   */
  public void present() {
    System.out.println("  QueryModalityUnit number " + getQueryModalityUnit().getNumber() +
		       ": orig. RSV = " + getRSV() + " max. RSV = " + maxRSV +
		       " RSV = " + getRSV()/maxRSV +
		       " Weight = " + getQueryModalityUnit().getWeight() + " weighted RSV = " +
		       (getRSV() / maxRSV) * getQueryModalityUnit().getWeight());
  }
  


   /** 
   * Creates a string representation for a 
   * ResultForDocMatchingQueryModalityUnit
   * @return The string representation. 
   */
  public String toString() {
    return("RSV: " + RSV +
	   "\tdocScore: " + docScore +
	   "\tnumQueryFeatures: " + numQueryFeatures +
	   "\tnumUniqueQueryFeatures: " + numUniqueQueryFeatures +
	   "\tQMU: " + QMU +
	   "\tmaxRSV: " + maxRSV);
  }
}
