package gov.nist.nlpir.irf.util;

/**
 * A circular queue of keys and associated values
 *
 * @version $Revision: 1.1 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 */
public class CircQueue {


  /** Array of keys to queue */
  private Object keys[] = null;

  /** Array of values, each associated with keys at same index */
  private Object values[] = null;

  /** Index of most recently added value */
  private int newestIndex = -1;
    
  /** Modulus to give queue its circularity */
  private int qMod = 0;

  /** Value to add to index to get previous */
  private int backOne = 0;



  /**
   * Constructs a new, empty circular queue of the size specified
   * @param      maxEntries maximum number of entries the queue can hold
   */
  public CircQueue(int maxEntries) {
    keys = new Object[maxEntries];
    values = new Object[maxEntries];
    qMod = maxEntries;
    backOne = maxEntries-1;
  }



  /**
   * Constructs a new, empty circular queue of the size specified
   * with as much data from the old as will fit
   * @param      maxEntries maximum number of entries the queue can hold
   */
  public synchronized void setMaxEntries(int newMaxEntries) {
    Object newKeys[] = new Object[newMaxEntries];
    Object newValues[] = new Object[newMaxEntries];
    System.arraycopy(keys, 0, newKeys, 0, newMaxEntries);
    System.arraycopy(values, 0, newValues, 0, newMaxEntries);
    keys = newKeys;
    values = newValues;
    qMod = newMaxEntries;
    backOne = newMaxEntries - 1;
    newestIndex = 
      newestIndex >= newMaxEntries ? newMaxEntries - 1 : newestIndex;
  }



  /**
   * Adds an object (key-value pair) to the queue
   * @param key of object to be added
   * @param value of object to be added
   */
  public final void add(Object key, Object value) {
    synchronized(keys) {
      newestIndex = (newestIndex+1) % qMod; 
      keys[newestIndex] = key;
      values[newestIndex] = value;
    }
  }



  /**
   * Finds the most recently added value with the given key 
   * @param key search key of object sought
   * @return the most recent value with the given key or null if none found
   */
  public final Object find(Object key) {
    synchronized (keys) {

      if (newestIndex == -1)
	return null;

      int i = newestIndex;
      do {
        if (keys[i].equals(key))
	  return values[i];
	i = (i+backOne) % qMod;
      } while (keys[i] != null && i != newestIndex);
      
      return null;
    }
  }
  


  /**
   * Finds the most recently added value with the given key 
   * @param key search key of object sought
   * @param value replacement value 
   */
  public final void replace(Object key, Object value) {
    synchronized (keys) {
      
      if (newestIndex == -1)
	return;
      
      int i = newestIndex;
      do {
        if (keys[i].equals(key))
	  values[i] = value;
	i = (i+backOne) % qMod;
      } while (keys[i] != null && i != newestIndex);
    }
  }
  


  /**
   * Presents the contents of the queue, most recent first
   */
  public void present() {
    synchronized (keys) {
      if (newestIndex == -1)
	System.out.println("Queue content, most recent first... None");
      else {
	int i = newestIndex;
	System.out.println("Queue contents, most recent first...");
	do {
          System.out.println("Key: "+keys[i]);
          System.out.println("Value: "+values[i]);
	  System.out.println("-----------------------");
	  i = (i+backOne) % qMod;
	} while (keys[i] != null && i != newestIndex);
      }
    } 
  }
  
}
