package gov.nist.nlpir.irf.util;

/**
 * This cache associates a value to a key, a bit like a hashtable does. 
 * When a couple is stored, the value may then be retrieved using the key. 
 * But this operation may not succeed in case a couple has been stored with 
 * a key having the same hashcode as the first one (collision case). This 
 * cache allows values of two different status
 * to be stored: those which are considered <i>dirty</i>, and those which 
 * aren't. This way, the client may store a persistent object in the cache, 
 * even if this one needs to be written. If the cache destroys this object 
 * because of a collision, it will return it as the return value of the 
 * destroying method so that the client can then do what's necessary. In 
 * the case an object is destroyed but the key associated to the new value 
 * is the same as the one for the disappearing object (regarding <code>
 * equals()</code>, the method returns nothing, as this is considered an 
 * update "in memory" that doesn't need to be reported to the disk. This 
 * feature may be deferred to the client in a later release.
 *
 * @version $Revision: 1.2 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the U.S. government, 
 * and by statute is not subject to copyright in the United States.  
 * Recipients of this software assume all responsibilities associated with 
 * its operation, modification and maintenance.
 */

public class HashingCache {

  /** Array keeping track of the stored keys. */
  private Object keys[] = null;

  /** Array containing the values. */
  private Object values[] = null;

  /** For every couple (keys[i], values[i]), must it be returned when it's 
      removed from the cache because of a collision? */
  private boolean dirty[] = null;

  /** Current size of cache, ie of each of its arrays. */
  private int size = 0;



  /**
   * Builds a new cache with the given size.
   */
  public HashingCache(int maxEntries) {
    size = maxEntries;
    keys = new Object[maxEntries];
    values = new Object[maxEntries];
    dirty = new boolean[maxEntries];
    for (int i = 0; i < maxEntries; i++)
      dirty[i] = false;
  }


  /**
   * This method allows to dynamically change the size of the cache. It 
   * returns the objects that were marked as <i>dirty</i> when they were 
   * put in the cache. The format of the returned array is the same as the 
   * one from <code>getDirtyValues()
   * </code>.
   * @see HashingCache#getDirtyValues
   */
  public synchronized Object[] setMaxEntries(int newMaxEntries) {
    Object[] result = getDirtyValues();

    Object newKeys[] = new Object[newMaxEntries];
    System.arraycopy(keys, 0, newKeys, 0, size);
    keys = newKeys;

    Object newValues[] = new Object[newMaxEntries];
    System.arraycopy(values, 0, newValues, 0, size);
    values = newValues;

    dirty = new boolean[newMaxEntries];
    for (int i = 0; i < newMaxEntries; i++)
      dirty[i] = false;

    size = newMaxEntries;

    return result;
  }


  /**
   * Adds an object (key-value pair) to the queue. This is considered a new 
   * object in the table, i.e. it should be written before leaving the cache.
   * For this, it will be returned with its key by the add() or put() method 
   * that will delocate it.
   * @param key of object to be added
   * @param value of object to be added
   * @return If a collision occured, an array of two objects containing the 
   * removed key and the removed object in [0] and [1]. This only if this 
   * couple was marked <i>dirty</i>, ie put by <code>add()</code>, and its 
   * key is different from the new one given.<br>
   * Otherwise, null.
   */
  public final synchronized Object[] add(Object key, Object value) {
    int range = (key.hashCode() & 0x7FFFFFFF) % size;
    return internPut(range, key, value, true);
  }


  /** Same as add with no dirty bit.
   * @return If a collision occured, an array of two objects containing the 
   * removed key and the removed object in [0] and [1]. This only if this 
   * couple was marked <i>dirty</i>, ie put by <code>add()</code>, and its 
   * key is different from the new one given.<br>
   * Otherwise, null.
   */
  public final synchronized Object[] put(Object key, Object value) {
    int range = (key.hashCode() & 0x7FFFFFFF) % size;
    return internPut(range, key, value, false);
  }


  /**
   * Common part to <code>put()</code> and <code>add()</code>, this method 
   * seeks the right place in the two tables. If an object is already 
   * present, AND it is marked <i>dirty</i> (it was introduced here with 
   * <code>add()</code> and not <code>put() </code>, AND THE KEY IS 
   * DIFFERENT OF THE NEW ONE, then it is returned by this method: the 
   * client should save the value if he wants to keep track of it.
   */
  private final Object[] internPut(int range, Object key, Object value, 
				   boolean willBeDirty) {
    Object[] result = null;
    if (dirty[range])
      if (!key.equals(keys[range])) {
	result = new Object[2];
	result[0] = keys[range];
	result[1] = values[range];
      }
    dirty[range] = willBeDirty;
    keys[range] = key;
    values[range] = value;
    return result;
  }


  /**
   * Find the most recently added value with the given key 
   * @param key search key of object sought
   * @return the most recent value with the given key or null if none found
   */
  public final synchronized Object find(Object key) {
    int possibleRange = (key.hashCode() & 0x7FFFFFFF) % size;
    if (key.equals(keys[possibleRange]))
      return values[possibleRange];
    return null;
  }
  

  /**
   * @return an array of couples of values that are marked <i>dirty</i>. 
   * Each Object of even indice is the key associated in the table to the 
   * value that is just after it in the array.
   */
  public synchronized Object[] getDirtyValues() {
    int numberOfDirtyValues = 0;
    for (int i = 0; i < size; i++) // Count the number of dirty values.
      if (dirty[i]) numberOfDirtyValues++;
    Object[] result = new Object[numberOfDirtyValues * 2];
    int current = 0;
    for (int i = 0; i < size; i++)
      if (dirty[i]) {
	result[current++] = keys[i];
	result[current++] = values[i];
      }
    return result;
  }


  /**
   * Present the contents of the cache, in the form:<br>
   * <code>d key1 <=> value1   </code>for couples that are marked dirty,<br>
   * <code>key2 <=> value2   </code>for the others.
   */
  public void present() {
    synchronized (keys) {
      for (int i = 0; i < size; i++)
	System.out.print((dirty[i] ? "d " : "") + keys[i] + " <=> " + values[i] + ", ");
    } 
    System.out.println();
  }
}
