package gov.nist.nlpir.irf.util;

import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This class implements a table of unique strings read from a file
 * in the format described below. 
 * removal.<br><br>
 * 
 * @version $Revision: 1.1 $ 	
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 */
public class StringTable {
  

  /** Table of strings - key is the string, value is always 0 */
  private Hashtable table;


  /** Working copy of end of file marker value */
  private static int EOF = -1;        
  


  /**
   *  Creates the StringTable from a file. The file must contain the list
   *  of terms. A term is a sequence of letters and/or digits. Terms must
   *  be separated by one or more chars which are not letters/digits.
   *  If file cannot be opened, constructor returns with empty
   *  stringtable (size() = 0)
   *  @param filename the name of the file used.
   *  @exception IOException if the file cannot be read
   */
  public StringTable(String fileName) throws IOException {
    BufferedReader f = null;
    table = new Hashtable(500); // initial capacity 500; default loadfactor
    int ch = 0;                 // temp storage for character from file
    
    
    // Handle the case that the file is not found
    
    try {
      f = new BufferedReader(new FileReader(fileName));
    } catch (FileNotFoundException error) {
      Support.errorMessage("StringTable::StringTable(): " + error +
			   ", Sorry, can't open file: " + fileName );
      return; /* return using empty string table; */
    }
    
    StringBuffer term = new StringBuffer(); // reuse for all terms
    
    // Read through the file finding terms and saving them in the table
    
    ch = f.read();
    while (ch != EOF) { 
      
      // ----- Discard leading non-term chars ---------------
      // ch cannot be EOF at this point   
      do {
	if (Character.isLetterOrDigit((char)ch))  
	  break;
	ch = f.read();            
      } while (ch!= EOF);
      
      // ----- Gather term chars-----------------------------
      while (ch != EOF) {
	if (!Character.isLetterOrDigit((char)ch))  
	  break;
	term.append((char)ch);
	ch = f.read();     
      }
      
      // ----- Discard trailing non-term chars---------------
      while (ch != EOF) {
	if (Character.isLetterOrDigit((char)ch))  
	  break;
	ch = f.read();     
      }
      
      // If we got a term, save it in the table
      if (term.length() > 0) {
	table.put(new String(term),new Integer(0));
	term.setLength(0); // reuse term so hashes for same strings are ==
      }                  // identical strings in different StringBufs
      // yield different hashes and multipel entries
      
    } // end of outer while (ch != EOF)
    
  }
  


  /**
   * Gets number of strings in string table.
   * @return number of strings in string table
   */
  public int size() {
    return table.size();
  }
  


  /**
   * Presents the contents of the string table.
   */
  public void present() {
    System.out.println( size() + " Strings in the table:");
    Enumeration keyEnum = table.keys();
    while (keyEnum.hasMoreElements())
      System.out.println(keyEnum.nextElement());     
  }
  


  /**
   *  Tests whether the given string is in the table or not.
   *  @return true if the word is in the table,<br>
   *  false otherwise.
   */
  public boolean containsString(String stringToLookUp) {
    return table.containsKey(new String(stringToLookUp));
  }
}
