package gov.nist.nlpir.irf.util;

import java.io.PrintWriter;
import java.io.PrintStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.util.Date;

/** 
 * Provides miscellaneous methods to support some logging, access to system
 * and runtime values, basic test mode functions, input methods, etc.
 * 
 * @version $Revision: 1.3 $
 * 
 * @author
 * This software was produced by NIST, an agency of the U.S. government,  
 * and by statute is not subject to copyright in the United States.        
 * Recipients of this software assume all responsibilities associated      
 * with its operation, modification and maintenance.  
 *                             
 * @see gov.nist.nlpir.irf.InfoObj
 */

public final class Support {
  
  /** Trace level constant */
  public static final int COARSE_TRACE = 4;

  /** Trace level constant */
  public static final int MEDIUM_TRACE = 6;

  /** Trace level constant */
  public static final int DETAIL_TRACE = 8;

  /** Working local copy of file separator character */
  public static final String fileSeparator = 
    System.getProperty("file.separator");
  
  /** Flag indicating if debugging output is to be produced or not */
  private static boolean outputState = false;

  /** Flag indicating if the user is to be prompted for input */
  private static boolean promptUser  = true;

  /** Flag indicating weather the application is in test mode */
  private static boolean testMode    = false;

  /** Current trace level */
  public static int traceLevel      = 0;

  /** Working PrintWriter for error message output */
  private static PrintWriter errorWriter  = new PrintWriter(System.err, true);

  /** Working PrintWriter for use by the local methods print and println  */
  private static PrintWriter outputWriter = new PrintWriter(System.out, true);

  /** Working PrintWriter for input prompt message output */
  private static PrintWriter promptWriter = new PrintWriter(System.out, true);

  /** Working PrintWriter for trace message output */
  private static PrintWriter traceWriter  = new PrintWriter(System.out, true);

  /** Working PrintWriter for warning message output */
  private static PrintWriter warningWriter= new PrintWriter(System.err, true);
  
  /** Working instance of a Runtime object */
  private static Runtime runtime = java.lang.Runtime.getRuntime();



  /** This reader should be the only one in the whole application listening
   * to STDIN. Thus use of input files isn't prohibited anymore.  */  
  private static BufferedReader sysRdr = new BufferedReader(new InputStreamReader(System.in));

  /** Everything beyond this sign is a comment that <code>readLine()</code> 
   * won't return. */
  public static final char COMMENT_CHAR = '#';



  /** 
   * Sets the writer to which errors messages will be directed.
   * @param writer where error messages are to be sent.
   */
  public static final void setErrorWriter(PrintWriter writer) {
    errorWriter = writer;
  }
  


  /**
   * Sets the writer to a writer for the stream to which errors messages 
   * will be directed.
   * @param stream where error messages are to be sent.
   */
  public static final void setErrorWriter(PrintStream stream) {
    Support.setErrorWriter(new PrintWriter(stream, true));
  }
  


  /** 
   * Gets the writer to which error messages are being directed.
   * @return PrintWriter where error messages are being sent.
   */
  public static final PrintWriter getErrorWriter() {
    return errorWriter;
  }  



  /** 
   * Writes message to the error writer.
   * @param message text to be printed on error writer.
   */
  public static final void errorMessage(String message)
  {
    errorWriter.println(message);
  }
  


  /** 
   * Sets the maximum level of trace messages to be generated.
   * @param level new maximum level of trace messages to be generated.
   */
  public static final void setTraceLevel(int level) {
    traceLevel = (level < 0) ? 0 : level;
  }
  


  /** 
   * Gets the maximum level of trace messages to be generated.
   * @return the maximum level of trace messages to be generated.
   */
  public static final int getTraceLevel() {
    return traceLevel;
  }


  
  /** 
   * Returns true if the supplied level would result in a trace message 
   * being generated.
   * @param level to be checked against current trace level.
   * @return true if the supplied level would result in a trace message 
   * being generated.
   */
  public static final boolean executeTrace(int level) {
    return level <= Support.getTraceLevel();
  }
  


  /** 
   * Sets the writer to which trace messages will be directed.
   * @param writer where trace messages are to be sent.
   */
  public static void setTraceWriter(PrintWriter writer) {
    traceWriter = writer;
  }
  


  /** 
   * Sets the writer to a writer for the stream to which trace messages
   * will be directed.
   * @param stream where trace messages are to be sent.
   */
  public static final void setTraceWriter(PrintStream stream) {
    Support.setTraceWriter(new PrintWriter(stream, true));
  }
  


  /** 
   * Gets the writer to which trace messages are being directed.
   * @return PrintWriter where trace messages are being sent.
   */
  public static final PrintWriter getTraceWriter() {
    return traceWriter;
  }
  


  /** 
   * Writes message to the trace writer.
   * @param level at which this trace messages is to be sent.
   * @param message text to be printed on trace writer.
   */
  public static final void traceMessage(int level,
				  String s1, int v1,
				  String s2, int v2,
				  String s3, int v3) {
    if (level <= traceLevel)
      traceWriter.println(s1 + " " + v1 + " " + s2 + " " + v2 + " " +
			  s3 + " " + v3);
  }
  


  /** 
   * Writes message to the trace writer.
   * @param level at which this trace messages is to be sent.
   * @param message text to be printed on trace writer.
   */
  public static final void traceMessage(int level, String message) {
    if (level <= traceLevel)
      traceWriter.println(message);
  }
  


  /** 
   * Writes message to the trace writer.
   * @param message text to be printed on trace writer.
   */
  public static void traceMessage(String message) {
    Support.traceMessage(0, message);
  }
  
  
  /** 
   * Sets the writer to which warning messages will be directed.
   * @param writer where warning messages are to be sent.
   */
  public static final void setWarningWriter(PrintWriter writer) {
    warningWriter = writer;
  }
  


  /** 
   * Sets the writer to a writer for the stream to which warning messages
   * will be directed.
   * @param stream where warning messages are to be sent.
   */
  public static final void setWarningWriter(PrintStream stream) {
    Support.setWarningWriter(new PrintWriter(stream, true));
  }
  


  /** 
   * Gets the writer to which warning messages are being directed.
   * @return Writer where warning messages are being sent.
   */
  public static final PrintWriter getWarningWriter() {
    return warningWriter;
  }
  


  /** 
   * Writes message to the warning stream.
   * @param level at which this warning messages is to be sent.
   * @param message text to be printed on warning stream.
   */
  public static final void warningMessage(String message) {
    warningWriter.println(message);
  }
  


  /** 
   * Sets the test mode.
   * @param mode test mode on/off.
   */
  public static final void setTestMode(boolean mode) {
    testMode = mode;
  }
  


  /** 
   * Gets the test mode.
   * @return current test mode.
   */
  public static final boolean getTestMode() {
    return testMode;
  }
  


  /** 
   * Sets the promptUser.
   * @param promptUser on/off.
   */
  public static final void setPromptUser(boolean mode) {
    promptUser = mode;
  }
  


  /** 
   * Gets value of promptUser.
   * @return current value promptUser.
   */
  public static final boolean getPromptUser() {
    return promptUser;
  }
  


  /** 
   * Sets the writer to which user prompts will be directed.
   * @param writer where user prompt are to be sent.
   */
  public static final void setPromptWriter(PrintWriter writer) {
    promptWriter = writer;
  }
  


  /**
   * Sets the writer to a writer for the stream to which user prompts 
   * will be directed.
   * @param stream where user prompts are to be sent.
   */
  public static final void setPromptWriter(PrintStream stream) {
    Support.setPromptWriter(new PrintWriter(stream, true));
  }
  


  /** 
   * Gets the writer to which user prompts are being directed.
   * @return PrintWriter where user prompts are being sent.
   */
  public static final PrintWriter getPromptWriter() {
    return promptWriter;
  }
  


  /** 
   * Writes the user prompt to the prompt writer.
   * @param message text to be printed on prompt writer.
   */
  public static final void userPrompt(String prompt) {
    if (promptUser)
      promptWriter.print(prompt);
    promptWriter.flush();
  }
  


  /** 
   * Writes the user prompt to the prompt writer.
   * @param message text to be printed on prompt writer.
   */
  public static final void userPromptLn(String prompt) {
    if (promptUser)
      promptWriter.println(prompt);
  }
  


  /**
   * Sets state of debuging output.
   * @param state new state of debugging output, set to true to
   * turn on debugging output, false to turn off.
   */
  public static final void debugOutputOn(boolean state) {
    outputState = state;
  }


  
  /**
   * Gets state of debugging output.
   * @return true if debugging output is on.
   */
  public static final boolean queryDebugOutput() {
    return outputState;
  }
  




  /**
   * Sets output print writer for debugging output.
   * @param writer print writer for debugging output
   */
  public static final void setPrintWriter(PrintWriter writer) {
    outputWriter = writer;
  }
  

  
  /**
   * Sets output writer to a writer for the stream for debugging
   * output.
   * @param pStream print stream for debugging output
   */
  public static final void setPrintWriter(PrintStream pStream) {
    Support.setPrintWriter(new PrintWriter(pStream, true));
  }
  

  
  /**
   * Writes a timestamp to standard out.
   */
  public static final void printTimestamp() {
    Date now = new Date();
    System.out.println("T i m e s t a m p : "+now.toString()+"; Millisecs since 1.1.1970 = "+new Long(now.getTime()).toString());
  }
  
  /**
   * Gets time in milliseconds since midnight Jan 1st 1970.
   * @return time in milliseconds
   */
  public static final long getTimeInMillis() {
    Date now = new Date();
    return now.getTime();
  }
  


  /**
   * Requests the garbage collector be run.
   */
  public static final void requestGarbageCollection() {
    runtime.gc();
  }
  


  /**
   * If printing debugging output, writes stats on memory usage
   * by the Java virtual machine to standard out.
   */
  public static final void printVM_MemoryStats() {
    System.out.println("Total memory (bytes) = " + runtime.totalMemory() +
		       "\nTotal free memory    = " + runtime.freeMemory());
  }
  


  /**
   * If printing debugging output, writes stats on memory usage
   * by the Java virtual machine to standard out after calling 
   * gabage collection.
   */
  public static final void printVM_MemoryStatsWithGC() {
    System.out.println("Total memory (bytes)      = " + runtime.totalMemory() +
		       "\nTotal free memory pre GC  = " + runtime.freeMemory());
    runtime.gc();   // request garbage collection
    System.out.println("Total free memory post GC = " + runtime.freeMemory());
  }



  /**
   * This method mainly embeds BufferedReader.readLine() so that it doesn't 
   * raise any exception and, moreover, the entire application can share a 
   * unique reader based on the standard output, thus allowing to have, for 
   * example, command files.<br>
   * NOTE ON COMMENTS: In order for those command files to be usable, a 
   * comment mechanism appeared. Everything you type after the comment char 
   * is considered a comment and thus is not parsed. The comment sign is 
   * defined by <code>COMMENT_CHAR</code>:
   * @see Support#COMMENT_CHAR
   */
  public static final String readLine() {
    try {
      String result = sysRdr.readLine();
      int commentOffset = result.indexOf(COMMENT_CHAR);
      if (commentOffset > 0)
	result = result.substring(0, commentOffset);
      result = result.trim();
      if (result.equals("")) 
	result = readLine(); // Allows to skip comment lines.
      return result;
    } catch (IOException error) {
      errorMessage("Error in Support.readLine: " + error);
    }
    return "";
  }
}
