package gov.nist.nlpir.irfapps.hci;

import gov.nist.nlpir.irf.conversion.IrfConverter;
import gov.nist.nlpir.irf.conversion.Ascii2HtmlConverter;
import gov.nist.nlpir.irf.document.ProxyDocument;
import gov.nist.nlpir.irf.de.DeString;
import gov.nist.nlpir.irf.de.ProxyDeString;
import gov.nist.nlpir.irf.de.DePersonName;
import gov.nist.nlpir.irf.de.ProxyDePersonName;
import gov.nist.nlpir.irf.de.DeHtml;
import gov.nist.nlpir.irf.de.ProxyDeHtml;
import gov.nist.nlpir.irf.de.DeDouble;
import gov.nist.nlpir.irf.de.ProxyDeDouble;
import gov.nist.nlpir.irf.de.DeInteger;
import gov.nist.nlpir.irf.de.ProxyDeInteger;
import gov.nist.nlpir.irf.util.Support;

import java.util.Vector;
import java.util.Random;
import java.io.FileReader;
import java.io.PushbackReader;
import java.io.IOException;
import java.io.FileNotFoundException;

/**
 * Converts TROFF Bibliogrphic documents in a file to HciDoc
 * documents using Bib2AppDocConv, which creates proxy data elements
 * as attributes rather than data elements
 *
 *<p>
 * The class was initially translated verbatim from C++.  This was
 * less than ideal.  To make this work better in Java, the lexical
 * scanning portion of the code was moved to BibToken.java; this has
 * simplified the method Convert1 significantly.  Methods for handling
 * personal names were added.
 *</p>
 *
 * @version $Revision: 1.2 $ 
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 */
public class Bib2AppDocConv extends IrfConverter {


  /** Reader for use in getting document data from a file */
  PushbackReader file = null;

  /**
   * Makes a new Bib to AppDoc2 Converter
   */
  public Bib2AppDocConv() {
    setRealDocClass("gov.nist.nlpir.irfapps.hci.HciDoc");
    setProxyDocClass("gov.nist.nlpir.irfapps.hci.ProxyHciDoc");
  }


  /**
   * Gives the converter the information it needs to access the raw
   * data for the documents to be converted. In the case of this 
   * converter this will be a fully qualified file name.
   *
   * @param loc fully qualified name of the file with the raw doc data
   */
  public void setRawDocLocation(String loc) throws FileNotFoundException {
    file = new PushbackReader(new FileReader(loc));
  }
  
 

  /** 
   * Skips numberOfDocsToSkip in the input file starting from the
   * current location
   *
   * @param numberOfDocsToSkip  number of documents to skip 
   * @return number of docs 
   */
  public int ignoreN(int numberOfDocsToSkip)
    throws IOException {
    return 0;
  }



  /** 
   * Converts one document from a file opened using a 
   * PushbackReader
   *
   * @return the document converted or null
   */
  public ProxyDocument convert1() 
    throws IOException {
    String name1 = "";		//  firstname
    String name2 = "";		//  surname
    String title = "";		//  title
    String abstract_text = "";	// abstract
    BibToken tok;
    char tagname;
    tok = BibToken.readBibToken(file); 
    while (tok.type == BibToken.SEP ) { tok = BibToken.readBibToken(file);  }
    while (tok.type != BibToken.EOF && tok.type != BibToken.SEP ) 
      {
      tagname = tok.tagname;
      switch (tagname) 
	{
	case 'T':
	  title = tok.value.trim();
	  break;
	case 'A':
	  if (name2 == "" && name1 == "") 
	    {
	    name1 = firstName(tok.value.trim()).trim();
	    name2 = surName(tok.value.trim());
	    }
	  break;
	case 'X':
	  abstract_text = tok.value.trim();
	  break;
	}
      tok = BibToken.readBibToken(file);
      }
    
    Ascii2HtmlConverter conv = new Ascii2HtmlConverter();
    return 
      new ProxyHciDoc(title,
		       name1,
		       name2,
		       conv.convert(abstract_text)); 
    }



  /** 
   * Converts one document from a file opened using a 
   * PushbackReader.
   *
   * @param f             PushbackReader file reference.
   * @param ignore        Return null document if true.
   */
  public ProxyDocument convert1x(PushbackReader f, int ignore)
    throws IOException {
    String name1 = "";		//  firstname
    String name2 = "";		//  surname
    String title = "";		//  title
    String abstract_text = "";	// abstract
    BibToken tok;
    char tagname;
    tok = BibToken.readBibToken(f); 
    while (tok.type == BibToken.SEP ) { tok = BibToken.readBibToken(f);  }
    while (tok.type != BibToken.EOF && tok.type != BibToken.SEP ) {
      tagname = tok.tagname;
      switch (tagname) {
      case 'T':
	title = tok.value.trim();
	break;
      case 'A':
	if (name2 == "" && name1 == "") {
	  name1 = firstName(tok.value.trim()).trim();
	  name2 = surName(tok.value.trim());
	}
	break;
      case 'X':
	abstract_text = tok.value.trim();
	break;
      }
      tok = BibToken.readBibToken(f);
    }
    
    if (ignore == 1) 
      {
      return (ProxyHciDoc)null; // means everything went OK
      } 
    else 
      {
      Ascii2HtmlConverter conv = new Ascii2HtmlConverter();
      return 
	new ProxyHciDoc(title,
			 name1,
			 name2,
			 conv.convert(abstract_text)); 
      }
  }
  


  /**
   * Extracts first name from personal name string.
   * defined as everything up to the last space.
   * @param name personal name
   */
  public static String firstName(String name) {
    return name.substring(0,name.lastIndexOf(' ')+1);
  }
  


  /**
   * Extracts surname from personal name string.
   * defined as everything starting after the last space and going to
   * the end
   * @param name personal name
   */
  public static String surName(String name) {
    return name.substring(name.lastIndexOf(' ')+1, name.length());
  }
  
} // Bib2AppDocConv


