package gov.nist.nlpir.irfapps.hci;

import java.io.IOException;
import java.io.PushbackReader;

/**
 * A class to represent a token in a TROFF Bibliographic
 * (refer) Document.  
 *<p>
 *
 * @version $Revision: 1.1 $
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.  
*/
public class BibToken {
  // token ids

  /** Token type value */
  static final int TAG     = 1;

  /** Token type value */
  static final int UNDEF   = 2;

  /** Token type value */
  static final int SEP     = 3;

  /** Token type value */
  static final int EOF     = -1;
  
  /** type of token */
  public int type;

  /** field abbreviation */
  public char tagname;

  /** value of token */
  public String value;	



  /**
   * Creates new biblio field token using the bibliography field
   * abbreviation character.
   * @param t    id of TOKEN 
   * @param name bibliography field abbreviation character */
  public BibToken(int t, char name, String v) {
    type = t;
    tagname = name;
    value = v;
  }



  /**
   * Sets BibToken
   * @param t id of TOKEN 
   * @param bibliography field abbreviation character
   */
  public void setToken(int t, char name, String v) {
    type = t;
    tagname = name;
  }
  


  /**
   * Returns a string representation of the BibToken object with the
   * token type, tag name, and token value.  
   * @return string representation of BibToken Object
   */
  public String toString() {
    return "BibToken: type = " + type + 
      ", tagname = " +tagname + ", value = " + value;
  }
  


  /**
   * Determines if supplied string contains only white space charaters.
   * @param line string containing one line of text ending in a
   * newline
   * @return true if line is blank.
   */
  static boolean isBlankLine(String line) {
    for (int i = 0; i < line.length(); i++) {
      if (! Character.isWhitespace(line.charAt(i))) {
	return false;
      }
    }
    return true;
  }
  


  /**
   * Scans a java PushbackReader of TROFF Bibliographic Documents
   * until a token is recognized, then returns the token.
   *
   * @param PushbackReader containing BIB Documents
   * @return newly created BibToken.
   * @exception IOException thrown if IO error
   * <p>
   * The function is based on the grammar shown below. 
   * </p>
   * <pre>
   * lambda_ws   -> lambda | WHITESPACE_NO_NL ;
   * blankline   -> NEWLINE lambda_ws NEWLINE ;
   * eod         -> blankline | eof ;
   * HCIdocument -> recordlist eod ;
   * recordlist  -> lambda | record recordlist ;
   * record      -> tag content 
   * record      -> title | section | author | bookname | date |
   * pages | copyright | abstract ;
   * title       -> "%T" content ;
   * section     -> "%S" content ;
   * author      -> "%A" content ;
   * bookname    -> "%B" content ;
   * date        -> "%D" content ;
   * pages       -> "%P" content ;
   * copyright   -> "%C" content ;
   * abstract    -> "%X" content ;
   * content       -> STRING | STRING SPACE content ;
   *</pre>
   * terminals are in UPPERCASE
   * <p>
   * scanner finite state machine
   * <pre>
   * state  input                output       next state
   * -------------------------------------------------------
   *   0     %                                    1
   *   0     eof                 separator        end                 
   * -------------------------------------------------------
   *   1     A-Z                 setrectype       2
   * -------------------------------------------------------
   *   2     space                                3
   * -------------------------------------------------------
   *   3     A-Za-z{punct, sp}                    3
   *   3     newline                              4
   *   3     eof                 newdoc,pushback  end
   * -------------------------------------------------------
   *   4     A-Za-z{punct, sp}                    3
   *   4     space                                4
   *   4     newline             newdoc,pushback  end
   *   4     eof                 newdoc,pushback  end
   *   4     %                   newdoc,pushback  end
   * -------------------------------------------------------
   * </pre>
   * */
  public static BibToken readBibToken(PushbackReader f) throws IOException {
    int state = 0;
    StringBuffer vbuf = new StringBuffer();
    char rectype= ' ';
    int c = 0;
    while (state < 5) {
      // System.out.println("c = " + (char)c + " => state " + state);
      c = f.read();
      switch (state) {
      case 0:
	switch (c) {
	case '\n':
	  return new BibToken(SEP, ' ', "Separator");
	case ' ':
	  state = 0;
	  break;
	case EOF:
	  return new BibToken(EOF, ' ', "End of File" );
	case '%':
	  state = 1;
	  break;
	}
	break;
      case 1:
	if (Character.isLetter((char)c) || '*' == (char)c ) {
	  rectype = (char)c;
	  state = 2;

	}
	break;
      case 2:
	if (Character.isWhitespace((char)c)) {
	  state = 3;
	}
	break;
      case 3:
	vbuf.append((char)c);
	if ('\n' == (char)c) {
	  state = 4;
	} else if (EOF == c) {
	  f.unread(c);
	  return new BibToken(TAG, rectype, vbuf.toString());
	}
	break;
      case 4:
	if ('\n' == (char)c) {
	  f.unread(c);
	  return new BibToken(TAG, rectype, vbuf.toString());
	} else if ('%' == (char)c) {
	  f.unread(c);
	  return new BibToken(TAG, rectype, vbuf.toString());
	} else if (EOF == c) {
	  f.unread(c);
	  return new BibToken(TAG, rectype, vbuf.toString());
	} else {
	  vbuf.append((char)c);
	  state = 3;
	}
	break;
      }
    }
    return new BibToken(UNDEF, 'U', "Undefined record");
  }

} // BibToken

