package gov.nist.nlpir.irfapps.trec.cran;

import gov.nist.nlpir.irfapps.trec.TrecDocument;

import gov.nist.nlpir.irf.document.Document;
import gov.nist.nlpir.irf.de.DeString;
import gov.nist.nlpir.irf.de.ProxyDeString;
import gov.nist.nlpir.irf.de.ProxyDeHtml;
import gov.nist.nlpir.irf.util.Support;

/**
 * An extension of TrecDocument that models a document from the Cranfield
 * collection (in TREC format)
 *
 * @see gov.nist.nlpir.irf.document.Document
 * @see gov.nist.nlpir.irfapps.trec.TrecDocument
 *
 * @version $Revision: 1.2 $
 *
 * @serial title
 * @serial author
 * @serial biblio
 * @serial text
 *
 *
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 */

public class CranfieldDocument extends TrecDocument {

  /** Document title
   * @serial */
  public ProxyDeString title;

  /** Document author
   * @serial */
  public ProxyDeString author;

  /** Bibliographic information
   * @serial */
  public ProxyDeString biblio;

  /** Document text
   * @serial */
  public ProxyDeHtml text;


  /** No-argument constructor*/
  public CranfieldDocument() {
    super();
  }



  /**
   * Constructs a document given a document number
   * @param docNr document number
   */
  public CranfieldDocument(String docNr) {
    super(docNr); // all TREC documents contain a document number
  }



  /**
   * Constructs a document given the pieces
   * @param docNr document number
   * @param ti title
   * @param auth author
   * @param bib bibliograhic info
   * @param txt text
   */
  public CranfieldDocument(String docNr, String ti, String auth, 
			   String bib, String txt) {
    super(docNr);
    title   = new ProxyDeString(ti);
    author  = new ProxyDeString(auth);
    biblio  = new ProxyDeString(bib);
    text    = new ProxyDeHtml(txt);
  }



  /**
   * Constructs a document given the pieces
   * @param docNr document number
   * @param ti title
   * @param auth author
   * @param bib bibliograhic info
   * @param txt text
   */
  public CranfieldDocument(ProxyDeString docNr, String ti, String auth, 
			   String bib, String txt) {
    super(docNr);
    title   = new ProxyDeString(ti);
    author  = new ProxyDeString(auth);
    biblio  = new ProxyDeString(bib);
    text    = new ProxyDeHtml(txt);
  }


  /**
   * Constructs a document given the pieces
   * @param docNr document number
   * @param ti title
   * @param auth author
   * @param bib bibliograhic info
   * @param txt text
   */
  public CranfieldDocument(ProxyDeString docNr, DeString ti, 
			   DeString auth,  DeString bib,
			   String txt) {
    super(docNr);
    title  = new ProxyDeString(ti);
    author  = new ProxyDeString(auth);
    biblio  = new ProxyDeString(bib);
    text    = new ProxyDeHtml(txt);
    System.err.println("CranfieldDocument 5");
  }


  /**
   * Constructs a document given the pieces
   * @param docNr document number
   * @param ti title
   * @param auth author
   * @param bib bibliograhic info
   * @param txt text
   */
  public CranfieldDocument(ProxyDeString docNr, ProxyDeString ti, 
			   ProxyDeString auth, ProxyDeString bib,
			   ProxyDeString txt) {
    super(docNr);
    title  = new ProxyDeString(ti.getValueAsString());
    author  = new ProxyDeString(auth.getValueAsString());
    biblio  = new ProxyDeString(bib.getValueAsString());
    text    = new ProxyDeHtml(txt.getValueAsString());
    System.err.println("CranfieldDocument 6");
  }



  /**
   * Returns string representation of CRANFIELD Document
   */
  public String toString() {
    return ("\nD o c n o :   " + docNumber +
	    "\nT i t l e :   " + title +
	    "\nA u t h o r : " + author +
	    "\nB i b l i o : " + biblio +
	    "\nT e x t :     " + text);
  }
  

  /**
   * Overrides 
   */
  public boolean equals(Object o) {
    if (o instanceof CranfieldDocument) {
      CranfieldDocument temp = (CranfieldDocument)o;
      return (this.docNumber.equals(temp.docNumber));
    }
    System.err.println("CranfieldDocument.equals(): was expecting"+
		       " a CranfieldDocument, received a " +
		       o.getClass().getName() + ": " + o);
    return false;
  }



  public Document getQueryDocument() {
    String docNr = "";
    String ti = "";
    String auth = "";
    String bib = "";
    String txt = "";
    System.out.print("Type . to leave a field empty (no value)\n" +
		     "Doc Number       : "); 
    docNr = Support.readLine();
    System.out.print("Title  : "); 
    ti = Support.readLine(); 
    System.out.print("Author  : "); 
    auth = Support.readLine(); 
    System.out.print("Biblio : "); 
    bib = Support.readLine();
    System.out.print("Text    : ");
    txt = Support.readLine();
    return new CranfieldDocument(docNr, ti, auth, bib, txt);
  }

  public void presentBrief() {
    if (docNumber != null) {
      System.out.print(docNumber.getValue() + " ");
    }

    if (title != null) {
      String ttl = title.getValue();
      String prnTtl = ttl.substring(0, Math.min(ttl.length(), 60));
      System.out.println(prnTtl);
    } 
    else if (biblio != null) {
	String abstr = biblio.getValue();
	String prnAbstr = abstr.substring(0, Math.min(abstr.length(), 50));
	System.out.println("Doc with no title (abstract: " + prnAbstr + "...)");
    } else if (text != null) {
	String text = this.text.getValue();
	String prnText = text.substring(0, Math.min(text.length(), 50));
	System.out.println("Doc with no title (text: " + prnText + "...)");
    } else
	System.out.println("Doc with no title, abstract or text.");
  }
}
